#!/bin/bash
SCRIPT=$0
SCRIPT_NAME=$(basename "${SCRIPT}")
LOG_FILE=/tmp/${SCRIPT_NAME}.log
exec > >(tee -a ${LOG_FILE} )
exec 2> >(tee -a ${LOG_FILE} >&2)
LOG_TIME=`date "+%Y-%m-%d %H:%M:%S"`

echo starting ${SCRIPT_NAME} at ${LOG_TIME} 

[ "`id -u`" = "0" ] || (echo "$0 must be run as root" && exit 1)

#Common dir is one level above where this script lives, and allows us to find the other peer folders
export COMMON_DIR=$(dirname "${SCRIPT}")/..
export EXT_FILE_DIR=$COMMON_DIR/external-files
export FILES_TO_COPY_DIR=$COMMON_DIR/files-to-copy-oracle
export SCRIPTS_DIR=$COMMON_DIR/scripts

# install all the dependencies required by Oracle 11G that are not already installed
# http://docs.oracle.com/cd/E11882_01/install.112/e24326/toc.htm#BHCGJCEA - see Red Hat Enterprise 7 list
echo "Installing linux package dependencies for Oracle DB"
REQUIRED_PACKAGES="binutils compat-libcap1 compat-libstdc++-33 compat-libcap1 gcc gcc-c+ glibc
glibc-devel glibc.i686 ksh libaio libaio-devel libgcc libstdc++ libstdc++-devel libXi libXtst make
sysstat"
yum --assumeyes install $REQUIRED_PACKAGES \
    || yum --assumeyes install $REQUIRED_PACKAGES \
    || yum --assumeyes install $REQUIRED_PACKAGES

# Fail immediately on any error
#set -e

#extract the two parts of the oracle DB install into the same dir - /tmp/database
echo "Unzipping oracle install files"
unzip -q $EXT_FILE_DIR/linux.x64_11gR2_database_1of2.zip -d /tmp  || (echo "Could not find $EXT_FILE_DIR/linux.x64_11gR2_database_1of2.zip"  && exit 2)
unzip -q $EXT_FILE_DIR/linux.x64_11gR2_database_2of2.zip -d /tmp  || (echo "Could not find $EXT_FILE_DIR/linux.x64_11gR2_database_2of2.zip"  && exit 2)

echo "Adding Groups and oracle user"
[ "$(getent group dba)" ] || groupadd dba
[ "$(getent group oinstall)" ] || groupadd oinstall
#add oracle user with primary group oinstall and secondary group dba
[ "$(getent passwd oracle)" ] || useradd -c 'Oracle DB User' -g oinstall -G dba oracle
#might be able to specify -p<password> on useradd, if I can figure out how to crypt the password.
#passwd oracle
#But, for now, we just need to set it later.  Script goes from root to oracle, so no password is needed during setup

echo "Changing kernel settings per Oracle requirements"

# copy sysctl.conf in, set owner and permissions, and then prod the system to use the new settings
cp $FILES_TO_COPY_DIR/sysctl.conf /etc
chmod -R 644 /etc/sysctl.conf
sysctl -p

#copy limits.conf in, set owner and permissions.  
# -rw-r--r--. 1 root root 2572 Feb  9 19:19 /etc/security/limits.conf
cp $FILES_TO_COPY_DIR/limits.conf /etc/security
chmod -R 644 /etc/security/limits.conf

echo "Creating ORACLE_BASE directory"
#Create ORACLE_BASE
ORACLE_BASE=/u01/dbs
mkdir -p $ORACLE_BASE
chown -R oracle:oinstall $ORACLE_BASE
chmod -R 775 $ORACLE_BASE

echo "Calling script to install Oracle Software"
su - oracle -- $SCRIPTS_DIR/oracle-db-install-sw.sh

echo "Running post install root scripts"
/u01/dbs/oraInventory/orainstRoot.sh
/u01/dbs/oracle/11.2.0/root.sh

echo "Calling script to create DB"
su - oracle -- $SCRIPTS_DIR/oracle-db-create-db.sh

echo "Calling script for post create DB root stuff"
$SCRIPTS_DIR/oracle-db-create-db-root.sh

echo "Calling script to import DB"
su - oracle -- $SCRIPTS_DIR/oracle-import.sh

#delete /tmp/database files to save space.
echo "Cleaning up install files"
rm -rf /tmp/database

echo "End of ${SCRIPT_NAME}"



