/**
 * Source file created in 2011 by Southwest Research Institute
 */


package gov.va.med.pharmacy.peps.common.utility;


import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;

import org.apache.commons.lang3.SystemUtils;
import org.apache.commons.lang3.builder.ToStringBuilder;
import org.apache.commons.lang3.builder.ToStringStyle;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.stereotype.Component;

import gov.va.med.pharmacy.peps.common.utility.security.UntrustedLogParam;


/**
 * Implementation of {@link PpsConfiguration} that reads configuration properties from a file. It uses default values if
 * the configuration file cannot be read or does not contain a particular property.
 */
@Component
public class ConfigFileUtility implements PpsConfiguration {

    /** IMAGE_PATH */
    public static final String DEFAULT_FDB_IMAGE_PATH = "/FDB_Images/";

    /** DEFAULT_TIMEOUT */
    public static final int DEFAULT_TIMEOUT = 60;

    /** DEFAULT_PROXY */
    public static final int DEFAULT_PROXY = 360;

    /** DEFAULT_DIVISION */
    public static final int DEFAULT_DIVISION = 521;

    /** DEFAULT_CONNSPEC */
    public static final String DEFAULT_CONNSPEC = "DUZ";

    /** DEFAULT_RPCTIMEOUT */
    public static final int DEFAULT_RPCTIMEOUT = 60000;

    /**
     * The default value for {@link #getPersistentDirectory()}.
     */
    private static final File DEFAULT_PERSISTENT_DIRECTORY = new File(
      SystemUtils.getUserDir(),   // the working directory should be the WebLogic domain directory
      "/config/");

    private static final Logger LOG = LogManager.getLogger(ConfigFileUtility.class);

    private static String START_TIMEOUT_TAG = "<TransactionTimeout>";
    private static String STOP_TIMEOUT_TAG = "</TransactionTimeout>";
    private static String START_PROXY_TAG = "<NDFProxyUserIEN>";
    private static String STOP_PROXY_TAG = "</NDFProxyUserIEN>";
    private static String START_DIV_TAG = "<NDFDivision>";
    private static String STOP_DIV_TAG = "</NDFDivision>";
    private static String START_CONNSPEC_TAG = "<ConnectionSpecName>";
    private static String STOP_CONNSPEC_TAG = "</ConnectionSpecName>";
    private static String START_RPCTIMEOUT_TAG = "<RPCTimeout>";
    private static String STOP_RPCTIMEOUT_TAG = "</RPCTimeout>";
    private static String START_FDBLOC_TAG = "<FDBImageLocation>";
    private static String STOP_FDBLOC_TAG = "</FDBImageLocation>";
    private static String START_STS_WEBSERVICE_TAG = "<STSWebServiceURL>";
    private static String STOP_STS_WEBSERVICE_TAG = "</STSWebServiceURL>";

    private int timeout;
    private int ndfProxyUserDUZ;
    private int division;
    private int rpcTimeout;
    private String connectionSpecName;
    private String fdbImageLocation;
    private String stsWebServiceURL;

    /**
     * Constructs an instance.
     */
    public ConfigFileUtility() {

        init();
    }

    @Override
    public String getConnectionSpecName() {

        return connectionSpecName;
    }

    /**
     * setConnectionSpecName
     * @param connectionSpecName : connectionSpecName
     */
    public void setConnectionSpecName(String connectionSpecName) {

        this.connectionSpecName = connectionSpecName;
    }

    @Override
    public int getRpcTimeout() {

        return rpcTimeout;
    }

    /**
     * setRpcTimeout
     * @param rpcTimeout : rpcTimeout
     */
    public void setRpcTimeout(int rpcTimeout) {

        this.rpcTimeout = rpcTimeout;
    }

    @Override
    public int getTimeout() {

        return timeout;
    }

    /**
     * setTimeout
     * @param timeout : timeout
     */
    public void setTimeout(int timeout) {

        this.timeout = timeout;
    }

    @Override
    public int getNdfProxyUserDUZ() {

        return ndfProxyUserDUZ;
    }

    /**
     * setNdfProxyUserDUZ
     * @param ndfProxyUserDUZ : ndfProxyUserDUZ
     */
    public void setNdfProxyUserDUZ(int ndfProxyUserDUZ) {

        this.ndfProxyUserDUZ = ndfProxyUserDUZ;
    }

    @Override
    public int getDivision() {

        return division;
    }

    /**
     * setDivision
     * @param division division
     */
    public void setDivision(int division) {

        this.division = division;
    }

    @Override
    public String getFdbImageLocation() {

        return fdbImageLocation;
    }

    /**
     * sets fdbImageLocation field
     * 
     * @param fdbImageLocation the fdbImageLocation to set
     */
    public void setFdbImageLocation(String fdbImageLocation) {

        this.fdbImageLocation = fdbImageLocation;
    }

    @Override
    public String getStsWebServiceURL() {
    
        return stsWebServiceURL;
    }

    /**
     * Sets the STS WebService URL.
     * 
     * @param stsWebServiceURL the STS WebService URL
     */
    public void setStsWebServiceURL(String stsWebServiceURL) {
    
        this.stsWebServiceURL = stsWebServiceURL;
    }
    
    @Override
    public File getPersistentDirectory() {
      // We do not currently support loading this property from the configuration file. There has not been a need to so
      // far.
      return DEFAULT_PERSISTENT_DIRECTORY;
    }

    private void init() {

        timeout = DEFAULT_TIMEOUT;
        ndfProxyUserDUZ = DEFAULT_PROXY;
        division = DEFAULT_DIVISION;
        connectionSpecName = DEFAULT_CONNSPEC;
        rpcTimeout = DEFAULT_RPCTIMEOUT;
        fdbImageLocation = DEFAULT_FDB_IMAGE_PATH;

        String s = new String();

        InputStream is = this.getClass().getClassLoader().getResourceAsStream("xml/PPS-NConfig.xml");
        // Letting the BufferedReader close the stream per
        // <http://docs.oracle.com/javase/tutorial/essential/exceptions/tryResourceClose.html>.
        try (BufferedReader reader = new BufferedReader(new InputStreamReader(is))) {
            int iTimeout;
            int iNDFProxy;
            int iDivision;
            int iRPCTimeout;
            String iConnSpec;
            String iFdbLocation;
            
            // Note: The below is _not_ a good way of parsing this file. We iterate over every line, checking each one
            // for every tag. Each iteration will throw multiple exceptions, one for each property that was not on that
            // line. The next time someone changes this code, it would be good to improve the parsing.

            while ((s = reader.readLine()) != null) {

                try {
                    iTimeout = this.getPropertyInt(s, START_TIMEOUT_TAG, STOP_TIMEOUT_TAG, DEFAULT_TIMEOUT);
                    timeout = iTimeout;
                } catch (PropertyNotFoundInXmlLineException e) {
                  LOG.trace("Property {} not found on line.", START_TIMEOUT_TAG, e);
                }

                try {
                    iNDFProxy = this.getPropertyInt(s, START_PROXY_TAG, STOP_PROXY_TAG, DEFAULT_PROXY);
                    ndfProxyUserDUZ = iNDFProxy;
                } catch (PropertyNotFoundInXmlLineException e) {
                  LOG.trace("Property {} not found on line.", START_PROXY_TAG, e);
                }

                try {
                    iDivision = this.getPropertyInt(s, START_DIV_TAG, STOP_DIV_TAG, DEFAULT_DIVISION);
                    division = iDivision;
                } catch (PropertyNotFoundInXmlLineException e) {
                  LOG.trace("Property {} not found on line.", START_DIV_TAG, e);
                }

                try {
                    iRPCTimeout = this.getPropertyInt(s, START_RPCTIMEOUT_TAG, STOP_RPCTIMEOUT_TAG, DEFAULT_RPCTIMEOUT);
                    rpcTimeout = iRPCTimeout;
                } catch (PropertyNotFoundInXmlLineException e) {
                  LOG.trace("Property {} not found on line.", START_RPCTIMEOUT_TAG, e);
                }

                try {
                    iConnSpec = getProperty(s, START_CONNSPEC_TAG, STOP_CONNSPEC_TAG, null);
                    connectionSpecName = iConnSpec;
                } catch (PropertyNotFoundInXmlLineException e) {
                  LOG.trace("Property {} not found on line.", START_CONNSPEC_TAG, e);
                }

                try {
                    iFdbLocation = getProperty(s, START_FDBLOC_TAG, STOP_FDBLOC_TAG, null);
                    fdbImageLocation = iFdbLocation;
                } catch (PropertyNotFoundInXmlLineException e) {
                  LOG.trace("Property {} not found on line.", START_FDBLOC_TAG, e);
                }

                try {
                    String value = getProperty(s, START_STS_WEBSERVICE_TAG, STOP_STS_WEBSERVICE_TAG, null);
                    setStsWebServiceURL(value);
                } catch (PropertyNotFoundInXmlLineException e) {
                  LOG.trace("Property {} not found on line.", START_STS_WEBSERVICE_TAG, e);
                }
            }

            
        } catch (Exception e) {
            LOG.warn("Failed to load configuration due to Exception.", e);
        }
        
        LOG.debug("Configuration initialized: {}", new UntrustedLogParam(this));
    }
    
    /**
     * Returns a string revealing all public properties of this object.
     */
    @Override
    public String toString() {
      return new ToStringBuilder(this, ToStringStyle.SHORT_PREFIX_STYLE)
          .append("timeout", timeout)
          .append("ndfProxyUserDUZ", ndfProxyUserDUZ)
          .append("division", division)
          .append("rpcTimeout", rpcTimeout)
          .append("connectionSpecName", connectionSpecName)
          .append("fdbImageLocation", fdbImageLocation)
          .append("stsWebServiceURL", stsWebServiceURL)
          .toString();
    }

    /**
     * getPropertyInt - gets an integer property from the xml line
     *
     * @param s String
     * @param startTag String
     * @param stopTag String
     * @param defaultValue int
     * @return int
     * @throws PropertyNotFoundInXmlLineException exception
     */
    private int getPropertyInt(String s, String startTag, String stopTag, int defaultValue)
        throws PropertyNotFoundInXmlLineException {

        int rv = defaultValue;

        if (s.contains(startTag) && s.contains(stopTag)) {
            int beginIndex = s.indexOf(startTag) + startTag.length();
            int endIndex = s.indexOf(stopTag);
            String strPort = s.substring(beginIndex, endIndex);

            try {
                rv = Integer.valueOf(strPort).intValue();
            } catch (NumberFormatException nfe) {
                rv = defaultValue;
            }

            return rv;
        }

        throw new PropertyNotFoundInXmlLineException();
    }

    /**
     * getProperty - gets a value from the xml line (a string)
     *
     * @param s String
     * @param startTag String
     * @param stopTag String
     * @param defaultValue String
     * @return String
     * @throws PropertyNotFoundInXmlLineException exception
     */
    private String getProperty(String s, String startTag, String stopTag, String defaultValue)
        throws PropertyNotFoundInXmlLineException {

        String rv = defaultValue;

        if (s.contains(startTag) && s.contains(stopTag)) {
            int beginIndex = s.indexOf(startTag) + startTag.length();
            int endIndex = s.indexOf(stopTag);
            rv = s.substring(beginIndex, endIndex);

            return rv;
        }

        throw new PropertyNotFoundInXmlLineException();
    }

    /**
     * this exception class is used to skip the assignment of a property.
     */
    private class PropertyNotFoundInXmlLineException extends Exception {

        /** serialVersionUID */
        private static final long serialVersionUID = 8015312492996833208L;
        private static final String DEFAULT_MSG = "Property not found in input string";

        /**
         * Constructor
         *
         */
        public PropertyNotFoundInXmlLineException() {

            super(DEFAULT_MSG);
        }

        /**
         * Constructor
         *
         * @param message String
         */
        @SuppressWarnings("unused")
        public PropertyNotFoundInXmlLineException(String message) {

            super(message);
        }

    }
}
