package gov.va.med.pharmacy.peps.domain.common.capability.impl;


import java.util.List;

import org.hibernate.Criteria;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.criterion.Restrictions;

import gov.va.med.pharmacy.peps.common.exception.ValidationException;
import gov.va.med.pharmacy.peps.common.utility.PPSConstants;
import gov.va.med.pharmacy.peps.common.vo.CopayTierVo;
import gov.va.med.pharmacy.peps.common.vo.ManagedItemVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.validator.ErrorKey;
import gov.va.med.pharmacy.peps.common.vo.validator.Errors;
import gov.va.med.pharmacy.peps.common.vo.validator.ValidationError;
import gov.va.med.pharmacy.peps.common.vo.validator.ValidationErrorLevel;
import gov.va.med.pharmacy.peps.domain.common.capability.CopayTierDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.dao.DataAccessObject;
import gov.va.med.pharmacy.peps.domain.common.dao.EplCopayTierDao;
import gov.va.med.pharmacy.peps.domain.common.model.EplCopayTierDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplProdCopayTierAssocDo;
import gov.va.med.pharmacy.peps.domain.common.utility.converter.Converter;
import gov.va.med.pharmacy.peps.domain.common.utility.converter.CopayTierConverter;
import gov.va.med.pharmacy.peps.domain.common.utility.querybuilder.HqlBuilder;


/**
 * Perform CRUD operations on CopayTier.
 */
public class CopayTierDomainCapabilityImpl extends
        ManagedDataDomainCapabilityImpl<CopayTierVo, EplCopayTierDo>
        implements CopayTierDomainCapability {
    
    /**
     * The data access object for Copay Tier.
     */
    private EplCopayTierDao eplCopayTierDao;
    
    /**
     * The converter for Copay Tier.
     */
    private CopayTierConverter copayTierConverter;
    
    public final EplCopayTierDao getEplCopayTierDao() {
        return eplCopayTierDao;
    }

    
    public final void setEplCopayTierDao(EplCopayTierDao eplCopayTierDao) {
        this.eplCopayTierDao = eplCopayTierDao;
    }

    
    public final CopayTierConverter getCopayTierConverter() {
        return copayTierConverter;
    }

    
    public final void setCopayTierConverter(CopayTierConverter copayTierConverter) {
        this.copayTierConverter = copayTierConverter;
    }

    
    /**
     * deleteItem.
     * @param copayTierVo copayTierVo
     * @throws ValidationException ValidationException
     */
    @Override
    public final void deleteItem(CopayTierVo copayTierVo) throws ValidationException  {
        List<EplCopayTierDo> eplCopayTierDo = getDataAccessObject().retrieve(EplCopayTierDo.EPL_ID, 
            Long.valueOf(copayTierVo.getId()));
       
        if (eplCopayTierDo.size() == 1) {
            if (eplCopayTierDo.get(0).getTierLevel() == null) {            
                getDataAccessObject().delete(eplCopayTierDo.get(0));
            } else  {
                throw new ValidationException(ValidationException.CANNOT_DELETE,
                   copayTierVo.getId() + " because it has already synched with PRODUCT.");
            }
        }
    }   
    
    /**
     * Warn if copayTierVo with associated products is about to be inactivated.
     * Constructs HQL to query the Copay Tier Associations table and determine
     * if there are products associated with the copay tier.
     *  
     * @param copayTierVo copayTierVo
     * @param user {@link UserVo} performing the action
     */
    @Override
    public Errors checkForInactivateWarnings(CopayTierVo copayTierVo) {
        Errors warnings = new Errors();
        Session session = this.getDataAccessObject().getCurrentSession();
        StringBuffer sql = new StringBuffer();
        sql.append(HqlBuilder.create("SELECT item FROM ").append(EplProdCopayTierAssocDo.class)
                .append(PPSConstants.ITEM));
        String whereClause = " WHERE EPL_ID_COPAY_TIER_FK = " + copayTierVo.getId();
        sql.append(whereClause);
        String sqlQuery = (sql.toString());
        Query query = session.createQuery(sqlQuery);
        query.setMaxResults(1);
        
        @SuppressWarnings("rawtypes")
        List assocs = query.list();
        
        if (assocs != null && assocs.size() > 0) {
            warnings.addError(
                new ValidationError(
                    ErrorKey.INACTIVE_COPAY_TIER_HAS_PRODUCTS_ASSIGNED,
                    ValidationErrorLevel.WARNING));
        }  

        return warnings;
    }

    
    
    /**
     * Create Hibernate {@link Criteria} that will find an item by its
     * uniqueness fields.
     * 
     * @param item
     *            {@link ManagedItemVo} for which to create uniqueness
     *            {@link Criteria}
     * @return {@link Criteria}
     */
    @Override
    protected final Criteria createUniquenessCriteria(CopayTierVo item) {
        Criteria criteria = getDataAccessObject().getCriteria();
        criteria.add(Restrictions.ilike(EplCopayTierDo.TIER_LEVEL,
                item.getTierLevel()));

        return criteria;
    }

    /**
     * Insert the given CopayTierVo.
     * 
     * @param drugClassVo drugClassVo
     * @param user {@link UserVo} performing the action
     * @return DrugClass with new ID
     */
    @Override
    public final CopayTierVo createWithoutDuplicateCheck(CopayTierVo copayTierVo,
            UserVo user) {

        if (copayTierVo.getId() == null) {
            copayTierVo.setId(getSeqNumDomainCapability().generateId(
                copayTierVo.getEntityType(), user));

            if ((copayTierVo.getVaDataFields() != null)
                    && (!(copayTierVo.getVaDataFields().isEmpty()))) {
                copayTierVo.getVaDataFields().setVaDfOwnerId(
                        getSeqNumDomainCapability().generatedOwnerId(user));
            }
        }

        EplCopayTierDo ingredientDo = copayTierConverter.convert(copayTierVo);

        eplCopayTierDao.insert(ingredientDo, user);


        return (copayTierVo);
    }

    /**
     * Get the {@link DataAccessObject} that this CopayTierClassCapabilityImpl uses.
     * 
     * @return {@link DataAccessObject}
     */
    public final DataAccessObject getDataAccessObject() {
        return eplCopayTierDao;
    }

    /**
     * Return the {@link Converter} instance for this capability.
     * 
     * @return {@link Converter}
     */
    public final CopayTierConverter getConverter() {
        return copayTierConverter;
    }

}
