package gov.va.med.pharmacy.peps.domain.common.capability.impl;

import java.math.BigDecimal;
import java.util.List;
import java.util.Optional;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import gov.va.med.pharmacy.peps.common.utility.security.UntrustedLogParam;
import gov.va.med.pharmacy.peps.common.vo.RematchVo;
import gov.va.med.pharmacy.peps.common.vo.UnmatchRematchVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.domain.common.capability.RematchMgtDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.dao.EplRematchDao;
import gov.va.med.pharmacy.peps.domain.common.model.EplRematchDo;
import gov.va.med.pharmacy.peps.domain.common.utility.converter.RematchConverter;


/**
 * Main implementation of {@link RematchMgtDomainCapability}.
 */
public class RematchMgtDomainCapabilityImpl implements RematchMgtDomainCapability {
  
  private static final Logger LOG = LogManager.getLogger(RematchMgtDomainCapabilityImpl.class);

    private final EplRematchDao eplRematchDao; 
    private final RematchConverter rematchConverter;
    
    /**
     * Constructs an instance with the required dependencies.
     * 
     * @param pEplRematchDao the {@link EplRematchDao} to use.
     * @param pRematchConverter the {@link RematchConverter} to use.
     */
    @Autowired
    public RematchMgtDomainCapabilityImpl(
        final EplRematchDao pEplRematchDao,
        final RematchConverter pRematchConverter) {
      eplRematchDao = pEplRematchDao;
      rematchConverter = pRematchConverter;
    }

    @Override
    public void removeObsoleteProducts() {
        
        removeObsoleteEntries(eplRematchDao.removeObsoleteOthers());
        
    }
    
    @Override
    public RematchVo saveOrUpdate(final RematchVo rematch, final UserVo user) {
      EplRematchDo savedDo;
      if (rematch.getEplRematchId() != null) {
        final Long id = Long.valueOf(rematch.getEplRematchId());
        LOG.debug("Rematch has already been saved (id={}): updating it.", id);
        final EplRematchDo dataObj = eplRematchDao.retrieve(id);
        rematchConverter.setDataObjectPropertiesFrom(dataObj, rematch);
        // We're not actually calling a DAO method here so we need to call this ourselves.
        eplRematchDao.updateCreatedAndLastModified(dataObj, user);
        // Just let Hibernate update things when the session closes.
        savedDo = dataObj;
      } else {
        LOG.debug("Saving new rematch.");
        final EplRematchDo newDo = rematchConverter.convert(rematch);
        savedDo = eplRematchDao.insert(newDo, user);
      }

      return rematchConverter.convert(savedDo);
    }
    
    @Override
    public void delete(final RematchVo rematch) {
      LOG.debug("Deleting rematch {}.", new UntrustedLogParam(rematch));
      // Use convert minimal because we should only need the primary key here.
      eplRematchDao.delete(rematchConverter.convertMinimal(rematch));
    }
    
    @Override
    public void addUnmatchedProducts(UserVo user) {
      
      try {
        addUnmatchedEntries(eplRematchDao.loadUnmatchCMOP(), user);
        addUnmatchedEntries(eplRematchDao.loadUnmatchDispense(), user);
        addUnmatchedEntries(eplRematchDao.loadUnmatchDosageForm(), user);
        addUnmatchedEntries(eplRematchDao.loadUnmatchDrugUnit(), user);
        addUnmatchedEntries(eplRematchDao.loadUnmatchFedSch(), user);
        addUnmatchedEntries(eplRematchDao.loadUnmatchGenericName(), user);
        addUnmatchedEntries(eplRematchDao.loadUnmatchPrintName(), user);
        addUnmatchedEntries(eplRematchDao.loadUnmatchStrength(), user);
      } catch (final Exception ex) {
        // Some of the above queries aren't yet updated for the new schema. addUnmatchedEntries() is already disabled
        // so we can just ignore failures here. This whole method will go away under Story 452145.
        LOG.warn("Ignoring failure while attempting to create \"other change\" rematches.", ex);
      }
        
    }
    
    @Override
    public List<RematchVo> retrieveUnmatched(Long matchTypeId) {
      
      List<EplRematchDo> rematchSuggestions = eplRematchDao.retrieveByMatchType(matchTypeId);
      return rematchConverter.convert(rematchSuggestions);         
      
  }
    
    @Override   
    public List<UnmatchRematchVo> findRematchProducts(Long productId) {
        return eplRematchDao.findRematchProducts(productId);
    } 
    
    
    private void addUnmatchedEntries(List<Object[]> addEntries, UserVo user ){
        // This code is only kept for reference. Should be removed under Story 452145
        /*
        if (addEntries!= null && !addEntries.isEmpty()){
            for (Object[] addEntry : addEntries) {
                String oldIen = (String)addEntry[0];
                BigDecimal matchType = (BigDecimal)addEntry[1];
                RematchVo unmatch = new RematchVo(); 
                throwing exception as this is not supported
                unmatch.setOldProductId(oldIen);
                unmatch.setMatchReason(createMatchType(matchType.longValue()));
                               
                insert(unmatch, user) ;              
            }
        }
        */
    }
    
    private void removeObsoleteEntries(List<Object[]> obsoleteEntries){
        
        if (obsoleteEntries!= null && !obsoleteEntries.isEmpty()){
            for (Object[] obsoleteEntry : obsoleteEntries) {
                BigDecimal id = (BigDecimal)obsoleteEntry[1];
                eplRematchDao.delete("eplRematchSuggestionId", id.longValue());                
            }
        }
        
    }
    
    @Override
    public Optional<RematchVo> getRematchForProduct(long productId) {
      
      Optional<EplRematchDo> retrievedRematch = eplRematchDao.retrieveForProduct(productId);
      if (retrievedRematch.isPresent()) {
        return Optional.of(rematchConverter.convert(retrievedRematch.get()));
      } else {
        return Optional.empty();
      }
    }
}
