package gov.va.med.pharmacy.peps.domain.common.dao.impl;


import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.hibernate.Criteria;
import org.hibernate.FlushMode;
import org.hibernate.ObjectNotFoundException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.criterion.Restrictions;
import org.springframework.beans.factory.annotation.Autowired;

import gov.va.med.pharmacy.peps.common.utility.VistaFileNumber;
import gov.va.med.pharmacy.peps.common.vista.VistaProductFile;
import gov.va.med.pharmacy.peps.common.vo.NdfUpdateFileVo;
import gov.va.med.pharmacy.peps.domain.common.dao.EplNdfOutgoingDifferencesDao;
import gov.va.med.pharmacy.peps.domain.common.model.EplNdfOutgoingDifferences;
import gov.va.med.pharmacy.peps.domain.common.utility.converter.NdfUpdateFileConverter;
import gov.va.med.pharmacy.peps.external.common.utility.MumpsConverter;
import gov.va.med.pharmacy.peps.service.common.update.CopayTierChange;

/**
 * Main implementation of {@link EplNdfOutgoingDifferencesDao}.
 */
//@Repository
public class EplNdfOutgoingDifferencesDaoImpl implements EplNdfOutgoingDifferencesDao {

    private static final org.apache.logging.log4j.Logger LOG = org.apache.logging.log4j.LogManager.getLogger(
        EplNdfOutgoingDifferencesDaoImpl.class);
    
    /** The session factory. */
    @Autowired
    private SessionFactory sessionFactory;

    /** The ndf update file converter. */
    private NdfUpdateFileConverter ndfUpdateFileConverter;
    
    /**
     * Find by id.
     *
     * @param id the id
     * @return the epl ndf outgoing differences
     */
    @Override
    public EplNdfOutgoingDifferences findById(Long id){
        if(id == null){
            throw new IllegalArgumentException("id cannot be null");
        }
        Session session = getSessionFactory().getCurrentSession();
        EplNdfOutgoingDifferences eplNdfOutgoingDifferences =
            (EplNdfOutgoingDifferences) session.get(EplNdfOutgoingDifferences.class, id);
        return eplNdfOutgoingDifferences;
    }
    
    /**
     * Find by id.
     *
     * @param date the date
     * @return the epl ndf outgoing differences
     */
    @Override
    public List<EplNdfOutgoingDifferences> retrieveDifferencesForOutputFile(Date date){ 
        
        Timestamp startDate = new Timestamp(date.getTime());
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("retrieveDifferencesForOutputFile")
          .setTimestamp("startDate", startDate);
            
        List<EplNdfOutgoingDifferences> returnedValues = query.list();
        
        return returnedValues;
    }
    
    /**
     * Find by file and field number for data that is not yet associated to an output file.
     *
     * @param vistaFileNumber the vista file number
     * @param vistaFieldNumber the vista field number
     * @param startDTM the start dtm
     * @return the epl ndf outgoing differences
     */
    @Override
    public List<Object[]> findByFileAndField(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("findByFileAndField")
            .setString("vistaFileNumber", vistaFileNumber)
            .setString("vistaFieldNumber", vistaFieldNumber)
            .setTimestamp("startDate", startDate);
            
        List<Object[]> returnedValues = query.list();
                
        return returnedValues;
    }
       
    @Override
    public List<Object[]> findDrugClassChanges(String vistaFileNumber, String vistaFieldNumber, Date startDTM) {
        
        Timestamp startDate = new Timestamp(startDTM.getTime());
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("findDrugClassChanges")
            .setString("vistaFileNumber", vistaFileNumber)
            .setString("vistaFieldNumber", vistaFieldNumber)
            .setTimestamp("startDate", startDate);

        List<Object[]> returnedValues = query.list();
        
        return returnedValues;
    }
    
    /**
     * Find by file,field, and ien were ndf update file is null.
     *
     * @param vistaFileNumber the vista file number
     * @param vistaFieldNumber the vista field number
     * @param vistaIen the vista ien
     * @return the epl ndf outgoing differences
     */
    @Override
    public EplNdfOutgoingDifferences findByFileFieldIen(String vistaFileNumber, String vistaFieldNumber, String vistaIen){
    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("findByFileFieldIen")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)
          .setString("vistaIen", vistaIen);
            
        EplNdfOutgoingDifferences returnedValue = (EplNdfOutgoingDifferences) query.uniqueResult();

        
        return returnedValue;    
   }
    
    @Override
    public EplNdfOutgoingDifferences findModifiedRecords(
        String vistaFileNumber, String vistaFieldNumber, String vistaIen, String actionType) {
    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("findModifiedRecords")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)
          .setString("vistaIen", vistaIen)
          .setString("actionType", actionType);
            
        EplNdfOutgoingDifferences returnedValue = (EplNdfOutgoingDifferences) query.uniqueResult();
        
        return returnedValue;    
   }

  @Override
  public List<EplNdfOutgoingDifferences> findPendingRecordsForFieldAndIen(String vistaFileNumber, String vistaFieldNumber, 
      String vistaIen) {
    Criteria criteria = getCriteria();
    criteria.add(Restrictions.isNull(EplNdfOutgoingDifferences.UPDATE_FILE_FK));
    criteria.add(Restrictions.eq(EplNdfOutgoingDifferences.VISTA_FILE_NUMBER, vistaFileNumber));
    criteria.add(Restrictions.eq(EplNdfOutgoingDifferences.VISTA_FIELD_NUMBER, vistaFieldNumber));
    criteria.add(Restrictions.eq(EplNdfOutgoingDifferences.VISTA_IEN, vistaIen));

    @SuppressWarnings("unchecked")
    List<EplNdfOutgoingDifferences> results = criteria.list();
    return results;
  }    
    
    /**
     * Create a {@link Criteria} without an alias.
     * 
     * @return {@link Criteria}
     */
    private Criteria getCriteria() {
        return getSessionFactory().getCurrentSession().createCriteria(EplNdfOutgoingDifferences.class);
    }
    
    /**
     * Find inactive generic name by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<String> findInactiveGenericName(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("genericNameInactivateReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)     
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<String>) query.list();
   } 
    
    /**
     * Find override exclusion by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findOverrideDFChkExcl(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("overrideDFChkExclReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)  
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find message report data by by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> messageGenericChange(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("messageGenericChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)     
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findCMOPIdChanges(Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("findCMOPIdChanges")
          .setString("vistaFileNumber",VistaProductFile.FILE_NUMBER)
          .setString("vistaFieldNumber", VistaProductFile.VAPRODUCTIDENTIFIER_FIELD_NUMBER)               
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find drug unit change by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> drugUnitsChange(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("drugUnitsChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)    
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find all generic name change by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> genNameAllChange(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("genericNameAllChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber) 
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find dosage form by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> dosageFormChange(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("dosageFormChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber) 
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find dispense unit change by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> dispenseUnitsChange(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("dispenseUnitsChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)   
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find class change by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> classChange(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("classChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)   
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find schedule change by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> scheduleAllChange(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("scheduleAllChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber) 
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find active national formulary product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findActiveNatFormularyProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("activeNatFormularyChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)     
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find find added med guide product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findAddedMedGuideProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("addedMedGuideReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)   
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find Added Drug Drug Interactions by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findAddedDDIs(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("addedDDIReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)  
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find Reactivated Drug Drug Interactions by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findReactivatedDDIs(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("reactivatedDDIReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)    
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find edited med guide product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findEditedMedGuideProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("editedMedGuideReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber) 
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find edited med guide product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findEditedDDIs(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("editedDDIReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)     
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find deleted med guide product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findDeletedMedGuideProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("deletedMedGuideReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)   
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find deleted med guide product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findInactivatedDDIs(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("inactivatedDDIReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)    
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   } 
    
    /**
     * Find excluded DDI product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findExcludedDDIProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("excludedDDIReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)      
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }  
    
    /**
     * Find excluded DDI product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findPrevExcludedDDIProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("prevExcludedDDIReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)      
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }  
    
    /**
     * Find active national formulary product by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findInactiveNatFormularyProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("inactiveNatFormularyChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)   
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findInactivatedCMOPProducts(Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());      
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("inactivatedCMOPReport")
          .setString("vistaFileNumber",VistaProductFile.FILE_NUMBER)
          .setString("vistaFieldNumber", VistaProductFile.CMOPINDICATOR_FIELD_NUMBER)                
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();             
   }
    
    /**
     * Find in-activated CMOP by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findReactivatedProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("reactivatedProductReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find schedule change by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findScheduleChangeProducts(String vistaFileNumber, String vistaFieldNumber){
    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("scheduleChangeReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)          
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find possible dosage change to NO by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findNoPossibleDosageProducts(String vistaFileNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());      
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("noPossibleDosageReport")
          .setString("vistaFileNumber",vistaFileNumber)    
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find possible dosage change to YES by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findYesPossibleDosageProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("yesPossibleDosageReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)     
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find unmatched products by file number.
     *
     * @param vistaFileNumber the file number
     * @return fileIEN
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findUnmatchedProducts(String vistaFileNumber){
    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("unmatchedProducts")
          .setString("vistaFileNumber",vistaFileNumber)                  
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findInactivatedProductsForLocal(String vistaFileNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());     
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("inactivatedProductsForLocal")
          .setString("vistaFileNumber",vistaFileNumber)      
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findInactivatedProducts(Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("inactivatedProductReport")
          .setString("vistaFileNumber", VistaProductFile.FILE_NUMBER)   
          .setString("vistaFieldNumber", VistaProductFile.STATUS_FIELD_NUMBER)
          .setTimestamp("startDate", startDate)
          // This should not be necessary here, but Hibernate was not doing the necessary flushing before running this
          // query.
          .setFlushMode(FlushMode.ALWAYS)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    /**
     * Find added products by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list
     */
    @SuppressWarnings("unchecked")
    @Override
    public List<Object[]> findAddedProducts(String vistaFileNumber, String vistaFieldNumber, Date startDTM){
        
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("addedProductsReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
   }
    
    
    @Override
    public EplNdfOutgoingDifferences save(EplNdfOutgoingDifferences eplNdfOutgoingDifferences) {        
        EplNdfOutgoingDifferences differences = new EplNdfOutgoingDifferences();
        Session session = getSessionFactory().getCurrentSession();
        Long id = (Long) session.save(eplNdfOutgoingDifferences);
        differences = findById(id);
        return differences;
    }
    
    @Override
    public EplNdfOutgoingDifferences update(EplNdfOutgoingDifferences eplNdfOutgoingDifferences) {
        Session session = getSessionFactory().getCurrentSession();
        session.update(eplNdfOutgoingDifferences);
        return eplNdfOutgoingDifferences;
    }
    
  @Override
  public void delete(EplNdfOutgoingDifferences diff) {
    // copied from DataAccessObjectImpl
    if (diff != null) {
      Session session = getSessionFactory().getCurrentSession();
      Object merged;

      try {
        merged = session.merge(diff);
      } catch (ObjectNotFoundException e) {
        LOG.trace("Unable to merge object, will try to delete it anyway.");
        merged = diff;
      }

      session.delete(merged);
      session.flush();
      session.clear();
    }
  }

    /**
     * Associate the specified outgoing differences records to associate them with the specified update file .
     * 
     * @param diffRecordsToUpdate the outgoing differences records to update
     * @param fileToAssociate the update file to associate
     */

    public void updateNdfUpdateFileFk(List<EplNdfOutgoingDifferences> diffRecordsToUpdate, NdfUpdateFileVo fileToAssociate) {
       
       for (EplNdfOutgoingDifferences eplNdfOutgoingDifferences : diffRecordsToUpdate) {
           
           eplNdfOutgoingDifferences.setNdcUpdateFileFk(ndfUpdateFileConverter.convert(fileToAssociate));
           update(eplNdfOutgoingDifferences);
       }
    }
    
    /**
     * Remove the fk association between outgoing differences table and the specified file on the ndf update file table.
     *
     * @param fileSeqNo the file id to process
     */
    public void removeNdfUpdateFileFk(Long fileSeqNo) {
              
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.createQuery("update EplNdfOutgoingDifferences set ndcUpdateFileFk = null "
            + " where ndcUpdateFileFk = " + fileSeqNo);
        query.executeUpdate();
    }
       

    /**
     * Gets the session factory.
     *
     * @return the session factory
     */
    public SessionFactory getSessionFactory() {
        return sessionFactory;
    }

    /**
     * Sets the session factory.
     *
     * @param sessionFactory the sessionFactory to set
     */
    public void setSessionFactory(SessionFactory sessionFactory) {
        this.sessionFactory = sessionFactory;
    }

    
    /**
     * Gets the ndf update file converter.
     *
     * @return the eplNdfUpdateFileDaoImpl
     */
    public NdfUpdateFileConverter getNdfUpdateFileConverter() {
        return ndfUpdateFileConverter;
    }

    
    /**
     * Sets the ndf update file converter.
     *
     * @param ndfUpdateFileConverter the new ndf update file converter
     */
    public void setNdfUpdateFileConverter(NdfUpdateFileConverter ndfUpdateFileConverter) {
        this.ndfUpdateFileConverter = ndfUpdateFileConverter;
    }

    /**
     * Find added products by file and field number.
     *
     * @param vistaFileNumber the file number
     * @param vistaFieldNumber the field number
     * @param startDTM the start dtm
     * @return report data list, including classification
     */
    @Override
    public List<Object[]> findNewClassProducts(
            String vistaFileNumber, String vistaFieldNumber, Date startDTM) {
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        Query query = session.getNamedQuery("newClassReport")
          .setString("vistaFileNumber",vistaFileNumber)
          .setString("vistaFieldNumber", vistaFieldNumber)
          .setTimestamp("startDate", startDate)
          .setReadOnly(true);
        
        return (List<Object[]>) query.list();
    }
    
  /**
   * The native SQL query to retrieve copay tier changes.
   */
  // Note: normally constants go at the top of the file but keep this one close to the tightly-coupled code that uses it.
  // Also, normally queries were defined in the model object, but starting to move them here to keep field mappings closer
  // to the queries being mapped. 
  private static final String COPAY_TIER_CHANGE_REPORT_QUERY =
      // The VISTA_IEN column contains both the product IEN and copay tier index/pseudo-IEN.
      "SELECT REGEXP_SUBSTR(tierLevelDiff.vista_ien, '[^,]+') AS productIen,"
      + "   prod.VA_PRODUCT_NAME AS productName, prod.CMOP_ID AS cmopId,"
      + "   TO_NUMBER(SUBSTR(tierLevelDiff.vista_ien, INSTR(tierLevelDiff.vista_ien, ',') + 1)) AS copayTierIndex,"
      + "   tierLevelDiff.new_value AS copayTierValue,"
      + "   effectiveDateDiff.new_value AS effectiveDateValue,"
      + "   endDateDiff.new_value AS endDateValue"
      + " FROM PPSNEPL.EPL_NDF_OUTGOING_DIFFERENCES tierLevelDiff"
      + " LEFT JOIN PPSNEPL.EPL_PRODUCTS prod ON prod.NDF_PRODUCT_IEN = REGEXP_SUBSTR(tierLeveLDiff.vista_ien, '[^,]+')"
      // Left join to get the possibly-present effective and end dates in the same query.
      + " LEFT JOIN PPSNEPL.EPL_NDF_OUTGOING_DIFFERENCES effectiveDateDiff ON ("
      + "   effectiveDateDiff.VISTA_IEN = tierLevelDiff.vista_ien AND"
      + "   effectiveDateDiff.NDC_UPDATE_FILE_FK IS NULL AND"
      + "   effectiveDateDiff.VISTA_FILE_NUMBER = tierLevelDiff.VISTA_FILE_NUMBER AND"
      + "   effectiveDateDiff.VISTA_FIELD_NUMBER = :effectiveDateFieldNumber AND"
      + "   effectiveDateDiff.CREATED_DTM <= :startDate)"
      + " LEFT JOIN PPSNEPL.EPL_NDF_OUTGOING_DIFFERENCES endDateDiff ON ("
      + "   endDateDiff.VISTA_IEN = tierLevelDiff.vista_ien AND"
      + "   endDateDiff.NDC_UPDATE_FILE_FK IS NULL AND"
      + "   endDateDiff.VISTA_FILE_NUMBER = tierLevelDiff.VISTA_FILE_NUMBER AND"
      + "   endDateDiff.VISTA_FIELD_NUMBER = :endDateFieldNumber AND"
      + "   endDateDiff.CREATED_DTM <= :startDate)"
      + " WHERE tierLevelDiff.NDC_UPDATE_FILE_FK IS NULL"
      + "   AND tierLevelDiff.VISTA_FILE_NUMBER = :vistaFileNumber"
      + "   AND tierLevelDiff.VISTA_FIELD_NUMBER = :tierLevelFieldNumber"
      + "   AND tierLevelDiff.created_Dtm <= :startDate"
      // Order by product name per business requirement, then IEN in case two products possibly
      // have the same names. We don't actually care about IEN order, just that the IENs are grouped
      // together.
      + " ORDER BY productName ASC, productIen, copayTierIndex ASC";

  /**
   * Factory method to create a {@link CopayTierChange} from a row returned from Hibernate.
   *
   * @param row the row
   * @return the copay tier change
   */
  private static CopayTierChange makeCopayTierChangeFromRow(Object[] row) {
    final String productIen = (String)row[0];  // The IEN comes back as a String, should not be null.
    final String productName = (String)row[1]; // Product Name may be null if we didn't match.
    final String cmopId = (String)row[2];      // CMOP ID may be null even if we did get a match.
    // Note: we ignore column 3 (copay tier index).
    final String copayTierValue = (String)row[4];     // Copay tier should never be null.
    final String effectiveDateValue = (String)row[5]; // Effective date will be null if it's a DELETED row.
    final String endDateValue = (String)row[6];       // End date may be null.
    return new CopayTierChange(
      productIen,
      StringUtils.defaultString(productName, "Unknown"),
      StringUtils.defaultString(cmopId, ""),
      copayTierValue,
      (effectiveDateValue == null) ? Optional.empty() : Optional.of(MumpsConverter.mumpsToLocalDate(effectiveDateValue)),
      (endDateValue == null) ? Optional.empty() : Optional.of(MumpsConverter.mumpsToLocalDate(endDateValue)));
  }
  
    /* (non-Javadoc)
     * @see gov.va.med.pharmacy.peps.domain.common.dao.EplNdfOutgoingDifferencesDao#findCopayTierChanges(java.util.Date)
     */
    @Override
    public List<CopayTierChange> findCopayTierChanges(Date startDTM) {
        Timestamp startDate = new Timestamp(startDTM.getTime());    
        Session session = getSessionFactory().getCurrentSession();
        
        // Note: we rely on this query returning the data in the necessary order.
        Query query = session.createSQLQuery(COPAY_TIER_CHANGE_REPORT_QUERY)
            .addScalar("productIen")
            .addScalar("productName")
            .addScalar("cmopId")
            .addScalar("copayTierIndex")
            .addScalar("copayTierValue")
            .addScalar("effectiveDateValue")
            .addScalar("endDateValue")
            // Native Query: Hibernate doesn't always flush beforehand. See 
            // <https://dzone.com/articles/dark-side-hibernate-auto-flush>.
            .setFlushMode(FlushMode.ALWAYS)  
            .setString("vistaFileNumber", VistaFileNumber.FIFTY_68.getFileNumber())
            .setString("tierLevelFieldNumber", VistaProductFile.COPAY_TIER_FIELD_NUMBER)
            .setString("effectiveDateFieldNumber", VistaProductFile.COPAY_TIER_START_DATE_FIELD_NUMBER)
            .setString("endDateFieldNumber", VistaProductFile.COPAY_TIER_END_DATE_FIELD_NUMBER)
            .setTimestamp("startDate", startDate)
            .setReadOnly(true);
        
        final List<?> rows = query.list();

        // Convert the Object arrays into CopayTierChange objects.
        final ArrayList<CopayTierChange> changes = new ArrayList<CopayTierChange>(rows.size());
        for (final Object row : rows) {
          changes.add(makeCopayTierChangeFromRow((Object[])row));
        }
        
        return changes;
    }    

  @Override
  public int removePendingOutgoingDifferencesForMultipleField(String vistaFileNumber, String vistaFieldNumberBase,
      String vistaIen) {
    int recordsAffected;
    Session session = getSessionFactory().getCurrentSession();
    Query query = session.getNamedQuery("removePendingOutgoingDifferencesForMultipleField")
        .setString("vistaFileNumber", vistaFileNumber)
        .setString("vistaFieldNumberBase", vistaFieldNumberBase + ",%")
        .setString("vistaIen", vistaIen + ",%");

    recordsAffected = query.executeUpdate();
    return recordsAffected;
  }
  
  /**
   * Find Formulary Designator product by file and field number.
   * @param startDTM the start dtm
   * @param vistaFileNumber the file number
   * @param vistaFieldNumber the field number
   *
   * @return report data list
   */
  @SuppressWarnings("unchecked")
  @Override
  public List<Object[]> findFormularyDesignatorProducts(Date startDTM){
      
      Timestamp startDate = new Timestamp(startDTM.getTime());    
      Session session = getSessionFactory().getCurrentSession();
      Query query = session.getNamedQuery("formularyDesignatorChangeReport")
          .setString("vistaFileNumber",VistaProductFile.FILE_NUMBER)
          .setString("vistaFieldNumber", VistaProductFile.FORMULARY_DESIGNATOR_FIELD_NUMBER)     
        .setTimestamp("startDate", startDate)
        .setReadOnly(true);
      
      return (List<Object[]>) query.list();
  }
  
}
