package gov.va.med.pharmacy.peps.domain.common.utility.converter;

import java.util.HashSet;
import java.util.List;

import gov.va.med.pharmacy.peps.common.vo.MatchReason;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.RematchVo;
import gov.va.med.pharmacy.peps.domain.common.model.DataObject;
import gov.va.med.pharmacy.peps.domain.common.model.EplMatchTypeDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplProductDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplRematchDo;

/**
 * Convert to/from {@link RematchVo} and {@link EplRematchDo}.
 */
public class RematchConverter extends Converter<RematchVo, EplRematchDo>{
    
    private NdfUpdateFileConverter ndfUpdateFileConverter;
    private ProductConverter productConverter;
    
    /**
     * Fully copies data from the given RematchVo into a {@link DataObject}.
     * 
     * @param valueObj RematchVo to convert
     */
    @Override
    protected EplRematchDo toDataObject(RematchVo valueObj) {
        EplRematchDo dataObj = new EplRematchDo();
        
        setDataObjectPropertiesFrom(dataObj, valueObj);
    
        return dataObj;
    }


    /**
     * Sets the properties on the given {@link EplRematchDo} from the given {@link RematchVo}.
     * @param dataObj must not be null
     * @param valueObj must not be null
     */
    public void setDataObjectPropertiesFrom(final EplRematchDo dataObj, final RematchVo valueObj) {
      if(valueObj.getEplRematchId() != null ){
          dataObj.setEplRematchId(Long.valueOf(valueObj.getEplRematchId()));            
      }

      // Need to use convertMinimal because the product is only guaranteed to be minimally populated.
      dataObj.setFromProduct(productConverter.convertMinimal(valueObj.getFromProduct()));

      if (valueObj.getToAutomaticProduct() != null) {
        // Need to use convertMinimal because product is only guaranteed to be minimally populated.
        final EplProductDo toProduct = productConverter.convertMinimal(valueObj.getToAutomaticProduct());
        dataObj.setToAutoProduct(toProduct);  
      }
      else {
        final List<EplProductDo> toProducts = productConverter.convertMinimalVOs(valueObj.getToSuggestedProducts());
        dataObj.setToSuggestedProducts(new HashSet<>(toProducts));
      }
      
      if (valueObj.getMatchReason() != null) {           
        dataObj.setMatchType(
            // We only need the ID in this object to link it in the database. Don't even populate the other
            // properties.
            new EplMatchTypeDo(valueObj.getMatchReason().getId(), "", ""));
      }
      
      if (valueObj.getNdfUpdateFile() != null) { 
          dataObj.setNdfUpdateFile(ndfUpdateFileConverter.convert(valueObj.getNdfUpdateFile()));            
      }
    }

  @Override
  protected RematchVo toValueObject(EplRematchDo dataObj) {
    // These properties are not nullable.
    ProductVo fromProduct = productConverter.convertMinimal(dataObj.getFromProduct());
    MatchReason matchReason = getMatchReasonById(dataObj.getMatchType().getMatchTypeId());
    
    // Create an automatic or suggested rematch, depending on whether there is an automatic product.
    RematchVo valueObj = (dataObj.getToAutoProduct() != null) 
        ? RematchVo.automaticRematch(
            fromProduct, productConverter.convertMinimal(dataObj.getToAutoProduct()), matchReason) 
        : RematchVo.suggestedRematch(
            fromProduct, new HashSet<>(productConverter.convertMinimal(dataObj.getToSuggestedProducts())), matchReason);
            
    // This property is not nullable
    valueObj.setEplRematchId(dataObj.getEplRematchId().toString());
    
    if (dataObj.getNdfUpdateFile() != null) {
      valueObj.setNdfUpdateFile(ndfUpdateFileConverter.convert(dataObj.getNdfUpdateFile()));
    }
    return valueObj;
  }


     /**
     * Returns the {@link MatchReason} for the given ID.
     * @return never null
     * @throws IllegalArgumentException if there is no MatchReason for the given ID. (This would be a coding error
     * because {@link MatchReason} should have all reasons defined in the database.)
     */
    private MatchReason getMatchReasonById(final long id) {
      for (final MatchReason matchReason : MatchReason.values()) {
        if (matchReason.getId() == id) {
          return matchReason;
        }
      }
      
      throw new IllegalArgumentException("Could not find MatchReason for ID " + id);
    }


    public NdfUpdateFileConverter getNdfUpdateFileConverter() {
        return ndfUpdateFileConverter;
    }

    
    public void setNdfUpdateFileConverter(NdfUpdateFileConverter ndfUpdateFileConverter) {
        this.ndfUpdateFileConverter = ndfUpdateFileConverter;
    }

    public ProductConverter getProductConverter() {
      return productConverter;
    }

    public void setProductConverter(ProductConverter productConverter) {
      this.productConverter = productConverter;
    }

}
