package gov.va.med.pharmacy.peps.domain.common.utility.impl;


import gov.va.med.pharmacy.peps.common.utility.EnvironmentUtility;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.common.vo.ItemAuditHistoryVo;
import gov.va.med.pharmacy.peps.common.vo.ModDifferenceVo;
import gov.va.med.pharmacy.peps.common.vo.NotificationVo;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.RequestItemStatus;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.Difference;
import gov.va.med.pharmacy.peps.domain.common.capability.NotificationDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.dao.EplProductDao;
import gov.va.med.pharmacy.peps.domain.common.model.EplProductDo;
import gov.va.med.pharmacy.peps.domain.common.utility.RxNormCapability;
import gov.va.med.pharmacy.peps.domain.common.utility.client.rxnorm.RxNormClient;
import gov.va.med.pharmacy.peps.domain.common.utility.converter.ProductConverter;
import gov.va.med.pharmacy.peps.service.common.update.NdfFileSyncProcessor;
import gov.va.med.pharmacy.peps.service.common.update.NdfFileSyncProcessorFactory;

import static gov.va.med.pharmacy.peps.common.utility.security.ESAPIValidationType.LOG_FORGING;
import static gov.va.med.pharmacy.peps.common.utility.security.ESAPIValidator.validateStringInput;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.logging.log4j.Logger;
import org.hibernate.Hibernate;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;


/**
 * This class handles the process of keeping RxNorm data up to date on the EPL_PRODUCTS table.
 * It uses RxNormClient to execute two web service calls to NLM and the product DAO to populate 
 * the data into the appropriate columns.
 * 
 * @author K. Edgecomb
 *
 */
@Component
public class RxNormCapabilityImpl implements RxNormCapability {

  private static final Logger LOG = org.apache.logging.log4j.LogManager.getLogger(RxNormCapabilityImpl.class);
  
    private RxNormClient rxNormClient;
    private EplProductDao eplProductDao;
    private ProductConverter productConverter;
    private NdfFileSyncProcessorFactory syncProcessorFactory;
    private NotificationDomainCapability notificationDomainCapability;
    private EnvironmentUtility environmentUtility;

    
    /* (non-Javadoc)
	 * @see gov.va.med.pharmacy.peps.domain.common.utility.impl.RxNormCapability#populateRxNormForProduct(gov.va.med.pharmacy.peps.common.vo.UserVo, java.lang.Long)
	 */
    @Override
	@Transactional
    public long populateRxNormForProduct(UserVo user, Long productId) {
    	Date runDate = new Date();
        EplProductDo product = eplProductDao.retrieve(productId);        
        populateAndPersist(user, product, runDate);
        return 0;
    }

    
	/* (non-Javadoc)
	 * @see gov.va.med.pharmacy.peps.domain.common.utility.RxNormCapability#fetchRxNormUpdateProducts()
	 */
	@Transactional
    public List<EplProductDo> fetchRxNormUpdateProducts() {
        List<EplProductDo> products = eplProductDao.retrieve();
	     // List<EplProductDo> products = eplProductDao.retrieveTestProductList();
        return products;
    }

    
    /* (non-Javadoc)
     * @see gov.va.med.pharmacy.peps.domain.common.utility.RxNormCapability#populateAndPersist(gov.va.med.pharmacy.peps.common.vo.UserVo, gov.va.med.pharmacy.peps.domain.common.model.EplProductDo, java.util.Date)
     */
    public void populateAndPersist(UserVo user, EplProductDo productDo, Date runDate) {

      LOG.debug("Start RX Norm update for product: " + validateStringInput(productDo.getVaProductName(), LOG_FORGING) 
                                                     + validateStringInput(", VUID: ", LOG_FORGING) + productDo.getVuid());
    	
        
        // Call the ISAAC web service with VUID to get a product if available
        String productCode = rxNormClient.getCode(productDo.getVuid());
    
        LOG.debug("Product Code System Code from ISAAC: " + validateStringInput(productCode, LOG_FORGING)); 
        // If the returned data is the same as the current product data, don't push to Vista
        
        List<Difference> differences = getDifferences(productCode, productDo);
      	if (!differences.isEmpty()) {
      	  
      	  LOG.debug("Processing differences on Code System info for: " + validateStringInput(productDo.getVaProductName(), LOG_FORGING) 
                    + validateStringInput(", VUID: ", LOG_FORGING) + productDo.getVuid());
      	  
          // Saving audit history records
          ProductVo productVo = getProductConverter().toCodingSystemValueObject(productDo);
          productVo.updateFromModDifferences(
                                          ModDifferenceVo.asModDifferenceList(differences, true, "Coding System Modification"), 
                                          user, 
                                          getEnvironmentUtility());
          
          // Save the Product to the PPS-N Database          
          productDo.setCodingSystemCode(productCode);
          if (productCode != null)
            productDo.setCodingSystem(EplProductDo.RX_NORM_CODING_SYSTEM);
          else
            productDo.setCodingSystem(null);
          productDo.setCodingSystemLastUpdateDate(runDate); 
          eplProductDao.update(productDo, user);
          
          saveItemAuditHistoryRecords(productVo.getNewAuditHistory(), user);
          
          // Sending modifications to VISTA                                     
          NdfFileSyncProcessor fileProcessor = getSyncProcessorFactory().getProcessor(productVo);
          //set the request item statUs to approve so the VISTA update gets processed without waiting for pending changes
          productVo.setRequestItemStatus(RequestItemStatus.APPROVED);
          fileProcessor.processModified(productVo, user, differences);
      	}
      	else{
      	  LOG.debug("No differences found in Product Code System information for: "+ validateStringInput(productDo.getVaProductName(), LOG_FORGING) 
                    + validateStringInput(", VUID: ", LOG_FORGING) + productDo.getVuid());
      	}            

        LOG.debug("Finish RX Norm update for product: " + validateStringInput(productDo.getVaProductName(), LOG_FORGING) 
                  + validateStringInput(", VUID: ", LOG_FORGING) + productDo.getVuid());
    }    
    
    
    private void saveItemAuditHistoryRecords(ItemAuditHistoryVo itemAudit, UserVo user) {
      if (!(itemAudit == null)) {
        NotificationVo notification = new NotificationVo();
        ArrayList<ItemAuditHistoryVo> itemAudits = new ArrayList<ItemAuditHistoryVo>();
        itemAudits.add(itemAudit);
        notification.setItemAudits(itemAudits);
        notification.determineNotificationType();
        notificationDomainCapability.create(notification, user);
      }
    }
    

    /**
     * Gets the differences between the ISAAC service code and the one currently stored in the 
     * product from the EPL_PRODUCTS table, uf any.
     * 
     * @param code
     * @param product
     * @return
     */
    public List<Difference> getDifferences (String code, EplProductDo product) {
      
      List<Difference> differences = new ArrayList<Difference>();
      
      if (different(product.getCodingSystemCode(), product.getCodingSystem(), code)){
        differences.add(new Difference(FieldKey.PRODUCT_CODING_SYTEM_CODE, product.getCodingSystemCode(), code));
        if (code == null)
          differences.add(new Difference(FieldKey.PRODUCT_CODING_SYTEM, product.getCodingSystem(), null));
        else
          differences.add(new Difference(FieldKey.PRODUCT_CODING_SYTEM, product.getCodingSystem(), EplProductDo.RX_NORM_CODING_SYSTEM));
      }      
          	            	
    	return differences;
    }

    private boolean different(String productCode, String codingSystem, String isaacCode){
      if (productCode == null && isaacCode == null)
        return false;
      
      if (productCode == null && isaacCode != null)
        return true;
      
      if (productCode != null && isaacCode == null)
        return true;
      
      if (codingSystem != null && !codingSystem.equals(EplProductDo.RX_NORM_CODING_SYSTEM))
        return true;
      
      return !productCode.trim().equals(isaacCode.trim());
    }
    

	/* (non-Javadoc)
	 * @see gov.va.med.pharmacy.peps.domain.common.utility.RxNormCapability#getRxNormClient()
	 */
	public RxNormClient getRxNormClient() {
		return rxNormClient;
	}
	/* (non-Javadoc)
	 * @see gov.va.med.pharmacy.peps.domain.common.utility.impl.RxNormCapability#setEplProductDao(gov.va.med.pharmacy.peps.domain.common.dao.EplProductDao)
	 */
	@Override
	public void setEplProductDao(EplProductDao eplProductDao) {
		this.eplProductDao = eplProductDao;
	}


	/* (non-Javadoc)
	 * @see gov.va.med.pharmacy.peps.domain.common.utility.RxNormCapability#getEplProductDao()
	 */
	public EplProductDao getEplProductDao() {
		return eplProductDao;
	}
	/* (non-Javadoc)
	 * @see gov.va.med.pharmacy.peps.domain.common.utility.impl.RxNormCapability#setRxNormClient(gov.va.med.pharmacy.peps.domain.common.utility.client.rxnorm.RxNormClient)
	 */
	@Override
	public void setRxNormClient(RxNormClient rxNormClient) {
		this.rxNormClient = rxNormClient;
	}


  public ProductConverter getProductConverter() {
    return productConverter;
  }


  public void setProductConverter(ProductConverter productConverter) {
    this.productConverter = productConverter;
  }


  public NdfFileSyncProcessorFactory getSyncProcessorFactory() {
    return syncProcessorFactory;
  }


  public void setSyncProcessorFactory(NdfFileSyncProcessorFactory syncProcessorFactory) {
    this.syncProcessorFactory = syncProcessorFactory;
  }


  public NotificationDomainCapability getNotificationDomainCapability() {
    return notificationDomainCapability;
  }


  public void setNotificationDomainCapability(NotificationDomainCapability notificationDomainCapability) {
    this.notificationDomainCapability = notificationDomainCapability;
  }


  public EnvironmentUtility getEnvironmentUtility() {
    return environmentUtility;
  }


  public void setEnvironmentUtility(EnvironmentUtility environmentUtility) {
    this.environmentUtility = environmentUtility;
  }

}
