package gov.va.med.pharmacy.peps.service.common.capability.impl;

import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.springframework.transaction.annotation.Transactional;

import gov.va.med.pharmacy.peps.common.exception.ValidationException;
import gov.va.med.pharmacy.peps.common.vista.VistaProductFile;
import gov.va.med.pharmacy.peps.domain.common.dao.EplNdfOutgoingDifferencesDao;
import gov.va.med.pharmacy.peps.domain.common.model.EplNdfOutgoingDifferences;
import gov.va.med.pharmacy.peps.service.common.capability.NdfFileSyncCapability;



/**
 * Contains methods for managing the outgoing records to sync with VistA in an update file.
 */
public class NdfFileSyncCapabilityImpl implements NdfFileSyncCapability{
  
  private static final org.apache.logging.log4j.Logger LOG = org.apache.logging.log4j.LogManager
      .getLogger(NdfFileSyncCapabilityImpl.class);
  
    private EplNdfOutgoingDifferencesDao eplNdfOutgoingDifferencesDao;
    
  @Override
  @Transactional
  // TODO: Transaction seems unnecessary at this level
  public void saveNewItem(EplNdfOutgoingDifferences eplNdfOutgoingDifferences) throws ValidationException {
    //For now, we branch off for CMOP Indicator and VA Product Id (CMOP Id), 
    //but eventually, all fields should be treated the same
    if (VistaProductFile.FILE_NUMBER.equals(eplNdfOutgoingDifferences.getVistaFileNumber())
        && (VistaProductFile.CMOPINDICATOR_FIELD_NUMBER.equals(eplNdfOutgoingDifferences.getVistaFieldNumber())
        ||  VistaProductFile.VAPRODUCTIDENTIFIER_FIELD_NUMBER.equals(eplNdfOutgoingDifferences.getVistaFieldNumber() ))) {
      updatePendingQueue(eplNdfOutgoingDifferences);
      return;
    }

    eplNdfOutgoingDifferencesDao.save(eplNdfOutgoingDifferences);

  }

  @Override
  @Transactional
  // TODO: Transaction seems unnecessary at this level
  public void saveModifiedItem(EplNdfOutgoingDifferences unsavedRecord) throws ValidationException {
    //For now, we branch off for just a subset of fields to limit the scope of implementing this new pattern, 
    //but eventually, all fields should be treated the same
    if (VistaProductFile.FILE_NUMBER.equals(unsavedRecord.getVistaFileNumber())
        && (VistaProductFile.CMOPINDICATOR_FIELD_NUMBER.equals(unsavedRecord.getVistaFieldNumber())
        ||  VistaProductFile.VAPRODUCTIDENTIFIER_FIELD_NUMBER.equals(unsavedRecord.getVistaFieldNumber())
        ||  VistaProductFile.PRODUCT_TEXT_FIELD_NUMBER.equals(unsavedRecord.getVistaFieldNumber())
        ||  VistaProductFile.FORMULARY_DESIGNATOR_FIELD_NUMBER.equals(unsavedRecord.getVistaFieldNumber()) )) {
      updatePendingQueue(unsavedRecord);
      return;
    }

    EplNdfOutgoingDifferences existingRecord = eplNdfOutgoingDifferencesDao.findModifiedRecords(
        unsavedRecord.getVistaFileNumber(), unsavedRecord.getVistaFieldNumber(), unsavedRecord.getVistaIen(),
        unsavedRecord.getActionType());
    if (existingRecord == null) {
      eplNdfOutgoingDifferencesDao.save(unsavedRecord);
    } else {
      existingRecord.setNewValue(unsavedRecord.getNewValue());
      existingRecord.setCreatedBy(unsavedRecord.getCreatedBy());
      existingRecord.setCreatedDtm(unsavedRecord.getCreatedDtm());
      eplNdfOutgoingDifferencesDao.update(existingRecord);
    }
  }

  /**
   * Update pending queue of outgoing difference records - records that have not been included in a file yet.
   * 
   * @param unsavedRecord the unsaved record
   */
  protected void updatePendingQueue(EplNdfOutgoingDifferences unsavedRecord) {
    List<EplNdfOutgoingDifferences> existingRecords = eplNdfOutgoingDifferencesDao.findPendingRecordsForFieldAndIen(
        unsavedRecord.getVistaFileNumber(), unsavedRecord.getVistaFieldNumber(), unsavedRecord.getVistaIen());

    if (existingRecords.isEmpty()) {
      eplNdfOutgoingDifferencesDao.save(unsavedRecord);
    } else {
      consolidatePendingRecords(unsavedRecord, existingRecords);
    }
  }

  /**
   * Consolidate pending records.  The result is that there should only ever be one pending record for a given
   * file/field/ien combination.
   *
   * @param unsavedDifference the unsaved difference
   * @param existingRecords the existing records
   */
  private void consolidatePendingRecords(EplNdfOutgoingDifferences unsavedDifference,
      List<EplNdfOutgoingDifferences> existingRecords) {
    if (existingRecords.size() > 1) {
      String msg = "Unsupported Data Situation - multiple pending EplNdfOutgoingDifferences exist for file: "           
          + unsavedDifference.getVistaFileNumber() 
          + ", field: " + unsavedDifference.getVistaFieldNumber() 
          + ", IEN: " + unsavedDifference.getVistaIen();
      LOG.error(msg);
      throw new IllegalStateException(msg);
    }

    EplNdfOutgoingDifferences existingRecord = existingRecords.get(0);
    //If value was reverted and it is a modify record, then VistA already has the same value as PPS-N.    
    if (wasValueReverted(existingRecord, unsavedDifference)
        && EplNdfOutgoingDifferences.MODIFY_ACTION.equals(existingRecord.getActionType())) {
      // change reverts the value back to where it started, we don't need either record
      eplNdfOutgoingDifferencesDao.delete(existingRecord);
    } else {
      // change does not revert value, or the existing record was an add so update record so latest value gets sent to VistA
      existingRecord.setNewValue(unsavedDifference.getNewValue());
      existingRecord.setCreatedBy(unsavedDifference.getCreatedBy());
      existingRecord.setCreatedDtm(unsavedDifference.getCreatedDtm());
      eplNdfOutgoingDifferencesDao.update(existingRecord);
    }

  }

  /**
   * Determines if value was reverted between two outgoing differences for the same field.
   * The value was reverted if the existing record's old value is equal to the unsaved record's new value OR
   * the existing record's old value is null or empty and the unsaved record's new value is null or empty.
   * The null or empty check is needed, since the existing record from the database may have had an empty string
   * value changed to null per the database vendor.
   *  
   * @param existingRecord the existing record for the field 
   * @param unsavedRecord the unsaved record for the same field
   * @return true, if the value was reverted; false if the value was not reverted.
   */
  private boolean wasValueReverted(EplNdfOutgoingDifferences existingRecord, EplNdfOutgoingDifferences unsavedRecord){
    boolean sameValue = StringUtils.equals(existingRecord.getOldValue(), unsavedRecord.getNewValue());
    boolean bothEmpty = (StringUtils.isEmpty(existingRecord.getOldValue()) && StringUtils.isEmpty(unsavedRecord.getNewValue()));
    return (sameValue || bothEmpty);
  }
  
  @Override
  public void removePendingOutgoingDifferencesForMultipleField(String vistaFileNumber, String vistaFieldNumberBase,
      String vistaIen) {
    eplNdfOutgoingDifferencesDao.removePendingOutgoingDifferencesForMultipleField(vistaFileNumber, vistaFieldNumberBase,
        vistaIen);

  }

    public EplNdfOutgoingDifferencesDao getEplNdfOutgoingDifferencesDao() {
        return eplNdfOutgoingDifferencesDao;
    }

    
    public void setEplNdfOutgoingDifferencesDao(EplNdfOutgoingDifferencesDao eplNdfOutgoingDifferencesDao) {
        this.eplNdfOutgoingDifferencesDao = eplNdfOutgoingDifferencesDao;
    }

}
