package gov.va.med.pharmacy.peps.service.common.update;

import java.time.LocalDate;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.builder.ToStringBuilder;
import org.apache.commons.lang3.builder.ToStringStyle;

/**
 * Represents a changed copay tier record on a product.
 */
public class CopayTierChange {

  private final String productIen;
  private final String productName;
  private final String cmopId;
  private final String copayTier;
  private final Optional<LocalDate> effectiveDate;
  private final Optional<LocalDate> endDate;

  /**
   * Constructs an instance with the given properties.
   * @param pProductIen See {@link #getProductIen()}. Must not be null.
   * @param pProductName See {@link #getProductName()}. Must not be null.
   * @param pCmopId See {@link #getCmopId()}. Must not be null.
   * @param pCopayTier See {@link #getCopayTier()}. Must not be null.
   * @param pEffectiveDate See {@link #getEffectiveDate()}. Must not be null.
   * @param pEndDate See {@link #getEndDate()}. Must not be null.
   */
  public CopayTierChange(
      final String pProductIen,
      final String pProductName,
      final String pCmopId,
      final String pCopayTier,
      final Optional<LocalDate> pEffectiveDate,
      final Optional<LocalDate> pEndDate) {
    productIen = Objects.requireNonNull(pProductIen);
    productName = Objects.requireNonNull(pProductName);
    cmopId = Objects.requireNonNull(pCmopId);
    copayTier = Objects.requireNonNull(pCopayTier);
    effectiveDate = Objects.requireNonNull(pEffectiveDate);
    endDate = Objects.requireNonNull(pEndDate);
  }
  
  /**
   * Returns the Product IEN of the associated product.
   * @return never null
   */
  public String getProductIen() {
    return productIen;
  }

  /**
   * The VA Product Name of the associated product.
   * @return never null
   */
  public String getProductName() {
    return productName;
  }

  /**
   * The CMOP ID of the associated product.
   * @return never null
   */
  public String getCmopId() {
    return cmopId;
  }

  /**
   * The new copay tier level.
   * @return never null
   */
  public String getCopayTier() {
    return copayTier;
  }
  
  /**
   * The optional effective date of the new copay tier level. At time of writing, effective date should always be defined
   * but don't enforce that business rule here.
   * @return never null
   */
  public Optional<LocalDate> getEffectiveDate() {
    return effectiveDate;
  }

  /**
   * The optional end date of the new copay tier level.
   * @return never null
   */
  public Optional<LocalDate> getEndDate() {
    return endDate;
  }

  @Override
  public String toString() {
    return new ToStringBuilder(this, ToStringStyle.SHORT_PREFIX_STYLE)
        .append("productIen", getProductIen())
        .append("productName", getProductName())
        .append("cmopId", getCmopId())
        .append("copayTier", getCopayTier())
        .append("effectiveDate", getEffectiveDate())
        .append("endDate", getEndDate())
        .toString();
  }
  
  /**
   * Returns true if and only if the given object is also a CopayTierChange with the same properties.
   * @return true if the given object is equivalent, false otherwise
   */
  @Override
  public final boolean equals(final Object obj) {
    // instanceof returns false for null.
    if (!(obj instanceof CopayTierChange)) {
      return false;
    }
    
    final CopayTierChange other = (CopayTierChange)obj;
    return new EqualsBuilder()
        .append(this.getProductIen(), other.getProductIen())
        .append(this.getProductName(), other.getProductName())
        .append(this.getCmopId(), other.getCmopId())
        .append(this.getCopayTier(), other.getCopayTier())
        .append(this.getEffectiveDate(), other.getEffectiveDate())
        .append(this.getEndDate(), other.getEndDate())
        .isEquals();
  }
  
  /**
   * Generates a hash code depending on the public properties (the same properties used in 
   * {@link #equals(Object)}).
   */
  @Override
  public final int hashCode() {
    return new HashCodeBuilder()
        .append(getProductIen())
        .append(getProductName())
        .append(getCmopId())
        .append(getCopayTier())
        .append(getEffectiveDate())
        .append(getEndDate())
        .toHashCode();
  }
}
