

package gov.va.med.pharmacy.peps.service.common.update.impl;


import java.util.Calendar;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;

import gov.va.med.pharmacy.peps.common.exception.ValidationException;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.domain.common.model.EplNdfOutgoingDifferences;
import gov.va.med.pharmacy.peps.service.common.capability.NdfFileSyncCapability;
import gov.va.med.pharmacy.peps.service.common.update.SyncProcessorElement;


/**
 * Abstract Base class of all NdfFileSyncProcessor classes to support queueing information to be included in an update file.
 */
public abstract class AbstractSimpleNdfFileSyncProcessor {
    
    /** Value to send on Required Fields and Modified Old/New Values if the value is null empty. */
    protected static final String BLANK = " ";
  
    private NdfFileSyncCapability ndfFileSyncCapability;

    private Set<FieldKey> fields;
    private String fileNumber;
    private String fieldAddNumber;

    /**
     * Instantiates a new AbstractSimpleNdfFileSyncProcessor.
     *
     * @param fields the FieldKeys of the fields that, if modified, would trigger the modification logic of this processor
     * @param fileNumber the VistA file number
     * @param fieldAddNumber the VistA field number
     * @param pNdfFileSyncCapability the ndf file sync capability
     * FIXME: Redesign logic that catches and squelches exceptions so errors are displayed to user
     */
    public AbstractSimpleNdfFileSyncProcessor(Set<FieldKey> fields, String fileNumber, String fieldAddNumber, 
        NdfFileSyncCapability pNdfFileSyncCapability) {
        this.fields = fields;
        this.fileNumber = fileNumber;
        this.fieldAddNumber = fieldAddNumber;
        this.ndfFileSyncCapability = pNdfFileSyncCapability;

    }
    
    /**
     * Insert new top-level Vista File element.
     *
     * @param vistaIen the vista ien
     * @param newValue the new value
     * @param userName the user name who made or approved the change
     */
    protected void insertNewElement(String vistaIen, String newValue, String userName){
        EplNdfOutgoingDifferences diffObj = new EplNdfOutgoingDifferences();
        diffObj.setVistaIen(vistaIen);
        diffObj.setNewValue(newValue);
        diffObj.setCreatedBy(userName);
        processInsert(diffObj); 
    }
   
    /**
     * Insert new top-level Vista File element.
     *
     * @param element SyncProcessorElement to insert as a new Element
     */
    protected void insertNewElement(SyncProcessorElement element){      
      EplNdfOutgoingDifferences diffObj = new EplNdfOutgoingDifferences();
        diffObj.setVistaIen(element.getVistaIen());
        diffObj.setNewValue(element.getNewValue());
        diffObj.setCreatedBy(element.getUserName());
        processInsert(diffObj); 
    }
    
    /**
     * Insert a modified element - making modifications to an item that was sent to VistA in a previous update file.
     *
     * @param vistaIen the vista ien
     * @param newValue the new value
     * @param userName the user name
     * @param fieldId the field id
     * @param oldValue the old value
     */
    protected void insertModifiedElement(String vistaIen, String newValue, String userName, String fieldId, String oldValue){
        EplNdfOutgoingDifferences diffObj = new EplNdfOutgoingDifferences();
        diffObj.setVistaIen(vistaIen);
        diffObj.setNewValue(newValue);
        diffObj.setOldValue(oldValue);
        diffObj.setCreatedBy(userName);
        diffObj.setVistaFieldNumber(fieldId);
        processUpdate(diffObj); 
    }  

    /**
     * Insert new child element, queueing an outgoing difference record.
     *
     * @param vistaIen the vista ien of the main element this field is associated to
     * @param newValue the new value
     * @param userName the user name who made or approved the change.
     * @param fieldId the VistA field id
     * 
     * @deprecated use insertNewRequiredChildElement or insertNewRequiredChildElement instead, this method will be made private.
     */  
    @Deprecated
    protected void insertNewChildElement(String vistaIen, String newValue, String userName, String fieldId){
        EplNdfOutgoingDifferences diffObj = new EplNdfOutgoingDifferences();
        diffObj.setVistaIen(vistaIen);
        diffObj.setNewValue(newValue);
        diffObj.setCreatedBy(userName);
        diffObj.setVistaFieldNumber(fieldId);        
        processInsertChild(diffObj); 
    }    

    /**
     * Insert new child element.  It will queue an outgoing difference record.  This differs from insertNewRequiredChildElement
     * in that this method just passes through the new value without checking or changing the value. 
     *
     * @param element the SyncProcessorElement with the VistA IEN, New Value, Vista Field Id, and User Name populated
     */    
    protected void insertNewChildElement(SyncProcessorElement element){
        insertNewChildElement(element.getVistaIen(), element.getNewValue(), element.getUserName(), element.getFieldId());
        
    }           
    
    /**
     * Insert new child element for required fields.  It will queue an outgoing difference record.  If the value is null or 
     * blank (including empty spaces), then a single blank space will be sent. 
     *
     * @param element the SyncProcessorElement with the VistA IEN, New Value, Vista Field Id, and User Name populated
     */    
    protected void insertNewRequiredChildElement(SyncProcessorElement element){
        String requiredNewValue = (StringUtils.isNotBlank(element.getNewValue())) ? element.getNewValue() : BLANK;
        insertNewChildElement(element.getVistaIen(), requiredNewValue, element.getUserName(), element.getFieldId());
        
    }        

    /**
     * Insert new child element for optional fields.  It will queue an outgoing difference record only if the value is not null
     * and not blank. 
     *
     * @param element the SyncProcessorElement with the VistA IEN, New Value, Vista Field Id, and User Name populated
     */    
    protected void insertNewOptionalChildElement(SyncProcessorElement element){
      insertNewOptionalChildElement(element.getVistaIen(), element.getNewValue(), element.getUserName(), 
          element.getFieldId());       
    }     
    
    /**
     * Insert new child element for optional fields.  If the value is not null and not blank, it will queue an outgoing 
     * difference record.
     *
     * @param vistaIen the vista ien of the main element this field is associated to
     * @param newValue the new value
     * @param userName the user name who made or approved the change.
     * @param fieldId the VistA field id 
     */
    protected void insertNewOptionalChildElement(String vistaIen, String newValue, String userName, String fieldId){
      if (StringUtils.isNotBlank(newValue)){
        insertNewChildElement(vistaIen, newValue, userName, fieldId);
      }

  }     

    /**
     * Inserts an outgoing difference record with {@code EplNdfOutgoingDifferences.ADD_ACTION} for a top-level element.
     *
     * @param diffObj the EplNdfOutgoingDifferences to save
     */
    public void processInsert(EplNdfOutgoingDifferences diffObj){
        try {            
            populateCommonFields(diffObj,EplNdfOutgoingDifferences.ADD_ACTION); 
            diffObj.setVistaFieldNumber(fieldAddNumber);
            ndfFileSyncCapability.saveNewItem(diffObj);            
        } catch (ValidationException e) {
            // FIXME: We should not be squelching this exception.
            e.printStackTrace();
        }
        
    }

    /**
     * Inserts an outgoing difference record with {@code EplNdfOutgoingDifferences.MODIFY_ACTION}.
     *
     * @param diffObj the EplNdfOutgoingDifferences to save
     */    
    public void processUpdate(EplNdfOutgoingDifferences diffObj){
        try {            
            populateCommonFields(diffObj, EplNdfOutgoingDifferences.MODIFY_ACTION );
            ndfFileSyncCapability.saveModifiedItem(diffObj);            
        } catch (ValidationException e) {
            // FIXME: We should not be squelching this exception.
            e.printStackTrace();
        }
        
    }
    
    /**
     * Inserts an outgoing difference record with {@code EplNdfOutgoingDifferences.ADD_ACTION} for a child element.
     *
     * @param diffObj the EplNdfOutgoingDifferences to save
     */
    public void processInsertChild(EplNdfOutgoingDifferences diffObj){
        try {            
            populateCommonFields(diffObj,EplNdfOutgoingDifferences.ADD_ACTION);
            ndfFileSyncCapability.saveNewItem(diffObj);            
        } catch (ValidationException e) {
            // FIXME: We should not be squelching this exception.
            e.printStackTrace();
        }
        
    }
       
    private EplNdfOutgoingDifferences populateCommonFields(EplNdfOutgoingDifferences diffObj, String action){        
        diffObj.setVistaFileNumber(fileNumber);
        diffObj.setActionType(action); 
        diffObj.setCreatedDtm(Calendar.getInstance().getTime());   
        return diffObj;
    }   
      
    public void setNdfFileSyncCapability(NdfFileSyncCapability ndfFileSyncCapability) {
        this.ndfFileSyncCapability = ndfFileSyncCapability;

    }

    public NdfFileSyncCapability getNdfFileSyncCapability() {
        return ndfFileSyncCapability;
    }
    
    protected Set<FieldKey> getFields() {
        return fields;
    }   

}
