package gov.va.med.pharmacy.peps.service.common.update.impl;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import gov.va.med.pharmacy.peps.common.vo.DispenseUnitPerDoseVo;
import gov.va.med.pharmacy.peps.common.vo.DosageFormUnitVo;
import gov.va.med.pharmacy.peps.common.vo.DosageFormVo;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.common.vo.ManagedItemVo;
import gov.va.med.pharmacy.peps.common.vo.PossibleDosagesPackageVo;
import gov.va.med.pharmacy.peps.common.vo.RequestItemStatus;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.Difference;
import gov.va.med.pharmacy.peps.external.common.preencapsulation.utility.update.item.DosageFormConverter;
import gov.va.med.pharmacy.peps.external.common.utility.MumpsConverter;
import gov.va.med.pharmacy.peps.external.common.vo.outbound.common.ItemAction;
import gov.va.med.pharmacy.peps.service.common.capability.NdfFileSyncCapability;
import gov.va.med.pharmacy.peps.service.common.update.NdfFileSyncProcessor;

/**
 * DosageFormNdfFileSyncProcessor generates outgoing difference records for the NDF Update File for Dosage Form changes.
 */
public class DosageFormNdfFileSyncProcessor extends AbstractSimpleNdfFileSyncProcessor implements NdfFileSyncProcessor{
    
    /** FieldKeys of the fields that, if modified, would be processed by the modification logic of this processor. */
    public static final Set<FieldKey> FIELDS = Collections.unmodifiableSet(new LinkedHashSet<FieldKey>(Arrays
        .asList(
            FieldKey.DOSAGE_FORM_NAME, 
            FieldKey.LOCAL_MED_ROUTES, 
            FieldKey.LOCAL_MEDICATION_ROUTE,
            FieldKey.VERB, 
            FieldKey.OTHER_LANGUAGE_VERB, 
            FieldKey.PREPOSITION,
            FieldKey.OTHER_LANGUAGE_PREPOSITION, 
            FieldKey.DF_NOUNS, 
            FieldKey.INACTIVATION_DATE,
            FieldKey.DF_UNITS, 
            FieldKey.DF_DISPENSE_UNITS_PER_DOSE, 
            FieldKey.EXCLUDE_FROM_DOSAGE_CHKS,
            FieldKey.CONJUNCTION, 
            FieldKey.ITEM_STATUS
            )));
    
    private static final String FILE_NUMBER = "50.606";
    
    private static final String VISTA_YES = "1";
    private static final String VISTA_NO = "0";

    private static final org.apache.logging.log4j.Logger LOG = org.apache.logging.log4j.LogManager
        .getLogger(DosageFormNdfFileSyncProcessor.class);
    
    private static final String NAME_FIELD = ".01";
    private static final String INACTIVATION_DATE_FIELD = "7";
    private static final String EXCLUDE_FROM_DOSE_CHECK_FIELD = "11";
    private static final String UNTIS_FIELD = "8,.01";
    private static final String UNITS_PACKAGE_FIELD = "8,1";
    private static final String DISPENSE_UNIT_PER_DOSE_FIELD = "9,.01";
    private static final String DISPENSE_UNIT_PACKAGE_FIELD = "9,1";
    private static final String I_INPATIENT = "I-Inpatient";
    
    private String userName;
    private String ien;
   
    /**
     * Instantiates a new DosageFormNdfFileSyncProcessor.
     *
     * @param pNdfFileSyncCapability the ndf file sync capability
     * @see AbstractSimpleNdfFileSyncProcessor#AbstractSimpleNdfFileSyncProcessor(Set, String, String)
     */
    public DosageFormNdfFileSyncProcessor(NdfFileSyncCapability pNdfFileSyncCapability) {
        super(FIELDS, FILE_NUMBER, NAME_FIELD, pNdfFileSyncCapability);
    }

    /* (non-Javadoc)
     * @see gov.va.med.pharmacy.peps.service.common.update.NdfFileSyncProcessor#processNew(
     * gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, gov.va.med.pharmacy.peps.common.vo.UserVo)
     */
    @Override
    public void processNew(ManagedItemVo managedItem, UserVo user) {
        
        DosageFormVo dosageFormVo = (DosageFormVo) managedItem;
        
        if (RequestItemStatus.APPROVED.equals(dosageFormVo.getRequestItemStatus()) && dosageFormVo.getValue() != null) {
            
            // IEN
            setIen(dosageFormVo.getDosageFormIen());
            
            //User
            setUserName(user.getUsername());
            
            
            //NAME_FIELD
            insertNewElement(getIen(),dosageFormVo.getDosageFormName(), getUserName() );
        
            insertNewChildElement(getIen(), dosageFormVo.getExcludeFromDosageChks() != null ? convertYesNo(
                    dosageFormVo.getExcludeFromDosageChks().toString()) : BLANK, getUserName(),EXCLUDE_FROM_DOSE_CHECK_FIELD);
            
            //process Dispense Unit if present
            if (dosageFormVo.getDfDispenseUnitsPerDose().size() > 0){
                processInsertDispenseUnitFields(dosageFormVo);
            }
            
            //process Units if present
            if (dosageFormVo.getDfUnits().size() > 0){
                processInsertDfUnitFields(dosageFormVo);
            }
        }

    }
     
    @Override
    public void processModified(ManagedItemVo managedItem, UserVo user, Collection<Difference> differences) {
        
        Map<FieldKey, Difference> setDifference =  Difference.toDifferenceMap(differences);
        boolean hasDifference = DosageFormConverter.hasNewOrModifiedFields(super.getFields(), setDifference, ItemAction.MODIFY);

        DosageFormVo dosageFormVo = (DosageFormVo) managedItem;

        if (hasDifference) 
        {
            if (RequestItemStatus.APPROVED.equals(dosageFormVo.getRequestItemStatus()) && dosageFormVo.getValue() != null) 
            {
               
                setIen(dosageFormVo.getDosageFormIen());
                
                setUserName(user.getUsername());
                
                //Exclude From Dosage Checks
                if(DosageFormConverter.fieldWasModified(FieldKey.EXCLUDE_FROM_DOSAGE_CHKS, differences)){
                            insertModifiedElement(getIen(),
                            dosageFormVo.getExcludeFromDosageChks() != null ? convertYesNo(
                                dosageFormVo.getExcludeFromDosageChks().toString()) : BLANK, 
                            getUserName(), EXCLUDE_FROM_DOSE_CHECK_FIELD, null);
                }
                // Inactivation date
                if (DosageFormConverter.fieldWasModified(FieldKey.ITEM_STATUS, differences)){
                        processActivationDate(dosageFormVo);
                 }
                //Dispense Units
                if(DosageFormConverter.fieldWasModified(FieldKey.DF_DISPENSE_UNITS_PER_DOSE, differences)) {
                    processModifyDispenseUnitFields(dosageFormVo, differences);
                }

                //Units
                if(DosageFormConverter.fieldWasModified(FieldKey.DF_UNITS, differences)){
                    processModifyDfUnitFields(dosageFormVo, differences);
                }
            }
       }
    }

    /**
     * Process activation date.
     *
     * @param dosageFormVo the dosage form vo
     */
    private void processActivationDate(DosageFormVo dosageFormVo) {
        if(dosageFormVo.getInactivationDate() != null){
            GregorianCalendar c = new GregorianCalendar();
            c.setTime(dosageFormVo.getInactivationDate());
            insertModifiedElement(getIen(), MumpsConverter.convertDate(c.getTime()), getUserName(), INACTIVATION_DATE_FIELD,
                null);
        }else
            processReactivated(dosageFormVo);
    }

    /**
     * process dispense unit children - vista file 50.6069 #9.
     *
     * @param dosageFormVo the dosage form vo
     */
    private void processInsertDispenseUnitFields(DosageFormVo dosageFormVo) {
        
        Iterator<DispenseUnitPerDoseVo> dfuIterator = dosageFormVo.getDfDispenseUnitsPerDose().iterator();
        while (dfuIterator.hasNext()){
            DispenseUnitPerDoseVo dupd = dfuIterator.next(); 
            //TODO: DEFECT 508784 - send a sequential index for ien part after the comma
            insertNewChildElement(getIen() + ","+ dupd.getStrDispenseUnitPerDose(), dupd.getStrDispenseUnitPerDose(),
                getUserName(), DISPENSE_UNIT_PER_DOSE_FIELD); 
             
            if(dupd.getPackages().size()>0){
                //TODO: DEFECT 508784 - pass in the ien used for DISPENSE_UNIT_PER_DOSE_FIELD above
                processInsertPackageFields(dupd.getPackages(), dupd.getStrDispenseUnitPerDose(), DISPENSE_UNIT_PACKAGE_FIELD);
            }
        }
    }
    
    /**
     * process units children - vista file 50.6068P #8.
     *
     * @param dosageFormVo the dosage form vo
     */
    private void processInsertDfUnitFields(DosageFormVo dosageFormVo) {
        
        Iterator<DosageFormUnitVo> dfUnitIterator = dosageFormVo.getDfUnits().iterator();
        while (dfUnitIterator.hasNext()) {
             DosageFormUnitVo dfUnit = dfUnitIterator.next();

             //TODO: DEFECT 508784 - the part concatenated to the ien after the comma should be a sequential ordered index
             insertNewChildElement(getIen() + "," + dfUnit.getDrugUnit().getDrugUnitIen(), 
                 dfUnit.getDrugUnit().getDrugUnitIen(), getUserName(),UNTIS_FIELD);

           if(dfUnit.getPackages().size()>0){
               //TODO:pass the ien that we use on the UNITS_FIELD above instead of drugUnitIen 
                processInsertPackageFields(dfUnit.getPackages(), dfUnit.getDrugUnit().getDrugUnitIen(), UNITS_PACKAGE_FIELD);
            }
          
        }
    }

    /**
     * process dispense unit children - vista file 50.6069 #9.
     *
     * @param dosageFormVo the dosage form vo
     * @param differences the differences
     */
    private void processModifyDispenseUnitFields(DosageFormVo dosageFormVo, Collection<Difference> differences) {
        checkForDeletedDispenseUnit(differences);
        for (DispenseUnitPerDoseVo dupd : dosageFormVo.getDfDispenseUnitsPerDose()) {
            //TODO: DEFECT 508784 - the part concatenated to the ien after the comma should be a sequential ordered index
            insertModifiedElement(getIen() + "," + dupd.getStrDispenseUnitPerDose(), dupd.getStrDispenseUnitPerDose(),
                getUserName(), DISPENSE_UNIT_PER_DOSE_FIELD, null);
             
            if(dupd.getPackages().size()>0){
                //TODO: DEFECT 508784 - Pass in the same ien as used above in DISPENSE_UNIT_PER_DOSE_FIELD
                processModifyPackageFields(dupd.getPackages(), dupd.getStrDispenseUnitPerDose(), DISPENSE_UNIT_PACKAGE_FIELD);
            }
        }
    }

    /**
     * Generate a blank record for any removed dispense units.
     *
     * @param differences the differences
     */
    private void checkForDeletedDispenseUnit(Collection<Difference> differences) {

        boolean foundMatch = false;
       for (Difference difference : differences){
             if(difference.getFieldKey().equals(FieldKey.DF_DISPENSE_UNITS_PER_DOSE )){
                List<DispenseUnitPerDoseVo> oldList =  (List<DispenseUnitPerDoseVo>) difference.getOldValue();
                List<DispenseUnitPerDoseVo> newList = (List<DispenseUnitPerDoseVo>) difference.getNewValue();
                for (DispenseUnitPerDoseVo oldItem : oldList) {
                    foundMatch = false;
                    for (DispenseUnitPerDoseVo newItem : newList) {
                        if(oldItem.getStrDispenseUnitPerDose().equals(newItem.getStrDispenseUnitPerDose())){
                            foundMatch = true;
                            return;
                        }
                    }
                    if (!foundMatch){
                      //TODO: DEFECT 508784 - part concatenated to ien after the comma should be a sequential ordered index
                        insertModifiedElement(getIen() + ","+ oldItem.getStrDispenseUnitPerDose(),BLANK,getUserName(),
                            DISPENSE_UNIT_PER_DOSE_FIELD, null);  
                        
                        if(oldItem.getPackages().size()>0){
                          //TODO: DEFECT 508784 - part concatenated to ien after the comma should be a sequential ordered index
                            insertModifiedElement(getIen() + ","+oldItem.getStrDispenseUnitPerDose(), BLANK,getUserName(),
                                DISPENSE_UNIT_PACKAGE_FIELD, null);
                        }
                    }
                }
            }
        }

    }
   
    /**
     * process units children - vista file 50.6068P #8.
     *
     * @param dosageFormVo the dosage form vo
     * @param differences the differences
     */
    private void processModifyDfUnitFields(DosageFormVo dosageFormVo, Collection<Difference> differences) {
        
        checkForDeletedUnit(differences);
        
        for (DosageFormUnitVo dfUnit : dosageFormVo.getDfUnits()) {          
            
             insertModifiedElement(getIen() + "," + dfUnit.getDrugUnit().getDrugUnitIen(), 
                                  dfUnit.getDrugUnit().getDrugUnitIen(),getUserName(),UNTIS_FIELD, null);
            if(dfUnit.getPackages().size()>0){
                processModifyPackageFields(dfUnit.getPackages(), dfUnit.getDrugUnit().getDrugUnitIen(), UNITS_PACKAGE_FIELD);
            }
        }
    }

    /**
     * Generate a blank record for any removed units.
     *
     * @param differences the differences
     */
    private void checkForDeletedUnit(Collection<Difference> differences) {
       boolean foundMatch = false;
        
        for (Difference difference : differences){
             if(difference.getFieldKey().equals(FieldKey.DF_UNITS )){
                List<DosageFormUnitVo> oldList =  (List<DosageFormUnitVo>) difference.getOldValue();
                List<DosageFormUnitVo> newList = (List<DosageFormUnitVo>) difference.getNewValue();
                for (DosageFormUnitVo oldItem : oldList) {
                    foundMatch = false;
                    //TODO: DEFECT 508784 - dfUnitIen is not needed, replace with index - see next tag below
                    String dfUnitIen = null;
                    for (DosageFormUnitVo newItem : newList) {
                        dfUnitIen = newItem.getDrugUnit().getDrugUnitIen();
                        if(oldItem.getDrugUnit().equals(newItem.getDrugUnit())){
                            foundMatch = true;
                            return;
                        }
                    }
                    if (!foundMatch){
                      //TODO: DEFECT 508784 - part concatenated to ien after the comma should be a sequential ordered index
                        insertModifiedElement(getIen() + "," + dfUnitIen, BLANK,getUserName(), UNTIS_FIELD, null);  
                        
                        if(oldItem.getPackages().size()>0){
                          //TODO: DEFECT 508784 - part concatenated to ien after the comma should be a sequential ordered index
                            insertModifiedElement(getIen() + "," + dfUnitIen, BLANK,getUserName(),UNITS_PACKAGE_FIELD, null);
                        }
                    }
                }
            }
        }
     }

    /**
     * Process dispense unit children - vista file 50.6069 #9.
     *
     * @param packages the packages
     * @param dfUnitIen the df unit ien
     * @param fieldName the field name
     */
    private void processInsertPackageFields(Collection<PossibleDosagesPackageVo> packages, String dfUnitIen, String fieldName) {
     
        String newPackageValue =BLANK;
        if (packages.size() == 2) {
            newPackageValue = "IO";
        } else if (packages.size() == 1 
                && packages.iterator().next().getValue().contains(I_INPATIENT)) {
            newPackageValue = "I";
        } else if (packages.size() == 1) {
            newPackageValue = "O";
        }
        
        insertNewChildElement(getIen() + "," + dfUnitIen, newPackageValue,getUserName(),fieldName);
    }

    
    /**
     * Process dispense unit children - vista file 50.6069 #9.
     *
     * @param packages the packages
     * @param dfUnitIen the df unit ien
     * @param fieldName the field name
     */
    private void processModifyPackageFields(Collection<PossibleDosagesPackageVo> packages, String dfUnitIen, String fieldName) 
    {
        
       String newPackageValue =BLANK;
        if (packages.size() == 2) {
            newPackageValue = "IO";
        } else if (packages.size() == 1 
                && packages.iterator().next().getValue().contains(I_INPATIENT)) {
            newPackageValue = "I";
        } else if (packages.size() == 1) {
            newPackageValue = "O";
        }
        
        insertModifiedElement(getIen() + "," + dfUnitIen, newPackageValue,getUserName(),fieldName, null);
    }
    
    /**
     * Process a reactivated DosageFormVo.
     *
     * @param dosageFormVo the dosage form vo that has been reactivated
     */
    public void processReactivated(DosageFormVo dosageFormVo) {
        
        insertModifiedElement(getIen(),dosageFormVo.getDosageFormName(), getUserName(),NAME_FIELD,null );
            
        insertModifiedElement(getIen(),
                dosageFormVo.getExcludeFromDosageChks() != null ? convertYesNo(
                    dosageFormVo.getExcludeFromDosageChks().toString()) : BLANK,
                getUserName(),EXCLUDE_FROM_DOSE_CHECK_FIELD, null);
            
        insertModifiedElement(getIen(), BLANK, getUserName(),INACTIVATION_DATE_FIELD, null);
            
        //process Dispense Unit if present
        if (dosageFormVo.getDfDispenseUnitsPerDose().size() > 0){
            processInsertDispenseUnitFields(dosageFormVo);
        }
            
        //process Units if present
        if (dosageFormVo.getDfUnits().size() > 0){
            processInsertDfUnitFields(dosageFormVo);
        }
    }
    
    /**
     * Convert YES to 1 and NO (or all other non-YES values) to 0 for VistA.  The method is case-insensitive.
     *  
     * @param value String containing YES/NO to convert to 1/0
     * @return String converted value 1 or 0
     */
    public static String convertYesNo(String value){
        return ("YES".equalsIgnoreCase(value) ?VISTA_YES : VISTA_NO);
    }
    
    /**
     * Get the name of the PPS-N user to list as the creator/modifier on the outgoing difference.
     * @return the user name
     */
    public String getUserName() {
        return userName;
    }

    /**
     * Set the name of the PPS-N user to list as the creator/modifier on the outgoing difference.
     *
     * @param pUserName the new user name
     */
    public void setUserName(String pUserName) {
        this.userName = pUserName;
    }
    
    /**
     * Get the Dosage Form IEN.
     * @return the ien
     */
    public String getIen() {
        return ien;
    }
    
    /**
     * Set the Dosage Form IEN.
     * @param ien the ien to set
     */
    public void setIen(String ien) {
        this.ien = ien;
    }
}   
