package gov.va.med.pharmacy.peps.domain.common.dao;

import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;

import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Optional;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import gov.va.med.pharmacy.peps.common.utility.test.generator.UserGenerator;
import gov.va.med.pharmacy.peps.common.vista.VistaProductFile;
import gov.va.med.pharmacy.peps.common.vo.MatchReason;
import gov.va.med.pharmacy.peps.domain.common.model.EplMatchTypeDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplNdfOutgoingDifferences;
import gov.va.med.pharmacy.peps.domain.common.model.EplProductDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplRematchDo;
import gov.va.med.pharmacy.peps.domain.common.utility.test.AbstractDomainIT;
import gov.va.med.pharmacy.peps.domain.common.utility.test.EplNdfOutgoingDifferencesGenerator;

/**
 * Automated integration tests for {@link EplRematchDao}.
 */
public class EplRematchDaoIT extends AbstractDomainIT {
  
  private static final Logger LOG = LogManager.getLogger(EplRematchDaoIT.class);
  
  /**
   * The EPL ID of "CIPROFLOXACIN 10MG/ML INJ IV". Some of these tests assume this product exists.
   */
  private static final long CIPRO_10MG_INJ_PRODUCT_ID = 9991350429193351L;
  
  /**
   * The EPL ID of "CIPROFLOXACIN 2MG/ML/DEXTROSE 5% INJ". Some of these tests assume this product exists.
   */
  private static final long CIPRO_2MG_INJ_PRODUCT_ID = 9991350429335118L;
  
  // Using field-based autowiring only in automated tests.
  @Autowired
  private EplRematchDao eplRematchDao;

  @Autowired
  private EplProductDao eplProductDao;
  
  @Autowired
  private EplNdfOutgoingDifferencesDao eplNdfOutgoingDifferencesDao;
  
  /**
   * Simply verifies that {@link EplRematchDao#findRematchProducts(Long)} returns a non-empty List. The actual
   * returned results are not verified since they are quite dependent on database data.
   */
  @Test
  @Transactional(readOnly = true)
  public void testFindRematchProducts() {
    
    // This product must exist and have at least one default possible rematch for this test to work.
    final long productWithPossibleRematches = CIPRO_10MG_INJ_PRODUCT_ID;

    /* Behavior & Verification */
    final List<?> returnedList = eplRematchDao.findRematchProducts(productWithPossibleRematches);
    LOG.debug("Returned list is: {}", returnedList);
    assertNotNull("should return a List", returnedList);
    assertNotEquals("should return a non-empty List", Collections.emptyList(), returnedList);
  }
  
  @Test
  @Transactional
  public void testRetrieveForProduct() {
    
    /* Setup */
    
    // Assumption: the old product does not already have an outgoing rematch.
    final EplRematchDo createdRematch = createTestRematch();
    final EplProductDo fromProduct = createdRematch.getFromProduct();
    final EplProductDo toAutoProduct = createdRematch.getToAutoProduct();
    
    /* Behavior & Verification */
    
    // Use Matchers here?
    assertThat("should not return a rematch before saving it",
        eplRematchDao.retrieveForProduct(fromProduct.getEplId()), equalTo(Optional.empty()));
    
    LOG.debug("Saving test rematch.");
    eplRematchDao.insert(createdRematch, new UserGenerator().getNationalManagerTwo());
    final Optional<EplRematchDo> returnedRematch = eplRematchDao.retrieveForProduct(fromProduct.getEplId());
    LOG.debug("Returned rematch was: {}", returnedRematch);

    // Can't just compare to an expected Optional here because we don't actually know the ID of the saved rematch.
    assertThat("should return a rematch after saving it", returnedRematch.isPresent(), equalTo(true));
    assertThat("returned rematch should be for old product", returnedRematch.get().getFromProduct(), equalTo(fromProduct));
    assertThat("returned rematch should be to correct product",
        returnedRematch.get().getToAutoProduct(), equalTo(toAutoProduct));
    assertThat("returned rematch should not have suggested products",
        returnedRematch.get().getToSuggestedProducts(), equalTo(Collections.emptySet()));
        
  }
    
  /**
   * Verifies that {@link EplNdfOutgoingDifferencesDaoImpl#findRematchesForInactivations(Date)} returns only rematches that
   * have the associated outgoing difference to VistA.
   */
  @Test
  @Transactional(readOnly = true)
  public void testfindInactiveRematchReport() {
    
    /* Setup */
    
    // Save off the original list because the database may already have outgoing rematches.
    final List<EplRematchDo> initialList = eplRematchDao.findRematchesForInactivations(new Date());
    
    // Create a rematch in the database.
    final EplRematchDo rematchDo = createTestRematch();
    eplRematchDao.insert(rematchDo, new UserGenerator().getNationalManagerTwo());
    
    /* Behavior & Verification */
    
    // Ensure just creating the rematch does not add anything to the list.
    final List<EplRematchDo> listAfterSavingRematch = eplRematchDao.findRematchesForInactivations(new Date());
    assertThat(listAfterSavingRematch, equalTo(initialList));
    
    // Now add the outgoing difference and ensure the list contains the rematch now.
    final EplNdfOutgoingDifferences inactivationDiff = EplNdfOutgoingDifferencesGenerator.createDiff(
        VistaProductFile.FILE_NUMBER, VistaProductFile.STATUS_FIELD_NUMBER,
        rematchDo.getFromProduct().getNdfProductIen().toString(),
        EplNdfOutgoingDifferences.MODIFY_ACTION);
    // The difference must be from no inactivation date to an inactivation date.
    inactivationDiff.setOldValue(null);
    inactivationDiff.setNewValue("3170201");
    eplNdfOutgoingDifferencesDao.save(inactivationDiff);
    final List<EplRematchDo> listAfterSavingDiff = eplRematchDao.findRematchesForInactivations(new Date());
    assertThat(listAfterSavingDiff.size(), equalTo(initialList.size() + 1));
    // Now ensure we actually have the new rematch in the list.
    listAfterSavingDiff.removeAll(listAfterSavingRematch);
    // Can't compare the whole thing because we don't have the real match type.
    assertThat(listAfterSavingDiff.get(0).getEplRematchId(), equalTo(rematchDo.getEplRematchId()));
  }

  /* Helpers */

  private EplRematchDo createTestRematch() {
    
    final EplMatchTypeDo matchTypeDo = new EplMatchTypeDo(MatchReason.INACTIVATION.getId(), "Inactivation", "Inactivation");
    return new EplRematchDo(
        eplProductDao.retrieve(CIPRO_10MG_INJ_PRODUCT_ID),
        eplProductDao.retrieve(CIPRO_2MG_INJ_PRODUCT_ID),
        matchTypeDo);
  }
}
