package gov.va.med.pharmacy.peps.domain.common.utility.impl;

import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.junit.*;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.support.DefaultTransactionDefinition;

import gov.va.med.pharmacy.peps.common.utility.impl.test.EnvironmentUtilityStub;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.Difference;
import gov.va.med.pharmacy.peps.domain.common.capability.impl.SeqNumDomainCapabilityImpl;
import gov.va.med.pharmacy.peps.domain.common.dao.impl.EplProductDaoImpl;
import gov.va.med.pharmacy.peps.domain.common.model.EplProductDo;
import gov.va.med.pharmacy.peps.domain.common.utility.RxNormCapability;
import gov.va.med.pharmacy.peps.domain.common.utility.client.rxnorm.RxNormClient;
import gov.va.med.pharmacy.peps.domain.common.utility.client.rxnorm.test.RxNormClientIT;
import gov.va.med.pharmacy.peps.domain.common.utility.test.AbstractDomainIT;

public class RxNormCapabilityImplIT extends AbstractDomainIT implements ApplicationContextAware{

  private ApplicationContext context;
  private UserVo user;
  
  @Autowired
  private EplProductDaoImpl eplProductDao;
  
  private RxNormCapability rxNormCapability;
  private RxNormClient rxnormClient;
  private EnvironmentUtilityStub environmentUtility;
  private static final int TIMEOUT = 200;
  private static final int LONG_TIMEOUT = 1200;
  
  
  @Override
  public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
      this.context = applicationContext;
      
  }
  
  @Before
  public void setup(){    
    user = new UserVo();
    user.setUsername("AutomatedTest");
    RxNormClientIT.trustSelfSignedSSL();
    
    rxNormCapability =  (RxNormCapability)context.getBean("rxNormCapability");
    rxnormClient =  (RxNormClient)context.getBean("rxNormClient");
    
    environmentUtility = new EnvironmentUtilityStub();
    environmentUtility.setNational();
    
    SeqNumDomainCapabilityImpl seqNumCapability =  (SeqNumDomainCapabilityImpl)context.getBean("seqNumDomainCapability");
    seqNumCapability.setEnvironmentUtility(environmentUtility);        
    
    rxNormCapability.setEnvironmentUtility(environmentUtility);
  }
  
  /* this is the list or Ids and VUID pairs we will try to test
   * we assume a db table that just added the coding columns
   * so they are empty and we'll populate the data
    9991350428094730  4001164
    9991350428536081  4003547
    9991350428218557  4001954
    9991350428315101  4002412
    9991350428834531  4005733
    9991350429068828  4007158
    9991350429073426  4007183
  */
  
  @Test
  public void testGetDifferences(){
    EplProductDo product = new EplProductDo();    
    
    // test both null
    product.setCodingSystemCode(null);
    List<Difference> differences = rxNormCapability.getDifferences(null, product);
    Assert.assertTrue(differences.isEmpty());;
    
    //test both equal
    product.setCodingSystemCode("1234");
    differences = rxNormCapability.getDifferences("1234", product);
    Assert.assertTrue(differences.isEmpty());
    
    // test different code
    product.setCodingSystemCode("1235");
    differences = rxNormCapability.getDifferences("1234", product);
    Assert.assertTrue(differences.size() == 2);
    
 // test different system
    product.setCodingSystemCode("1235");
    product.setCodingSystem("Not RX Norm");
    differences = rxNormCapability.getDifferences("1235", product);
    Assert.assertTrue(differences.size() == 2);
    
    // test new code
    differences = rxNormCapability.getDifferences("1235", product);
    Assert.assertTrue(differences.size() == 2);

    // test deleted code
    product.setCodingSystemCode("1235");
    differences = rxNormCapability.getDifferences(null, product);
    Assert.assertTrue(differences.size() == 2);

  }
  
  
  @Test
  @Transactional
  public void testPopulateAndPersistNullCoding(){    
    EplProductDo eplProductDo = retrieveProduct(9991350428094730L);    
    Assert.assertNull(eplProductDo.getCodingSystemCode());
    
    String code = rxnormClient.getCode(eplProductDo.getVuid()); 
    rxNormCapability.populateRxNormForProduct(user, 9991350428094730L);
    
    eplProductDo = retrieveProduct(9991350428094730L);
    Assert.assertEquals(code, eplProductDo.getCodingSystemCode());
    Assert.assertEquals(EplProductDo.RX_NORM_CODING_SYSTEM, eplProductDo.getCodingSystem());    
  }
  
  @Test
  @Transactional
  public void testPopulateAndPersistUpdateCoding(){
    EplProductDo eplProductDo = retrieveProduct(9991350428536081L);    
    Assert.assertNotNull(eplProductDo.getCodingSystemCode());
    Assert.assertEquals("99999", eplProductDo.getCodingSystemCode());
    Assert.assertEquals("Test", eplProductDo.getCodingSystem());
    
    String code = rxnormClient.getCode(eplProductDo.getVuid()); 
    rxNormCapability.populateRxNormForProduct(user, 9991350428536081L);
    
    Assert.assertEquals(code, eplProductDo.getCodingSystemCode());
    Assert.assertEquals(EplProductDo.RX_NORM_CODING_SYSTEM, eplProductDo.getCodingSystem());
  }
  
  
  @Test
  @Transactional
  public void testPopulateAndPersistNoUpdateCoding(){
    EplProductDo eplProductDo = retrieveProduct(9991350428218557L);    
    String code = rxnormClient.getCode(eplProductDo.getVuid()); 
    
    Assert.assertNotNull(eplProductDo.getCodingSystemCode());
    Assert.assertEquals(code, eplProductDo.getCodingSystemCode());
    Assert.assertEquals(EplProductDo.RX_NORM_CODING_SYSTEM, eplProductDo.getCodingSystem());
    
    rxNormCapability.populateRxNormForProduct(user, 9991350428218557L);
    
    Assert.assertEquals(code, eplProductDo.getCodingSystemCode());
    Assert.assertEquals(EplProductDo.RX_NORM_CODING_SYSTEM, eplProductDo.getCodingSystem());

  }
  
  @Test
  @Transactional
  public void testPopulateAndPersistDeleteCoding(){
    EplProductDo eplProductDo = retrieveProduct(9991350429068828L);    
    String code = rxnormClient.getCode("1223"); //fake vuid to get a null
    
    Assert.assertNotNull(eplProductDo.getCodingSystemCode());
    Assert.assertNull(code);
    Assert.assertEquals(EplProductDo.RX_NORM_CODING_SYSTEM, eplProductDo.getCodingSystem());
    
    rxNormCapability.populateRxNormForProduct(user, 9991350429068828L);
    
    eplProductDo = retrieveProduct(9991350429068828L);
    Assert.assertNull(eplProductDo.getCodingSystemCode());
    Assert.assertNull(eplProductDo.getCodingSystem());

  } 
  
  public EplProductDo retrieveProduct(long id){
     return eplProductDao.retrieve(id);
  }
  
}
