/**
 * 
 */
package gov.va.med.pharmacy.peps.service.common.capability.impl;

import static org.easymock.EasyMock.capture;
import static org.easymock.EasyMock.mock;
import static org.easymock.EasyMock.newCapture;
import static org.easymock.EasyMock.replay;
import static org.easymock.EasyMock.verify;
import static org.hamcrest.CoreMatchers.equalTo;
import static org.hamcrest.CoreMatchers.is;
import static org.junit.Assert.assertThat;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang3.time.DateUtils;
import org.easymock.Capture;
import org.easymock.CaptureType;
import org.easymock.EasyMock;
import org.hamcrest.core.IsNull;
import org.junit.Test;

import gov.va.med.pharmacy.peps.domain.common.dao.EplNdfOutgoingDifferencesDao;
import gov.va.med.pharmacy.peps.domain.common.model.EplNdfOutgoingDifferences;
import gov.va.med.pharmacy.peps.domain.common.utility.test.EplNdfOutgoingDifferencesGenerator;

/**
 * JUnit Test class for NdfFileSyncCapabilityImpl.
 */
public class NdfFileSyncCapabilityImplTest {

  private static final long EXISTING_RECORD_ID = 9999L;

  /**
   * Test method for
   * {@link NdfFileSyncCapabilityImpl#removePendingOutgoingDifferencesForMultipleField(String, String, String)}
   * . Scenario is an new record with Add Action and no existing records. Expect save to be called top insert method.
   */
  @Test
  public void testUpdatePendingQueue_AddNoExisting() {
    // Data
    final String VISTA_FILE_NUMBER = "321.123";
    final String VISTA_FIELD_NUMBER = "123,456";
    final String VISTA_IEN = "12345,6789";
    final String ACTION = EplNdfOutgoingDifferences.ADD_ACTION;
    final EplNdfOutgoingDifferences diff = EplNdfOutgoingDifferencesGenerator.createDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER,
        VISTA_IEN, ACTION);

    // Expectations
    NdfFileSyncCapabilityImpl impl = new NdfFileSyncCapabilityImpl();
    Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
    EplNdfOutgoingDifferencesDao dao = mock(EplNdfOutgoingDifferencesDao.class);
    impl.setEplNdfOutgoingDifferencesDao(dao);
    EasyMock.expect(dao.findPendingRecordsForFieldAndIen(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN))
        .andReturn(Collections.emptyList());
    EasyMock.expect(dao.save(capture(capturedOutgoingDiffs))).andReturn(diff).once();
    replay(impl.getEplNdfOutgoingDifferencesDao());

    // Test
    impl.updatePendingQueue(diff);
    assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(1));
    // object without id is the object we passed in. The mock won't populate it like the real DAO,
    // but this makes it easy to tell our existing and unsaved records apart.
    assertThat(capturedOutgoingDiffs.getValues().get(0).getId(), is(IsNull.nullValue()));
    verify(dao);
  }

  /**
   * Test method for
   * {@link NdfFileSyncCapabilityImpl#removePendingOutgoingDifferencesForMultipleField(String, String, String)}
   * . Scenario is an new record with Add Action and an existing Add record. Expect update to be called to update the existing
   * record.
   */
  @Test
  public void testUpdatePendingQueue_AddWithExistingAdd() {
    // Data
    final String VISTA_FILE_NUMBER = "321.123";
    final String VISTA_FIELD_NUMBER = "123,456";
    final String VISTA_IEN = "12345,6789";
    final String ACTION = EplNdfOutgoingDifferences.ADD_ACTION;
    final EplNdfOutgoingDifferences diff = EplNdfOutgoingDifferencesGenerator.createDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER,
        VISTA_IEN, ACTION);
    final EplNdfOutgoingDifferences existingDiff = createExistingDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN, 
        EplNdfOutgoingDifferences.ADD_ACTION);
    List<EplNdfOutgoingDifferences> existingDiffs = new ArrayList<EplNdfOutgoingDifferences>();
    existingDiffs.add(existingDiff);

    String expectedNewValue = diff.getNewValue();
    String expectedCreatedBy = diff.getCreatedBy();
    Date expectedCreatedDtm = diff.getCreatedDtm();

    // Expectations
    NdfFileSyncCapabilityImpl impl = new NdfFileSyncCapabilityImpl();
    Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
    EplNdfOutgoingDifferencesDao dao = mock(EplNdfOutgoingDifferencesDao.class);
    impl.setEplNdfOutgoingDifferencesDao(dao);
    EasyMock.expect(dao.findPendingRecordsForFieldAndIen(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN))
        .andReturn(existingDiffs);
    EasyMock.expect(dao.update(capture(capturedOutgoingDiffs))).andReturn(diff).once();
    replay(impl.getEplNdfOutgoingDifferencesDao());

    // Test
    impl.updatePendingQueue(diff);
    assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(1));
    //We have two objects - the unsaved one without an id, and an "existing" one with an id.
    //Since the mock dao save method doesn't populate the id, we can tell whether the unsaved
    //record or the existing record was saved.  We expect the existing record to be saved.
    assertThat(capturedOutgoingDiffs.getValues().get(0).getId(), equalTo(EXISTING_RECORD_ID));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getNewValue(), equalTo(expectedNewValue));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getCreatedBy(), equalTo(expectedCreatedBy));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getCreatedDtm(), equalTo(expectedCreatedDtm));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getActionType(), equalTo(EplNdfOutgoingDifferences.ADD_ACTION));
    verify(dao);
  }

  /**
   * Test method for
   * {@link NdfFileSyncCapabilityImpl#removePendingOutgoingDifferencesForMultipleField(String, String, String)}
   * . Scenario is an new record with Modify Action and an existing Add record. Expect update to be called to update the
   * existing record.
   */
  @Test
  public void testUpdatePendingQueue_ModifyWithExistingAdd() {
    // Data
    final String VISTA_FILE_NUMBER = "321.123";
    final String VISTA_FIELD_NUMBER = "123,456";
    final String VISTA_IEN = "12345,6789";
    final EplNdfOutgoingDifferences diff = EplNdfOutgoingDifferencesGenerator.createDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER,
        VISTA_IEN, EplNdfOutgoingDifferences.MODIFY_ACTION);
    final EplNdfOutgoingDifferences existingDiff = createExistingDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN, 
        EplNdfOutgoingDifferences.ADD_ACTION);
    List<EplNdfOutgoingDifferences> existingDiffs = new ArrayList<EplNdfOutgoingDifferences>();
    existingDiffs.add(existingDiff);

    String expectedNewValue = diff.getNewValue();
    String expectedCreatedBy = diff.getCreatedBy();
    Date expectedCreatedDtm = diff.getCreatedDtm();

    // Expectations
    NdfFileSyncCapabilityImpl impl = new NdfFileSyncCapabilityImpl();
    Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
    EplNdfOutgoingDifferencesDao dao = mock(EplNdfOutgoingDifferencesDao.class);
    impl.setEplNdfOutgoingDifferencesDao(dao);
    EasyMock.expect(dao.findPendingRecordsForFieldAndIen(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN))
        .andReturn(existingDiffs);
    EasyMock.expect(dao.update(capture(capturedOutgoingDiffs))).andReturn(diff).once();
    replay(impl.getEplNdfOutgoingDifferencesDao());

    // Test
    impl.updatePendingQueue(diff);
    assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(1));
    //We have two objects - the unsaved one without an id, and an "existing" one with an id.
    //Since the mock dao save method doesn't populate the id, we can tell whether the unsaved
    //record or the existing record was saved.  We expect the existing record to be saved.
    assertThat(capturedOutgoingDiffs.getValues().get(0).getId(), equalTo(EXISTING_RECORD_ID));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getNewValue(), equalTo(expectedNewValue));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getCreatedBy(), equalTo(expectedCreatedBy));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getCreatedDtm(), equalTo(expectedCreatedDtm));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getActionType(), equalTo(EplNdfOutgoingDifferences.ADD_ACTION));
    verify(dao);
  }

  /**
   * @param vistFileNumber
   * @param action TODO
   * @param VISTA_FIELD_NUMBER
   * @param VISTA_IEN
   * @return
   */
  private EplNdfOutgoingDifferences createExistingDiff(final String vistFileNumber, final String vistaFileNumber,
      final String vistaIen, String action) {
    final int TEN_DAYS_AGO = -10;
    final EplNdfOutgoingDifferences existingDiff = EplNdfOutgoingDifferencesGenerator.createDiff(vistFileNumber,
        vistaFileNumber, vistaIen, action);
    existingDiff.setId(EXISTING_RECORD_ID);
    existingDiff.setNewValue("ExistingNewValue");
    existingDiff.setCreatedBy("ExistingCreatedBy");
    existingDiff.setCreatedDtm(DateUtils.addDays(new Date(), TEN_DAYS_AGO));
    return existingDiff;
  }

  /**
   * Test method for
   * {@link NdfFileSyncCapabilityImpl#removePendingOutgoingDifferencesForMultipleField(String, String, String)}
   * . Scenario is an new record with Modify Action and an existing Modify record, and it does not revert the value. Expect
   * update to be called to update the existing record.
   */
  @Test
  public void testUpdatePendingQueue_ModifyWithExistingModifyDiffValue() {
    // Data
    final String VISTA_FILE_NUMBER = "321.123";
    final String VISTA_FIELD_NUMBER = "123,456";
    final String VISTA_IEN = "12345,6789";
    final EplNdfOutgoingDifferences diff = EplNdfOutgoingDifferencesGenerator.createDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER,
        VISTA_IEN, EplNdfOutgoingDifferences.MODIFY_ACTION);
    final EplNdfOutgoingDifferences existingDiff = createExistingDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN, 
        EplNdfOutgoingDifferences.MODIFY_ACTION);
    List<EplNdfOutgoingDifferences> existingDiffs = new ArrayList<EplNdfOutgoingDifferences>();
    existingDiffs.add(existingDiff);

    String expectedNewValue = diff.getNewValue();
    String expectedCreatedBy = diff.getCreatedBy();
    Date expectedCreatedDtm = diff.getCreatedDtm();

    // Expectations
    NdfFileSyncCapabilityImpl impl = new NdfFileSyncCapabilityImpl();
    Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
    EplNdfOutgoingDifferencesDao dao = mock(EplNdfOutgoingDifferencesDao.class);
    impl.setEplNdfOutgoingDifferencesDao(dao);
    EasyMock.expect(dao.findPendingRecordsForFieldAndIen(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN))
        .andReturn(existingDiffs);
    EasyMock.expect(dao.update(capture(capturedOutgoingDiffs))).andReturn(diff).once();
    replay(impl.getEplNdfOutgoingDifferencesDao());

    // Test
    impl.updatePendingQueue(diff);
    assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(1));
    //We have two objects - the unsaved one without an id, and an "existing" one with an id.
    //Since the mock dao save method doesn't populate the id, we can tell whether the unsaved
    //record or the existing record was saved.  We expect the existing record to be saved.
    assertThat(capturedOutgoingDiffs.getValues().get(0).getId(), equalTo(EXISTING_RECORD_ID));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getNewValue(), equalTo(expectedNewValue));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getCreatedBy(), equalTo(expectedCreatedBy));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getCreatedDtm(), equalTo(expectedCreatedDtm));
    assertThat(capturedOutgoingDiffs.getValues().get(0).getActionType(), equalTo(EplNdfOutgoingDifferences.MODIFY_ACTION));
    verify(dao);
  }

  /**
   * Test method for
   * {@link NdfFileSyncCapabilityImpl#removePendingOutgoingDifferencesForMultipleField(String, String, String)}
   * . Scenario is an new record with Modify Action and an existing Modify record, reverting the value. Expect delete to be
   * called to delete the existing record.
   */
  @Test
  public void testUpdatePendingQueue_ModifyWithExistingModifyRevertValue() {
    // Data
    final String VISTA_FILE_NUMBER = "321.123";
    final String VISTA_FIELD_NUMBER = "123,456";
    final String VISTA_IEN = "12345,6789";
    final EplNdfOutgoingDifferences diff = EplNdfOutgoingDifferencesGenerator.createDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER,
        VISTA_IEN, EplNdfOutgoingDifferences.MODIFY_ACTION);
    diff.setNewValue("");
    final EplNdfOutgoingDifferences existingDiff = createExistingDiff(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN, 
        EplNdfOutgoingDifferences.MODIFY_ACTION);
    List<EplNdfOutgoingDifferences> existingDiffs = new ArrayList<EplNdfOutgoingDifferences>();
    existingDiffs.add(existingDiff);

    // Expectations
    NdfFileSyncCapabilityImpl impl = new NdfFileSyncCapabilityImpl();
    Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
    EplNdfOutgoingDifferencesDao dao = mock(EplNdfOutgoingDifferencesDao.class);
    impl.setEplNdfOutgoingDifferencesDao(dao);
    EasyMock.expect(dao.findPendingRecordsForFieldAndIen(VISTA_FILE_NUMBER, VISTA_FIELD_NUMBER, VISTA_IEN))
        .andReturn(existingDiffs);
    dao.delete(capture(capturedOutgoingDiffs));
    EasyMock.expectLastCall().once();
    replay(impl.getEplNdfOutgoingDifferencesDao());

    // Test
    impl.updatePendingQueue(diff);
    assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(1));
    //We have two objects - the unsaved one without an id, and an "existing" one with an id.
    //Since the mock dao save method doesn't populate the id, we can tell whether the unsaved
    //record or the existing record was saved.  We expect the existing record to be saved.
    assertThat(capturedOutgoingDiffs.getValues().get(0).getId(), equalTo(EXISTING_RECORD_ID));
    verify(dao);
  }
}
