package gov.va.med.pharmacy.peps.service.common.update.impl;

import static org.easymock.EasyMock.capture;
import static org.easymock.EasyMock.expectLastCall;
import static org.easymock.EasyMock.mock;
import static org.easymock.EasyMock.newCapture;
import static org.easymock.EasyMock.replay;
import static org.hamcrest.CoreMatchers.equalTo;
import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.CoreMatchers.notNullValue;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.fail;

import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.lang.StringUtils;
import org.easymock.Capture;
import org.easymock.CaptureType;
import org.easymock.Mock;
import org.junit.Test;

import gov.va.med.pharmacy.peps.common.utility.PPSConstants;
import gov.va.med.pharmacy.peps.common.utility.test.generator.CopayTierAssocVoGenerator;
import gov.va.med.pharmacy.peps.common.vista.VistaProductFile;
import gov.va.med.pharmacy.peps.common.vo.ActiveIngredientVo;
import gov.va.med.pharmacy.peps.common.vo.CopayTierAssocVo;
import gov.va.med.pharmacy.peps.common.vo.CsFedScheduleVo;
import gov.va.med.pharmacy.peps.common.vo.DispenseUnitVo;
import gov.va.med.pharmacy.peps.common.vo.DosageFormVo;
import gov.va.med.pharmacy.peps.common.vo.DrugClassGroupVo;
import gov.va.med.pharmacy.peps.common.vo.DrugClassVo;
import gov.va.med.pharmacy.peps.common.vo.DrugUnitVo;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.common.vo.FieldKey.FieldEnvironment;
import gov.va.med.pharmacy.peps.common.vo.FormularyDesignatorVo;
import gov.va.med.pharmacy.peps.common.vo.GenericNameVo;
import gov.va.med.pharmacy.peps.common.vo.IngredientVo;
import gov.va.med.pharmacy.peps.common.vo.ItemStatus;
import gov.va.med.pharmacy.peps.common.vo.OrderableItemVo;
import gov.va.med.pharmacy.peps.common.vo.PossibleDosagesAutoCreate;
import gov.va.med.pharmacy.peps.common.vo.ProductPackage;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.RequestItemStatus;
import gov.va.med.pharmacy.peps.common.vo.SingleMultiSourceProductVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.VuidStatusHistoryVo;
import gov.va.med.pharmacy.peps.common.vo.diff.Difference;
import gov.va.med.pharmacy.peps.common.vo.diff.GroupDifference;
import gov.va.med.pharmacy.peps.domain.common.model.EplNdfOutgoingDifferences;
import gov.va.med.pharmacy.peps.external.common.utility.MumpsConverter;
import gov.va.med.pharmacy.peps.service.common.capability.impl.NdfFileSyncCapabilityImpl;
import gov.va.med.pharmacy.peps.service.common.update.SyncProcessorElement;

/**
 * Unit test class for VaProductNdfFileSyncProcessor.
 */
public class VaProductNdfFileSyncProcessorTest {

    private static final org.apache.logging.log4j.Logger LOG = org.apache.logging.log4j.LogManager.getLogger(
        VaProductNdfFileSyncProcessorTest.class);
    
    /** Product File Number - should match the one used in VaProductNdfFileSyncProcessor. */
    private static final String PRODUCT_FILE_NUMBER = "50.68";
    /** Field Number for Copay Tier End Date - should match the one used in VaProductNdfFileSyncProcessor. */
    private static final String FIELD_COPAY_TIER_END_DATE = "45,2";
    /** Field Number for Copay Tier Start Date - should match the one used in VaProductNdfFileSyncProcessor. */
    private static final String FIELD_COPAY_TIER_START_DATE = "45,1";
    /** Field Number for Copay Tier Value - should match the one used in VaProductNdfFileSyncProcessor. */
    private static final String FIELD_COPAY_TIER = "45,.01";

    /** Fields FieldKey has assigned to Product National or Both that this processor will NOT process. */
    protected static final Set<FieldKey> FIELDKEYS_NOT_PROCESSED = Collections.unmodifiableSet(new LinkedHashSet<FieldKey>(
        Arrays.asList(new FieldKey[] { 
            FieldKey.FSN, 
            FieldKey.APPROVED_FOR_SPLITTING,
            FieldKey.AR_WS_AMIS_CATEGORY,
            FieldKey.AR_WS_CONVERSION_NUMBER,
            FieldKey.DAW_CODE,
            FieldKey.DAY_ND_OR_DOSE_NL_LIMIT,
            FieldKey.DEA_SCHEDULE,
            FieldKey.DEFAULT_MAIL_SERVICE,
            FieldKey.DISPENSE_DAYS_SUPPLY_LIMIT,
            FieldKey.DISPENSE_LIMIT_FOR_ORDER,
            FieldKey.DISPENSE_LIMIT_FOR_SCHEDULE,
            FieldKey.DISPENSE_OVERRIDE,
            FieldKey.DISPENSE_OVERRIDE_REASON,
            FieldKey.DISPENSE_QUANTITY_LIMIT,
            FieldKey.DISPENSE_QUANTITY_OVERRIDE,
            FieldKey.DISPENSE_QUANTITY_OVERRIDE_REASON,
            FieldKey.DISPLAYABLE_INGREDIENT_STRENGTH,
            FieldKey.DISPLAYABLE_INGREDIENT_UNIT,
            FieldKey.DISPLAYABLE_SYNONYM_NAME,
            FieldKey.DO_NOT_HANDLE_IF_PREGNANT,
            FieldKey.DO_NOT_MAIL,
            FieldKey.FOLLOW_UP_TIME,
            FieldKey.FORMULARY_ALTERNATIVE,
            FieldKey.HAZARDOUS_TO_PATIENT,
            FieldKey.HIGH_RISK,
            FieldKey.HIGH_RISK_FOLLOWUP,
            FieldKey.HIGH_RISK_FOLLOWUP_TIME_PERIOD,
            FieldKey.ID,
            FieldKey.INPATIENT_MEDICATION_EXPIRED_ORDER_MAX_TIME,
            FieldKey.INPATIENT_MEDICATION_EXPIRED_ORDER_MIN_TIME,
            FieldKey.LONG_TERM_OUT_OF_STOCK,
            FieldKey.LOW_SAFETY_MARGIN,
            FieldKey.MASTER_ENTRY_FOR_VUID,
            FieldKey.MAX_DISPENSE_LIMIT,
            FieldKey.MAX_DOSE_PER_DAY,
            FieldKey.MESSAGE,
            FieldKey.MIGRATED_DOSAGE_FORM_NAME,
            FieldKey.MONITOR_MAX_DAYS,
            FieldKey.MONITOR_ROUTINE,
            FieldKey.NDF_PRODUCT_IEN,
            FieldKey.NDC_COUNT,
            FieldKey.NDC_LIST,
            FieldKey.NON_RENEWABLE,
            FieldKey.SERVICE_CODE,
            FieldKey.NDCDP_QUANTITY_MULTIPLIER,            
            FieldKey.NUMBER_OF_DOSES,
            FieldKey.OVERRIDE_REASON_ENTERER,
            FieldKey.PATIENT_SPECIFIC_LABEL,
            FieldKey.PRIMARY_DRUG_CLASS,
            FieldKey.PRIMARY_DRUG_CLASS2,
            FieldKey.SECONDARY_DRUG_CLASS,
            FieldKey.PRODUCT_DISPENSE_UNIT_PER_ORDER_UNIT,
            FieldKey.CATEGORIES,
            FieldKey.SUB_CATEGORIES,
            FieldKey.MEDICATION,
            FieldKey.INVESTIGATIONAL,
            FieldKey.COMPOUND,
            FieldKey.SUPPLY,
            FieldKey.HERBAL,
            FieldKey.CHEMOTHERAPY,
            FieldKey.OTC,
            FieldKey.VETERINARY,
            FieldKey.PROPOSED_INACTIVATION_DATE,
            FieldKey.PROTECT_FROM_LIGHT,
            FieldKey.MONITOR_MAX_DAYS,
            FieldKey.LAB_MONITOR_MARK,
            FieldKey.QUANTITY_DISPENSE_MESSAGE,
            FieldKey.RECALL_LEVEL,
            FieldKey.REFRIGERATION,
            FieldKey.REQUEST_ITEM_STATUS,
            FieldKey.REQUEST_REJECTION_REASON,
            FieldKey.REJECTION_REASON_TEXT,
            FieldKey.NATIONAL_DRUG_TEXTS,
            FieldKey.TALL_MAN_LETTERING,
            FieldKey.TOTAL_DISPENSE_QUANTITY,
            FieldKey.UNIT_DOSE_SCHEDULE,
            FieldKey.UNIT_DOSE_SCHEDULE_TYPE,
            FieldKey.NCPDP_DISPENSE_UNIT,
            FieldKey.VA_DATA_FIELDS,
            FieldKey.VUID,
            FieldKey.WITNESS_TO_ADMINISTRATION,
            FieldKey.EFFECTIVE_DATES,
            FieldKey.ORDER_UNIT,
            FieldKey.SPECIAL_HANDLINGS,
            FieldKey.SYNONYMS,
            FieldKey.RXNORM_Name,
            FieldKey.RXNORM_ID,
            FieldKey.RXNORM_TTY,
            FieldKey.RXNORM_LANGUAGE,
            FieldKey.RXNORM_SUPPRESS,
            FieldKey.RXNORM_EXCLUDE,
            FieldKey.RXNORM_LAST_UPD,
            FieldKey.RXNORM_UMLSCUI,
            FieldKey.RXNORM_SYN           
            })));

    /**
     * Test method for modifyElement.
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#modifyElement(
     * FieldKey, gov.va.med.pharmacy.peps.common.vo.diff.Difference, String, String, 
     * gov.va.med.pharmacy.peps.common.vo.ProductVo)())}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testModifyElement() throws Exception {
      NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
      VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
      String vistaIen = "123";
      String userValue = "TESTUSER";
      ProductVo productVo = new ProductVo();

      Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
      fileProcessor.getNdfFileSyncCapability().saveModifiedItem(capture(capturedOutgoingDiffs));
      //expectLastCall().times(1);
      expectLastCall().anyTimes();

      replay(fileProcessor.getNdfFileSyncCapability());
           
      //List of FieldKeys that FieldKey says is not editable after creation, but still can be edited indirectly 
      //and need to be sent to VistA.
      final Set<FieldKey> NON_EDITABLEFIELDS_PROCESSED = Collections.unmodifiableSet(new LinkedHashSet<FieldKey>(Arrays
          .asList(new FieldKey[] { FieldKey.CMOP_ID})));              
            
      for (FieldKey fieldKey : VaProductNdfFileSyncProcessor.FIELDS) {     
        if (!fieldKey.isEditableAfterCreation() && !NON_EDITABLEFIELDS_PROCESSED.contains(fieldKey)){
          continue;
        }
        //TODO: for now, we need to skip this field, until we figure out Defect RTC 470632
        if (fieldKey == FieldKey.GCN_SEQUENCE_NUMBER ){
          continue;
        }
        
        Difference diff = null; 
        String expectedOldValue = null;
        String expectedNewValue = null;

        if (fieldKey == FieldKey.EXCLUDE_DRG_DRG_INTERACTION_CHECK ){
          diff = new Difference(fieldKey, Boolean.TRUE, Boolean.FALSE);
          expectedOldValue = "1";
          expectedNewValue = " ";           
        }
        else if (fieldKey == FieldKey.CREATE_POSSIBLE_DOSAGE ){
          diff = new Difference(fieldKey, Boolean.TRUE, Boolean.FALSE);
          expectedOldValue = "Y";
          expectedNewValue = "N";           
        }        
        else if (fieldKey.getFieldClass() == String.class){
          diff = new Difference(fieldKey, "oldValue", "newValue");
          expectedOldValue = "oldValue";
          expectedNewValue = "newValue";
        }
        else if (fieldKey.getFieldClass() == Boolean.class){
          diff = new Difference(fieldKey, Boolean.TRUE, Boolean.FALSE);
          expectedOldValue = "1";
          expectedNewValue = "0"; 
        }
        //TODO: Add the rest of the fields, which will require individual cases to build expected and actual values
//        else if (fieldKey == FieldKey.GENERIC_NAME){
//          
//        }
        else{
          LOG.debug(fieldKey.getKey() + " - not tested");
          continue;
        }
        LOG.debug(fieldKey.getKey());
        capturedOutgoingDiffs.reset();
        fileProcessor.modifyElement(fieldKey, diff, userValue, vistaIen, productVo);
        
        assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(1));
  
        EplNdfOutgoingDifferences savedDiff = capturedOutgoingDiffs.getValues().get(0);
        assertThat(savedDiff.getVistaFileNumber(), equalTo(PRODUCT_FILE_NUMBER));
        //TODO: Figure out a way to assert field number
        //assertThat(savedDiff.getVistaFieldNumber(), equalTo(""));
        assertThat(savedDiff.getVistaIen(), equalTo(vistaIen));
        assertThat(savedDiff.getOldValue(), equalTo(expectedOldValue));
        assertThat(savedDiff.getNewValue(), equalTo(expectedNewValue));
        assertThat(savedDiff.getActionType(), equalTo("M"));
        assertThat(savedDiff.getCreatedBy(), equalTo(userValue));
      }      
    }
    
    
    /**
     * Test method for inserting multiple copay tiers (as on a new product).
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#insertCopayTiers(
     * java.util.Collection, java.lang.String, java.lang.String)}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testInsertCopayTiersMultiRecords() throws Exception {
        List<CopayTierAssocVo> copayTierAssocs = new CopayTierAssocVoGenerator().getList();
        testInsertCopayTiersCommon(copayTierAssocs);
    }

    /**
     * Test method for passing an empty list to the insertCopayTiers process.  Expect no outgoing difference records and 
     * no exceptions.
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#insertCopayTiers(
     * java.util.Collection, java.lang.String, java.lang.String)}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testInsertCopayTiersNoRecords() throws Exception {
        List<CopayTierAssocVo> copayTierAssocs = new ArrayList<CopayTierAssocVo>();
        testInsertCopayTiersCommon(copayTierAssocs);
    }

    /**
     * Test method for inserting one copay tier with no end date (as on a new product).
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#insertCopayTiers(
     * java.util.Collection, java.lang.String, java.lang.String)}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testInsertCopayTiersOneRecordNoEndDate() throws Exception {
        List<CopayTierAssocVo> copayTierAssocs = new CopayTierAssocVoGenerator().getList();
        copayTierAssocs = copayTierAssocs.subList(copayTierAssocs.size() - 1, copayTierAssocs.size());
        testInsertCopayTiersCommon(copayTierAssocs);
    }

    /**
     * Test method for inserting one copay tier with an end date (as on a new product).
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#insertCopayTiers(
     * java.util.Collection, java.lang.String, java.lang.String)}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testInsertCopayTiersOneRecordWithEndDate() throws Exception {
        List<CopayTierAssocVo> copayTierAssocs = new CopayTierAssocVoGenerator().getList();
        copayTierAssocs = copayTierAssocs.subList(copayTierAssocs.size() - 2, copayTierAssocs.size() - 1);
        testInsertCopayTiersCommon(copayTierAssocs);
    }

    /**
     * Common logic supporting the testInsertCopayTiers tests.  Uses EasyMock to capture the outgoing 
     * differences records created and assert them based on the list of copay tiers passed in.
     * 
     * @param copayTierAssocs List of copay tiers to use in the test
     * @throws Exception if an error is encountered during the test
     */
    private void testInsertCopayTiersCommon(List<CopayTierAssocVo> copayTierAssocs) throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
        String vistaIen = "123";
        String userValue = "TESTUSER";

        Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
        int numDiffRecordsExpected = calculateNumberOfExpectedRecordsForCopayTest(copayTierAssocs);
        fileProcessor.getNdfFileSyncCapability().saveModifiedItem(capture(capturedOutgoingDiffs));

        if (numDiffRecordsExpected > 0) {
            expectLastCall().times(numDiffRecordsExpected);
        }

        replay(fileProcessor.getNdfFileSyncCapability());

        fileProcessor.insertCopayTiers(copayTierAssocs, vistaIen, userValue, true);

        assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(numDiffRecordsExpected));

        // Uncomment out the following line to test the test logic for failure
        // if not all expected fields are captured.
        // capturedOutgoingDiffs.reset();

        Comparator<EplNdfOutgoingDifferences> byVistaIen = (EplNdfOutgoingDifferences o1,
                EplNdfOutgoingDifferences o2) -> o1.getVistaIen().compareTo(o2.getVistaIen());

        capturedOutgoingDiffs.getValues().sort(byVistaIen);

        String expectedActionType = "M";

        assertResultsForCopayTest(copayTierAssocs, vistaIen, userValue, capturedOutgoingDiffs, expectedActionType);
    }

    /**
     * Test method for inserting multiple copay tiers (as on a new product).
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#insertCopayTiers(
     * java.util.Collection, java.lang.String, java.lang.String)}
     * .
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNewCopayTiersMultiRecords() throws Exception {
        List<CopayTierAssocVo> copayTierAssocs = new CopayTierAssocVoGenerator().getList();
        testProcessNewCopayTiersCommon(copayTierAssocs);
    }

    /**
     * Test method for passing an empty list to the insertCopayTiers process.  Expect no outgoing difference records and 
     * no exceptions.
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#insertCopayTiers(
     * java.util.Collection, java.lang.String, java.lang.String)}
     * .
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNewCopayTiersNoRecords() throws Exception {
        List<CopayTierAssocVo> copayTierAssocs = new ArrayList<CopayTierAssocVo>();
        testProcessNewCopayTiersCommon(copayTierAssocs);
    }

    /**
     * Test method for inserting one copay tier with no end date (as on a new product).
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#insertCopayTiers(
     * java.util.Collection, java.lang.String, java.lang.String)}
     * .
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNewCopayTiersOneRecordNoEndDate() throws Exception {
        List<CopayTierAssocVo> copayTierAssocs = new CopayTierAssocVoGenerator().getList();
        copayTierAssocs = copayTierAssocs.subList(copayTierAssocs.size() - 1, copayTierAssocs.size());
        testProcessNewCopayTiersCommon(copayTierAssocs);
    }

    /**
     * Test method for inserting one copay tier with an end date (as on a new product).
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#insertCopayTiers(
     * java.util.Collection, java.lang.String, java.lang.String)}
     * .
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNewCopayTiersOneRecordWithEndDate() throws Exception {
        List<CopayTierAssocVo> copayTierAssocs = new CopayTierAssocVoGenerator().getList();
        copayTierAssocs = copayTierAssocs.subList(copayTierAssocs.size() - 2, copayTierAssocs.size() - 1);
        testProcessNewCopayTiersCommon(copayTierAssocs);
    }

    /**
     * Common logic supporting the testProcessNewCopayTiers tests.  Uses EasyMock to capture the outgoing 
     * differences records created and assert them based on the list of copay tiers passed in.
     * 
     * @param copayTierAssocs List of copay tiers to use in the test
     * @throws Exception if an error is encountered during the test
     */
    private void testProcessNewCopayTiersCommon(List<CopayTierAssocVo> copayTierAssocs) throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
        String vistaIen = "123";
        String userValue = "TESTUSER";

        int numDiffRecordsExpected = calculateNumberOfExpectedRecordsForCopayTest(copayTierAssocs);

        List<SyncProcessorElement> elementsToSendList = new ArrayList<SyncProcessorElement>();
        fileProcessor.processNewCopayTiers(elementsToSendList, copayTierAssocs, vistaIen, userValue);

        assertThat(elementsToSendList.size(), equalTo(numDiffRecordsExpected));

        Comparator<SyncProcessorElement> byVistaIen = (SyncProcessorElement o1,
            SyncProcessorElement o2) -> o1.getVistaIen().compareTo(o2.getVistaIen());

        elementsToSendList.sort(byVistaIen);        

        assertResultsForCopayTest(copayTierAssocs, vistaIen, userValue, elementsToSendList);
    }
    
    
    /**
     * Asserts the results of insert and modify copay tier test methods.
     * @param copayTierAssocs List of copay tiers that were specified in the test
     * @param vistaIen Product VistA IEN that was specified in the test
     * @param userValue User that was specified in the test
     * @param capturedOutgoingDiffs The EplNdfOutgoingDifferences that were captured by EasyMock during the test
     * @param expectedActionType The expected action type (A, M)
     */
    private void assertResultsForCopayTest(List<CopayTierAssocVo> copayTierAssocs, String vistaIen, String userValue,
            Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs, String expectedActionType) {
        Iterator<EplNdfOutgoingDifferences> diffIter = capturedOutgoingDiffs.getValues().iterator();

        int assocIndex = 1;
        for (CopayTierAssocVo copayTierAssocVo : copayTierAssocs) {
            Set<String> expectedFields = createSetOfExpectedFieldsForCopayTest();

            while (diffIter.hasNext() && !expectedFields.isEmpty()) {
                EplNdfOutgoingDifferences diff = diffIter.next();
                assertThat(diff.getVistaFileNumber(), equalTo(PRODUCT_FILE_NUMBER));
                assertThat(diff.getVistaIen(), equalTo(vistaIen + "," + assocIndex));
                assertThat(diff.getOldValue(), equalTo(null));
                assertThat(diff.getActionType(), equalTo(expectedActionType));
                assertThat(diff.getCreatedBy(), equalTo(userValue));

                if (diff.getVistaFieldNumber().equals(FIELD_COPAY_TIER)) {
                    assertThat(diff.getNewValue(), equalTo(copayTierAssocVo.getCopayTier().getValue()));
                    expectedFields.remove(FIELD_COPAY_TIER);
                } else if (diff.getVistaFieldNumber().equals(FIELD_COPAY_TIER_START_DATE)) {
                    assertThat(diff.getNewValue(),
                            equalTo(MumpsConverter.convertDate(copayTierAssocVo.getEffectiveDate())));
                    expectedFields.remove(FIELD_COPAY_TIER_START_DATE);
                } else if (diff.getVistaFieldNumber().equals(FIELD_COPAY_TIER_END_DATE)) {
                    assertThat(diff.getNewValue(), equalTo(MumpsConverter.convertDate(copayTierAssocVo.getEndDate())));
                    expectedFields.remove(FIELD_COPAY_TIER_END_DATE);
                } else {
                    fail("Unexpected Field in the EplNdfOutgoingDifferences");
                }
            }

            assertExpectedFieldsFoundForCopayTest(copayTierAssocs, assocIndex, expectedFields);
            assocIndex++;
        }
    }

    /**
     * Asserts the results of processNewCopayTiers test methods.
     * @param copayTierAssocs List of copay tiers that were specified in the test
     * @param vistaIen Product VistA IEN that was specified in the test
     * @param userValue User that was specified in the test
     * @param capturedOutgoingDiffs The EplNdfOutgoingDifferences that were captured by EasyMock during the test
     */
    private void assertResultsForCopayTest(List<CopayTierAssocVo> copayTierAssocs, String vistaIen, String userValue,
            List<SyncProcessorElement> elementsToSendList) {
        Iterator<SyncProcessorElement> elementsIter = elementsToSendList.iterator();

        int assocIndex = 1;
        for (CopayTierAssocVo copayTierAssocVo : copayTierAssocs) {
            Set<String> expectedFields = createSetOfExpectedFieldsForCopayTest();

            while (elementsIter.hasNext() && !expectedFields.isEmpty()) {
              SyncProcessorElement element = elementsIter.next();
                assertThat(element.getVistaIen(), equalTo(vistaIen + "," + assocIndex));
                assertThat(element.getOldValue(), equalTo(null));
                assertThat(element.getUserName(), equalTo(userValue));

                if (element.getFieldId().equals(FIELD_COPAY_TIER)) {
                    assertThat(element.getNewValue(), equalTo(copayTierAssocVo.getCopayTier().getValue()));
                    expectedFields.remove(FIELD_COPAY_TIER);
                } else if (element.getFieldId().equals(FIELD_COPAY_TIER_START_DATE)) {
                    assertThat(element.getNewValue(),
                            equalTo(MumpsConverter.convertDate(copayTierAssocVo.getEffectiveDate())));
                    expectedFields.remove(FIELD_COPAY_TIER_START_DATE);
                } else if (element.getFieldId().equals(FIELD_COPAY_TIER_END_DATE)) {
                    assertThat(element.getNewValue(), equalTo(MumpsConverter.convertDate(copayTierAssocVo.getEndDate())));
                    expectedFields.remove(FIELD_COPAY_TIER_END_DATE);
                } else {
                    fail("Unexpected Field in the List of SyncProcessorElements");
                }
            }

            assertExpectedFieldsFoundForCopayTest(copayTierAssocs, assocIndex, expectedFields);
            assocIndex++;
        }
    }    
    
    /**
     * Asserts that all the expected fields had outgoing difference records.
     * @param copayTierAssocs List of copay tiers that were specified in the test
     * @param assocIndex How many records were captured for a specific copay tier
     * @param expectedFields Set of expected fields that should have outgoing differences records for each copay tier
     */
    private void assertExpectedFieldsFoundForCopayTest(List<CopayTierAssocVo> copayTierAssocs, int assocIndex,
            Set<String> expectedFields) {
        if (!expectedFields.isEmpty()) {
            if (!(assocIndex == copayTierAssocs.size() && expectedFields.contains(FIELD_COPAY_TIER_END_DATE)
                    && expectedFields.size() == 1)) {
                // It's ok for the last record to not have end date, but any
                // other condition is a failure
                StringBuilder sb = new StringBuilder("Expected Outgoing Differences for Fields: ");
                for (String field : expectedFields) {
                    sb.append("[");
                    sb.append(field);
                    sb.append("] ");
                }
                fail(sb.toString());
            }
        }
    }

    /**
     * Test for processModifiedCopayTiers when all copay tiers have been deleted.
     * Expects that one outgoing difference record is recorded for field 45,.01 with a value
     * of 'DELETED'.
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessModifiedCopayTiersAllDeleted() throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
        String vistaIen = "123";
        String userValue = "TESTUSER";

        Collection<CopayTierAssocVo> oldValue = new CopayTierAssocVoGenerator().getList();
        Collection<CopayTierAssocVo> newValue = Collections.emptyList();
        GroupDifference difference = new GroupDifference(FieldKey.COPAY_TIER_ASSOC, oldValue, newValue);        
        
        Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
        int numDiffRecordsExpected = 1;
        fileProcessor.getNdfFileSyncCapability().removePendingOutgoingDifferencesForMultipleField(PRODUCT_FILE_NUMBER, 
            VistaProductFile.COPAY_TIER_FIELD_NUMBER_BASE, vistaIen);
        fileProcessor.getNdfFileSyncCapability().saveModifiedItem(capture(capturedOutgoingDiffs));
        expectLastCall().times(numDiffRecordsExpected);

        replay(fileProcessor.getNdfFileSyncCapability());

        fileProcessor.processModifiedCopayTiers(difference, vistaIen, userValue);

        assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(numDiffRecordsExpected));

        EplNdfOutgoingDifferences diff = capturedOutgoingDiffs.getValues().get(0);
        assertThat(diff.getVistaFileNumber(), equalTo(PRODUCT_FILE_NUMBER));
        assertThat(diff.getVistaIen(), equalTo(vistaIen + "," + 1));
        assertThat(diff.getOldValue(), equalTo(null));
        assertThat(diff.getNewValue(), equalTo("DELETED"));
        assertThat(diff.getActionType(), equalTo("M"));
        assertThat(diff.getCreatedBy(), equalTo(userValue));
    }

    /**
     * Test for processModifiedCopayTiers when a copay tier has been added to an existing product.
     * Expects that all copay tiers will be sent in outgoing difference records, all with action 'M'.
     * 
     * @param copayTierAssocs
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessModifiedCopayTiersAddCopayTier() throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
        String vistaIen = "123";
        String userValue = "TESTUSER";
        
        List<CopayTierAssocVo> newValue = new CopayTierAssocVoGenerator().getList();
        List<CopayTierAssocVo> oldValue = newValue.subList(0, newValue.size() -1);
        GroupDifference difference = new GroupDifference(FieldKey.COPAY_TIER_ASSOC, oldValue, newValue);

        Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
        int numDiffRecordsExpected = calculateNumberOfExpectedRecordsForCopayTest(newValue);
        fileProcessor.getNdfFileSyncCapability().removePendingOutgoingDifferencesForMultipleField(PRODUCT_FILE_NUMBER, 
            VistaProductFile.COPAY_TIER_FIELD_NUMBER_BASE, vistaIen);
        fileProcessor.getNdfFileSyncCapability().saveModifiedItem(capture(capturedOutgoingDiffs));
        expectLastCall().times(numDiffRecordsExpected);        

        replay(fileProcessor.getNdfFileSyncCapability());


        fileProcessor.processModifiedCopayTiers(difference, vistaIen, userValue);

        assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(numDiffRecordsExpected));

        String expectedActionType = "M";

        assertResultsForCopayTest(newValue, vistaIen, userValue, capturedOutgoingDiffs, expectedActionType);
    }   
    
    /**
     * Creates a set of expected outgoing differences record fields for each copay tier.
     * If the end date was not populated, then the FIELD_END_DATE would not exist. 
     * @return Set of Strings containing the field numbers that are expected in the outgoing difference records.
     */
    private Set<String> createSetOfExpectedFieldsForCopayTest() {
        Set<String> expectedFields = new HashSet<String>();
        expectedFields.add(FIELD_COPAY_TIER);
        expectedFields.add(FIELD_COPAY_TIER_START_DATE);
        expectedFields.add(FIELD_COPAY_TIER_END_DATE);
        return expectedFields;
    }

    /**
     * Calculates the number of expected outgoing difference records for Copay Tier.
     * A record is expected for each:
     * <ul>
     * <li>tier level</li>
     * <li>effective date</li>
     * <li>end date - if populated</li>
     * 
     * @param copayTierAssocs List of copay tiers that were specified in the test
     * @return int containing the number of expected outgoing difference records.
     */
    private int calculateNumberOfExpectedRecordsForCopayTest(List<CopayTierAssocVo> copayTierAssocs) {
        int numDiffRecordsExpected = 0;
        if (copayTierAssocs != null) {
            for (CopayTierAssocVo copayTierAssocVo : copayTierAssocs) {
                numDiffRecordsExpected += 2;
                if (copayTierAssocVo.getEndDate() != null) {
                    numDiffRecordsExpected++;
                }
            }
        }
        return numDiffRecordsExpected;
    }
    
    /**
     * Test all FieldKeys are accounted for, to make sure .
     * This test class maintains a list of fields not processed.  This will help 
     */
    @Test
    public void testAllNationalProductFieldKeysAccountedFor(){

      List<FieldKey> nationalAndBothFieldKeys = FieldKey.getProductDataFields().stream().filter(
          fieldKey -> !fieldKey.getFieldEnvironment().equals(FieldEnvironment.LOCAL)).collect(Collectors.toList());
      
      //fields in FieldKey's National or Both for Product that are not in the list of processor fields
      List<FieldKey> nationalAndBothListFieldsNotProcessed = nationalAndBothFieldKeys.stream().filter(
          fkField -> VaProductNdfFileSyncProcessor.FIELDS.stream().noneMatch(procField -> procField == fkField))
          .collect(Collectors.toList());
      
      //fields in FieldKey's National or Both for Product that are not in the list of processor fields AND are not in the list 
      //of expected not processed fields
      List<FieldKey> unexpectedNotProcessedNationalAndBothFields = nationalAndBothListFieldsNotProcessed.stream().filter(
          fkField -> FIELDKEYS_NOT_PROCESSED.stream().noneMatch(expectedField -> expectedField == fkField))
          .collect(Collectors.toList());

      if (unexpectedNotProcessedNationalAndBothFields.size() > 0){        
        LOG.debug("The following fields are in the FieldKey National or Both List and are not processed and not in the "  
            + "FIELDKEYS_NOT_PROCESSED List:");
        unexpectedNotProcessedNationalAndBothFields.forEach(fieldKey -> LOG.debug(fieldKey));        
      }      

      //fields in the processor's list of fields that are not in the FieldKey's National or Both for Product
      List<FieldKey> processorFieldsNotInNationalAndBothList = VaProductNdfFileSyncProcessor.FIELDS.stream().filter(
          procField -> nationalAndBothFieldKeys.stream().noneMatch(fkField -> procField == fkField))
          .collect(Collectors.toList());  
            
      if (processorFieldsNotInNationalAndBothList.size() > 0){        
        LOG.debug("The following fields processed by the Processor are not in the FieldKey National or Both List:");
        processorFieldsNotInNationalAndBothList.forEach(fieldKey -> LOG.debug(fieldKey));        
      }
      
      //dosage.form is not on the Product National/Both list, but it is sent by the product processor
      assertThat(processorFieldsNotInNationalAndBothList.size(), equalTo(1));
      assertThat(processorFieldsNotInNationalAndBothList.contains(FieldKey.DOSAGE_FORM), is(true));      
      
      List<FieldKey> unmatchedNotProcessed = FIELDKEYS_NOT_PROCESSED.stream().filter(
          expected -> nationalAndBothListFieldsNotProcessed.stream().noneMatch(fkField -> expected == fkField))
          .collect(Collectors.toList());
      if(unmatchedNotProcessed.size() > 0){
        LOG.debug("The following expected not processed fields were not in the list of fields not processed:" );
        unmatchedNotProcessed.forEach(fieldKey -> LOG.debug(fieldKey));
      }
      
      assertThat(nationalAndBothListFieldsNotProcessed.size(), equalTo(FIELDKEYS_NOT_PROCESSED.size()));      
    } 
    
    /**
     * Test method for processNewVistaRequiredFields where no values are set on a ProductVo.  Required String fields should
     * be stored as single blank space.  Booleans will be based on the default value on ProductVo.
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#processNewVistaRequiredFields()}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNewVistaRequiredFields_NothingSet() throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
      String vistaIen = "123";
      String userName = "TESTUSER";
      ProductVo productVo = new ProductVo();

      List<EplNdfOutgoingDifferences> expectedQueue = Collections.unmodifiableList(Stream.of(
          expectedAddDiff(null, ".01", vistaIen),          
          expectedAddDiff("1", "99.98", vistaIen),
          expectedAddDiff(" ", "99.99", vistaIen),
          expectedAddDiff(" ", "1", vistaIen),
          expectedAddDiff(" ", "2", vistaIen),
          expectedAddDiff(" ", "3", vistaIen),
          expectedAddDiff(" ", ".05", vistaIen),
          expectedAddDiff("0", "31", vistaIen)
          ).sorted(compareByFieldIenAndNewValue()).collect(Collectors.toList()));           

      Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
      fileProcessor.getNdfFileSyncCapability().saveNewItem(capture(capturedOutgoingDiffs));
      expectLastCall().anyTimes();

      replay(fileProcessor.getNdfFileSyncCapability());
      
      //the method under test      
      fileProcessor.processNewVistaRequiredFields(productVo, vistaIen, userName);
     
      LOG.debug("ProductVo: {}", productVo.toString());      
             
      assertCapturedEqualToExpectedQueue(capturedOutgoingDiffs, expectedQueue);      
      
    }

    /**
     * Test method for processNewVistaRequiredFields where all required values are set on a ProductVo.  Required String fields 
     * should be stored as single blank space.  VistA Optional Booleans will be based on the default value on ProductVo.
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#processNewVistaRequiredFields()}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNewRequiredFields_ValuesSet() throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
      String vistaIen = "123";
      String userName = "TESTUSER";
      ProductVo productVo = new ProductVo();
      populateVistARequiredFieldsOnProduct(productVo);
      
      //Note: the .01 expected value is all uppercase of what was set in vaProductName  
      List<EplNdfOutgoingDifferences> expectedQueue = Collections.unmodifiableList(Stream.of(
          expectedAddDiff("VAPRODUCTNAMEVALUE", ".01", vistaIen),          
          expectedAddDiff("0", "99.98", vistaIen),
          expectedAddDiff("VUIDVALUE", "99.99", vistaIen),
          expectedAddDiff("3170329", "99.991,.01", vistaIen + ",1"),
          expectedAddDiff("1", "99.991,.02", vistaIen + ",1"),
          expectedAddDiff("DosageFormIenValue", "1", vistaIen),
          expectedAddDiff("StrengthValue", "2", vistaIen),
          expectedAddDiff("DrugUnitIenValue", "3", vistaIen),
          expectedAddDiff("GenericIenValue", ".05", vistaIen),
          expectedAddDiff("IngredientIen1", "14,.01", vistaIen + ",IngredientIen1"),
          expectedAddDiff("IngredientStrength1", "14,1", vistaIen + ",IngredientIen1"),
          expectedAddDiff("IngredientIen2", "14,.01", vistaIen + ",IngredientIen2"),
          expectedAddDiff("IngredientStrength2", "14,1", vistaIen + ",IngredientIen2"),
          expectedAddDiff("DrugUnitIenIngredient2", "14,2", vistaIen + ",IngredientIen2"),
          expectedAddDiff("DrugUnitIenIngredient1", "14,2", vistaIen + ",IngredientIen1"),
          expectedAddDiff("drugClassPrimaryIEN", "15", vistaIen),
          expectedAddDiff("1", "31", vistaIen)
          ).sorted(compareByFieldIenAndNewValue()).collect(Collectors.toList()));      
    
      Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
      fileProcessor.getNdfFileSyncCapability().saveNewItem(capture(capturedOutgoingDiffs));
      //expectLastCall().times(1);
      expectLastCall().anyTimes();
    
      replay(fileProcessor.getNdfFileSyncCapability());
      
      //the method under test
      fileProcessor.processNewVistaRequiredFields(productVo, vistaIen, userName);
      
      LOG.debug("ProductVo: {}", productVo.toString());           
     
      List<String> distinctExpectedFields = expectedQueue.stream().map(
          expected -> expected.getVistaFieldNumber()).distinct().collect(Collectors.toList()); 
      
      assertThat(VistaProductFile.VISTA_REQUIRED_FIELDS.size(), equalTo(distinctExpectedFields.size()));
      assertThat(VistaProductFile.VISTA_REQUIRED_FIELDS.containsAll(distinctExpectedFields), is(true));            
      
      assertCapturedEqualToExpectedQueue(capturedOutgoingDiffs, expectedQueue);                       
    }
    
 
    /**
     * Test method for processNewVistaOptionalFields where no values are set on a ProductVo.  Null Optional String fields 
     * should not result in outgoing difference records.  Booleans will be based on the default value on ProductVo.
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#processNewVistaOptionalFields()}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNewVistaOptionalFields_NothingSet() throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
      String vistaIen = "123";
      String userName = "TESTUSER";
      ProductVo productVo = new ProductVo();

      List<EplNdfOutgoingDifferences> expectedQueue = new ArrayList<EplNdfOutgoingDifferences>(Stream.of(
          expectedAddDiff("1", "7", vistaIen),
          expectedAddDiff("0", "17", vistaIen),
          expectedAddDiff(" ", "23", vistaIen),          
          expectedAddDiff("Y", "40", vistaIen)
          ).sorted(compareByFieldIenAndNewValue()).collect(Collectors.toList()));  

      Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
      fileProcessor.getNdfFileSyncCapability().saveNewItem(capture(capturedOutgoingDiffs));
      //expectLastCall().times(1);
      expectLastCall().anyTimes();

      replay(fileProcessor.getNdfFileSyncCapability());
      
      fileProcessor.processNewVistAOptionalFields(productVo, vistaIen, userName);
      
      LOG.debug("ProductVo: {}", productVo.toString());
                 
      assertCapturedEqualToExpectedQueue(capturedOutgoingDiffs, expectedQueue);      
      
    }

    /**
     * Test method for processNewVistaOptionalFields where no values are set on a ProductVo.  Populated optional fields should
     * result in outgoing difference records.  Booleans will be based on the default value on ProductVo.
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#processNewVistaOptionalFields()}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNewVistaOptionalFields_ValuesSet() throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
      String vistaIen = "123";
      String userName = "TESTUSER";
      ProductVo productVo = new ProductVo();
      populateVistAOptionalFieldsOnProduct(productVo);
      
      List<EplNdfOutgoingDifferences> expectedQueue = new ArrayList<EplNdfOutgoingDifferences>(Stream.of(
          expectedAddDiff("NATIONALFORMULARYNAMEVALUE", "4", vistaIen),          
          expectedAddDiff("VAPRINTNAMEVALUE", "5", vistaIen),
          expectedAddDiff("CMOPID1", "6", vistaIen),
          expectedAddDiff("1", "7", vistaIen),
          expectedAddDiff("dispenseUnitIen", "8", vistaIen),
          expectedAddDiff("GCN_SeqNo", "11", vistaIen),
          expectedAddDiff("drugClassSecondaryIEN", "16,.01", vistaIen + ",drugClassSecondaryIEN"),
          expectedAddDiff("1", "17", vistaIen),
          expectedAddDiff("3n", "19", vistaIen),          
          expectedAddDiff("S", "20", vistaIen),
          expectedAddDiff(" ", "23", vistaIen),          
          expectedAddDiff("N", "40", vistaIen),
          expectedAddDiff("N", "41", vistaIen),
          expectedAddDiff("O", "42", vistaIen),                 
          expectedAddDiff("fdaMedGuideValue", "100", vistaIen),
          expectedAddDiff("serviceCodeValue", "2000", vistaIen),
          expectedAddDiff("1", "101", vistaIen),
          expectedAddDiff("1", "102", vistaIen),
          expectedAddDiff("primaryEpaValue", "103", vistaIen),
          expectedAddDiff("wasteSortCodeValue", "104", vistaIen),
          expectedAddDiff("dotShippingNameValue", "105", vistaIen),
          expectedAddDiff("FORMULARYDESIGNATORVALUE", "109", vistaIen),
          expectedAddDiff("productTextValue", "110", vistaIen)
          ).collect(Collectors.toList()));  
      
      addCopayTiersToExpected(vistaIen, productVo, expectedQueue);

      Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
      fileProcessor.getNdfFileSyncCapability().saveNewItem(capture(capturedOutgoingDiffs));
      expectLastCall().anyTimes();

      replay(fileProcessor.getNdfFileSyncCapability());
      
      //the method under test
      fileProcessor.processNewVistAOptionalFields(productVo, vistaIen, userName);
      
      LOG.debug("ProductVo: {}", productVo.toString());        
      
      
      expectedQueue.sort(compareByFieldIenAndNewValue());
      
      List<String> distinctExpectedFields = expectedQueue.stream().map(
          expected -> expected.getVistaFieldNumber()).distinct().collect(Collectors.toList()); 

      List<String> optionalFieldsNotInExpected = VistaProductFile.VISTA_OPTIONAL_FIELDS.stream().filter(
          optional -> distinctExpectedFields.stream().noneMatch(expected -> optional.equals(expected)))
          .collect(Collectors.toList());
      if(optionalFieldsNotInExpected.size() > 0){
        LOG.debug("The following optional fields are not in the expected field list:" );
        optionalFieldsNotInExpected.forEach(field -> LOG.debug(field));
      }      
      
      assertThat(VistaProductFile.VISTA_OPTIONAL_FIELDS.size(), equalTo(distinctExpectedFields.size()));
      assertThat(VistaProductFile.VISTA_OPTIONAL_FIELDS.containsAll(distinctExpectedFields), is(true));                  
      
      assertCapturedEqualToExpectedQueue(capturedOutgoingDiffs, expectedQueue);             
      
    }

    /**
     * @param capturedOutgoingDiffs
     * @param expectedQueue
     */
    private void assertCapturedEqualToExpectedQueue(Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs,
        List<EplNdfOutgoingDifferences> expectedQueue) {
      
      
      capturedOutgoingDiffs.getValues().sort(compareByFieldIenAndNewValue());
      
      LOG.debug("Actuals:");
      capturedOutgoingDiffs.getValues().forEach(expected -> LOG.debug(expected));
      
      List<EplNdfOutgoingDifferences> unmatchedCapturedNotInExpected = capturedOutgoingDiffs.getValues().stream().filter(
          actual -> expectedQueue.stream().noneMatch(
          (expected -> actual.getVistaFileNumber().equals(expected.getVistaFileNumber())
          && StringUtils.equals(actual.getVistaFieldNumber(), expected.getVistaFieldNumber())
          && StringUtils.equals(actual.getVistaIen(), expected.getVistaIen())
          && StringUtils.equals(actual.getNewValue(), expected.getNewValue())
          && StringUtils.equals(actual.getActionType(), expected.getActionType())))).collect(Collectors.toList());          
      
      if (unmatchedCapturedNotInExpected.size() > 0){
        LOG.debug("The following actual results were not expected:");
        unmatchedCapturedNotInExpected.forEach(diff -> LOG.debug(diff));
      }
      
      List<EplNdfOutgoingDifferences> unmatchedExpectedNotInCaptured = expectedQueue.stream().filter(
          expected -> capturedOutgoingDiffs.getValues().stream().noneMatch(
          (actual -> actual.getVistaFileNumber().equals(expected.getVistaFileNumber())
          && StringUtils.equals(actual.getVistaFieldNumber(), expected.getVistaFieldNumber())
          && StringUtils.equals(actual.getVistaIen(), expected.getVistaIen())
          && StringUtils.equals(actual.getNewValue(), expected.getNewValue())
          && StringUtils.equals(actual.getActionType(), expected.getActionType())))).collect(Collectors.toList());          
      
      if (unmatchedExpectedNotInCaptured.size() > 0){
        LOG.debug("The following expected records were not found in the actual results:");
        unmatchedExpectedNotInCaptured.forEach(diff -> LOG.debug(diff));
      }
      
//      if ()
//        Iterator<EplNdfOutgoingDifferences> expectedIter = expectedQueue.iterator();
//        Iterator<EplNdfOutgoingDifferences> actualIter = capturedOutgoingDiffs.getValues().iterator();
//        
//        EplNdfOutgoingDifferences expected;
//        EplNdfOutgoingDifferences actual;
//        
//        while(expectedIter.hasNext()){
//          if (actualIter.hasNext()){
//            expected = expectedIter.next();
//            actual = actualIter.next();
//            LOG.debug("Expected: " + expected);
//            LOG.debug("  Actual: " + actual);
//            assertThat(actual.getVistaFileNumber(), equalTo(expected.getVistaFileNumber()));
//            assertThat(actual.getVistaFieldNumber(), equalTo(expected.getVistaFieldNumber()));
//            assertThat(actual.getVistaIen(), equalTo(expected.getVistaIen()));
//            assertThat(actual.getNewValue(), equalTo(expected.getNewValue()));
//            assertThat(actual.getActionType(), equalTo(expected.getActionType()));
//            assertThat(actual.getCreatedDtm(), is(notNullValue()));
//            assertThat(actual.getCreatedBy(), is(notNullValue()));  
//          }                  
//        }
//      }
      
      assertThat(capturedOutgoingDiffs.getValues().size(), equalTo(expectedQueue.size()));
      assertThat(unmatchedCapturedNotInExpected.size(), equalTo(0));
    }

    /**
     * @param vistaIen
     * @param productVo
     * @param expectedQueue
     */
    private void addCopayTiersToExpected(String vistaIen, ProductVo productVo, List<EplNdfOutgoingDifferences> expectedQueue) {
      int inx = 1;
      for (CopayTierAssocVo cta : productVo.getCopayTierAssocs()) {
        expectedQueue.add(expectedAddDiff(cta.getCopayTier().getTierLevel(), "45,.01", vistaIen + "," + inx));
        
        if (cta.getEffectiveDate() != null){
          expectedQueue.add(expectedAddDiff(MumpsConverter.convertDate(cta.getEffectiveDate()), "45,1", vistaIen + "," + inx));
        }

        if (cta.getEndDate() != null){
          expectedQueue.add(expectedAddDiff(MumpsConverter.convertDate(cta.getEndDate()), "45,2", vistaIen + "," + inx));
        }
        
        inx++;
      }
    }    
    
    /**
     * Test method for processNew where required values are set on a ProductVo.  Required String fields should
     * be stored as single blank space.  VistA optional Booleans will be based on the default value on ProductVo.
     * {@link gov.va.med.pharmacy.peps.service.common.update.impl.VaProductNdfFileSyncProcessor#processNewVistaRequiredFields()}
     * 
     * @throws Exception if an error is encountered during the test
     */
    @Test
    public void testProcessNew_RequiredFieldsValuesSet() throws Exception {
        NdfFileSyncCapabilityImpl ndfFileSyncCapability = mock(NdfFileSyncCapabilityImpl.class);
        VaProductNdfFileSyncProcessor fileProcessor = new VaProductNdfFileSyncProcessor(ndfFileSyncCapability);
      String vistaIen = "123";
      String userName = "TESTUSER";
      ProductVo productVo = new ProductVo();
      populateVistARequiredFieldsOnProduct(productVo);
      
      //these need to be set to get through processNew
      productVo.setRequestItemStatus(RequestItemStatus.APPROVED);
      productVo.setVaProductName("VAProductNameValue");
      productVo.setNdfProductIen(Long.valueOf(vistaIen));
    
      //Note: the .01 expected value is all uppercase of what was set in vaProductName  
      List<EplNdfOutgoingDifferences> expectedQueue = Collections.unmodifiableList(Stream.of(
          expectedAddDiff("VAPRODUCTNAMEVALUE", ".01", vistaIen),          
          expectedAddDiff("0", "99.98", vistaIen),
          expectedAddDiff("VUIDVALUE", "99.99", vistaIen),
          expectedAddDiff("3170329", "99.991,.01", vistaIen + ",1"),
          expectedAddDiff("1", "99.991,.02", vistaIen + ",1"),
          expectedAddDiff("DosageFormIenValue", "1", vistaIen),
          expectedAddDiff("StrengthValue", "2", vistaIen),
          expectedAddDiff("DrugUnitIenValue", "3", vistaIen),
          expectedAddDiff("GenericIenValue", ".05", vistaIen),
          expectedAddDiff("IngredientIen1", "14,.01", vistaIen + ",IngredientIen1"),
          expectedAddDiff("IngredientStrength1", "14,1", vistaIen + ",IngredientIen1"),
          expectedAddDiff("IngredientIen2", "14,.01", vistaIen + ",IngredientIen2"),
          expectedAddDiff("IngredientStrength2", "14,1", vistaIen + ",IngredientIen2"),
          expectedAddDiff("DrugUnitIenIngredient2", "14,2", vistaIen + ",IngredientIen2"),
          expectedAddDiff("DrugUnitIenIngredient1", "14,2", vistaIen + ",IngredientIen1"),
          expectedAddDiff("drugClassPrimaryIEN", "15", vistaIen),
          expectedAddDiff("1", "31", vistaIen),
          //the fields below are optional boolean fields which will always be populated
          expectedAddDiff("1", "7", vistaIen),
          expectedAddDiff("0", "17", vistaIen),
          expectedAddDiff(" ", "23", vistaIen),
          expectedAddDiff("Y", "40", vistaIen)
          ).sorted(compareByFieldIenAndNewValue()).collect(Collectors.toList()));
      
      //Create a set of required fields and optional boolean fields 
      List<String> requiredAndBooleanFieldList = new ArrayList<String>();
      requiredAndBooleanFieldList.addAll(VistaProductFile.VISTA_REQUIRED_FIELDS);
      //Add in optional boolean fields.  Required boolean fields should already be in VISTA_REQUIRED_FIELDS
      requiredAndBooleanFieldList.addAll(Arrays.asList(new String[] {
          VistaProductFile.CMOPINDICATOR_FIELD_NUMBER,
          VistaProductFile.NATFORMULARYIND_FIELD_NUMBER, 
          VistaProductFile.EXCLUDEDRUGINT_FIELD_NUMBER, 
          VistaProductFile.DEFAULTPOSDOSAGE_FIELD_NUMBER
      }));      
      
      Capture<EplNdfOutgoingDifferences> capturedOutgoingDiffs = newCapture(CaptureType.ALL);
      fileProcessor.getNdfFileSyncCapability().saveNewItem(capture(capturedOutgoingDiffs));
      //expectLastCall().times(1);
      expectLastCall().anyTimes();
    
      replay(fileProcessor.getNdfFileSyncCapability());
           
      UserVo userVo = new UserVo();
      userVo.setUsername(userName);
      
      //the method under test
      fileProcessor.processNew(productVo, userVo);
      
      LOG.debug("ProductVo: {}", productVo.toString());     
      
      List<String> distinctExpectedFields = expectedQueue.stream().map(
          expected -> expected.getVistaFieldNumber()).distinct().collect(Collectors.toList()); 
      
      assertThat(requiredAndBooleanFieldList.size(), equalTo(distinctExpectedFields.size()));
      assertThat(requiredAndBooleanFieldList.containsAll(distinctExpectedFields), is(true));            
      assertCapturedEqualToExpectedQueue(capturedOutgoingDiffs, expectedQueue);                  
      
    }

    /**
     * Convenience method to create an EplNdfOutgoingDifferences.  Defaults the File number and action type.
     *
     * @param newValue the new value
     * @param vistaFieldNumber the vista field number
     * @param vistaIen the vista ien
     * @return the epl ndf outgoing differences
     */
    private static EplNdfOutgoingDifferences expectedAddDiff(String newValue, String vistaFieldNumber, String vistaIen) {
      EplNdfOutgoingDifferences diff = new EplNdfOutgoingDifferences();
      diff.setNewValue(newValue);
      diff.setVistaFileNumber(VistaProductFile.FILE_NUMBER);
      diff.setVistaFieldNumber(vistaFieldNumber);
      diff.setVistaIen(vistaIen);
      diff.setActionType("A");
      return diff;
    }

   
    /**
     * Comparator to sort EplNdfOutgoingDifferences by field number, ien, and new value.
     *
     * @return the comparator
     */
    private Comparator<EplNdfOutgoingDifferences> compareByFieldIenAndNewValue(){
      return Comparator.comparing(EplNdfOutgoingDifferences::getVistaFieldNumber)
      .thenComparing(EplNdfOutgoingDifferences::getVistaIen)
      .thenComparing(EplNdfOutgoingDifferences::getNewValue);
      
    }
    /**
     * Populates all the fields on a ProductVo that VistA requires.  Note, this is different from the set of fields PPS-N
     * requires.
     * @param productVo
     */
    private void populateVistARequiredFieldsOnProduct(ProductVo productVo) {
      productVo.setMasterEntryForVuid(false);
      productVo.setVuid("VUIDVALUE");
      VuidStatusHistoryVo statusHistory = new VuidStatusHistoryVo();
      statusHistory.setEffectiveDateTime(Date.from(MumpsConverter.mumpsToLocalDate("3170329").atStartOfDay().
          atZone(ZoneId.systemDefault()).toInstant()));
      statusHistory.setItemStatus(ItemStatus.ACTIVE);
      productVo.setEffectiveDates(Arrays.asList(statusHistory));
      productVo.setOverrideDfDoseChkExclusn(true);
      GenericNameVo genericNameVo = new GenericNameVo();
      genericNameVo.setGenericIen("GenericIenValue");      
      productVo.setGenericName(genericNameVo);
      OrderableItemVo oi = new OrderableItemVo();
      DosageFormVo dosageFormVo = new DosageFormVo();
      dosageFormVo.setDosageFormIen("DosageFormIenValue");
      oi.setDosageForm(dosageFormVo);
      productVo.setOrderableItem(oi);
      productVo.setProductStrength("StrengthValue");
      DrugUnitVo drugUnitVo = new DrugUnitVo();
      drugUnitVo.setDrugUnitIen("DrugUnitIenValue");
      productVo.setProductUnit(drugUnitVo);
      
      //ingredients
      Collection<ActiveIngredientVo> ingredients = new ArrayList<ActiveIngredientVo>();
      ActiveIngredientVo ai1 = new ActiveIngredientVo();
      ai1.setActive(true);
      IngredientVo ing1 = new IngredientVo();
      ing1.setIngredientIen("IngredientIen1");
      ai1.setIngredient(ing1);
      ai1.setStrength("IngredientStrength1");
      DrugUnitVo drugUnit1 = new DrugUnitVo();
      drugUnit1.setDrugUnitIen("DrugUnitIenIngredient1");
      ai1.setDrugUnit(drugUnit1);
      ingredients.add(ai1);

      ActiveIngredientVo ai2 = new ActiveIngredientVo();
      ai2.setActive(true);
      IngredientVo ing2 = new IngredientVo();
      ing2.setIngredientIen("IngredientIen2");
      ai2.setIngredient(ing2);
      ai2.setStrength("IngredientStrength2");
      DrugUnitVo drugUnit2 = new DrugUnitVo();
      drugUnit2.setDrugUnitIen("DrugUnitIenIngredient2");
      ai2.setDrugUnit(drugUnit2);
      ingredients.add(ai2);      
      
      productVo.setActiveIngredients(ingredients);
      
      //drug classes
      DrugClassGroupVo primary = new DrugClassGroupVo();
      DrugClassVo drugClassPrimary = new DrugClassVo();
      drugClassPrimary.setDrugClassIen("drugClassPrimaryIEN");
      primary.setDrugClass(drugClassPrimary);
      primary.setPrimary(true);

      Collection<DrugClassGroupVo> drugClasses = productVo.getDrugClasses();
      if (drugClasses == null){
        drugClasses = new ArrayList<DrugClassGroupVo>();
      }
      drugClasses.add(primary);
      productVo.setDrugClasses(drugClasses);
      //this is technically not required by VistA, but it is the .01 level value sent.  It's easier to test to have it here.
      productVo.setVaProductName("VAProductNameValue");
    }

    /**
     * Populates all the fields on a ProductVo that VistA does not require.  Note, this is different from the set of fields 
     * PPS-N requires.
     * @param productVo
     */
    private void populateVistAOptionalFieldsOnProduct(ProductVo productVo) {
      productVo.setNationalFormularyName("NationalFormularyNameValue");
      productVo.setVaPrintName("VaPrintNameValue");
      productVo.setCmopId("CMOPID1");
      productVo.setCmopDispense(true);
      DispenseUnitVo dispenseUnitVo = new DispenseUnitVo();
      dispenseUnitVo.setDispenseUnitIen("dispenseUnitIen");
      productVo.setDispenseUnit(dispenseUnitVo);
      productVo.setGcnSequenceNumber("GCN_SeqNo");
      productVo.setNationalFormularyIndicator(true);
      
      CsFedScheduleVo fs = new CsFedScheduleVo();
      fs.setValue("3nFedScheduleValue");
      productVo.setCsFedSchedule(fs);
      
      SingleMultiSourceProductVo multiSource = new SingleMultiSourceProductVo();
      multiSource.setValue(PPSConstants.MULTISOURCE_SINGLE);
      productVo.setSingleMultiSourceProduct(multiSource); //PPSConstants.MULTISOURCE_SINGLE

      productVo.setExcludeDrgDrgInteractionCheck(false);
      productVo.setCreatePossibleDosage(false);
      
      productVo.setPossibleDosagesAutoCreate(PossibleDosagesAutoCreate.NO_POSSIBLE_DOSAGES); //N
      
      productVo.setProductPackage(ProductPackage.OUTPATIENT);  //O
      
      productVo.setCopayTierAssocs(new CopayTierAssocVoGenerator().getList());
      productVo.setFdaMedGuide("fdaMedGuideValue");
      productVo.setServiceCode("serviceCodeValue");
      productVo.setHazardousToHandle(Boolean.valueOf(true)); //TODO: also should test null, which would result in BLANK
      productVo.setHazardousToDispose(Boolean.valueOf(true));
      productVo.setPrimaryEpa("primaryEpaValue");
      productVo.setWasteSortCode("wasteSortCodeValue");
      productVo.setDotShippingName("dotShippingNameValue");
      FormularyDesignatorVo formularyDesignatorVo = new FormularyDesignatorVo();
      formularyDesignatorVo.setName("FormularyDesignatorValue");
      productVo.setProductFormularyDesignator(formularyDesignatorVo);
      productVo.setProductText("productTextValue");      

      //drug classes
      DrugClassGroupVo secondary = new DrugClassGroupVo();
      DrugClassVo drugClassSecondary = new DrugClassVo();
      drugClassSecondary.setDrugClassIen("drugClassSecondaryIEN");
      secondary.setDrugClass(drugClassSecondary);
      secondary.setPrimary(false);

      Collection<DrugClassGroupVo> drugClasses = productVo.getDrugClasses();
      if (drugClasses == null){
        drugClasses = new ArrayList<DrugClassGroupVo>();
      }
      drugClasses.add(secondary);
      productVo.setDrugClasses(drugClasses);      
      
    }
    
}
