package gov.va.med.pharmacy.peps.external.common.utility;

import java.text.SimpleDateFormat;
import java.time.DateTimeException;
import java.time.LocalDate;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Objects;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A utility class to handle conversions from Java objects to FileMan formats and vice versa.
 */
public final class MumpsConverter {

  /**
   * The base year of FileMan dates. FileMan dates cannot represent dates before this year.
   */
  public static final int BASE_YEAR = 1700;

  private static final SimpleDateFormat PARTIAL_DATE_FORMAT = new SimpleDateFormat("YYMMdd");
   
  /**
   * Regular expression that matches FileMan dates.
   */
  private static final Pattern FILEMAN_DATE_REGEX = Pattern.compile("^(?<year>\\d\\d\\d)(?<month>\\d\\d)(?<day>\\d\\d)");

    /**
     * Hide the constructor so an instance of this class cannot be created.
     */
    private MumpsConverter() {
        
    }
    
    /**
     * Converts the Time for the passed in Date object in to Mumps format.
     * 
     * @param time
     * @return String
     */
    public static String convertDate(Date date) {
        StringBuilder convertedDate = new StringBuilder();
        
        GregorianCalendar cal = new GregorianCalendar();
        cal.setTime(date);
        int year = cal.get(Calendar.YEAR);
        
        convertedDate.append("" + ((year - BASE_YEAR)/100));
        convertedDate.append(PARTIAL_DATE_FORMAT.format(date));
        
        return convertedDate.toString();
    }
    
    /**
     * Converts the given date object to FileMan format.
     * @param date the date to convert. Must not be null.
     * @return a date in FileMan format.
     */
    public static String convertDate(final LocalDate date) {
      Objects.requireNonNull(date);
      return String.format("%03d%02d%02d", date.getYear() - BASE_YEAR, date.getMonthValue(), date.getDayOfMonth());
    }
    
    /**
     * Converts the Time for the passed in Date object in to Mumps format.
     * 
     * @param time
     * @return String
     */
    public static String convertTime(Date time) {
        GregorianCalendar time2 = new GregorianCalendar();
        time2.setTime(time);
        time2.set(Calendar.HOUR, 0);
        time2.set(Calendar.MINUTE, 0);
        time2.set(Calendar.SECOND, 0);
        time2.set(Calendar.MILLISECOND, 0);
        
        return Long.toString((time.getTime() - time2.getTimeInMillis())/1000);
    }
    
    /**
     * Converts the Date and Time for the passed in Date object in to Mumps format.
     * 
     * @param dateTime
     * @return String
     */
    public static String convertDateTime(Date dateTime) {
        StringBuilder sb = new StringBuilder();
        
        sb.append(convertDate(dateTime));
        sb.append(".");
        sb.append(convertTime(dateTime));
        
        return sb.toString();
        
    }
    
  /**
   * Converts the given FileMan date to a Java LocalDateTime object.
   * @param mumpsDate the FileMan date. Must not be null.
   * @return a LocalDate instance
   * @throws IllegalArgumentException if the given argument does not represent a valid date
   */
  public static LocalDate mumpsToLocalDate(final String mumpsDate) {
    Objects.requireNonNull(mumpsDate);
    final Matcher matcher = FILEMAN_DATE_REGEX.matcher(mumpsDate);

    if (!matcher.matches()) {
      throw new IllegalArgumentException("Supplied argument is not in FileMan date format.");
    }

    final String yearString = matcher.group("year");
    final String monthString = matcher.group("month");
    final String dayString = matcher.group("day");

    // We should not get a NumberFormatException from any of these because we just verified they are all numeric.
    final int year = Integer.parseInt(yearString) + BASE_YEAR;
    final int month = Integer.parseInt(monthString);
    final int day = Integer.parseInt(dayString);

    try {
      return LocalDate.of(year, month, day);
    } catch (final DateTimeException ex) {
      throw new IllegalArgumentException("Argument was not a valid FileMan date.", ex);
    }
  }
}
