/**
 * Source file created in 2008 by Southwest Research Institute
 */

package gov.va.med.pharmacy.peps.service.common.capability;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Objects;
import java.util.Optional;

import gov.va.med.pharmacy.peps.common.vo.ManagedItemVo;
import gov.va.med.pharmacy.peps.common.vo.ModDifferenceVo;
import gov.va.med.pharmacy.peps.common.vo.ModificationSummaryVo;
import gov.va.med.pharmacy.peps.common.vo.RematchVo;
import gov.va.med.pharmacy.peps.common.vo.validator.Errors;


/**
 * Act as a "Map" of {@link ManagedItemVo} to its Collection of differences and an optional rematch.
 * <p>
 * Used while performing a multi-edit of a List of {@link ManagedItemVo}. Struts/OGNL would have a difficult (though not
 * impossible) time of using the {@link ManagedItemVo} as a key, so to simplify things there we'll use this object in a
 * Collection.
 * </p>
 * <p>
 * This class contains similar data to {@link ModificationSummaryVo} and could probably be combined with that one.
 * </p>
 */
public class ItemModDifferences {

    /** The differences. */
    private Collection<ModDifferenceVo> differences;
    
    /** The unmodified item. */
    private ManagedItemVo item;
    
    /** The errors and warnings associated to this item. */
    private Errors errors;

    /** The optional rematch **/
    private Optional<RematchVo> optionalRematch;
    
    /**
     * Set the unmodified {@link ManagedItemVo} and Collection of {@link ItemModDifferences}.
     * 
     * @param pItem the unmodified item
     * @param pDifferences Collection of {@link ModDifferenceVo}
     */
    public ItemModDifferences(ManagedItemVo pItem, Collection<ModDifferenceVo> pDifferences) {
        this(pItem, pDifferences, new Errors());
    }

    /**
     * Set the {@link ManagedItemVo} and Collection of {@link ItemModDifferences}.
     *
     * @param pItem the unmodified item
     * @param pDifferences Collection of {@link ModDifferenceVo}
     * @param pErrors the errors and warnings for this item
     */
    public ItemModDifferences(ManagedItemVo pItem, Collection<ModDifferenceVo> pDifferences, Errors pErrors) {
        item = pItem;
        differences = pDifferences;
        errors = pErrors;
        optionalRematch = Optional.empty();
    }
    
    /**
     * Sets the item.
     *
     * @param item ValueObject
     */
    public void setItem(ManagedItemVo item) {
        this.item = item;
    }

    /**
     * Returns the unmodified item.
     *
     * @return ValueObject
     */
    public ManagedItemVo getItem() {
        return this.item;
    }

    /**
     * Gets the differences.
     *
     * @return Collection of ModDifferenceVos
     */
    public Collection<ModDifferenceVo> getDifferences() {
        return differences;
    }

    /**
     * Sets the differences.
     *
     * @param differences differences property
     */
    public void setDifferences(Collection<ModDifferenceVo> differences) {
        if (this.differences == null) {
            this.differences = new ArrayList<ModDifferenceVo>();
        }

        if (differences != null && !differences.isEmpty()) {
            this.differences.addAll(differences);
        }
    }

    /**
     * Returns the same as {@link #getDifferences()} except that the returned collection only includes "accepted"
     * differences (i.e., {@link ModDifferenceVo#isAcceptChange()} returns true).
     * @return a new collection. Never null.
     */
    public Collection<ModDifferenceVo> getAcceptedDifferences() {
        final Collection<ModDifferenceVo> acceptedDifferences = new ArrayList<ModDifferenceVo>();

        for (ModDifferenceVo difference : differences) {
            if (difference.isAcceptChange()) {
                acceptedDifferences.add(difference);
            }
        }

        return acceptedDifferences;
    }
    
    /**
     * Gets the errors.
     *
     * @return the errors
     */
    public Errors getErrors() {
    
        return errors;
    }

    
    /**
     * Sets the errors.
     *
     * @param errors the errors to set
     */
    public void setErrors(Errors errors) {
    
        this.errors = errors;
    }

    /**
     * Returns the optional rematch.
     * @return never null
     */
    public Optional<RematchVo> getOptionalRematch() {
      return optionalRematch;
    }

    /**
     * Sets the optional rematch.
     * @param optionalRematch must not be null
     */
    public void setOptionalRematch(Optional<RematchVo> optionalRematch) {
      this.optionalRematch = Objects.requireNonNull(optionalRematch);
    }
}
