package gov.va.med.pharmacy.peps.service.common.capability;

import gov.va.med.pharmacy.peps.common.exception.ValidationException;
import gov.va.med.pharmacy.peps.common.vo.NdfUpdateFileVo;
import gov.va.med.pharmacy.peps.common.vo.PaginatedList;
import gov.va.med.pharmacy.peps.common.vo.StatusVo;
import gov.va.med.pharmacy.peps.common.vo.UpdateSearchCriteria;
import gov.va.med.pharmacy.peps.common.vo.UserVo;

import java.io.IOException;
import java.util.List;

/**
 * <p>Contains the main logic to create the NDF Update File and send email notifications.</p>
 * 
 * <p>All methods that modify state (e.g., {@link #createOnDemandFileAsync(UserVo)} and {@link #createScheduledFile(UserVo)})
 * are mutually exclusive from a thread-safety perspective. That is, a thread trying to call one of these methods while
 * another thread is already calling one will block.</p>
 */
public interface NdfUpdateCapability {

    /**
     * Creates and sends a new on-demand update file asynchronously. Similar {@link #createScheduledFile(UserVo)} but
     * asynchronous. Only one thread can execute this method at a time: other threads will block.
     * 
     * @param user the user generating the file
     * @return the created {@link NdfUpdateFileVo}. Will have status {@link StatusVo#INITIATED}.
     * @throws ValidationException if there is already a file going through the approval process
     */
    NdfUpdateFileVo createOnDemandFileAsync(UserVo user) throws ValidationException;

    /**
     * Creates and sends a new scheduled update file. Only one thread can execute this method at a time: other threads
     * will block.
     *
     * @param user the user
     * @throws IOException if there is an error generating or transferring the update file.
     */    
    public void createScheduledFile(UserVo user) throws IOException;
    
    /**
     * Retrieve all NdfUpdateFileVo.
     * @return List<FileUpdateVo> updateFiles
     */
    List<NdfUpdateFileVo> retrieveFileUpdates();

    /**
     * Retrieve paged NdfUpdateFileVo.
     *
     * @param searchCriteria
     *            the search criteria
     * @return List<FileUpdateVo> updateFiles
     */
    PaginatedList<NdfUpdateFileVo> retrievePagedFileUpdates(UpdateSearchCriteria searchCriteria);

    /**
     * Retrieve NdfUpdateFileVo ready for approval/rejection.
     * @return List<NdfUpdateFileVo> pendingUpdateFiles
     */
    List<NdfUpdateFileVo> retrievePendingUpdate();

    /**
     * Retrieve NdfUpdateFileVo by NDFUpdateFileId.
     * @param id NDFUpdateFileId 
     * @return NdfUpdateFileVo The found file row, or null if the id is null or no row found with that id.
     */
    NdfUpdateFileVo retrieveById(Long id);

    /**
     * Retrieve a list of FileUpdateVo with initiated status.  Realistically, there should only ever be one.
     *
     * @return List<FileUpdateVo> initiatedUpdateFiles
     */
    List<NdfUpdateFileVo> retrieveInitiatedUpdate();

    /**
     * Retrieve all the status values (EPL_STATUS).
     * @return List<StatusVo> all update file statuses
     */
    List<StatusVo> retrieveAllFileStatuses();  
    
    /**
     * Retrieves the latest NdfUpdateFileVo, meaning the one with the max file id.
     *
     * @return the latest NdfUpdateFileVo
     */    
    NdfUpdateFileVo retrieveCurrentStatus();
    
    /**
     * Approve.  Updates the status and transmits the file to the production folder.
     *
     * @param toUpdate the update file to update
     * @param user the user
     * @throws ValidationException if the file transmission to the production folder failed
     */
    public void approve(NdfUpdateFileVo toUpdate, UserVo user) throws ValidationException;
    
    /**
     * Reject.  Updates the status, moves the file to the reject folder, and remove the associations between the File 
     * and the data.
     *
     * @param toUpdate the update file to update
     * @param user the user
     * @throws ValidationException if the given metadata was invalid or the file transmission to the reject folder
     * failed
     */
    public void reject(NdfUpdateFileVo toUpdate, UserVo user) throws ValidationException;

}
