package gov.va.med.pharmacy.peps.service.common.capability;

import gov.va.med.pharmacy.peps.common.vo.ManagedItemVo;
import gov.va.med.pharmacy.peps.common.vo.RematchVo;
import gov.va.med.pharmacy.peps.common.vo.UnmatchRematchVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.Difference;

import java.util.Collection;
import java.util.List;
import java.util.Optional;

/**
 * Business operations pertaining to rematches.
 * 
 * @see RematchVo
 */
public interface RematchCapability {
    
    public void removeObsoleteProducts();
    
    public void addUnmatchedProducts(UserVo user);

    public List<RematchVo> retrieveUnmatched(Long matchTypeId);

    public List<UnmatchRematchVo> findRematchProducts(Long productId);

    /**
     * Updates the specified Rematch with the given new product IEN and auto match flag.
     * @param rematchId the database ID of the rematch.
     * @param matchIen the new product IEN
     * @param autoMatch auto match flag
     */
    public void updateRematch(Long rematchId, String matchIen, Boolean autoMatch);

  /**
   * Returns the existing rematch for the given item and differences, if any.
   * @param item the item 
   * @param differences the differences that triggered the rematch
   * @return never null
   */
  public Optional<RematchVo> getExistingRematchForDifferences(
      final ManagedItemVo item, final Collection<Difference> differences);
  
  /**
   * Updates rematches as necessary when the given item has been inactivated per a proposed inactivation.
   * @param item the updated item, now inactive
   * @param user the user updating the item (probably a batch job user)
   */
  public void updateUponProposedInactivation(final ManagedItemVo item, UserVo user);
  
  /**
   * Saves or updates the given rematch for the given changes to the item. This may involve, for example, deleting the
   * rematch if the changes no longer trigger one. {@link #saveOrUpdateCommittedRematch(ManagedItemVo, Collection,
   * Optional, UserVo)} should be called instead if actually committing the changes to the item.
   * 
   * @param diffs changes to the item
   * @param rematch the rematch associated with the changes. May be updated or deleted.
   * @param user the user performing changes
   */
  public Optional<RematchVo> saveOrUpdateUncommittedRematch(
      final Collection<Difference> diffs, final RematchVo rematch, final UserVo user);

  /**
   * Like {@link #saveOrUpdateUncommittedRematch(Collection, Optional, UserVo)} but also removes any
   * pre-existing rematch if the changes reverse it. This one should be called when actually committing the changes to
   * the item.
   * 
   * @param item the item being modified. The item may or may not have the modifications applied; it does not matter.
   * @param committedDiffs changes being committed to the item
   * @param rematch the rematch associated with the changes. May be updated or dropped.
   * @param user the user performing changes
   */
  public Optional<RematchVo> saveOrUpdateCommittedRematch(
      final ManagedItemVo item, final Collection<Difference> committedDiffs, final Optional<RematchVo> rematch,
      final UserVo user);

}
