package gov.va.med.pharmacy.peps.service.common.capability.impl;

import gov.va.med.pharmacy.peps.common.utility.VistaFileNumber;
import gov.va.med.pharmacy.peps.common.vista.VistaFile;
import gov.va.med.pharmacy.peps.common.vo.MatchReason;
import gov.va.med.pharmacy.peps.common.vo.NationalSetting;
import gov.va.med.pharmacy.peps.domain.common.dao.EplNdfOutgoingDifferencesDao;
import gov.va.med.pharmacy.peps.domain.common.dao.EplProductDao;
import gov.va.med.pharmacy.peps.domain.common.dao.EplRematchDao;
import gov.va.med.pharmacy.peps.domain.common.model.EplProductDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplRematchDo;
import gov.va.med.pharmacy.peps.domain.common.model.dto.FutureInactiveProductDto;
import gov.va.med.pharmacy.peps.service.common.capability.MessageDataCapability;
import gov.va.med.pharmacy.peps.service.common.update.CopayTierChange;
import gov.va.med.pharmacy.peps.service.common.utility.MessageFormatter;
import gov.va.med.pharmacy.peps.service.common.utility.NDFUpdateFileMapping;
import gov.va.med.pharmacy.peps.service.common.utility.NdfUpdateProcessFile;
import gov.va.med.pharmacy.peps.service.national.updatefile.ReportMessageCode;
import gov.va.med.pharmacy.peps.service.national.updatefile.UpdateFileReportGenerator;

import java.io.IOException;
import java.time.Instant;
import java.time.LocalDate;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;

/**
 * Primary implementation of {@link MessageDataCapability}.
 */
public class MessageDataCapabilityImpl implements MessageDataCapability {

    private static final int COPAY_TIER_END_DATE_WIDTH = 10;
    private static final int COPAY_TIER_START_DATE_WIDTH = 13;
    private static final int COPAY_TIER_LEVEL_WIDTH = 4;
    private static final String END_SQUARE = "]";
    private static final String OPENING_SQUARE = "[";
    private static final String NEWLINE = System.getProperty("line.separator");
    private static final String EMPTYLINE = "";
    private static final String SPACER = "     ";
    private static final String SPACER2 = "          ";
    private static final String SPACER3 = "                 ";
    private static final String DOUBLEDSPACE = "    ";
    private static final String SPACE = "  ";
    private static final String ONE_SPACE = " ";
    private static final String OLD = "      Old Value: ";
    private static final String NEW = "      New Value: ";

    /**
     * Message returned for reports when generating them failed.
     */
    protected static final String REPORT_FAILURE_MESSAGE = "Failed to generate the report.";

    /**
     * The message placed in the file when all of a product's copay tiers have been removed.
     */
    public static final String COPAY_TIERS_REMOVED_MESSAGE = "COPAY TIER LEVEL has been Removed.";

    /** Logger for this class. */
    private static final org.apache.logging.log4j.Logger LOG = org.apache.logging.log4j.LogManager
            .getLogger(MessageDataCapabilityImpl.class);

    private final EplNdfOutgoingDifferencesDao eplNdfOutgoingDifferencesDao;
    private final EplRematchDao eplRematchDao;
    private final EplProductDao eplProductDao;
    private final UpdateFileReportGenerator reportGenerator;

    /** Title for Formulary Section of report. */
    public static final String FORMULARY = "FORMULARY ITEMS";

    /** Title for Non-Formulary Section of report. */
    public static final String NONFORMULARY = "NON-FORMULARY ITEMS";

    /** Verbiage used when no records found for a section. */
    public static final String NONE = "NONE";

    /** Ending parentheses to match the starting parentheses in some of the verbiage constants below. */
    public static final String ENDROUNDBRAC = ")";

    /** The Constant CMOP. */
    public static final String CMOP = "(CMOP - ";

    /** The Constant DISPENSE. */
    public static final String DISPENSE = "(DISPENSE UNIT - ";

    /** The Constant AUTOCREATEQUEST. */
    public static final String AUTOCREATEQUEST = "  Auto-Create Default Possible Dosage? ";

    /** The Constant AUTOCREATENEG. */
    public static final String AUTOCREATENEG = "  Auto-Create Default Possible Dosage? No ";

    /** The Constant POSSIBLEDOSAGESTMT. */
    public static final String POSSIBLEDOSAGESTMT = "  Possible Dosages To Auto-Create: ";

    /** The Constant PACKAGE. */
    public static final String PACKAGE = "  Package: ";

    /** The Constant ADD. */
    public static final String ADD = "ADDED";

    /** The Constant EDIT. */
    public static final String EDIT = "EDITED";

    /** The Constant DELETE. */
    public static final String DELETE = "DELETED";

    /** The Constant ADDINT. */
    public static final String ADDINT = "ADDED INTERACTION";

    /** The Constant EDITINT. */
    public static final String EDITINT = "EDITED INTERACTIONS";

    /** The Constant INACTIVATEINT. */
    public static final String INACTIVATEINT = "INACTIVATED INTERACTIONS";

    /** The Constant USE. */
    public static final String USE = "  USE";

  /**
   * The standard date format string used in the reports.
   */
  private static final String STANDARD_DATE_FORMAT = "MM/dd/yyyy";
  
  /**
   * A {@link DateTimeFormatter} for {@link #STANDARD_DATE_FORMAT} for convenience and efficiency.
   */
  private static final DateTimeFormatter STANDARD_DATE_TIME_FORMATTER = DateTimeFormatter.ofPattern(STANDARD_DATE_FORMAT);
  
  /**
   * Constructs an instance with the required dependencies.
   * @param pEplNdfOutgoingDifferencesDao The {@link EplNdfOutgoingDifferencesDao} to use.
   * @param pEplRematchSuggestionDao The {@link EplRematchDao} to use.
   * @param pEplProductDao The {@link EplProductDao} to use.
   */
  @Autowired
  public MessageDataCapabilityImpl(
      final EplNdfOutgoingDifferencesDao pEplNdfOutgoingDifferencesDao,
      final EplRematchDao pEplRematchSuggestionDao,
      final EplProductDao pEplProductDao,
      final UpdateFileReportGenerator pUpdateFileReportGenerator) {
    eplNdfOutgoingDifferencesDao = pEplNdfOutgoingDifferencesDao;
    eplRematchDao = pEplRematchSuggestionDao;
    eplProductDao = pEplProductDao;
    reportGenerator = pUpdateFileReportGenerator;
  }

    /*
     * (non-Javadoc)
     *
     * @see gov.va.med.pharmacy.peps.service.common.capability.MessageDataCapability#messageReportData(java.util.Date,
     * gov.va.med.pharmacy.peps.service.common.utility.NdfUpdateProcessFile)
     */
    @Override
    public void messageReportData(Date startDTM, NdfUpdateProcessFile ndfUpdateFile) throws IOException {
      
        String report = makeReport(startDTM, new Date()).stream().collect(Collectors.joining(NEWLINE));
        
        ndfUpdateFile.putNextRow(report);

    }

    /**
     * Assembles all of the "MESSAGE" (not "MESSAGE2") report data.
     * @param startDate the start time of the update file generation process
     * @return a list with one element per line. May be unmodifiable: do not try to add elements to the list.
     */
    private List<String> getReportData(Date startDate) {
      try {
        List<String> result = new ArrayList<String>();
        result.addAll(addedProductReport(startDate));
        result.addAll(inactivatedCMOPReport(startDate));
        result.addAll(futureInactivationReport());
        result.addAll(inactivatedRematchReport(startDate));
        result.addAll(inactivatedProductsReport(startDate));
        result.addAll(reactivatedProductReport(startDate));
        result.addAll(nationalFormularyChangeReport(startDate));
        result.addAll(formularyDesignatorChangeReport(startDate));
        result.addAll(copayTierChangesReport(startDate));
        result.addAll(genNameAllChangeReport(startDate));
        result.addAll(printNameChangeReport(startDate));
        result.addAll(cmopChangeReport(startDate));
        result.addAll(strengthChangeReport(startDate));
        result.addAll(drugUnitsChangeReport(startDate));
        result.addAll(dispenseUnitsChangeReport(startDate));
        result.addAll(scheduleAllChangeReport(startDate));
        result.addAll(possibleDosageReport(startDate));
        result.addAll(overrideDFChckExclReport(startDate));
        result.addAll(dosageFormChangeReport(startDate));
        try {
          result.addAll(otherRematchedReport(startDate));
          result.addAll(otherRematchSuggReport(startDate));
          result.addAll(otherRematchNoSuggReport(startDate));
        } catch (final Exception ex) {
          LOG.warn(
              "Ignoring exception thrown while generating other rematch sections. Some or all of these section will be "
              + "missing from the report.",
              ex);
        }
        result.addAll(classChangeReport(startDate));
        result.addAll(newClassReport(startDate));
       
        result.add(EMPTYLINE);
        
        return result;

      } catch (final Exception ex) {
        LOG.error(REPORT_FAILURE_MESSAGE, ex);
        return Arrays.asList(REPORT_FAILURE_MESSAGE);
      }
    }

    private List<String> makeReport(Date startDTM, Date today) {
        List<String> reportData = getReportData(startDTM);
        MessageFormatter formatter = new MessageFormatter(today, reportData);
        return formatter.apply(NDFUpdateFileMapping.MESSAGE);
    }

    

    /*
     * (non-Javadoc)
     *
     * @see gov.va.med.pharmacy.peps.service.common.capability.MessageDataCapability#message2ReportData(java.util.Date,
     * gov.va.med.pharmacy.peps.service.common.utility.NdfUpdateProcessFile)
     */
    @Override

    public void message2ReportData(Date startDTM, NdfUpdateProcessFile ndfUpdateFile) throws IOException {
            try {
                    
            List<String> result = message2ReportData(startDTM);
             
            MessageFormatter formatter = new MessageFormatter(new Date(), result);
             
            List<String> formattedResult =  formatter.apply(NDFUpdateFileMapping.MESSAGE2);
            
            String formattedString = formattedResult.stream().collect(Collectors.joining(NEWLINE));
            
            ndfUpdateFile.putNextRow(formattedString);
            
        } catch (Exception e) {
            LOG.error("error processing row for message data ");
        }

    }

    
    /**
     * Adds the strength change report to the message for changes to File 50.68 Field 2.
     * 
     * @param startDTM
     * @return list
     */
    protected List<String> strengthChangeReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> strengthChangeProducts =
                eplNdfOutgoingDifferencesDao.messageGenericChange(VistaFileNumber.FIFTY_68.getFileNumber(), "2", startDTM);
        if (strengthChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.STRENGTH));
        for (Object[] strengthChangeProduct : strengthChangeProducts) {
            StrengthChangeProductMapper mapper = new StrengthChangeProductMapper(strengthChangeProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(OLD + mapper.getOldValue());
            result.add(NEW + mapper.getNewValue());
            result.add(EMPTYLINE);
        }
        return result;
    }

    static class StrengthChangeProductMapper {
        //These index values must match the order of fields in the messageGenericChangeReport query in the DAO 
        final int NEW_VALUE = 0;
        final int OLD_VALUE = 1;
        final int PRODUCT_NAME = 2;
        final int CMOP_ID = 3;
        Object[] data;
        StrengthChangeProductMapper(Object[] row) {
            this.data = row;
        }
        String getOldValue() {
            return (String) data[OLD_VALUE];
        }
        String getNewValue() {
            return (String) data[NEW_VALUE];
        }
        String getProductName() {
            return (String) data[PRODUCT_NAME];
        }
        String getCmopId() {
            return (String) data[CMOP_ID];
        }
    }

        
    /**
     * Adds the override dosage form checks report to the message for changes to File 50.68 Field 31.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> overrideDFChckExclReport(Date startDTM) {
        
        List<String> result = new ArrayList<String>();
        List<Object[]> overrideDFCheckkExcls =
                eplNdfOutgoingDifferencesDao.findOverrideDFChkExcl(VistaFileNumber.FIFTY_68.getFileNumber(), "31", startDTM);
        if (overrideDFCheckkExcls.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.OVERRIDEDOSECHECK));
        for (Object[] overrideDFCheckkExcl : overrideDFCheckkExcls) {
            OverrideDfCheckExclsMapper mapper = new OverrideDfCheckExclsMapper(overrideDFCheckkExcl);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(getChangeString(mapper.getNewValue()));
            result.add(EMPTYLINE);
        }
      
        return result;
    }

    static class OverrideDfCheckExclsMapper {
        Object[] data;
        OverrideDfCheckExclsMapper(Object[] row) {
            this.data = row;
        }
        String getNewValue() {
            return (String) data[0];
        }
        String getProductName() {
            return (String) data[1];
        }
        String getCmopId() {
            return (String) data[2];
        }
    }

   
    /**
     * Adds the National Formulary Change Report to the message for changes to 50.68 Field 17.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> nationalFormularyChangeReport(
            Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> activeNationalFormularyProds =
                eplNdfOutgoingDifferencesDao.findActiveNatFormularyProducts(VistaFileNumber.FIFTY_68.getFileNumber(), "17",
                        startDTM);
        List<Object[]> inactiveNationalFormularyProds =
                eplNdfOutgoingDifferencesDao.findInactiveNatFormularyProducts(VistaFileNumber.FIFTY_68.getFileNumber(), "17",
                        startDTM);

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.NATIONALFORMULARY));

        result.add(SPACE + FORMULARY);
        if (!activeNationalFormularyProds.isEmpty()) {
            for (Object[] activeNationalFormularyProd : activeNationalFormularyProds) {
                NationalFormularyProdMapper mapper = new NationalFormularyProdMapper(activeNationalFormularyProd);
                result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
                result.add(EMPTYLINE);
            }
        } else {
            result.add(doubleIndent(NONE));
            result.add(EMPTYLINE);
        }

        result.add(SPACE + NONFORMULARY);
        if (!inactiveNationalFormularyProds.isEmpty()) {
            for (Object[] inactiveNationalFormularyProd : inactiveNationalFormularyProds) {
                NationalFormularyProdMapper mapper = new NationalFormularyProdMapper(inactiveNationalFormularyProd);
                result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
                result.add(EMPTYLINE);
            }
        } else {
            result.add(doubleIndent(NONE));
            result.add(EMPTYLINE);
        }

        return result;
    }

    static class NationalFormularyProdMapper {
        Object[] data;
        NationalFormularyProdMapper(Object[] row) {
            this.data = row;
        }
        String getNewValue() {
            return (String) data[0];
        }
        String getProductName() {
            return (String) data[1];
        }
        String getCmopId() {
            return (String) data[2];
        }
    }

    
    /**
     * Adds the inactivated CMOP report to the message for changes to File 50.68 Field 7.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> inactivatedCMOPReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> inactiveCMOPProds = eplNdfOutgoingDifferencesDao.findInactivatedCMOPProducts(startDTM);
        if (inactiveCMOPProds.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.INACTIVATEDCMOP));
        for (Object[] inactiveCMOPProd : inactiveCMOPProds) {
            result.add(formatPairWithSquareBraces(
                    (String) inactiveCMOPProd[1],
                    (String) inactiveCMOPProd[2]));
            result.add(EMPTYLINE);
        }
        return result;
    }

  /**
   * Adds the Formulary Designator Report to the message for changes to File {@code VistaProductFile.FILE_NUMBER} 
   * Field {@code VistaProductFile.FORMULARY_DESIGNATOR_FIELD_NUMBER}.
   * 
   * @param startDTM The date the report started.  All records before this date which have not been associated to a file will
   *    be included in the report section. 
   * @return list containing formatted output of section header and data, or section header and NONE if no records changed.
   */
  private List<String> formularyDesignatorChangeReport(Date startDTM) {
    List<String> result = new ArrayList<String>();
    List<Object[]> activeFormularyDesignators = eplNdfOutgoingDifferencesDao.findFormularyDesignatorProducts(startDTM);

    result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.FORMULARYDESIGNATOR));

    if (!activeFormularyDesignators.isEmpty()) {
      for (Object[] activeFormularyDesignator : activeFormularyDesignators) {
        FormularyDesignatorMapper mapper = new FormularyDesignatorMapper(activeFormularyDesignator);
        result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
        result.add(indentLevel1(mapper.getNewValue()));
        result.add(EMPTYLINE);
      }
    } else {
      result.add(doubleIndent(NONE));
      result.add(EMPTYLINE);
    }

    return result;
  }  

    /**
     * Class for mapping findFormularyDesignatorProducts results to getter methods for easier readability.
     */
    static class FormularyDesignatorMapper {
        private final int NEW_VALUE = 0;
        private final int PRODUCT_NAME = 1;
        private final int CMOP_ID = 2;
        private Object[] data;
        FormularyDesignatorMapper(Object[] row) {
            this.data = row;
        }
        String getNewValue() {
            return (String) data[NEW_VALUE];
        }
        String getProductName() {
            return (String) data[PRODUCT_NAME];
        }
        String getCmopId() {
            return (String) data[CMOP_ID];
        }
    }
    
    /**
     * Adds the added product report to the message for changes to File 50.68 Field .01.
     * 
     * @param startDTM
     * @return
     */
    protected List<String> addedProductReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> addedProds =
                eplNdfOutgoingDifferencesDao.findAddedProducts(VistaFileNumber.FIFTY_68.getFileNumber(), ".01", startDTM);
        boolean initialData = true;

        if ((addedProds != null) && !addedProds.isEmpty()) {
            result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.ADDEDPRODUCT));

            String processedParent = "None";
            int maxDataLimit = 4;
            int currentDataIndex = 0;

            for (Object[] addedProd : addedProds) {
                AddedProductMapper mapper = new AddedProductMapper(addedProd);
                String currentParent = mapper.getProductName();
                if (processedParent.equalsIgnoreCase(currentParent)) {
                    if (currentDataIndex < maxDataLimit) {
                        if ((mapper.getNdcNumber() != null) && !mapper.getNdcNumber().isEmpty()) {
                            result.add(SPACER + mapper.getNdcNumber());
                            currentDataIndex++;
                        }
                    } else {
                        if ((mapper.getNdcNumber() != null) && !mapper.getNdcNumber().isEmpty()) {
                            currentDataIndex = 0;
                            result.add(EMPTYLINE);
                            result.add(doubleIndent(mapper.getNdcNumber()));
                            currentDataIndex++;
                        }
                    }
                    processedParent = currentParent;
                } else {
                    currentDataIndex = 0;
//                    if (!initialData) result.add(EMPTY_STRING); //
                    result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));

                    String data2 = DISPENSE + mapper.getDispenseUnitName() + ENDROUNDBRAC;
                    result.add(doubleIndent(data2));

                    if ((mapper.getNdcNumber() != null) && !mapper.getNdcNumber().isEmpty()) {
                        result.add(doubleIndent(mapper.getNdcNumber()));
//                        result.add(EMPTY_STRING); //
                        currentDataIndex++;
                    }

                    processedParent = currentParent;
                    initialData = false;
                }
                result.add(EMPTYLINE);
            }
        }

        return result;
    }

    static class AddedProductMapper {
        Object[] data;
        AddedProductMapper(Object[] row) {
            this.data = row;
        }
        String getNewValue() {
            return (String) data[0];
        }
        String getProductName() {
            return (String) data[1];
        }
        String getCmopId() {
            return (String) data[2];
        }
        String getDispenseUnitName() {
            return (String) data[3];
        }
        String getNdcNumber() {
            return (String) data[4];
        }
    }

    static class AddedProductMapper2 {
        Object[] data;
        AddedProductMapper2(Object[] row) {
            this.data = row;
        }
        String getNewValue() {
            return (String) data[0];
        }
        String getProductName() {
            return (String) data[1];
        }
        String getCmopId() {
            return (String) data[2];
        }
        String getClassification() {
            return (String) data[3];
        }
        // just an alias..
        String getCode() {
            return getNewValue();
        }
    }

    
    
    /**
     * Possible dosage report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> possibleDosageReport(Date startDTM) {
       
        List<String> result = new ArrayList<String>();
        List<Object[]> yesPossibleDosageProds =
                eplNdfOutgoingDifferencesDao.findYesPossibleDosageProducts(VistaFileNumber.FIFTY_68.getFileNumber(), "40",
                        startDTM);
        List<Object[]> noPossibleDosageProds =
                eplNdfOutgoingDifferencesDao.findNoPossibleDosageProducts(VistaFileNumber.FIFTY_68.getFileNumber(), startDTM);

        if (yesPossibleDosageProds.isEmpty() && noPossibleDosageProds.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.POSSIBLEDOSAGE));

        for (Object[] yesPossibleDosageProd : yesPossibleDosageProds) {
            YesPossibleDosageMapper mapper = new YesPossibleDosageMapper(yesPossibleDosageProd);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(doubleIndent(AUTOCREATEQUEST + getChangeStringPossibleDosage(mapper.getNewValue())));
            result.add(EMPTYLINE);
        }

        for (Object[] noPossibleDosageProd : noPossibleDosageProds) {
            NoPossibleDosageMapper mapper = new NoPossibleDosageMapper(noPossibleDosageProd);
            if ("41".equals(mapper.getVistaFieldNumber())) {
                result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
                result.add(doubleIndent(AUTOCREATENEG));
                result.add(doubleIndent(POSSIBLEDOSAGESTMT + getChangeStringDosageAutoCreate(mapper.getNewValue())));
            } else if ("42".equals(mapper.getVistaFieldNumber())) {
                result.add(doubleIndent(PACKAGE + getChangeStringPackage(mapper.getNewValue())));
            }
            result.add(EMPTYLINE);
        }

        return result;
    }

    static class YesPossibleDosageMapper {
        Object[] data;
        YesPossibleDosageMapper(Object[] row) {
            this.data = row;
        }
        String getNewValue() {
            return (String) data[0];
        }
        String getProductName() {
            return (String) data[1];
        }
        String getCmopId() {
            return (String) data[2];
        }
    }

    static class NoPossibleDosageMapper {
        Object[] data;
        NoPossibleDosageMapper(Object[] row) {
            this.data = row;
        }
        String getVistaFieldNumber() {
            return (String) data[0];
        }
        String getNewValue() {
            return (String) data[1];
        }
        String getVistaIen() {
            return (String) data[2];
        }
        String getProductName() {
            return (String) data[3];
        }
        String getCmopId() {
            return (String) data[4];
        }
    }

   

    /**
     * Prints the name change report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> printNameChangeReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> printNameChangeProducts =
                eplNdfOutgoingDifferencesDao.messageGenericChange(VistaFileNumber.FIFTY_68.getFileNumber(), "5", startDTM);
        if (printNameChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.PRINTNAMECHANGE));
        for (Object[] printNameChangeProduct : printNameChangeProducts) {
            ChangeProductMapper mapper = new ChangeProductMapper(printNameChangeProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(),
                    mapper.getCmopId()));
            result.add(OLD + mapper.getOldValue());
            result.add(NEW + mapper.getNewValue());
            result.add(EMPTYLINE);
        }
        return result;
    }

    static class ChangeProductMapper {
        Object[] data;
        ChangeProductMapper(Object[] row) {
            this.data = row;
        }
        String getNewValue() {
            return (String) data[0];
        }
        String getOldValue() {
            return (String) data[1];
        }
        String getProductName() {
            return (String) data[2];
        }
        String getCmopId() {
            return (String) data[3];
        }
    }

    
    
    /**
     * Drug units change report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> drugUnitsChangeReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> unitsChangeProducts =
                eplNdfOutgoingDifferencesDao.drugUnitsChange(VistaFileNumber.FIFTY_68.getFileNumber(), "3", startDTM);
        if (unitsChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.DRUGUNITCHANGE));
        for (Object[] unitsChangeProduct : unitsChangeProducts) {
            GenNameChangeProductMapper mapper = new GenNameChangeProductMapper(unitsChangeProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(OLD + mapper.getOldName());
            result.add(NEW + mapper.getNewName());
            result.add(EMPTYLINE);
        }
        return result;
    }

    
    /**
     * Dispense units change report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> dispenseUnitsChangeReport(Date startDTM) {
        
        List<String> result = new ArrayList<String>();
        List<Object[]> unitsChangeProducts =
                eplNdfOutgoingDifferencesDao.dispenseUnitsChange(VistaFileNumber.FIFTY_68.getFileNumber(), "8", startDTM);
        if (unitsChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.DISPENSEUNITCHANGE));
        for (Object[] unitsChangeProduct : unitsChangeProducts) {
            GenNameChangeProductMapper mapper = new GenNameChangeProductMapper(unitsChangeProduct);
            result.add(doubleIndent(mapper.getProductName()));
            result.add(OLD + mapper.getOldName());
            result.add(NEW + mapper.getNewName());
            result.add(EMPTYLINE);
        }
        return result;
    }

    
    /**
     * Class change report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> classChangeReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> classChangeProducts =
                eplNdfOutgoingDifferencesDao.classChange(VistaFileNumber.FIFTY_68.getFileNumber(), "15", startDTM);
        if (classChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.VACLASSCHANGE));
        for (Object[] classChangeProduct : classChangeProducts) {
            ClassChangeProductMapper mapper = new ClassChangeProductMapper(classChangeProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(OLD + mapper.getOldClass());
            result.add(NEW + mapper.getNewClass());
            result.add(EMPTYLINE);
        }
        return result;
    }

    /**
     * Class change report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> newClassReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> newClassProducts =
                eplNdfOutgoingDifferencesDao.findNewClassProducts(VistaFileNumber.FIFTY_605.getFileNumber(), ".01", startDTM);
        if (newClassProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.NEWVADRUGCLASS));
        for (Object[] newClassProduct : newClassProducts) {
            AddedProductMapper2 mapper = new AddedProductMapper2(newClassProduct);
            result.add(formatPairWithSquareBraces(mapper.getCode(), mapper.getClassification()));
            result.add(EMPTYLINE);
        }
        return result;
    }

    static class ClassChangeProductMapper {
        Object[] data;
        ClassChangeProductMapper(Object[] row) {
            this.data = row;
        }
        String getOldClass() {
            String result = (String) data[5];
            if (result == null) return ONE_SPACE;
            return result;
        }
        String getNewClass() {
            String result = (String) data[4];
            if (result == null) return ONE_SPACE;
            return result;
        }
        String getProductName() {
            return (String) data[2];
        }
        String getCmopId() {
            return (String) data[3];
        }
    }

       
    /**
     * Cmop change report.
     * @param startDTM
     * @return
     */
    private List<String> cmopChangeReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> cmopChangeProducts = eplNdfOutgoingDifferencesDao.findCMOPIdChanges(startDTM);
        
        if (cmopChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.CMOPCHANGE));
        for (Object[] cmopChangeProduct : cmopChangeProducts) {
            ChangeProductMapper mapper = new ChangeProductMapper(cmopChangeProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(),
                    mapper.getCmopId()));
            result.add(OLD + mapper.getOldValue());
            result.add(NEW + mapper.getNewValue());
            result.add(EMPTYLINE);
        }
        return result;
    }


    /**
     * Generates the future inactivation reports, both for products with rematches and without rematches.
     */
    private List<String> futureInactivationReport() {

        // Find all future product inactivations.
        final List<FutureInactiveProductDto> futureInactivatedProducts = eplProductDao.findFutureInactive();
        LOG.debug("{} products have future inactivations.", futureInactivatedProducts.size());

        // Load all future inactivation rematches.
        final List<EplRematchDo> futureRematches =
            eplRematchDao.retrieveByMatchType(MatchReason.FUTURE_INACTIVATION.getId());
        LOG.debug("There are {} future inactivation rematches.", futureRematches.size());
        final Map<Long, EplRematchDo> productToRematchMap = futureRematches.stream()
            .collect(Collectors.toMap(rematch -> rematch.getFromProduct().getEplId(), Function.identity()));
        
        /*-
         * The below code builds the reports in two stages:
         * 
         * 1) We iterate through all of the product inactivations and build the content of each section (inactivations
         *    with rematches and inactivations without rematches.)
         * 
         * 2) We then add the content sections with their appropriate headers to the report itself.
         */

        /* Stage 1 */
        final ArrayList<String> withAutomaticRematchContent = new ArrayList<>();
        final ArrayList<String> withSuggestedRematchContent = new ArrayList<>();
        final ArrayList<String> withoutRematchContent = new ArrayList<>();
        for (final FutureInactiveProductDto futureInactive : futureInactivatedProducts) {
          // This list will contain the report content for this particular product whether it has a rematch or not.
          final ArrayList<String> productContent = new ArrayList<>();
          // This variable will be set below to point to the correct content section. We will use it later to append
          // productContent to the correct content section.
          final ArrayList<String> targetContent;
          
          // This part is the same whether with a rematch or without.
          productContent.add(formatPairWithSquareBraces(futureInactive.getVaProductName(), futureInactive.getCmopId()));
          productContent.add(doubleIndent(formatDate(futureInactive.getFutureDate())));

          // This part depends on whether there is a rematch or not.
          final long productId = futureInactive.getEplId().longValue();
          final EplRematchDo productRematch = productToRematchMap.get(productId);
          if (productRematch != null) {
            // With rematch.
            productContent.add(doubleIndent(USE));
            if (productRematch.getToAutoProduct() != null) {
              LOG.debug("Product {} has an automatic rematch.", productId);
              targetContent = withAutomaticRematchContent;
              productContent.add(formatPairWithSquareBraces(
                  productRematch.getToAutoProduct().getVaProductName(), productRematch.getToAutoProduct().getCmopId()));
            } else {
              LOG.debug("Product {} has a suggested rematch.", productId);
              targetContent = withSuggestedRematchContent;
              for (EplProductDo productDo : productRematch.getToSuggestedProducts()) {
                productContent.add(formatPairWithSquareBraces(productDo.getVaProductName(), productDo.getCmopId()));
              }
            }
          } else {
            // Without rematch.
            LOG.debug("Product {} does not have a rematch.", productId);
            targetContent = withoutRematchContent;
          }
          
          // This part is the same either way again.
          productContent.add(EMPTYLINE);
          
          // Add the whole product content to the correct content list.
          targetContent.addAll(productContent);
        }
        
        /* Stage 2 */

        final List<String> result = new ArrayList<String>();

        // Inactivations with automatic rematches. This report is incomplete and will be implemented fully under Story
        // 376721.
        if (!withAutomaticRematchContent.isEmpty()) {
          LOG.info("Skipping automatic rematch content until we define the header text.");
        }

        // Inactivations with suggested rematches.
        if (!withSuggestedRematchContent.isEmpty()) {
          result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.PROPOSEDINACTIVATEDWSUGG));
          result.addAll(withSuggestedRematchContent);
        }

        // Inactivations without rematches.
        if (!withoutRematchContent.isEmpty()) {
          result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.FUTUREINACTIVE));
          result.addAll(withoutRematchContent);
        }

        return result;
    }

    private String formatDate(final Instant date) {
        // Assumption: the date was stored in the current time zone.
        return date.atZone(ZoneId.systemDefault()).format(STANDARD_DATE_TIME_FORMATTER);
    }
    
  /**
   * <p>
   * Formats the given optional date using the standard format. If the Optional is "empty", returns a string of spaces of the
   * same length as an actual date string.
   * </p>
   * 
   * <p>
   * For example, if a "present" date would return <code>"12/25/2016"</code>, then an "empty" date would return
   * <code>"          "</code>.
   * </p>
   * 
   * @param optionalDate a date that may not be present
   * @return
   */
  private String formatOptionalDate(Optional<LocalDate> optionalDate) {
    if (optionalDate.isPresent()) {
      return optionalDate.get().format(STANDARD_DATE_TIME_FORMATTER);
    } else {
      // Ensure it's the correct length by formatting a sample date.
      final int length = LocalDate.now().format(STANDARD_DATE_TIME_FORMATTER).length();
      return StringUtils.repeat(" ", length);
    }
  }
    
    private String doubleIndent(String data) {
        return DOUBLEDSPACE + data;
    }

    private String indentLevel1(String data) {
        return DOUBLEDSPACE + data;
    }    

    private String indentLevel2(String data) {
        return DOUBLEDSPACE + DOUBLEDSPACE + data;
    }    
    
    protected final String formatPairWithSquareBraces(String item1, String item2) {
        
        StringBuffer buf = new StringBuffer();
        
        buf.append(DOUBLEDSPACE);
        buf.append(item1);
        buf.append(SPACER);
        buf.append(inSquareBraces(item2));
        
      return buf.toString();
    }
    
    private String formatWithDoubleSpace(String item1, String item2) {
        
        StringBuffer buf = new StringBuffer();
        buf.append(DOUBLEDSPACE);
        buf.append(item1);
        buf.append(SPACER3);        
        buf.append(item2);
        
        return buf.toString();  
        
        
    }

    private String inSquareBraces(String item) {
        return OPENING_SQUARE + item + END_SQUARE;
    }

    
    /**
     * Schedule all change report.
     * @param startDTM
     * @return list
     */
    private List<String> scheduleAllChangeReport(Date startDTM) {
      
        List<String> result = new ArrayList<String>();
        List<Object[]> scheduleChangeProducts =
                eplNdfOutgoingDifferencesDao.scheduleAllChange(VistaFileNumber.FIFTY_68.getFileNumber(), "19", startDTM);
        if (scheduleChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.SCHEDULECHANGEALL));
       
        for (Object[] scheduleChangeProduct : scheduleChangeProducts) {
            ScheduleChangeProductMapper mapper = new ScheduleChangeProductMapper(scheduleChangeProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(OLD + mapper.getOldSchedule());
            result.add(NEW + mapper.getNewSchedule());
            result.add(EMPTYLINE);
        }
      
        return result;
    }

    static class ScheduleChangeProductMapper {
        Object[] data;
        ScheduleChangeProductMapper(Object[] row) {
            this.data = row;
        }
        String getOldSchedule() {
            return (String) data[5];
        }
        String getNewSchedule() {
            return (String) data[4];
        }
        String getProductName() {
            return (String) data[2];
        }
        String getCmopId() {
            return (String) data[3];
        }
    }

        
    /**
     * Gen name all change report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> genNameAllChangeReport(Date startDTM) {
       
        List<String> result = new ArrayList<String>();
        List<Object[]> genNameChangeProducts =
                eplNdfOutgoingDifferencesDao.genNameAllChange(VistaFileNumber.FIFTY_68.getFileNumber(), ".05", startDTM);
        if (genNameChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.GENERICNAMECHANGE));
        for (Object[] genNameChangeProduct : genNameChangeProducts) {
            GenNameChangeProductMapper mapper = new GenNameChangeProductMapper(genNameChangeProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(OLD + mapper.getOldName());
            result.add(NEW + mapper.getNewName());
            result.add(EMPTYLINE);
        }
        return result;
    }

    static class GenNameChangeProductMapper {
        Object[] data;
        GenNameChangeProductMapper(Object[] row) {
            this.data = row;
        }
        String getOldName() {
            String result = (String) data[5];
            if (result == null) return ONE_SPACE;
            return result;
        }
        String getNewName() {
            String result = (String) data[4];
            if (result == null) return ONE_SPACE;
            return result;
        }
        String getProductName() {
            return (String) data[2];
        }
        String getCmopId() {
            return (String) data[3];
        }
    }

   
    /**
     * Dosage form change report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> dosageFormChangeReport(Date startDTM) {
        
        List<String> result = new ArrayList<String>();
        List<Object[]> dosageFormChangeProducts =
                eplNdfOutgoingDifferencesDao.dosageFormChange(VistaFileNumber.FIFTY_68.getFileNumber(), "1", startDTM);
        if (dosageFormChangeProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.DOSAGEFORMCHANGE));
        
        for (Object[] dosageFormChangeProduct : dosageFormChangeProducts) {
            GenNameChangeProductMapper mapper = new GenNameChangeProductMapper(dosageFormChangeProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(OLD + mapper.getOldName());
            result.add(NEW + mapper.getNewName());
            result.add(EMPTYLINE);
        }
       
        return result;
    }

  /**
   * Generates the suggested and automatic rematch reports for inactivations.
   * 
   * @return list of Strings to include in the report, one for each line
   */

  private List<String> inactivatedRematchReport(Date startDTM) {

    List<String> automaticRematches = new ArrayList<String>();
    List<String> suggestedRematches = new ArrayList<String>();
    List<EplRematchDo> rematchesForInactivation = eplRematchDao.findRematchesForInactivations(startDTM);

    /*-
     * The below code builds the reports in two stages:
     * 
     * 1) We iterate through all of the product rematches and build the content of each section (suggested rematches
     * and automatic rematches).
     * 
     * 2) We then add the content sections with their appropriate headers to the report itself.
     */

    /* Stage 1 */
    for (EplRematchDo rematch : rematchesForInactivation) {
      // Adds the inactivated products with automatic rematches to the report.
      if (rematch.getToAutoProduct() != null) {
        automaticRematches.addAll(formatRematchForReport(rematch));
      }
      // Adds the inactivated products with suggested rematches to the report.
      else {
        suggestedRematches.addAll(formatRematchForReport(rematch));
      }
    }

    /* Stage 2 */
    List<String> result = new ArrayList<String>();

    if (!automaticRematches.isEmpty()) {
      result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.INACTIVATEDREMATCHED));
      result.addAll(automaticRematches);
    }

    if (!suggestedRematches.isEmpty()) {
      result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.INACTIVATEDWSUGG));
      result.addAll(suggestedRematches);
    }

    return result;
  }

  /**
   * Returns the rematch formatted for the Data Update for NDF report.
   */
  private List<String> formatRematchForReport(final EplRematchDo rematch) {
    final ArrayList<String> result = new ArrayList<>();
    result.add(formatPairWithSquareBraces(
        rematch.getFromProduct().getVaProductName(), rematch.getFromProduct().getCmopId()));
    result.add(doubleIndent(USE));
    for (final EplProductDo product : rematch.getToProducts()) {
      result.add(formatPairWithSquareBraces(product.getVaProductName(), product.getCmopId()));
    }
    result.add(EMPTYLINE);
    return result;
  }

    private static class InactiveProductMapper {
        Object[] data;
        InactiveProductMapper(Object[] row) {
            this.data = row;
        }
        String getProductName() {
            return (String) data[3];
        }
        String getCmopId() {
            return (String) data[4];
        }
    }

    
    /**
     * Other rematched report.
     * @param startDTM
     * @return list
     */
    private List<String> otherRematchedReport(Date startDTM) {
       
        List<String> result = new ArrayList<String>();
        List<Object[]> rematchProducts = eplRematchDao.findRematchAuto(startDTM);
        if ((rematchProducts != null) && rematchProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.OTHERREMATCHED));
      
        for (Object[] rematchProduct : rematchProducts) {
            RematchProductMapper mapper = new RematchProductMapper(rematchProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(), mapper.getCmopId()));
            result.add(doubleIndent(USE));
            result.add(formatPairWithSquareBraces(mapper.getRematchedProductName(), mapper.getRematchedCmopId()));
            result.add(EMPTYLINE);
        }
       
        return result;
    }

    static class RematchProductMapper {
        Object[] data;
        RematchProductMapper(Object[] row) {
            this.data = row;
        }
        String getRematchedProductName() {
            return (String) data[4];
        }
        String getRematchedCmopId() {
            return (String) data[5];
        }
        String getProductName() {
            return (String) data[2];
        }
        String getCmopId() {
            return (String) data[3];
        }
    }

   
    /**
     * Other rematch sugg report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> otherRematchSuggReport(Date startDTM) {
        
        List<String> result = new ArrayList<String>();
        List<Object[]> rematchProducts = eplRematchDao.findRematchSugg(startDTM);
        if (rematchProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.OTHERREMATCHSUGG));
       
        for (Object[] rematchProduct : rematchProducts) {
            RematchProductMapper mapper = new RematchProductMapper(rematchProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(),
                    mapper.getCmopId()));
            result.add(doubleIndent(USE));
            result.add(formatPairWithSquareBraces(mapper.getRematchedProductName(),
                    mapper.getRematchedCmopId()));
            result.add(EMPTYLINE);
        }
       
        return result;
    }

    /**
     * Other rematch no sugg report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> otherRematchNoSuggReport(Date startDTM) {
        
        List<String> result = new ArrayList<String>();
        List<Object[]> rematchProducts = eplRematchDao.findRematchNoSugg(startDTM);
        
        if (rematchProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.OTHERNOREMATCH));
        
        for (Object[] rematchProduct : rematchProducts) {
            RematchProductMapper mapper = new RematchProductMapper(rematchProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(),
                    mapper.getCmopId()));
            result.add(EMPTYLINE);
        }
        
        return result;
    }

   

    /**
     * 
     * @param startDTM
     * @return list
     */
    private List<String> inactivatedProductsReport(
            Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> inactiveProducts = eplNdfOutgoingDifferencesDao.findInactivatedProducts(startDTM);
        if (inactiveProducts.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.INACTIVATEDPRODUCT));
        for (Object[] inactiveProduct : inactiveProducts) {
            InactiveProductMapper mapper = new InactiveProductMapper(inactiveProduct);
            result.add(formatPairWithSquareBraces(mapper.getProductName(),
                    mapper.getCmopId()));
            result.add(EMPTYLINE);
        }
        return result;
    }


        
    /**
     * Reactivated product report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> reactivatedProductReport(Date startDTM) {
        List<String> result = new ArrayList<String>();
        List<Object[]> reactivatedProds = eplNdfOutgoingDifferencesDao
                .findReactivatedProducts(
                        VistaFileNumber.FIFTY_68.getFileNumber(), "21",
                        startDTM);
        if (reactivatedProds.isEmpty()) return result;

        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.REACTIVATEDPRODUCT));
        for (Object[] reactivatedProd : reactivatedProds) {
            ReactivatedProductMapper mapper = new ReactivatedProductMapper(
                    reactivatedProd);
            result.add(formatPairWithSquareBraces(mapper.getProductName(),
                    mapper.getCmopId()));
            result.add(EMPTYLINE);
        }
        return result;
    }

    /**
     * Maps the query results for Reactivated Products to easy to read getters.
     * Note that this code is dependent on the order of the fields in the select portion of the query.
     */
    static class ReactivatedProductMapper {
        private static final int PRODUCT_NAME = 2;
        private static final int CMOP_ID = 3;
        private Object[] data;
        
        /**
         * Instantiates a new ReactivatedProductMapper.
         *
         * @param row Object[] containing one row's results from the query.
         */
        ReactivatedProductMapper(Object[] row) {
            this.data = row;
        }
        String getProductName() {
            return (String) data[PRODUCT_NAME];
        }
        String getCmopId() {
            return (String) data[CMOP_ID];
        }
    }

       
    /**
     *  Med guide change report.
     *  
     * @param startDTM
     * @return list
     */
    private List<String> medGuideChangeReport( Date startDTM) {

         List<String> result = new ArrayList<String>();
         
        List<Object[]> addedMedGuideProds =
                eplNdfOutgoingDifferencesDao.findAddedMedGuideProducts(VistaFileNumber.FIFTY_68.getFileNumber(), "100",
                        startDTM);
        List<Object[]> editedMedGuideProds =
                eplNdfOutgoingDifferencesDao.findEditedMedGuideProducts(VistaFileNumber.FIFTY_68.getFileNumber(), "100",
                        startDTM);
        List<Object[]> deletedMedGuideProds =
                eplNdfOutgoingDifferencesDao.findDeletedMedGuideProducts(VistaFileNumber.FIFTY_68.getFileNumber(), "100",
                        startDTM);

        if (((addedMedGuideProds != null) && !addedMedGuideProds.isEmpty())
                || ((editedMedGuideProds != null) && !editedMedGuideProds.isEmpty())
                || ((deletedMedGuideProds != null) && !deletedMedGuideProds.isEmpty())) {

            result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.MEDGUIDE) );

            if (!addedMedGuideProds.isEmpty()) {

                result.add(SPACE + ADD);

               

                for (Object[] addedMedGuideProd : addedMedGuideProds) {
                    result.add(formatPairWithSquareBraces((String) addedMedGuideProd[1], (String) addedMedGuideProd[2]));
                    result.add(EMPTYLINE);                    
                }
            }

            if (!editedMedGuideProds.isEmpty()) {
                
                 result.add(SPACE + EDIT);

                for (Object[] editedMedGuideProd : editedMedGuideProds) {                    
                    result.add(formatPairWithSquareBraces((String) editedMedGuideProd[1], (String) editedMedGuideProd[2]));
                    result.add(EMPTYLINE);
                }
            }

            if (!deletedMedGuideProds.isEmpty()) {
               
                result.add(SPACE + DELETE);
                
                for (Object[] deletedMedGuideProd : deletedMedGuideProds) {
                    result.add(formatPairWithSquareBraces((String) deletedMedGuideProd[1], (String) deletedMedGuideProd[2]));
                    result.add(EMPTYLINE);    
                }
            }
        }
        return result;
    }

        
    
    /**
     * Ddi change report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> ddiChangeReport(Date startDTM) {

        List<String> result = new ArrayList<String>();
        
        List<Object[]> addedDDIs =
                eplNdfOutgoingDifferencesDao.findAddedDDIs(VistaFileNumber.FIFTY_SIX.getFileNumber(), ".01", startDTM);
        List<Object[]> reactivatedDDIs =
                eplNdfOutgoingDifferencesDao.findReactivatedDDIs(VistaFileNumber.FIFTY_SIX.getFileNumber(), "7", startDTM);
        List<Object[]> editedDDIs =
                eplNdfOutgoingDifferencesDao.findEditedDDIs(VistaFileNumber.FIFTY_SIX.getFileNumber(), "3", startDTM);
        List<Object[]> inactivatedDDIs =
                eplNdfOutgoingDifferencesDao.findInactivatedDDIs(VistaFileNumber.FIFTY_SIX.getFileNumber(), "7", startDTM);

       
        result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.DDI));
        
        
        result.add(SPACE + ADDINT);


        if (((addedDDIs != null) && !addedDDIs.isEmpty()) || ((reactivatedDDIs != null) && !reactivatedDDIs.isEmpty())) {

            if (!addedDDIs.isEmpty()) {

                for (Object[] addedDDI : addedDDIs) {

                    result.add(formatWithDoubleSpace( (String) addedDDI[0], (String) addedDDI[1])  );
                    result.add(EMPTYLINE);
                }
            }

            if (!reactivatedDDIs.isEmpty()) {
                for (Object[] reactivatedDDI : reactivatedDDIs) {

                    result.add(formatWithDoubleSpace( (String) reactivatedDDI[3], (String) reactivatedDDI[4])  );
                    result.add(EMPTYLINE);                    
                }
            }

        } else {

            result.add(doubleIndent(NONE));
            result.add(EMPTYLINE);
        }


        result.add(SPACE + EDITINT );

        if ((editedDDIs != null) && !editedDDIs.isEmpty()) {

            for (Object[] editedDDI : editedDDIs) {
                result.add(formatWithDoubleSpace( (String) editedDDI[1], (String) editedDDI[2] ));
                result.add(EMPTYLINE);
            }
        } else {
           
            result.add(doubleIndent(NONE));
            result.add(EMPTYLINE);
        }

        result.add(SPACE + INACTIVATEINT);

        if (((inactivatedDDIs != null) && !inactivatedDDIs.isEmpty())) {
            for (Object[] inactivatedDDI : inactivatedDDIs) {               
                result.add( formatWithDoubleSpace( (String) inactivatedDDI[3],  (String) inactivatedDDI[4] )  );
                result.add(EMPTYLINE);                
            }
        } else {
             result.add(doubleIndent(NONE));
             result.add(EMPTYLINE);          
        }

        return result;
    }

        
    /**
     * Excluded ddi report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> excludedDDIReport(Date startDTM) {

         List<String> result = new ArrayList<String>();
         
        List<Object[]> excludedDDIProds =
                eplNdfOutgoingDifferencesDao.findExcludedDDIProducts(VistaFileNumber.FIFTY_68.getFileNumber(), "23", startDTM);

        if (((excludedDDIProds != null) && !excludedDDIProds.isEmpty())) {

            result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.EXCLUEDEDDDI));            

            for (Object[] excludedDDIProd : excludedDDIProds) {

                result.add(formatPairWithSquareBraces((String) excludedDDIProd[1], (String) excludedDDIProd[2]));
                result.add(EMPTYLINE);
            }
        }

        return result;
    }

       
    /**
     * Previously excluded ddi report.
     * 
     * @param startDTM
     * @return list
     */
    private List<String> previouslyExcludedDDIReport(Date startDTM) {

        List<String> result = new ArrayList<String>();
        
        List<Object[]> prevExcludedDDIProds =
                eplNdfOutgoingDifferencesDao.findPrevExcludedDDIProducts(VistaFileNumber.FIFTY_68.getFileNumber(), "23",
                        startDTM);

        if (((prevExcludedDDIProds != null) && !prevExcludedDDIProds.isEmpty())) {

            result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.PREVEXCLUEDEDDDI));
            
            for (Object[] prevExcludedDDIProd : prevExcludedDDIProds) {
                result.add(formatPairWithSquareBraces((String) prevExcludedDDIProd[1], (String)  prevExcludedDDIProd[2]));
                result.add(EMPTYLINE);
            }
        }

        return result;
    }

    /**
     * Gets the change string.
     *
     * @param strConversion
     *            the str conversion
     * @return the change string
     */
    private String getChangeString(String strConversion) {
        String actionString = EMPTYLINE;
        if (strConversion.equals("1")) {
            actionString = "  changed from NO to YES";
        } else {
            actionString = "  changed from YES to NO";
        }
        return actionString;

    }

    /**
     * Gets the change string possible dosage.
     *
     * @param strConversion
     *            the str conversion
     * @return the change string possible dosage
     */
    private String getChangeStringPossibleDosage(String strConversion) {
        String actionString = EMPTYLINE;
        if (strConversion.equals("Y")) {
            actionString = "YES";
        } else {
            actionString = "NO";
        }
        return actionString;
    }

    /**
     * Gets the change string dosage auto create.
     *
     * @param strConversion
     *            the str conversion
     * @return the change string dosage auto create
     */
    private String getChangeStringDosageAutoCreate(String strConversion) {
        String actionString = EMPTYLINE;
        if (strConversion.equals("N")) {
            actionString = "NO POSSIBLE DOSAGES";
        } else if (strConversion.equals("B")) {
            actionString = "1x AND 2x POSSIBLE DOSAGES";
        } else if (strConversion.equals("O")) {
            actionString = "1x POSSIBLE DOSAGES";
        }
        return actionString;
    }

    /**
     * Gets the change string package.
     *
     * @param strConversion
     *            the str conversion
     * @return the change string package
     */
    private String getChangeStringPackage(String strConversion) {
        String actionString = EMPTYLINE;
        if (strConversion.equals("I")) {
            actionString = "Inpatient";
        } else if (strConversion.equals("O")) {
            actionString = "Outpatient";

        } else if (strConversion.equals("IO")) {
            actionString = "Both";
        }
        return actionString;

    }
    
  private List<String> message2ReportData(Date startDTM) {

    try {
      final List<String> result = new ArrayList<String>();

      result.addAll(medGuideChangeReport(startDTM));

      result.addAll(ddiChangeReport(startDTM));

      result.addAll(excludedDDIReport(startDTM));

      result.addAll(previouslyExcludedDDIReport(startDTM));

      return result;

    } catch (final Exception e) {
      LOG.error(REPORT_FAILURE_MESSAGE, e);
      return Arrays.asList(REPORT_FAILURE_MESSAGE);
    }
  }
    
    @Override
    public List<String>  generateUpdateFilePreviewReport(Date startDTM) {
        
        // message report data
        ArrayList<String> reportList = new ArrayList<>(getReportData(startDTM));
        
        // append message2 report data.
        reportList.addAll(message2ReportData(startDTM));
      
        return reportList;
    }
    
  /**
   * <p>
   * Generates the lines for the Copay Tier Change Report section. If there were no changes, this section will be blank. Each
   * product may have one or more copay tiers, which will be grouped under a product heading. A special message is shown if all
   * copay tiers have been removed from a product.
   * </p>
   * 
   * <p>
   * The results are similar to the following:
   * </p>
   * 
   * <pre>
   * The COPAY TIER LEVEL settings have been edited for the following VA Products.
   * 
   *     FLUCONAZOLE 50MG TAB     [F0026]
   *         COPAY TIER LEVEL:     1          Start Date:     XX/XX/XXXX          End Date:     XX/XX/XXXX
   *         COPAY TIER LEVEL:     2          Start Date:     XX/XX/XXXX          End Date:     XX/XX/XXXX
   * 
   *     FLUCONAZOLE 100MG TAB     [F0024]
   *         COPAY TIER LEVEL has been Removed.
   * </pre>
   * 
   * @param startDate the start time of the update file generation process
   * @return a List of Strings, with one item per line
   */
  // This method is protected to facilitate testing. In the future we should just test
  // messageReportData() overall and not specifically this method.
  protected final List<String> copayTierChangesReport(Date startDate) {
    List<String> result = new ArrayList<String>();
    List<CopayTierChange> copayTierChanges =
        eplNdfOutgoingDifferencesDao.findCopayTierChanges(startDate);
    
    // Don't include the report at all if there were no changes.
    if (copayTierChanges.isEmpty())
      return result;

    result.addAll(reportGenerator.formatHeaderTextWithPadding(ReportMessageCode.COPAYTIER));

    String previousRecordProductIen = "";

    for (int i = 0; i < copayTierChanges.size(); i++) {
      CopayTierChange tierChange = copayTierChanges.get(i);

      if (!previousRecordProductIen.equals(tierChange.getProductIen())) {
        if (i > 0) {
          // add a line between products
          result.add(EMPTYLINE);
        }
        result.add(formatPairWithSquareBraces(tierChange.getProductName(), tierChange.getCmopId()));
      }

      previousRecordProductIen = tierChange.getProductIen();

      if (VistaFile.ALL_DELETED.equals(tierChange.getCopayTier())) {
        result.add(indentLevel2(COPAY_TIERS_REMOVED_MESSAGE));
      } else {
        result.add(indentLevel2(formatCopayTierRow(tierChange)));
      }
    }

    return result;
  }

  /**
   * Formats the given CopayTierChange for including in the Report. Does not indent the row.
   * @param change the change to format. Should not be a DELETED indicator.
   * @return String representing an unindented CopayTierChange detail line for the report.
   */
  protected String formatCopayTierRow(final CopayTierChange change) {
    return "COPAY TIER LEVEL: " + padRight(change.getCopayTier(), COPAY_TIER_LEVEL_WIDTH)
        + "Start Date: " + padRight(formatOptionalDate(change.getEffectiveDate()), COPAY_TIER_START_DATE_WIDTH)
        + "End Date: " + padRight(formatOptionalDate(change.getEndDate()), COPAY_TIER_END_DATE_WIDTH);
  }

  /**
   * Pad the given value on the right with spaces to ensure a minimum
   * width.
   * 
   * @param value to pad
   * @param width minimum width including padding spaces
   * @return value padded on the right with spaces
   */
  private String padRight(String value, int width) {
    return String.format("%1$-" + width + "s", value);
  }
}
