/**
 * Source file created in 2011 by Southwest Research Institute
 */


package gov.va.med.pharmacy.peps.service.common.capability.impl;


import java.util.Collection;
import java.util.Date;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionException;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;

import gov.va.med.pharmacy.peps.common.exception.ValidationException;
import gov.va.med.pharmacy.peps.common.vo.EntityType;
import gov.va.med.pharmacy.peps.common.vo.EventCategory;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.common.vo.ItemAuditHistoryVo;
import gov.va.med.pharmacy.peps.common.vo.ItemStatus;
import gov.va.med.pharmacy.peps.common.vo.ManagedItemVo;
import gov.va.med.pharmacy.peps.common.vo.ModDifferenceVo;
import gov.va.med.pharmacy.peps.common.vo.NationalSetting;
import gov.va.med.pharmacy.peps.common.vo.NdcVo;
import gov.va.med.pharmacy.peps.common.vo.OrderableItemVo;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.RequestVo;
import gov.va.med.pharmacy.peps.common.vo.Role;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.datafield.DataField;
import gov.va.med.pharmacy.peps.domain.common.capability.ItemAuditHistoryDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.NdcDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.OrderableItemDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.ProductDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.RequestDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.DatabaseDomainCapability;
import gov.va.med.pharmacy.peps.service.common.capability.ItemModDifferences;
import gov.va.med.pharmacy.peps.service.common.capability.ManagedItemCapability;
import gov.va.med.pharmacy.peps.service.common.capability.ProposedInactivateCapability;
import gov.va.med.pharmacy.peps.service.common.capability.RematchCapability;
import gov.va.med.pharmacy.peps.service.common.scheduler.ProcessStatus;
import gov.va.med.pharmacy.peps.service.common.scheduler.ProcessStatusUpdater;

import static gov.va.med.pharmacy.peps.common.utility.security.ESAPIValidationType.LOG_FORGING;
import static gov.va.med.pharmacy.peps.common.utility.security.ESAPIValidator.validateStringInput;


/**
 * Main implementation of {@link ProposedInactivateCapability}.
 */
public class ProposedInactivateCapabilityImpl implements ProposedInactivateCapability {

    private static final org.apache.logging.log4j.Logger LOG = org.apache.logging.log4j.LogManager
        .getLogger(ProposedInactivateCapabilityImpl.class);

  /**
   * The national setting for the job run state.
   */
  static final NationalSetting RUN_STATE_SETTING = NationalSetting.INACTIVATION_RUN_STATE;

  /**
   * The national setting for the job last run time.
   */
  static final NationalSetting LAST_RUN_SETTING = NationalSetting.INACTIVATION_CHECK_LAST_RUN;

    private static final int TIMEOUT = 600;
    private static final Long USER_ID = 42L;
    private static final String STATION = "999";

    private ManagedItemCapability managedItemCapability;
    private NdcDomainCapability ndcDomainCapability;
    private ProductDomainCapability productDomainCapability;
    private OrderableItemDomainCapability orderableItemDomainCapability;
    private RequestDomainCapability requestDomainCapability;
    private ProcessStatusUpdater processStatusUpdater;
    private ItemAuditHistoryDomainCapability itemAuditHistoryDomainCapability;
    private DatabaseDomainCapability databaseDomainCapability;
    private RematchCapability rematchCapability;

    @Autowired
    private PlatformTransactionManager transactionManager;

    @Override
    public void synchronizedProposedInactivationDate() {

        try {

            processStatusUpdater.updateStatus(RUN_STATE_SETTING, ProcessStatus.RUNNING, getUser());

            // Shared transaction template for all of these transactions.
            final TransactionTemplate txTemplate = new TransactionTemplate(transactionManager);
            txTemplate.setPropagationBehavior(TransactionTemplate.PROPAGATION_REQUIRES_NEW);
            txTemplate.setTimeout(TIMEOUT);
            
            // NDC transaction
            txTemplate.execute(new TransactionCallbackWithoutResult() {
              @Override
              protected void doInTransactionWithoutResult(TransactionStatus status) {
                final List<Long> ndcList = databaseDomainCapability.getIdsPropInact(EntityType.NDC);

                LOG.info("Will check these NDCs: {}", ndcList);

                for (Long ndc : ndcList) {
                  checkAndInactivateNdc(ndc);
                }
              }
            });

            // Product transaction
            txTemplate.execute(new TransactionCallbackWithoutResult() {
              @Override
              protected void doInTransactionWithoutResult(TransactionStatus status) {
                final List<Long> productList = databaseDomainCapability.getIdsPropInact(EntityType.PRODUCT);

                LOG.info("Will check these products: {}", productList);

                for (Long product : productList) {
                  checkAndInactivateProduct(product);
                }
              }
            });
            
            // Orderable Item transaction
            txTemplate.execute(new TransactionCallbackWithoutResult() {
              @Override
              protected void doInTransactionWithoutResult(TransactionStatus status) {
                final List<Long> orderableList = databaseDomainCapability.getIdsPropInact(EntityType.ORDERABLE_ITEM);

                LOG.info("Will check these orderable items: {}", orderableList);

                for (Long orderable : orderableList) {
                  checkAndInactivateOrderable(orderable);
                }

              }
            });
            
            // Delete completed requests transaction
            // As part of the proposed inactivation task we will also clear out completed requests.
            txTemplate.execute(new TransactionCallbackWithoutResult() {
              @Override
              protected void doInTransactionWithoutResult(TransactionStatus status) {
                LOG.debug("Before delete completed requests");
                requestDomainCapability.deleteCompletedRequest();
                LOG.debug("after delete completed requests");
              }
            });
            LOG.debug("after delete completed commit.");
            
        } catch (TransactionException e) {
            LOG.error("TransactionException in synchronizedProposedInactivationDate().", e);
        } catch (Exception e) {
            LOG.debug("Non transaction exception in ProposedInactivation.", e);
        }
        
        processStatusUpdater.updateStatusAndRuntime(
            RUN_STATE_SETTING, ProcessStatus.COMPLETED, LAST_RUN_SETTING, getUser());

    }

    /**
     * inactivateNdc. 
     * @param ndcId ndc
     */
    private void checkAndInactivateNdc(Long ndcId) {
      LOG.debug("Checking NDC {}", ndcId);

        // Inactivate the NDC.
        try {
            NdcVo ndcVo = ndcDomainCapability.retrieve(ndcId.toString());

            // Technical debt: could probably use checkAndInactivateItem() here. Also assuming that inactivating an NDC
            // does not require second approval.
            DataField<Date> proposedInativationDate = 
                ndcVo.getVaDataFields().getDataField(FieldKey.PROPOSED_INACTIVATION_DATE);
            
            if (proposedInativationDate.getValue() == null || 
                proposedInativationDate.getValue().before(new Date())) {
            
                ndcVo.setInactivationDate(new Date());
                ndcVo.setItemStatus(ItemStatus.INACTIVE);
                ndcVo.getVaDataFields().getDataField(FieldKey.PROPOSED_INACTIVATION_DATE).selectValue(null);
                ndcDomainCapability.update(ndcVo, getUser());
    
                createItemAuditHistory(EntityType.NDC, ndcId.toString(),
                    "NDC Inactivated based on proposed Inacativation being reached!", "", "");
            }

        } catch (ValidationException e) {
            // Just logging the error: will not roll back the transaction.
            LOG.error(validateStringInput(e.getMessage(), LOG_FORGING));
        }

    }

    /**
     * Checks whether the identified product should be inactivated and, if so, inactivates it. Inactivation is usually
     * a second-approved field so "inactivate" here means "generate a request to inactivate".
     */
    private void checkAndInactivateProduct(Long productId) {
      LOG.debug("Checking product {}", productId);

      try {
        // 1. need to check if product is already on the pending list, if is then ignore this and do nothing.
        Collection<RequestVo> requestList = requestDomainCapability.retrieve(productId.toString(), EntityType.PRODUCT);
        if (!requestList.isEmpty()) {
          LOG.debug("There is already a request for this item. Aborting.");
          return;
        }

        // 2. No request, go ahead an inactivate it if the date has passed
        ProductVo productVo = productDomainCapability.retrieve(productId.toString());
        checkAndInactivateItem(productVo);
      } catch (final ValidationException ex) {
        // Just logging the error: will not roll back the whole transaction.
        LOG.error("Could not check/inactivate product " + productId + " due to validation error.", ex);
      }
    }

    /**
     * inactivateOrderable. 
     * @param orderableId orderable
     */
    private void checkAndInactivateOrderable(Long orderableId) {
      LOG.debug("Checking orderable item {}", orderableId);

      try {
        // 1. need to check if product is already on the pending list, if is then ignore this and do nothing.
        Collection<RequestVo> requestList = requestDomainCapability.retrieve(orderableId.toString(),
            EntityType.ORDERABLE_ITEM);

        if (!requestList.isEmpty()) {
          return;
        }

        // 2. Now, update the product by adding it to the pending request list properly.
        OrderableItemVo orderableItemVo = orderableItemDomainCapability.retrieve(orderableId.toString());
        checkAndInactivateItem(orderableItemVo);
      } catch (final ValidationException ex) {
        // Just logging the error: will not roll back the whole transaction.
        LOG.error("Could not check/inactivate OI " + orderableId + " due to validation error.", ex);
      }

    }

    /**
     * Checks that the item's proposed inactivation date has passed and, if so, inactivates it.
     * @throws ValidationException if error validating data in ManagedItemVo
     */
    private void checkAndInactivateItem(final ManagedItemVo item) throws ValidationException {
      final ManagedItemVo originalItem = item.copy();
      
      DataField<Date> proposedInativationDate = 
          item.getVaDataFields().getDataField(FieldKey.PROPOSED_INACTIVATION_DATE);
      
      if (proposedInativationDate.getValue() == null
          || proposedInativationDate.getValue().before(new Date())) {
        
          LOG.debug("Date has hit: inactivating item.");
      
          item.setInactivationDate(new Date());
          item.setItemStatus(ItemStatus.INACTIVE);
          item.getVaDataFields().getDataField(FieldKey.PROPOSED_INACTIVATION_DATE).selectValue(null);
  
          Collection<ModDifferenceVo> modDiffs = originalItem.compareDifferences(item);
  
          for (ModDifferenceVo modDiff : modDiffs) {
              modDiff.setAcceptChange(true);
              modDiff.setComments("Proposed Inactivation Date reached");
              modDiff.setModificationReason("Proposed Inactivation Date has been reached.");
          }
          
          final UserVo user = getUser();
  
          managedItemCapability.commitModifications(new ItemModDifferences(originalItem, modDiffs), user);
          rematchCapability.updateUponProposedInactivation(item, user);
      }
    }

    /**
     * Returns the {@link UserVo} representing this job.
     * @return see description
     */
    protected final UserVo getUser() {

        UserVo user = new UserVo();
        user.setFirstName("PropInact");
        user.setLastName("Update");
        user.setStationNumber(STATION);
        user.setUsername("ProposedInactivationReached");
        user.setLocation(STATION);
        user.setId(USER_ID);

        user.addRole(Role.PSS_PPSN_MANAGER /*NATIONAL_SERVICE_MANAGER*/);

        return user;

    }

    /**
     * createItemAuditHistory record with the ProposedInactiveCapablityImpl.java
     * @param type EntityType
     * @param id EPL_ID
     * @param message message.
     * @param oldValue oldValue.
     * @param newValue newValue.
     */
    private void createItemAuditHistory(EntityType type, String id, String message, String oldValue, String newValue) {
        try {
            ItemAuditHistoryVo audit = new ItemAuditHistoryVo();
            audit.setReason(message);
            audit.setAuditItemId(id);
            audit.setCreatedBy("MEDROUTE UPDATE");
            audit.setEventCategory(EventCategory.SYSTEM_EVENT);
            audit.setAuditItemType(type);
            audit.setOldValue(oldValue);
            audit.setNewValue(newValue);
            audit.setSiteName(STATION);
            audit.setUsername(getUser().getUsername());
            itemAuditHistoryDomainCapability.create(audit, null, getUser());
        } catch (Exception e) {
            String str =
                    "Exception with audit history record of type type " + type + " and id of " + id + " the exception is "
                            + e.getMessage();
            LOG.debug("ProposedInactiveCapabilityImp: " + str);
        }
    }

    /**
     * setItemAuditHistoryDomainCapability in ProposedInactivateCapabilityImpl.
     * @param itemAuditHistoryDomainCapability the itemAuditHistoryDomainCapability to set
     */
    public void setItemAuditHistoryDomainCapability(ItemAuditHistoryDomainCapability itemAuditHistoryDomainCapability) {
        this.itemAuditHistoryDomainCapability = itemAuditHistoryDomainCapability;
    }

    /**
     * setManagedItemCapability.
     * @param managedItemCapability the managedItemCapability to set
     */
    public void setManagedItemCapability(ManagedItemCapability managedItemCapability) {
        this.managedItemCapability = managedItemCapability;
    }

    public void setProcessStatusUpdater(ProcessStatusUpdater processStatusUpdater) {
      this.processStatusUpdater = processStatusUpdater;
    }

    /**
     * setNdcDomainCapability in ProposedInactivateCapabilityImpl.
     * @param ndcDomainCapability the ndcDomainCapability to set
     */
    public void setNdcDomainCapability(NdcDomainCapability
                                       ndcDomainCapability) {
        this.ndcDomainCapability = ndcDomainCapability;
    }

    /**
     * setRequestDomainCapability.
     * @param requestDomainCapability the requestDomainCapability to set
     */
    public void setRequestDomainCapability(RequestDomainCapability
                                           requestDomainCapability) {
        this.requestDomainCapability = requestDomainCapability;
    }

    /**
     * setProductDomainCapability.
     * @param productDomainCapability the productDomainCapability to set
     */
    public void setProductDomainCapability(ProductDomainCapability
                                           productDomainCapability) {
        this.productDomainCapability = productDomainCapability;
    }

    /**
     * setOrderableItemDomainCapability.
     * @param orderableItemDomainCapability the orderableItemDomainCapability to set
     */
    public void setOrderableItemDomainCapability(OrderableItemDomainCapability
                                                 orderableItemDomainCapability) {
        this.orderableItemDomainCapability = orderableItemDomainCapability;
    }

    /**
    * setDatabaseDomainCapability.
    * @param databaseDomainCapability the databaseDomainCapability to set
    */
    public void setDatabaseDomainCapability(DatabaseDomainCapability
                                                         databaseDomainCapability) {
        this.databaseDomainCapability = databaseDomainCapability;
    }

    public void setRematchCapability(final RematchCapability rematchCapability) {
      this.rematchCapability = rematchCapability;
    }

    /**
     * gets the TransactionManager for the ProposedInactivateCapabilityImpl.
     * @return the transactionManager
     */
    public PlatformTransactionManager getTransactionManager() {
        return transactionManager;
    }

    /**
     * sets the TransactionManager for the ProposedInactivateCapabilityImpl.
     * @param transactionManager the transactionManager to set
     */
    public void setTransactionManager(PlatformTransactionManager transactionManager) {
        this.transactionManager = transactionManager;
    }
}
