package gov.va.med.pharmacy.peps.service.common.capability.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.apache.logging.log4j.LogManager;

import gov.va.med.pharmacy.peps.common.utility.security.UntrustedLogParam;
import gov.va.med.pharmacy.peps.common.vo.ManagedItemVo;
import gov.va.med.pharmacy.peps.common.vo.MatchReason;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.RematchVo;
import gov.va.med.pharmacy.peps.common.vo.UnmatchRematchVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.Difference;
import gov.va.med.pharmacy.peps.domain.common.capability.RematchMgtDomainCapability;
import gov.va.med.pharmacy.peps.service.common.capability.RematchCapability;

/**
 * Main implementation of {@link RematchCapability}.
 */
public class RematchCapabilityImpl implements RematchCapability{
    
  private static final Logger LOG = LogManager.getLogger(RematchCapabilityImpl.class);

  private final RematchMgtDomainCapability rematchMgtDomainCapability;

  /**
   * Constructs an instance with the required dependencies.
   * @param pRematchMgtDomainCapability the {@link RematchMgtDomainCapability} to use. Must not be null.
   */
  @Autowired
  public RematchCapabilityImpl(final RematchMgtDomainCapability pRematchMgtDomainCapability) {
    rematchMgtDomainCapability = Objects.requireNonNull(pRematchMgtDomainCapability);
  }

    @Override
    public void removeObsoleteProducts() {
        rematchMgtDomainCapability.removeObsoleteProducts();
        
    }
    
    @Override
    public void addUnmatchedProducts(UserVo user) {
        
        rematchMgtDomainCapability.addUnmatchedProducts(user);      
        
    }
    
    @Override
    public List<RematchVo> retrieveUnmatched(Long matchTypeId){
        return rematchMgtDomainCapability.retrieveUnmatched(matchTypeId);
        
    }
    
    @Override   
    public List<UnmatchRematchVo> findRematchProducts(Long productId) {
        return rematchMgtDomainCapability.findRematchProducts(productId);
    }  
    
    @Override
    public void updateRematch(Long rematchId, String matchIen, Boolean autoMatch){
      // TODO: This operation supports the rematch tab and should be removed with the tab.
      throw new UnsupportedOperationException("This operation is no longer supported.");
    }
    
  @Override
  public Optional<RematchVo> getExistingRematchForDifferences(
      final ManagedItemVo item, final Collection<Difference> differences) {

    final Set<MatchReason> matchReasons = item.determineRematchTriggers(differences);

    if (!matchReasons.isEmpty()) {  // Optimization: don't pull the rematch if we don't trigger one at all.
      final long itemId = Long.parseLong(item.getId());
      final Optional<RematchVo> existingRematch = rematchMgtDomainCapability.getRematchForProduct(itemId);
      if (existingRematch.isPresent() && matchReasons.contains(existingRematch.get().getMatchReason())) {
        return existingRematch;
      }
    }

    // Either there was no existing rematch or these diffs didn't trigger it.
    return Optional.empty();
  }
  
  @Override
  public void updateUponProposedInactivation(final ManagedItemVo item, UserVo user) {
    if (item instanceof ProductVo) {
      final long itemId = Long.parseLong(item.getId());
      final Optional<RematchVo> existingRematch = rematchMgtDomainCapability.getRematchForProduct(itemId);
      
      if (existingRematch.isPresent() && existingRematch.get().getMatchReason() == MatchReason.FUTURE_INACTIVATION) {
        final MatchReason newReason = MatchReason.INACTIVATION;
        LOG.debug("Updating match reason to {} to associate it with the generated request.", MatchReason.INACTIVATION);
        existingRematch.get().setMatchReason(newReason);
        rematchMgtDomainCapability.saveOrUpdate(existingRematch.get(), user);
      }
    }
  }
  
  @Override
  public Optional<RematchVo> saveOrUpdateUncommittedRematch(
      final Collection<Difference> committedDiffs, final RematchVo rematch, final UserVo user) {

    final Set<MatchReason> newTriggers = rematch.getFromProduct().determineRematchTriggers(committedDiffs);
    
    if (newTriggers.isEmpty()) {
      LOG.debug("The approved changes no longer trigger a rematch: dropping the rematch.");
      if (rematch.getEplRematchId() != null) {
        // The rematch has previously been saved to the database: delete it.
        rematchMgtDomainCapability.delete(rematch);
      }
      return Optional.empty();
    
    } else {
      final MatchReason newReason = newTriggers.iterator().next();
      if (newReason != rematch.getMatchReason()) {
        LOG.debug("The approved changes have changed the rematch reason: updating from {} to {}.",
            rematch.getMatchReason(), newReason);
        rematch.setMatchReason(newReason);
      }
      return Optional.of(rematchMgtDomainCapability.saveOrUpdate(rematch, user));
    }
  }

  @Override
  public Optional<RematchVo> saveOrUpdateCommittedRematch(
      final ManagedItemVo item, final Collection<Difference> committedDiffs, Optional<RematchVo> optionalRematch,
      final UserVo user) {
    
    LOG.debug("Updating rematches for changes {}.", new UntrustedLogParam(committedDiffs));
    
    deleteReversedRematch(item, committedDiffs);
    
    if (!optionalRematch.isPresent()) {
      return optionalRematch;
    }
    
    return saveOrUpdateUncommittedRematch(committedDiffs, optionalRematch.get(), user);
  }

  /**
   * Deletes an unsent rematch for the given item that the given diffs are reversing. E.g., if there is an unsent
   * rematch for product inactivation and the given diffs are re-activating the product, then the rematch will be
   * deleted.
   * @param item the affected item. May or may not have the changes applied.
   * @param committedDiffs the {@link Difference}s being applied to to the item
   */
  private void deleteReversedRematch(final ManagedItemVo item, final Collection<Difference> committedDiffs) {
    final ArrayList<Difference> reversedDiffs = new ArrayList<>(committedDiffs.size());
    for (final Difference committedDiff : committedDiffs) {
      reversedDiffs.add(new Difference(
          committedDiff.getFieldKey(), committedDiff.getNewValue(), committedDiff.getOldValue()));
    }
    
    // If we are reversing the differences that created an existing rematch, delete it.
    final Optional<RematchVo> existingReversedRematch = getExistingRematchForDifferences(item, reversedDiffs);
    if (existingReversedRematch.isPresent()) {
      rematchMgtDomainCapability.delete(existingReversedRematch.get());
    }
  }

}
