package gov.va.med.pharmacy.peps.service.common.migration;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.DateUtil;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.xssf.usermodel.XSSFSheet;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.apache.xmlbeans.impl.regex.ParseException;

import gov.va.med.pharmacy.peps.common.vo.MigrationCopayTierVo;

/**
 * Reads {@link MigrationCopayTierVo}s from Excel spreadsheets.
 */
public class CopayTierImportReader {
  
  private static final Logger LOG = LogManager.getLogger(CopayTierImportReader.class);
  
  /**
   * The index of the workbook sheet containing the copay tier data.
   */
  private static final int COPAY_TIER_SHEET_INDEX = 0;
  
  /**
   * Returns the number of data rows (excluding the header row) in the given Excel workbook. Performs no validation on
   * the data.
   * @param is the stream providing the Excel workbook. Will be closed before the method returns.
   * @return the number of data rows
   * @throws IOException if an I/O error occurs while reading the file
   */
  public int countCopayTierImport(final InputStream is) throws IOException {
      try (final XSSFWorkbook workbook = new XSSFWorkbook(is)) { 
          final XSSFSheet sheet = workbook.getSheetAt(COPAY_TIER_SHEET_INDEX);
          // getLastRowNum() is already 0-based so a 2-row spreadsheet will return 1. That is what we want.
          return sheet.getLastRowNum();
      }
  }

  /**
   * Converts all rows of the first sheet of the given Excel workbook.
   * @param is the stream providing the Excel workbook. Will be closed before the method returns.
   * @return a List containing one {@link MigrationCopayTierVo} for each row. Never null.
   */
  public List<MigrationCopayTierVo> getCopayTierImportList(InputStream is) {
      
      List<MigrationCopayTierVo> list = new ArrayList<MigrationCopayTierVo>();
      try (XSSFWorkbook workbook = new XSSFWorkbook(is)) { 
          XSSFSheet sheet = workbook.getSheetAt(COPAY_TIER_SHEET_INDEX);
          Iterator<Row> iterator = sheet.iterator();
          int processCount = 0;          
          MigrationCopayTierVo copayTier = null;
           
          while (iterator.hasNext()) {
              processCount=processCount+1;
              Row nextRow = iterator.next();
              
              if (processCount > 1) {
                  Iterator<Cell> cellIterator = nextRow.cellIterator();
                   
                  copayTier = new MigrationCopayTierVo();
                  while (cellIterator.hasNext()) {
                      Cell cell = cellIterator.next();
                      switch (cell.getCellType()) {
                          case Cell.CELL_TYPE_STRING:                                          
                                  copayTier.setVaProductName(cell.getStringCellValue());                              
                              break;
                          case Cell.CELL_TYPE_NUMERIC:
                              if (DateUtil.isCellDateFormatted(cell)) {
                                  try {
                                      copayTier.setEffectiveDate(cell.getDateCellValue());
                                  } catch (ParseException e) {
                                      LOG.error("Exception caught when parsing input content");
                                      LOG.error(e.getLocalizedMessage());
                                  }
                              } else {
                                  int num = ((Double) cell.getNumericCellValue()).intValue();
                                  if (cell.getColumnIndex() == 0) {
                                       copayTier.setVUID(num);
                                  } else if (cell.getColumnIndex() == 2) {
                                             copayTier.setCopayTierLevel(Integer.toString(num));
                                         }
                              }
                              break;
                        default:
                          // Ignore other cell types.
                          break;
                      }
                  } // end column
                  list.add(copayTier);
              } // end-if
          } // end row
      } catch (IOException ex) {
          LOG.error("Exception caught when reading input content");
          LOG.error(ex.getLocalizedMessage());
      } 
      finally {
          if (is!=null) {
              try {
                  is.close();
              } catch (IOException ioE) {
                  LOG.error("Caught exception during closing the input stream");
                  LOG.error(ioE.getLocalizedMessage());
              }
          }
      }
      
      return list;
      
  }
}
