package gov.va.med.pharmacy.peps.service.common.migration;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.poi.xssf.usermodel.XSSFCell;
import org.apache.poi.xssf.usermodel.XSSFRow;
import org.apache.poi.xssf.usermodel.XSSFSheet;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

import gov.va.med.pharmacy.peps.common.vo.FormularyDesignatorImportRow;

/**
 * This class writes the result file for the formulary designator
 * migration. The result file contains every row from the import file, 
 * plus an error flag column to indicate whether a validation error
 * occurred for that row, as well as an error message column
 * describing the error.
 */
public class FormularyDesignatorMigrationResultFileWriter {
    /**
    * The class's logger.
    */
    private static final Logger LOG = 
            LogManager.getLogger(FormularyDesignatorMigrationResultFileWriter.class);
    
    /**
     * Write the given rows to the given resultFile, per business rules.
     * 
     * @param rows to write
     * @param resultFile which should contain rows in error
     */
    public void write(List<FormularyDesignatorImportRow> rows, File resultFile) {
        try(XSSFWorkbook workbook = new XSSFWorkbook()) {
            XSSFSheet sheet = workbook.createSheet("Formulary Designator Data"); 
            sheet = addHeaderRow(sheet);
            sheet = addDetailRows(sheet, rows);
            
            try(FileOutputStream out = new FileOutputStream(resultFile)) {
                workbook.write(out);
            }
            catch (IOException e) {
                LOG.error("Failed to write Formulary Designator Migration error file.", e);
            }
        } catch (IOException e) {
            LOG.error("Failed to close Formulary Designator Migration error workbook.", e);
        }
    }
    
    private XSSFSheet addDetailRows(XSSFSheet sheet, List<FormularyDesignatorImportRow> rows) {
        int outputRowNumber = 0;
        
        for(FormularyDesignatorImportRow row : rows) {
            sheet = addDetailRow(sheet, row, ++outputRowNumber);
        }
        
        return sheet;
    }

    private XSSFSheet addDetailRow(XSSFSheet sheet, FormularyDesignatorImportRow importRow, int rowNumber) {
        XSSFRow row = sheet.createRow(rowNumber);
        XSSFCell cell = row.createCell(FormularyDesignatorMigrationConstants.EPL_ID_COLUMN);
        
        String eplId = importRow.getEplId() == null ? "" : String.valueOf(importRow.getEplId());
        
        cell.setCellValue(eplId);
        cell = row.createCell(FormularyDesignatorMigrationConstants.OI_NAME_COLUMN);
        cell.setCellValue(importRow.getOrderableItemName());
        cell = row.createCell(FormularyDesignatorMigrationConstants.FORMULARY_DESIGNATOR_COLUMN);
        cell.setCellValue(importRow.getFormularyDesignatorName());
        cell = row.createCell(FormularyDesignatorMigrationConstants.OI_TEXT_COLUMN);
        cell.setCellValue(importRow.getOrderableItemText());        
        cell = row.createCell(FormularyDesignatorMigrationConstants.ERROR_FLAG_COLUMN);
        cell.setCellValue(importRow.getErrorFlag());        
        cell = row.createCell(FormularyDesignatorMigrationConstants.ERROR_COLUMN);
        cell.setCellValue(importRow.getErrorMessage());
        return sheet;
    }

    private XSSFSheet addHeaderRow(XSSFSheet sheet) {
        XSSFRow row = sheet.createRow(0);
        XSSFCell cell = row.createCell(FormularyDesignatorMigrationConstants.EPL_ID_COLUMN);
        cell.setCellValue("EPL ID");
        cell = row.createCell(FormularyDesignatorMigrationConstants.OI_NAME_COLUMN);
        cell.setCellValue("Orderable Item Name");
        cell = row.createCell(FormularyDesignatorMigrationConstants.FORMULARY_DESIGNATOR_COLUMN);
        cell.setCellValue("Formulary Designator");
        cell = row.createCell(FormularyDesignatorMigrationConstants.OI_TEXT_COLUMN);
        cell.setCellValue("Orderable Item Text");
        cell = row.createCell(FormularyDesignatorMigrationConstants.ERROR_FLAG_COLUMN);
        cell.setCellValue("Error Y/N");
        cell = row.createCell(FormularyDesignatorMigrationConstants.ERROR_COLUMN);
        cell.setCellValue("Error Message");
        return sheet;
    }
}
