package gov.va.med.pharmacy.peps.service.common.scheduler;

import java.util.Date;
import java.util.Objects;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import gov.va.med.pharmacy.peps.common.vo.NationalSetting;
import gov.va.med.pharmacy.peps.common.vo.NationalSettingVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.domain.common.capability.NationalSettingDomainCapability;

/**
 * Updates jobs' {@link ProcessStatus}es and run times.
 */
public class ProcessStatusUpdater {
  
  private static final Logger LOG = LogManager.getLogger(ProcessStatusUpdater.class);

  private final NationalSettingDomainCapability nationalSettingDomainCapability;
  
  /**
   * Constructs an instance with the required dependencies.
   * 
   * @param pNationalSettingDomainCapability the {@link NationalSettingDomainCapability} to use
   */
  @Autowired
  public ProcessStatusUpdater(final NationalSettingDomainCapability pNationalSettingDomainCapability) {
    nationalSettingDomainCapability = pNationalSettingDomainCapability;
  }

  /**
   * Updates the given job run state setting to the given status. Performs the update in an independent transaction.
   * 
   * @param runStateSetting the setting to update. Must not be null.
   * @param processStatus the status to set. Must not be null.
   * @param user the user updating the status. Must not be null.
   */
  @Transactional(propagation = Propagation.REQUIRES_NEW)
  public void updateStatus(final NationalSetting runStateSetting, final ProcessStatus processStatus, final UserVo user) {
    Objects.requireNonNull(runStateSetting);
    Objects.requireNonNull(processStatus);
    Objects.requireNonNull(user);
    
    LOG.debug("Updating {} status to {}.", runStateSetting, processStatus);

    final NationalSettingVo returnedSetting = nationalSettingDomainCapability.retrieve(runStateSetting.toString());
    returnedSetting.setStringValue(processStatus.toString());
    nationalSettingDomainCapability.update(returnedSetting, user);
  }
  
  /**
   * Updates the given job run state to {@link ProcessStatus#RUNNING} and returns the last run time.
   * 
   * @param runStateSetting the setting to update. Must not be null.
   * @param runTimeSetting the run time setting to return. Must not be null.
   * @param user the user updating the status. Must not be null.
   * @return the last run time. Usually not null, but may be if the database did not have one set.
   */
  @Transactional(propagation = Propagation.REQUIRES_NEW)
  public Date startJob(final NationalSetting runStateSetting, final NationalSetting runTimeSetting, final UserVo user) {
    Objects.requireNonNull(runStateSetting);
    Objects.requireNonNull(runTimeSetting);
    Objects.requireNonNull(user);
    
    // Note that calling this member method will _not_ create a new transaction.
    updateStatus(runStateSetting, ProcessStatus.RUNNING, user);
    
    return nationalSettingDomainCapability.retrieve(runTimeSetting.toString()).getDateValue();
  }

  /**
   * Updates the given job run time setting to the given time. This method is private because jobs always want to
   * update their status and last run time at the same time, but it could be made public if there is a valid use case.
   * 
   * @param runTimeSetting the setting to update. Must not be null.
   * @param user the user updating the run time. Must not be null.
   * @param newTime the last run time to set. Must not be null.
   * @return the previous last run date.
   */
  private Date updateRuntime(final NationalSetting runTimeSetting, final UserVo user, final Date newTime) {
    Objects.requireNonNull(runTimeSetting);
    Objects.requireNonNull(user);
    Objects.requireNonNull(newTime);
    
    LOG.debug("Updating {} run time to {}.", runTimeSetting, newTime);

    final NationalSettingVo returnedSetting = nationalSettingDomainCapability.retrieve(runTimeSetting.toString());
    final Date previousTime = returnedSetting.getDateValue();
    returnedSetting.setDateValue(newTime);
    nationalSettingDomainCapability.update(returnedSetting, user);
    return previousTime;
  }

  /**
   * <p>Updates both the given job status setting and the given job run time setting. A combination of {@link
   * #updateStatus(NationalSetting, ProcessStatus, UserVo) updateStatus()} and a similar operation to update the last
   * run setting. Most clients will want to call {@link #updateStatusAndRuntime(NationalSetting, ProcessStatus,
   * NationalSetting, UserVo)}, but this overload is useful if you want precise control over the last run time.</p>
   * 
   * <p>Performs the update in an independent transaction.</p>
   * 
   * @param runStateSetting the setting to update. Must not be null.
   * @param processStatus the status to set. Must not be null.
   * @param runTimeSetting the setting to update. Must not be null.
   * @param newTime the last run time to set. Must not be null.
   * @param user the user updating the status. Must not be null.
   * @return the previous run time setting value
   */
  @Transactional(propagation = Propagation.REQUIRES_NEW)
  public Date updateStatusAndRuntime(
      final NationalSetting runStateSetting,
      final ProcessStatus processStatus,
      final NationalSetting runTimeSetting,
      final Date newTime,
      final UserVo user) {
    
    LOG.debug("Updating both job status and runtime.");
    
    // Note that calling this member method will _not_ create a new transaction.
    updateStatus(runStateSetting, processStatus, user);
    return updateRuntime(runTimeSetting, user, newTime);
  }
  
  /**
   * Convenience method equivalent to calling {@link #updateStatusAndRuntime(NationalSetting, ProcessStatus,
   * NationalSetting, Date, UserVo)} with the current date.
   * 
   * @param runStateSetting the setting to update. Must not be null.
   * @param processStatus the status to set. Must not be null.
   * @param runTimeSetting the setting to update. Must not be null.
   * @param user the user updating the status. Must not be null.
   * @return the previous run time setting value
   */
  @Transactional(propagation = Propagation.REQUIRES_NEW)
  public Date updateStatusAndRuntime(
      final NationalSetting runStateSetting,
      final ProcessStatus processStatus,
      final NationalSetting runTimeSetting,
      final UserVo user) {
    // Note that calling this member method will _not_ create a new transaction.
    return updateStatusAndRuntime(runStateSetting, processStatus, runTimeSetting, new Date(), user);
  }
}
