package gov.va.med.pharmacy.peps.service.common.session.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.DateUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.poi.xssf.usermodel.XSSFCell;
import org.apache.poi.xssf.usermodel.XSSFRow;
import org.apache.poi.xssf.usermodel.XSSFSheet;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import gov.va.med.pharmacy.peps.common.utility.PpsConfiguration;
import gov.va.med.pharmacy.peps.common.vo.CopayTierAssocVo;
import gov.va.med.pharmacy.peps.common.vo.EntityType;
import gov.va.med.pharmacy.peps.common.vo.EventCategory;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.common.vo.ItemAuditHistoryDetailsVo;
import gov.va.med.pharmacy.peps.common.vo.ItemAuditHistoryVo;
import gov.va.med.pharmacy.peps.common.vo.MigrationCopayTierVo;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.GroupDifference;
import gov.va.med.pharmacy.peps.common.vo.validator.ErrorKey;
import gov.va.med.pharmacy.peps.domain.common.capability.CopayTierMigrationDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.ItemAuditHistoryDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.dao.EplVuidStatusHistoryDao;
import gov.va.med.pharmacy.peps.domain.common.model.EplCopayTierDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplProdCopayTierAssocDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplProdDrugClassAssocDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplProductDo;
import gov.va.med.pharmacy.peps.domain.common.model.EplVaGenNameDo;
import gov.va.med.pharmacy.peps.service.common.migration.CopayTierImportReader;
import gov.va.med.pharmacy.peps.service.common.migration.MigrationStatus;
import gov.va.med.pharmacy.peps.service.common.session.CopayTierMigrationService;
import gov.va.med.pharmacy.peps.service.common.update.NdfFileSyncProcessorFactory;

/**
 * Main implementation of {@link CopayTierMigrationService}.
 */
public class CopayTierMigrationServiceImpl implements CopayTierMigrationService {
  
  /**
   * The number of seconds to use for the import transaction timeout. This value should be high enough to import 30,000
   * rows. (The actual largest expected file would be 27,068 rows.)
   */
  protected static final int IMPORT_TRANSACTION_TIMEOUT_SECONDS = 180 * 60;
  
  /**
   * The class's logger.
   */
  private static final Logger LOG = LogManager.getLogger(CopayTierMigrationServiceImpl.class);

  private static String PRODUCT_EPL_ID = "eplProduct.eplId";

  private static final String RESULTS_FILE_NAME = "MigrationCopayTierResults.xlsx";

  /** Excel column index. */
  private static final int COL_IDX_EFFECTIVE_DATE = 3;
  /** Excel column index. */
  private static final int COL_IDX_ERROR_FLAG = 4;
  /** Excel column index. */
  private static final int COL_IDX_ERROR_MESSAGE = 5;
  
  private final CopayTierMigrationDomainCapability copayTierMigrationDomainCapability;
  private final ItemAuditHistoryDomainCapability itemAuditHistoryDomainCapability;
  private final PpsConfiguration ppsConfiguration;
  private final NdfFileSyncProcessorFactory syncProcessorFactory;
  
  /**
   * The EplVuidStatusHistoryDao.
   */
  private final EplVuidStatusHistoryDao eplVuidStatusHistoryDao;
  
  private final MigrationStatus migrationStatus;
  
  /**
   * The character which separates error messages when multiple
   * errors are concatenated into a single message.
   */
  private static final String ERROR_MESSAGE_DELIMITER = ";";
  
  /**
   * Constructs an instance.
   * 
   * @param pCtMigrationDomainCapability the {@link CopayTierMigrationDomainCapability} to use
   * @param pItemAuditHistoryDomainCapability the {@link ItemAuditHistoryDomainCapability} to use
   * @param pEplVuidStatusHistoryDao the {@link EplVuidStatusHistoryDao} to use
   * @param pPpsConfiguration the {@link PpsConfiguration} to use
   * @param pSyncProcessorFactory the {@link NdfFileSyncProcessorFactory} to use
   */
  @Autowired
  public CopayTierMigrationServiceImpl(
      final CopayTierMigrationDomainCapability pCtMigrationDomainCapability,
      final ItemAuditHistoryDomainCapability pItemAuditHistoryDomainCapability,
      final EplVuidStatusHistoryDao pEplVuidStatusHistoryDao,
      final PpsConfiguration pPpsConfiguration,
      final NdfFileSyncProcessorFactory pSyncProcessorFactory) {
    copayTierMigrationDomainCapability = pCtMigrationDomainCapability;
    itemAuditHistoryDomainCapability = pItemAuditHistoryDomainCapability;
    eplVuidStatusHistoryDao = pEplVuidStatusHistoryDao;
    ppsConfiguration = pPpsConfiguration;
    syncProcessorFactory = pSyncProcessorFactory;
    
    migrationStatus = initializeMigrationStatus();
  }
    
  private ItemAuditHistoryDomainCapability getItemAuditHistoryDomainCapability() {
      return itemAuditHistoryDomainCapability;
  }

  @Override
  public File getResultFilePath() {
    return new File(ppsConfiguration.getPersistentDirectory(), RESULTS_FILE_NAME);
  }


  /**
   * Processing the import copay tier list.
   * @param list List<MigrationCopayTierVo>
   * @param user UserVo
   */
  @Override
  @Transactional(timeout = IMPORT_TRANSACTION_TIMEOUT_SECONDS) // This operation requires a higher-than-default timeout.
  public final void processCopayTierImport(List<MigrationCopayTierVo> list, UserVo user)
  {
      List<EplCopayTierDo> copayTiers = null;
      Map<String,EplCopayTierDo> copayTierMap = null;

      if (!CollectionUtils.isEmpty(list)) {
          try {
              copayTiers = copayTierMigrationDomainCapability.retrieveCopayTiers();
              if (copayTiers==null) {
                  list.get(0).setError(true);
                  list.get(0).setErrorMessage("NO Copay Tier Defined.  End Migration.");
                  updateProgress(0,0,MigrationStatus.DONE);
              }
          } catch (Exception ex) {
              list.get(0).setError(true);
              list.get(0).setErrorMessage("NO Copay Tier Defined.  End Migration.");
              updateProgress(0,0,MigrationStatus.DONE);
          }
          if (copayTiers != null && copayTiers.size() > 0) {
              copayTierMap = getCopayTierMap(copayTiers);
              for (int i = 0; i < list.size(); i++) {
                  final MigrationCopayTierVo copayTierVo = list.get(i);
                  try {
                      List<String> errorMessages = validates(copayTierMap, copayTierVo);
                      EplProductDo existingProductDo = null;
                      
                      try {
                          existingProductDo = copayTierMigrationDomainCapability.retrieveProduct(copayTierVo.getVUID());
                          errorMessages.addAll(validateExistingProduct(existingProductDo, user));
                      }
                      catch(Exception ex) {
                          errorMessages.add(
                              "Exception retrieving product by vuid " 
                              + copayTierVo.getVUID() + ": " 
                              + ex.getMessage());
                      }
                      
                      if (errorMessages.size() == 0) {
                          validateAndImport(existingProductDo, user, copayTierVo, copayTierMap);
                      }
                      else {
                          copayTierVo.setError(true);
                          copayTierVo.setErrorMessage(
                              String.join(ERROR_MESSAGE_DELIMITER, errorMessages));
                      }
                      
                      if ((i+1)==list.size()) {
                          updateProgress(i+1,list.size(),MigrationStatus.DONE);
                      } else {
                          updateProgress(i+1,list.size(),MigrationStatus.INPROGRESS);
                      }
                  } catch (Exception ex) {
                      LOG.warn("Unexpected exception while processing copay tier.", ex);
                      copayTierVo.setError(true);
                      copayTierVo.setErrorMessage("NO Copay Tier Defined.  End Migration.");
                      updateProgress(i+1,list.size(),MigrationStatus.DONE);
                  }
              }
          }   
          
          saveResultFile(list, getResultFilePath());
      }
      
  }

  
  /**
   * Check for hard errors present on the existing product.
   * 
   * @param existingProductDo The existing product.
   * @param user The current user
   * @return A list containing messages corresponding to any errors found, or an empty list
   * if there are no errors.
   */
  private List<String> validateExistingProduct(EplProductDo existingProductDo, UserVo user) {
      List<String> errorMessages = new ArrayList<String>();
      
      if(existingProductDo == null) {
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_VUID.getLocalizedError());
      }
      else {   
          errorMessages = validateVistaRequiredFields(existingProductDo, errorMessages);
      }
      
      return errorMessages;
  }
  
  /**
   * Check that all of the fields VistA requires are populated on the existingProductDo.
   * @param existingProductDo the product to check
   * @param errorMessages list where any error messages should be added
   * @return list with any error messages related to the VistA required fields appended.
   */
  private List<String> validateVistaRequiredFields(EplProductDo existingProductDo, List<String> errorMessages) {
      if(existingProductDo.getVaProductName() == null) {
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_VA_PRODUCT_NAME.getLocalizedError());
      }
      
      if(existingProductDo.getMasterEntryForVuid() == null) {
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_MASTER_ENTRY_VUID.getLocalizedError());
      }
      
      errorMessages = validateVuidStatusHistory(existingProductDo, errorMessages);
      
      if(existingProductDo.getOverrideDfDoseChkExclusn() == null) {
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_OVERRIDE_DF_DOSE_CHK_EXCLUSION.getLocalizedError());
      }
      
      EplVaGenNameDo genNameDo = existingProductDo.getEplVaGenName();
      if(genNameDo == null) {
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_VA_GENERIC_NAME.getLocalizedError());
      }
      
      errorMessages = validatePrimaryDrugClassCode(existingProductDo, errorMessages);
      
      if(existingProductDo.getEplOrderableItem().getEplDosageForm().getDfName() == null) {
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_DOSAGE_FORM_NAME.getLocalizedError());
      }
      
      return errorMessages;
  }

  /**
   * Check that the existingProductDo has at least one VuidStatusHistory record associated, and
   * that that record has a non-null effective date and status.
   * @param existingProductDo the product to check
   * @param errorMessages list where any error messages should be added
   * @return list with any error messages related to the VuidStatusHistory appended.
   */
  private List<String> validateVuidStatusHistory(EplProductDo existingProductDo, List<String> errorMessages) {
      Map<String, String> bindParameters = new HashMap<String, String>();
      bindParameters.put("vuid", existingProductDo.getVuid());
      List<?> results = eplVuidStatusHistoryDao.executeHqlQuery(
          "select count(*) from EplVuidStatusHistoryDo where vuid = :vuid", bindParameters);
      long countVuidStatusHistory = ((Long) results.get(0)).longValue();

      if(countVuidStatusHistory < 1) {
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_VUID_STATUS_HISTORY_EFFECTIVE_DATE.getLocalizedError());
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_VUID_STATUS_HISTORY_STATUS.getLocalizedError());
      }
      
      return errorMessages;
  }

  /**
   * Check that the existingProductDo has a primary drug class code defined.
   * @param existingProductDo the product to check
   * @param errorMessages list where any error messages should be added
   * @return list with any error messages related to the primary drug class code
   * appended.
   */
  private List<String> validatePrimaryDrugClassCode(EplProductDo existingProductDo, List<String> errorMessages) {
      Set<EplProdDrugClassAssocDo> drugClassAssocs = 
          existingProductDo.getEplProdDrugClassAssocs();
      
      boolean hasPrimary = false;
      
      for(EplProdDrugClassAssocDo drugClassAssoc : drugClassAssocs) {                
          if("Y".equals(drugClassAssoc.getPrimaryYn())) {
              hasPrimary = true;
          }
      }
      
      if(!hasPrimary) {
          errorMessages.add(ErrorKey.COPAY_TIER_MIGRATION_PRIMARY_DRUG_CLASS.getLocalizedError());
      }
      
      return errorMessages;
  }
  
  /**
   * Write the import result file back out to disk with messages.
   * @param importList List<MigrationCopayTierVo>
   * @param fullPath String
   */
  protected final void saveResultFile(List<MigrationCopayTierVo> importList, final File fullPath) {
      
      final int numColumns = 6;
      FileOutputStream out = null;
      // We don't need to close the Workbook per POI examples:
      // <https://poi.apache.org/spreadsheet/quick-guide.html#NewWorkbook>. We are closing the output file stream below.
      @SuppressWarnings("resource")
      XSSFWorkbook workbook = new XSSFWorkbook(); 
      XSSFSheet sheet = workbook.createSheet("Copay Tier Data");
      XSSFRow row = createFirstRow(sheet);
      //Iterate over data and write to sheet
      for (int i = 0; i < importList.size(); i++) {
          MigrationCopayTierVo item = importList.get(i);
          row = sheet.createRow(i+1);
          for (int j = 0; j < numColumns; j++) {
              XSSFCell cell = row.createCell(j);
              switch (j) {
                  case 0:
                      cell.setCellValue(Integer.toString(item.getVUID()));
                      break;
                  case 1:
                      cell.setCellValue(item.getVaProductName());
                      break;
                  case 2:
                      cell.setCellValue(item.getCopayTierLevel());
                      break;
                  case COL_IDX_EFFECTIVE_DATE:
                      if (item.getEffectiveDate()!=null) {
                          SimpleDateFormat sdf = new SimpleDateFormat("dd-MMM-yyyy");
                          String importDate = sdf.format(item.getEffectiveDate());
                          cell.setCellValue(importDate);
                      }
                      break;
                  case COL_IDX_ERROR_FLAG:
                      if (item.isError()) {
                          cell.setCellValue("Y");
                      } else {
                          cell.setCellValue("N");
                      }
                      break;
                  case COL_IDX_ERROR_MESSAGE:
                      cell.setCellValue(item.getErrorMessage());
                      break;
                  default:
                      break;
              }
          }
      } //end workbook
      try {
          //Write the workbook in file system
          out = new FileOutputStream(fullPath);
          workbook.write(out);
          LOG.info("Import file is written successfully to disk.");
      } 
      catch (IOException e) {
          LOG.error("Exception caught when writing results to disk");
          LOG.error(e.getLocalizedMessage());
      } finally {
          if (out!=null) {
              try {
              out.close();
              } catch (IOException ioE) {
                  LOG.error("Exception caught when writing results to disk");
                  LOG.error(ioE.getLocalizedMessage());
              }
          }
      }
  }

  /**
   * Create output spreadsheet first row.
   * @param sheet XSSFSheet
   * @return row Row
   */
  protected final XSSFRow createFirstRow(XSSFSheet sheet) {
      
      XSSFRow row = sheet.createRow(0);
      XSSFCell cell = row.createCell(0);
      cell.setCellValue("VUID");
      cell = row.createCell(1);
      cell.setCellValue("VA_PRODUCT_NAME");
      cell = row.createCell(2);
      cell.setCellValue("COPAY_TIER_LEVEL");
      cell = row.createCell(COL_IDX_EFFECTIVE_DATE);
      cell.setCellValue("EFFECTIVE_DATE");
      cell = row.createCell(COL_IDX_ERROR_FLAG);
      cell.setCellValue("ERROR");
      cell = row.createCell(COL_IDX_ERROR_MESSAGE);
      cell.setCellValue("ERROR_MESSAGE");
      
      return row;
      
  }
  
  /**
   * Validating input copay tier records.
   * @param copayTierMap Map<String,EplCopayTierDo>
   * @param copayTierVo MigrationCopayTierVo
   * @return boolean
   */
  protected final List<String> validates(Map<String,EplCopayTierDo> copayTierMap, MigrationCopayTierVo copayTierVo) {
      
      List<String> errorMessages = new ArrayList<String>();
      
      if (Integer.toString(copayTierVo.getVUID()) == null 
          || copayTierVo.getVUID()==0) {
          errorMessages.add("No VUID Entered");
      } 
      
      if (copayTierVo.getCopayTierLevel()==null 
          || copayTierVo.getCopayTierLevel().equals("")) {
          errorMessages.add("No Copay Tier Level Entered");
      } 
      else if (!copayTierMap.containsKey(copayTierVo.getCopayTierLevel())) {
          errorMessages.add("Copay Tier value in Column C does not match any Copay Tier domain value defined in PPS-N");  
      }
      
      if (copayTierVo.getEffectiveDate() == null) {
          errorMessages.add("No Effective Date Entered");
      }
      
      if (copayTierVo.getEffectiveDate()!=null) {
          Date tomorrow = DateUtils.truncate(DateUtils.addDays(new Date(), 1),Calendar.DATE);
          Date effectiveDate = DateUtils.truncate(copayTierVo.getEffectiveDate(),Calendar.DATE);
          if (effectiveDate.before(tomorrow)) {
              errorMessages.add("Effective Date is in the past");
          }
      }
      
      return errorMessages;
  }
  
  /**
   * Generate a map with EplCopayTierDo objects.
   * @param copayTiers List<EplCopayTierDo>
   * @return Map<String,EplCopayTierDo>
   */
  protected final Map<String,EplCopayTierDo> getCopayTierMap(List<EplCopayTierDo> copayTiers) {
      
      Map<String,EplCopayTierDo> map = new HashMap<String,EplCopayTierDo>();
      
      for (EplCopayTierDo item : copayTiers) {
          map.put(item.getTierLevel(),item);
      }
      
      return map;
  }
  
  /**
   * Generate a map with EplProdCopayTierAssocDo objects.
   * @param copayTierAssocs Set<EplProdCopayTierAssocDo>
   * @return Map<String,EplProdCopayTierAssocDo>
   */
  protected final Map<String,EplProdCopayTierAssocDo> getCopayTierAssocMap(Set<EplProdCopayTierAssocDo> copayTierAssocs) {
      
      Map<String,EplProdCopayTierAssocDo> map = new HashMap<String,EplProdCopayTierAssocDo>();
      
      for (EplProdCopayTierAssocDo item : copayTierAssocs) {
          map.put(item.getEplCopayTier().getTierLevel(),item);
      }
      
      return map;
  }
  
  /**
   * Import copay tier records.
   * @param eplProductDo EplProductDo
   * @param user UserVo
   * @param copayTierVo MigrationCopayTierVo
   * @param copayTierMap Map<String,EplCopayTierDo>
   */
  protected final void validateAndImport(EplProductDo eplProductDo, UserVo user, 
              MigrationCopayTierVo copayTierVo,
              Map<String,EplCopayTierDo> copayTierMap) {
      
      try {
          // Get initial ProductVo to save off the "before" list.
          final ProductVo productVo = copayTierMigrationDomainCapability.getProductConverter().convert(eplProductDo);
          final Collection<CopayTierAssocVo> copayTiersBefore = productVo.getCopayTierAssocs();
          
          // Update copay tier assocs. This code deals with DOs. Architecturally, it should probably deal with VOs, but
          // it was written this way in v1.2 and wasn't worth changing.
          final EplProdCopayTierAssocDo newCopayTierAssoc = 
              makeCopayTierAssoc(copayTierVo,eplProductDo, copayTierMap.get(copayTierVo.getCopayTierLevel()));
      
          List<EplProdCopayTierAssocDo> copayTiersAfter = null;
          if (CollectionUtils.isNotEmpty(eplProductDo.getEplProdCopayTierAssocs())) {
              final Map<String,EplProdCopayTierAssocDo> copayTierAssocMap = getCopayTierAssocMap(
                  eplProductDo.getEplProdCopayTierAssocs());
              copayTiersAfter = processImportCopayTierAssocs(newCopayTierAssoc,eplProductDo,
                              copayTierVo,user,copayTierAssocMap);
              if (!copayTierVo.isError()) {
                  copayTiersAfter = sortCopayTierList(copayTiersAfter);
              }
          } else {
              copayTiersAfter = new ArrayList<EplProdCopayTierAssocDo>();
              copayTiersAfter.add(newCopayTierAssoc);
          }
          
          if (!copayTierVo.isError()) {
            
              // Update productVo with the new copay tier assocs.
              copayTierMigrationDomainCapability.getProductConverter().setCopayTierAssocs(productVo, copayTiersAfter);

              // Save new copay tier assocs.
              updateCopayTierAssocs(copayTiersAfter, eplProductDo, user, copayTierVo);
              
              // Create GroupDifference.
              final GroupDifference ctDifference = new GroupDifference(FieldKey.COPAY_TIER_ASSOC, copayTiersBefore,
                  productVo.getCopayTierAssocs());
              // Create audit history.
              createAuditHistory(productVo, ctDifference, user);
              // Create Outgoing Differences.
              syncProcessorFactory.getProcessor(productVo)
                .processModified(productVo, user, Arrays.asList(ctDifference));
          }
      } catch (Exception ex) {
          final String errorMessage = "Exception occurred while processing the copay tier record.";
          copayTierVo.setError(true);
          copayTierVo.setErrorMessage(errorMessage);
          LOG.warn(errorMessage, ex);
      } 
  }
  
  /**
   * Update copay tier product association records.
   * 
   * @param merged List<EplProdCopayTierAssocDo>
   * @param eplProductDo EplProductDo
   * @param user UserVo
   * @param copayTierVo MigrationCopayTierVo
   * 
   */
  protected final void updateCopayTierAssocs(List<EplProdCopayTierAssocDo> merged, 
                                  EplProductDo eplProductDo, UserVo user,
                                  MigrationCopayTierVo copayTierVo) {
    
      try {
          copayTierMigrationDomainCapability.deleteCopayTierAssocs(PRODUCT_EPL_ID, eplProductDo.getEplId());
          if (merged!=null && merged.size()>0) {
              for(int i=0; i < merged.size(); i++) {
                  insertCopayTierAssoc(merged.get(i),user);
              }
          }
      } catch (Exception ex) {
          copayTierVo.setError(true);
          copayTierVo.setErrorMessage("Exception caught creating copay tier");
          LOG.error("Caught exception in deleting or inserting copay tier association.", ex);
      }
      
  }
  
  /**
   * Insert the copay tier product association record.
   * 
   * @param copayTierAssoc EplProdCopayTierAssocDo
   * @param user UserVo
   */
  protected final void insertCopayTierAssoc(EplProdCopayTierAssocDo copayTierAssoc, UserVo user) {
      
    copayTierAssoc.setId(Long.valueOf(
        copayTierMigrationDomainCapability.getSeqNumDomainCapability().generateId(EntityType.COPAY_TIER_ASSOC, user)));
    copayTierMigrationDomainCapability.insertCopayTierAssoc(copayTierAssoc, user);
  }
  
  /**
   * Creates a new {@link EplProdCopayTierAssocDo} with the given information.
   * 
   * @param copayTierVo MigrationCopayTierVo
   * @param eplProductDo EplProductDo
   * @param eplCopayTier EplCopayTierDo
   * @return EplProdCopayTierAssocDo
   */
  protected final EplProdCopayTierAssocDo makeCopayTierAssoc(MigrationCopayTierVo copayTierVo,
          EplProductDo eplProductDo, EplCopayTierDo eplCopayTier) {
      
      EplProdCopayTierAssocDo eplProdCopayTierAssocDo = new EplProdCopayTierAssocDo();
      eplProdCopayTierAssocDo.setEffectiveDate(copayTierVo.getEffectiveDate());
      eplProdCopayTierAssocDo.setEplCopayTier(eplCopayTier);
      eplProdCopayTierAssocDo.setEplProduct(eplProductDo);
      
      return eplProdCopayTierAssocDo;
     
  }
  
  /**
   * <p>Tries to add the given new copay tier association to the product's copay tier associations. Does <em>not</em>
   * actually modify the product, but returns the new copay tier association list instead.</p>
   * 
   * <p>If the new copay tier association cannot be added due to a business rule, the returned list will not be modified
   * and the error will be added to the given {@link MigrationCopayTierVo}.</p>
   * 
   * @param newCopayTierAssoc EplProdCopayTierAssocDo
   * @param eplProductDo EplProductDo
   * @param copayTierVo MigrationCopayTierVo
   * @param user UserVo
   * @param copayTierAssocMap Map<String,EplProdCopayTierAssocDo>
   * @return the updated copay tier association list
   */
  protected final List<EplProdCopayTierAssocDo> processImportCopayTierAssocs(EplProdCopayTierAssocDo 
                      newCopayTierAssoc, EplProductDo eplProductDo, 
                      MigrationCopayTierVo copayTierVo, UserVo user,
                      Map<String,EplProdCopayTierAssocDo> copayTierAssocMap) {
      
      List<EplProdCopayTierAssocDo> list = 
          new ArrayList<EplProdCopayTierAssocDo>(eplProductDo.getEplProdCopayTierAssocs());
      list = sortCopayTierList(list);
      List<EplProdCopayTierAssocDo> merged = new ArrayList<EplProdCopayTierAssocDo>();
      if (list.size()==1) {
          EplProdCopayTierAssocDo dbCopayTier = list.get(0);
          if (isActive(dbCopayTier)) {
              processActive(dbCopayTier,newCopayTierAssoc,copayTierVo,merged);
          } else {
              merged.add(newCopayTierAssoc);
          }
      } else if (list.size() > 1) {
          processMultiCopayTierAssocs(merged,list,newCopayTierAssoc,copayTierVo,copayTierAssocMap);
      }

      return merged;
      
  }
  
  /**
   * Process multiple copay tier product association records.
   * @param merged List<EplProdCopayTierAssocDo>
   * @param list List<EplProdCopayTierAssocDo>
   * @param newCopayTierAssoc EplProdCopayTierAssocDo
   * @param copayTierVo MigrationCopayTierVo
   * @param copayTierAssocMap Map<String,EplProdCopayTierAssocDo>
   */
  protected final void processMultiCopayTierAssocs(List<EplProdCopayTierAssocDo> merged,
              List<EplProdCopayTierAssocDo> list,
              EplProdCopayTierAssocDo newCopayTierAssoc,
              MigrationCopayTierVo copayTierVo,
              Map<String,EplProdCopayTierAssocDo> copayTierAssocMap) {
      
      String tierLevel = copayTierVo.getCopayTierLevel();
      EplProdCopayTierAssocDo tCopayTierAssoc;
          if (copayTierAssocMap.containsKey(tierLevel)) {
              tCopayTierAssoc = copayTierAssocMap.get(tierLevel);
              if (isActive(tCopayTierAssoc)) { //active record and matches tier level
                  copayTierVo.setError(true);
                  copayTierVo.setErrorMessage("Copay tier record already exists");
              } else {
                  for (EplProdCopayTierAssocDo dbAssoc : list) {
                      if (isExpired(dbAssoc)) {  //no match so keep the expired
                          merged.add(dbAssoc);
                      } else if (isActive(dbAssoc)) {  //update the current end date
                          Date newEndDate = DateUtils.addDays(newCopayTierAssoc.getEffectiveDate(), -1);
                          dbAssoc.setEndDate(newEndDate);
                          merged.add(dbAssoc);
                      } 
                  }
                  merged.add(newCopayTierAssoc); //replace the future
              }
          } else {
              for (EplProdCopayTierAssocDo dbAssoc : list) {
                  if (isExpired(dbAssoc)) {  //keep the expired
                      merged.add(dbAssoc);
                  } else if (isActive(dbAssoc)) {  //update the current end date and add new
                      Date newEndDate = DateUtils.addDays(newCopayTierAssoc.getEffectiveDate(), -1);
                      dbAssoc.setEndDate(newEndDate);
                      merged.add(dbAssoc);
                      merged.add(newCopayTierAssoc);
                      break;
                  } else if (isFuture(dbAssoc)) {
                      merged.add(newCopayTierAssoc);
                      break;
                  }
              }
          }
      
  }
  
  /**
   * Process copay tier active records.
   * @param dbCopayTier EplProdCopayTierAssocDo
   * @param newCopayTierAssoc EplProdCopayTierAssocDo
   * @param copayTierVo MigrationCopayTierVo
   * @param merged List<EplProdCopayTierAssocDo>
   */
  protected final void processActive(EplProdCopayTierAssocDo dbCopayTier,
                      EplProdCopayTierAssocDo newCopayTierAssoc,
                      MigrationCopayTierVo copayTierVo,
                      List<EplProdCopayTierAssocDo> merged) {
      
      String dbTierLevel = dbCopayTier.getEplCopayTier().getTierLevel();
      String newTierLevel = newCopayTierAssoc.getEplCopayTier().getTierLevel();
      if (newTierLevel.equals(dbTierLevel)) { //current copay tier exists and tier level matched            
             copayTierVo.setError(true);
             copayTierVo.setErrorMessage("Copay tier record already exists");                                 
      } else { //active but tier level doesn't match, add new record and set end date
          Date yesterday = DateUtils.addDays(newCopayTierAssoc.getEffectiveDate(), -1);
          dbCopayTier.setEndDate(yesterday);
          merged.add(dbCopayTier);
          merged.add(newCopayTierAssoc);        
      }
  }
  
  /**
   * Sort the copay tier list by effective date.
   * 
   * @param list List<eplProdCopayTierAssocDo>
   * @return the passed List, now sorted
   * 
  */
  protected final List<EplProdCopayTierAssocDo> sortCopayTierList(List<EplProdCopayTierAssocDo> list) {
      
      if (list.size() > 1) {
          Collections.sort(list, new Comparator<EplProdCopayTierAssocDo>() {
              @Override
              public int compare(EplProdCopayTierAssocDo o1, EplProdCopayTierAssocDo o2) {
                  if (o1.getEffectiveDate()!=null && o2.getEffectiveDate()!=null) {
                      return o1.getEffectiveDate().compareTo(o2.getEffectiveDate());
                  } else return 1;
              }
          });
      }
      
      return list;
      
  }
  
  /**
   * Returns true or false to indicate whether this copay tier 
   * association is currently in effect.
   * 
   * @param copayTierAssoc EplProdCopayTierAssocDo
   * @return true if and only if the effective date is before or
   * equal to the current date and the end date is blank, equal to, or
   * after the current date; otherwise, return false.
   */
  public final boolean isActive(EplProdCopayTierAssocDo copayTierAssoc) {
      Date today = DateUtils.truncate(new Date(),Calendar.DATE);
      Date effectiveDate = DateUtils.truncate(copayTierAssoc.getEffectiveDate(), Calendar.DATE);
      Date endDate=null;
      if (copayTierAssoc.getEndDate()!=null) {
          endDate = DateUtils.truncate(copayTierAssoc.getEndDate(), Calendar.DATE);
      }
      return (effectiveDate.compareTo(today) <= 0)
          && (endDate == null || endDate.compareTo(today) >= 0);
  }
  
  /**
   * Returns true or false to indicate whether this copay tier 
   * association is expired.
   * 
   * @param copayTierAssoc EplProdCopayTierAssocDo
   * @return true if and only if the effective date is before or
   * equal to the current date and the end date is blank, equal to, or
   * after the current date; otherwise, return false.
   */
  public final boolean isExpired(EplProdCopayTierAssocDo copayTierAssoc) {
      Date today = DateUtils.truncate(new Date(),Calendar.DATE);
      Date effectiveDate = DateUtils.truncate(copayTierAssoc.getEffectiveDate(), Calendar.DATE);
      Date endDate=null;
      if (copayTierAssoc.getEndDate()!=null) {
          endDate = DateUtils.truncate(copayTierAssoc.getEndDate(), Calendar.DATE);
      }
      return (effectiveDate.compareTo(today) < 0)
          && (endDate!=null && endDate.compareTo(today) <0);
          
  }
  
  /**
   * Returns true or false to indicate whether this copay tier 
   * association is expired.
   * 
   * @param copayTierAssoc EplProdCopayTierAssocDo
   * @return true if and only if the effective date is before or
   * equal to the current date and the end date is blank, equal to, or
   * after the current date; otherwise, return false.
   */
  public final boolean isFuture(EplProdCopayTierAssocDo copayTierAssoc) {
      Date today = DateUtils.truncate(new Date(),Calendar.DATE);
      Date effectiveDate = DateUtils.truncate(copayTierAssoc.getEffectiveDate(), Calendar.DATE);
      Date endDate=null;
      if (copayTierAssoc.getEndDate()!=null) {
          endDate = DateUtils.truncate(copayTierAssoc.getEndDate(), Calendar.DATE);
      }
      return (effectiveDate.compareTo(today) > 0)
          && (endDate==null || endDate.compareTo(today) > 0);
          
  }

  @Override
  public synchronized MigrationStatus getProgress() {
    return migrationStatus;
  }
  
  @Override
  public synchronized void updateProgress(int progressCount, int totalCount, String status) {
    migrationStatus.setProgressCount(progressCount);
    migrationStatus.setTotalCount(totalCount);
    migrationStatus.setStatus(status);
  }
  
  /** 
   * Create audit history for the given Product difference.
   */
  private void createAuditHistory(ProductVo productVo, GroupDifference difference, UserVo user) {
    createItemAuditHistory(EntityType.PRODUCT, productVo.getId(), "", difference.getFieldKey().getKey(),
        difference.getOldValueShortString(), difference.getNewValueShortString(), user);
  }
  
  /**
   * createItemAuditHistory record
   * @param type EntityType
   * @param id EPL_ID
   * @param message message.
   * @param column column
   * @param oldValue oldVAlue
   * @param newValue newValue
   */
  private void createItemAuditHistory(EntityType type, String id, String message, String column,
                                      String oldValue, String newValue, UserVo user) {
      try {
          ItemAuditHistoryVo audit = new ItemAuditHistoryVo();
          audit.setReason(message);

          if (StringUtils.isNotEmpty(column)) {
              Collection<ItemAuditHistoryDetailsVo> collection = new ArrayList<ItemAuditHistoryDetailsVo>();
              ItemAuditHistoryDetailsVo detail = new ItemAuditHistoryDetailsVo();
              detail.setColumnName(column);
              detail.setNewValue(newValue);
              detail.setEventCategory(EventCategory.NATIONAL_MODIFICATION);
              detail.setDetailReason(message);

              if (oldValue == null) {
                  detail.setOldValue("");
              } else {
                  detail.setOldValue(oldValue);
              }

              collection.add(detail);
              audit.setDetails(collection);
          }

          audit.setAuditItemId(id);
          audit.setCreatedBy(user.getUsername());
          audit.setEventCategory(EventCategory.NATIONAL_MODIFICATION);
          audit.setDetailCategory(EventCategory.NATIONAL_MODIFICATION);
          audit.setAuditItemType(type);
          audit.setSiteName(user.getStationNumber());
          audit.setUsername(user.getUsername());
          getItemAuditHistoryDomainCapability().create(audit, null, user);
      } catch (Exception e) {
          String str =
                  "Exception with audit history record of type type " + type + " and id of " + id + " the exception is "
                          + e.getMessage();
          LOG.debug(str);
      }
  }
  
  /**
   * <p>Returns a MigrationStatus object based on an existing result file from the last import. Returns a default object
   * if no result file exists or it could not be read.</p>
   * 
   * <p>Note that this method should never be overridden as it is called from the constructor.</p>
   * @return never null.
   */
  private MigrationStatus initializeMigrationStatus() {
    final File resultFilePath = getResultFilePath();
    try {
      if (resultFilePath.exists()) {
        try (final FileInputStream is = new FileInputStream(resultFilePath)) {
          final int lastRowCount = new CopayTierImportReader().countCopayTierImport(is);
          final MigrationStatus status = new MigrationStatus();
          status.setStatus(MigrationStatus.DONE);
          status.setProgressCount(lastRowCount);
          status.setTotalCount(lastRowCount);
          LOG.debug("Initialized MigrationStatus based on existing result file: {}", status);
          return status;
        }
      } else {
        LOG.debug("Import result file does not exist. Assuming no copay tier migration has been run.");
      }
    } catch (final Exception ex) {
      LOG.warn("Failed to read result file. Returning default MigrationStatus.", ex);
    }
    
    return new MigrationStatus();
  }

}
