/**
 * Source file created in 2008 by Southwest Research Institute
 */


package gov.va.med.pharmacy.peps.service.common.session.impl;


import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import org.apache.logging.log4j.Logger;
import org.springframework.stereotype.Service;

import gov.va.med.pharmacy.peps.common.exception.CsvFileException;
import gov.va.med.pharmacy.peps.common.vo.DataUpdateFileVo;
import gov.va.med.pharmacy.peps.common.vo.NationalSetting;
import gov.va.med.pharmacy.peps.common.vo.ReportDataUpdateVo;
import gov.va.med.pharmacy.peps.common.vo.UpdateSiteVo;
import gov.va.med.pharmacy.peps.service.common.session.AbstractReportExportProcess;
import gov.va.med.pharmacy.peps.service.common.session.ReportDataUpdateProcess;
import gov.va.med.pharmacy.peps.service.common.session.VistaUpdateSitesService;
import gov.va.med.pharmacy.peps.service.common.utility.ReportDataUpdateCsvFile;


/** Generates the Data Update Compliance report in a background process. */
@Service
public class ReportDataUpdateProcessImpl 
    extends AbstractReportExportProcess implements ReportDataUpdateProcess {

    private static final int MILLISECONDS_PER_DAY = 24 * 60 * 60 * 1000;
    private static final Logger LOG = org.apache.logging.log4j.LogManager.getLogger(ReportDataUpdateProcessImpl.class);
    private static final int DAYS_IN_COMPLIANCE = 7;

    private VistaUpdateSitesService vistaUpdateSitesService;

    /** 
     * Generates Data Update Compliance Report.
     * @throws InterruptedException 
     */
    @Override
    public final void doExport() throws InterruptedException {
        LOG.debug("Exporting Data Update Compliance Report");
        
        try {
            //Set start time
            setDateValue(NationalSetting.REPORT_DATA_UPDATE_START, getDateTime());

            ReportDataUpdateCsvFile reportDataUpdateCsvFile = new ReportDataUpdateCsvFile();

            try {
                reportDataUpdateCsvFile.createFile();
            } catch (CsvFileException e) {
                LOG.error("Cound create Data Update Compliance file for export: " + e);
            }
            
            ReportDataUpdateVo dataUpdateReport = new ReportDataUpdateVo();
            DataUpdateFileVo updateFile = getReportDomainCapability().getLatestUpdateFileInfo(); 
            dataUpdateReport.setFilename(updateFile.getFileName());
            dataUpdateReport.setReleaseDt(updateFile.getReleaseDt());
            Calendar cal = new GregorianCalendar();
            cal.setTime(updateFile.getReleaseDt());
            cal.add(Calendar.DAY_OF_MONTH, DAYS_IN_COMPLIANCE);
            Date complianceDt = cal.getTime();
            dataUpdateReport.setComplianceDt(complianceDt);
            
            // find / set sites in compliance
            List<UpdateSiteVo> sitesInCompliance = getReportDomainCapability().getSitesInCompliance(
                    updateFile.getId(), complianceDt);
            List<Long> siteNbrsInCompliance = new ArrayList<Long>();
            for (UpdateSiteVo siteVo : sitesInCompliance) {
                siteNbrsInCompliance.add(siteVo.getSiteNbr());
                siteVo.setSiteName(vistaUpdateSitesService.getNameBySiteNumber(siteVo.getSiteNbr()));
                siteVo.setDaysOfInstall(daysBetween(siteVo.getInstallDt(), updateFile.getReleaseDt()));
            }
            Collections.sort(sitesInCompliance);    // sort site names alphabetically
            dataUpdateReport.setSitesInCompliance(sitesInCompliance);
            
            // find / set sites out of compliance
            List<UpdateSiteVo> sitesOutOfCompliance = vistaUpdateSitesService.findSitesOutOfCompliances(siteNbrsInCompliance);
            for (UpdateSiteVo oocSite : sitesOutOfCompliance) {
                UpdateSiteVo siteVo = getReportDomainCapability().getSiteOutOfComplianceInfo(oocSite.getSiteNbr());
                oocSite.setFileName(siteVo == null ? "" : siteVo.getFileName());
                oocSite.setInstallDt(siteVo == null ? null : siteVo.getInstallDt());
                oocSite.setDaysOutOfCompliance(siteVo == null ? 0 : daysBetween(siteVo.getInstallDt(), complianceDt));
            }
            Collections.sort(sitesOutOfCompliance);    // sort site names alphabetically
            dataUpdateReport.setSitesOuOfCompliance(sitesOutOfCompliance);
            
            // print report
            reportDataUpdateCsvFile.printDataUpdateComplianceReport(dataUpdateReport);
            
            reportDataUpdateCsvFile.closeExport();

            //Set end time
            setDateValue(NationalSetting.REPORT_DATA_UPDATE_COMPLETE, getDateTime());

        } 
        catch (CsvFileException e) {
            LOG.error("FAILED TO generate Data Update Compliance Report. Error Message is ", e);
            this.stopProcess();
        }

    }
    
    private int daysBetween(Date d1, Date d2) {
        if ( d1 == null ) {
            LOG.warn("The d1 parameter passed to daysBetween() is null.");
            return 0;
        }
        if ( d2 == null ) {
            LOG.warn("The d2 parameter passed to daysBetween() is null.");
            return 0;
        }
        
        long diff = d2.getTime() - d1.getTime();
        return (int)(Math.abs(diff / MILLISECONDS_PER_DAY));
    }
    
    public VistaUpdateSitesService getVistaUpdateSitesService() {
        return vistaUpdateSitesService;
    }
    
    public void setVistaUpdateSitesService(VistaUpdateSitesService vistaUpdateSitesService) {
        this.vistaUpdateSitesService = vistaUpdateSitesService;
    }
}
