/**
 * Source file created in 2008 by Southwest Research Institute
 */


package gov.va.med.pharmacy.peps.service.common.session.impl;


import java.util.ArrayList;
import java.util.List;

import org.apache.logging.log4j.Logger;
import org.springframework.stereotype.Service;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionException;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.DefaultTransactionDefinition;

import gov.va.med.pharmacy.peps.common.exception.CsvFileException;
import gov.va.med.pharmacy.peps.common.vo.NationalSetting;
import gov.va.med.pharmacy.peps.common.vo.ReportCaptureNdfVo;
import gov.va.med.pharmacy.peps.common.vo.ReportType;
import gov.va.med.pharmacy.peps.service.common.reports.ReportProcessStatus;
import gov.va.med.pharmacy.peps.service.common.session.AbstractReportExportProcess;
import gov.va.med.pharmacy.peps.service.common.session.ReportNdfProcess;
import gov.va.med.pharmacy.peps.service.common.utility.ReportNdfCsvFile;


/** Generates the Capture NDF Data report in a background process. */
@Service
public class ReportNdfProcessImpl 
    extends AbstractReportExportProcess implements ReportNdfProcess {

    private static final Logger LOG = org.apache.logging.log4j.LogManager.getLogger(ReportNdfProcessImpl.class);
    private static final String ERROR = "Cound not open NDC file for export: ";

    /**
     * Executes Capture NDF Data Report.
     * @throws InterruptedException thrown if process interrupted 
     */
    public void doExport() throws InterruptedException {
        LOG.debug("Exporting NDF Capture Report");

        this.getExportState().setReportProcessStatus(ReportProcessStatus.RUNNING);

        DefaultTransactionDefinition def = null;
        TransactionStatus status = null;
        List<Long> eplIds = new ArrayList<Long>();

        try {
            //Set start time
            setDateValue(NationalSetting.REPORT_CAPTURE_START, getDateTime());
            ReportNdfCsvFile reportNdfCsvFile = new ReportNdfCsvFile();

            try {
                reportNdfCsvFile.createFile();
            } catch (CsvFileException e) {
                /*
                 * TODO: There is no indication to the user that an exception has occurred.
                 */
                LOG.error(ERROR + e);
            }

            def = new DefaultTransactionDefinition(TransactionDefinition.PROPAGATION_REQUIRES_NEW);
            def.setTimeout(TIMEOUT);
            status = getTransactionManager().getTransaction(def);
            eplIds = getReportDomainCapability().getIds(ReportType.NDC_LIST_PRINT_TEMPLATE);
            getExportState().setRecordTotal(eplIds.size());
            getTransactionManager().commit(status);

            // Notify observer
            notifyObservers();

            List<Long> actionList = new ArrayList<Long>(BATCH_SIZE);
            int i = 0;

            for (Long eplId : eplIds) {
                actionList.add(eplId);
                i++;

                if (actionList.size() >= BATCH_SIZE) {
                    processBatch(eplIds, reportNdfCsvFile, actionList, i);
                }
            }

            if (actionList.size() > 0) {
                processBatch(eplIds, reportNdfCsvFile, actionList, i);
            }

            reportNdfCsvFile.closeExport();

        } catch (TransactionException e) {
            LOG.error("FAILED TO retrieve List. Error Message is " + e.getMessage());
            this.stopProcess();
        }

        //Set stop time
        setDateValue(NationalSetting.REPORT_CAPTURE_COMPLETE, getDateTime());
    }

    private void processBatch(List<Long> eplIds, ReportNdfCsvFile reportNdfCsvFile, List<Long> actionList, int recordCounter) {
        DefaultTransactionDefinition def;
        TransactionStatus status;
        LOG.debug("Report Processing: " + recordCounter + " of " + eplIds.size());
        def = new DefaultTransactionDefinition(TransactionDefinition.PROPAGATION_REQUIRES_NEW);
        def.setTimeout(TIMEOUT);
        status = getTransactionManager().getTransaction(def);
        List<ReportCaptureNdfVo> rtList = getReportDomainCapability().getCaptureNdfData(actionList);
        getTransactionManager().commit(status);

        // Now call the report NDC File to print the list of NDCS
        try {
            reportNdfCsvFile.printNDC(rtList);
        } catch (CsvFileException e) {
            /*
             * TODO: There is no indication to the user that an exception has occurred.
             * This exception would occur if there was an error writing to the file
             * on disk. The report might be incomplete or missing records, and the user
             * wouldn't know.
             */
            LOG.error("Cound not open NDC file for export: " + e);
        }

        actionList.clear();

        // Update record count.                  
        getExportState().setRecordCounter(recordCounter);
        notifyObservers();
    }
}
