package gov.va.med.pharmacy.peps.service.common.capability.impl;

import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import static org.easymock.EasyMock.*;

import org.easymock.EasyMockSupport;
import org.junit.Before;
import org.junit.Test;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;

import gov.va.med.pharmacy.peps.common.utility.DateTimeUtility;
import gov.va.med.pharmacy.peps.common.vo.FdbAutoAddVo;
import gov.va.med.pharmacy.peps.common.vo.NationalSetting;
import gov.va.med.pharmacy.peps.common.vo.NdcVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.domain.common.capability.FdbAddDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.FdbAutoAddDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.FdbAutoUpdateDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.FdbUpdateDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.NdcDomainCapability;
import gov.va.med.pharmacy.peps.external.common.drugdatavendor.outbound.capability.DrugReferenceAutoCapability;
import gov.va.med.pharmacy.peps.external.common.drugdatavendor.outbound.capability.DrugReferenceCapability;
import gov.va.med.pharmacy.peps.service.common.scheduler.ProcessStatus;
import gov.va.med.pharmacy.peps.service.common.scheduler.ProcessStatusUpdater;

public class AutoAddUpdateCapabilityImplTest extends EasyMockSupport {

  /**
   * {@link DateTimeFormatter} for {@link DrugReferenceAutoCapability#FDB_DATE_PATTERN}.
   */
  private static final DateTimeFormatter FDB_DATE_FORMATTER =
      DateTimeFormatter.ofPattern(DrugReferenceAutoCapability.FDB_DATE_PATTERN);

  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private ProcessStatusUpdater mockProcessStatusUpdater;

  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private PlatformTransactionManager mockTxManager;

  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private FdbAutoAddDomainCapability mockAutoAddCapability;

  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private FdbAddDomainCapability mockAddCapability;
  
  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private FdbAutoUpdateDomainCapability mockAutoUpdateCapability;
  
  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private FdbUpdateDomainCapability mockUpdateCapability;
  
  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private DrugReferenceCapability mockDrugRefCapability;

  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private DrugReferenceAutoCapability mockDrugRefAutoCapability;
  
  /**
   * Mock stored as a member variable for convenient creation in {@link #setup()}.
   */
  private NdcDomainCapability mockNdcDomainCapability;
  
  @Before
  public void createMocks() {
    mockProcessStatusUpdater = strictMock(ProcessStatusUpdater.class);
    mockTxManager = strictMock(PlatformTransactionManager.class);
    mockAutoAddCapability = strictMock(FdbAutoAddDomainCapability.class);
    mockAddCapability = strictMock(FdbAddDomainCapability.class);
    mockAutoUpdateCapability = strictMock(FdbAutoUpdateDomainCapability.class);
    mockUpdateCapability = strictMock(FdbUpdateDomainCapability.class);
    mockDrugRefCapability = strictMock(DrugReferenceCapability.class);
    mockDrugRefAutoCapability = strictMock(DrugReferenceAutoCapability.class);
    mockNdcDomainCapability = strictMock(NdcDomainCapability.class);
  }
  
  /**
   * Creates the object under test with the mocks stored as member variables. Note that only the properties needed for
   * the below tests are set! Others will have their default values (probably null).
   */
  private AutoAddUpdateCapabilityImpl createObjUnderTest() {
    final AutoAddUpdateCapabilityImpl objUnderTest = new AutoAddUpdateCapabilityImpl(mockProcessStatusUpdater);
    objUnderTest.setTransactionManager(mockTxManager);
    objUnderTest.setFdbAutoAddDomainCapability(mockAutoAddCapability);
    objUnderTest.setFdbAddDomainCapability(mockAddCapability);
    objUnderTest.setFdbAutoUpdateDomainCapability(mockAutoUpdateCapability);
    objUnderTest.setFdbUpdateDomainCapability(mockUpdateCapability);
    objUnderTest.setDrugReferenceCapability(mockDrugRefCapability);
    objUnderTest.setDrugReferenceAutoCapability(mockDrugRefAutoCapability);
    objUnderTest.setNdcDomainCapability(mockNdcDomainCapability);
    return objUnderTest;
  }
  
  /**
   * Tests {@link AutoAddUpdateCapabilityImpl#processFDBAddedPackagedDrugsIntoEpl()} when there are no new NDCs to
   * check.
   */
  @Test
  public void testProcessFDBAddedPackagedDrugsIntoEpl_NoNewNdcs() {
    
    /* Setup */
    
    final NationalSetting jobRunStateSetting = NationalSetting.FDB_ADD_RUN_STATE;
    final NationalSetting jobLastRunSetting = NationalSetting.FDB_ADD_LAST_RUN;
    final Date previousLastRunTime = Date.from(DateTimeUtility.toInstant(LocalDate.of(2017, 2, 4)));
    
    final AutoAddUpdateCapabilityImpl objUnderTest = createObjUnderTest();
    
    /* Expected method calls */
    
    final UserVo user = objUnderTest.getUser();
    expectDeleteAddsOver90();
    expect(mockProcessStatusUpdater.startJob(jobRunStateSetting, jobLastRunSetting, user))
      .andReturn(previousLastRunTime);
    expect(mockDrugRefAutoCapability.getFdbAddedItems(previousLastRunTime)).andReturn(Collections.emptyList());
    expect(mockProcessStatusUpdater.updateStatusAndRuntime(
        // We will not change the previous last run time since we did not process any added NDCs.
        jobRunStateSetting, ProcessStatus.COMPLETED, jobLastRunSetting, previousLastRunTime, user))
      .andReturn(previousLastRunTime);
    
    replayAll();
    
    /* Behavior & Verification */
    
    objUnderTest.processFDBAddedPackagedDrugsIntoEpl();
    
    verifyAll();
    
    // All verification was done via expectations above: we don't check any postconditions.
  }
  
  @Test
  public void testProcessFDBAddedPackagedDrugsIntoEpl_NdcsToAdd() throws Exception {
    
    /* Setup */
    
    final NationalSetting jobRunStateSetting = NationalSetting.FDB_ADD_RUN_STATE;
    final NationalSetting jobLastRunSetting = NationalSetting.FDB_ADD_LAST_RUN;
    final LocalDate lastRunDate = LocalDate.of(2017, 2, 5);
    final Date previousLastRunTime = Date.from(DateTimeUtility.toInstant(lastRunDate));
    final long existingFdbDrugEplId = 15000L;
    
    /*-
     * We test two cases:
     * a) an FDB drug that already has an NDC in PPS-N
     * b) an FDB drug that for some reason causes an Exception and a transaction rollback
     * 
     * It would be very useful to test more, but not enough time right now.
     */

    // Case (a)
    final FdbAutoAddVo existingFdbDrug = new FdbAutoAddVo();
    existingFdbDrug.setNdc("00115165651");
    existingFdbDrug.setGcnSeqno(47636L);
    // This drug was added 1 day after the last run.
    final LocalDate existingFdbDrugDate = lastRunDate.plusDays(1L);
    existingFdbDrug.setAddDate(existingFdbDrugDate.format(FDB_DATE_FORMATTER));

    // Case (b)
    final FdbAutoAddVo fdbDrugForRollback = new FdbAutoAddVo();
    fdbDrugForRollback.setNdc("00065041663");
    fdbDrugForRollback.setGcnSeqno(63562L);
    // This drug was added 1 day after the above drug.
    final LocalDate fdbDrugForRollbackDate = existingFdbDrugDate.plusDays(1L);
    fdbDrugForRollback.setAddDate(fdbDrugForRollbackDate.format(FDB_DATE_FORMATTER));

    final List<FdbAutoAddVo> addedDrugs = Arrays.asList(existingFdbDrug, fdbDrugForRollback);

    final AutoAddUpdateCapabilityImpl objUnderTest = createObjUnderTest();
    
    /* Expected method calls */
    
    final UserVo user = objUnderTest.getUser();
    expectDeleteAddsOver90();
    expect(mockProcessStatusUpdater.startJob(jobRunStateSetting, jobLastRunSetting, user))
      .andReturn(previousLastRunTime);
    expect(mockDrugRefAutoCapability.getFdbAddedItems(previousLastRunTime)).andReturn(addedDrugs);

    // Expected calls for case (a)
    final TransactionStatus caseATxn = strictMock(TransactionStatus.class);
    expect(mockTxManager.getTransaction(anyObject())).andReturn(caseATxn);
    expect(mockNdcDomainCapability.getIdFromNDC(existingFdbDrug.getNdc())).andReturn(existingFdbDrugEplId);
    // We don't care what we return here, just need to return something non-null.
    expect(mockNdcDomainCapability.retrieve(String.valueOf(existingFdbDrugEplId))).andReturn(new NdcVo());
    mockTxManager.commit(caseATxn);

    // Expected calls for case (b)
    final TransactionStatus caseBTxn = strictMock(TransactionStatus.class);
    expect(mockTxManager.getTransaction(anyObject())).andReturn(caseBTxn);
    // Pretend that this NDC doesn't exist.
    expect(mockNdcDomainCapability.getIdFromNDC(fdbDrugForRollback.getNdc())).andReturn(0L);
    expect(mockDrugRefCapability.populateFdbAddVoFields(fdbDrugForRollback.getNdc())).andThrow(
        new RuntimeException("mocked exception", new IllegalStateException("mocked IllegalStateException")));
    mockTxManager.rollback(caseBTxn);

    // We should update the last run date to the latest _successfully_ processed NDC.
    final Date expectedNewLastRunTime = Date.from(DateTimeUtility.toInstant(existingFdbDrugDate));
    expect(mockProcessStatusUpdater.updateStatusAndRuntime(
        jobRunStateSetting, ProcessStatus.COMPLETED, jobLastRunSetting, expectedNewLastRunTime, user))
      .andReturn(previousLastRunTime);
    
    replayAll();
    
    /* Behavior & Verification */
    
    objUnderTest.processFDBAddedPackagedDrugsIntoEpl();
    
    verifyAll();
    
    // All verification was done via expectations above: we don't check any postconditions.
  }
  
  /**
   * Tests {@link AutoAddUpdateCapabilityImpl#synchronizeFdbUpdate()} when there are no new NDCs to check.
   */
  @Test
  public void testSynchronizeFdbUpdate_NoNewNdcs() {
    
    /* Setup */
    
    final NationalSetting jobRunStateSetting = NationalSetting.FDB_UPDATE_RUN_STATE;
    final NationalSetting jobLastRunSetting = NationalSetting.FDB_UPDATE_LAST_RUN;
    final Date previousLastRunTime = Date.from(DateTimeUtility.toInstant(LocalDate.of(2017, 1, 15)));
    
    final AutoAddUpdateCapabilityImpl objUnderTest = createObjUnderTest();
    
    /* Expected method calls */
    
    final UserVo user = objUnderTest.getUser();
    expectDeleteUpdatesOver90();
    expect(mockProcessStatusUpdater.startJob(jobRunStateSetting, jobLastRunSetting, user))
      .andReturn(previousLastRunTime);
    expect(mockDrugRefAutoCapability.getFdbUpdatedItems(previousLastRunTime)).andReturn(Collections.emptyList());
    expect(mockProcessStatusUpdater.updateStatusAndRuntime(
        // We will not change the previous last run time since we did not process any updated NDCs.
        jobRunStateSetting, ProcessStatus.COMPLETED, jobLastRunSetting, previousLastRunTime, user))
      .andReturn(previousLastRunTime);
    
    replayAll();
    
    /* Behavior & Verification */
    
    objUnderTest.synchronizeFdbUpdate();
    
    verifyAll();
    
    // All verification was done via expectations above: we don't check any postconditions.
  }

  /**
   * Records expectations for deleting FDB adds over 90 days old.
   */
  private void expectDeleteAddsOver90() {
    final TransactionStatus delete90Transaction = strictMock(TransactionStatus.class);
    expect(mockTxManager.getTransaction(anyObject())).andReturn(delete90Transaction);
    mockAutoAddCapability.deleteOver90();
    mockAddCapability.deleteOver90();
    mockTxManager.commit(delete90Transaction);
  }
  
  /**
   * Records expectation for deleting FDB updates over 90 days old.
   */
  private void expectDeleteUpdatesOver90() {
    final TransactionStatus delete90Transaction = strictMock(TransactionStatus.class);
    expect(mockTxManager.getTransaction(anyObject())).andReturn(delete90Transaction);
    mockAutoUpdateCapability.deleteOver90();
    mockUpdateCapability.deleteOver90();
    mockTxManager.commit(delete90Transaction);
  }

}
