

package gov.va.med.pharmacy.peps.service.common.capability.impl;


import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.CoreMatchers.notNullValue;
import static org.junit.Assert.assertThat;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang3.time.DateUtils;
import org.easymock.EasyMock;
import org.hamcrest.CoreMatchers;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.apache.logging.log4j.Logger; 
import org.apache.logging.log4j.LogManager;

import gov.va.med.pharmacy.peps.common.exception.ItemNotFoundException;
import gov.va.med.pharmacy.peps.common.utility.EnvironmentUtility;
import gov.va.med.pharmacy.peps.common.utility.test.generator.CopayTierGenerator;
import gov.va.med.pharmacy.peps.common.utility.test.generator.DrugClassGenerator;
import gov.va.med.pharmacy.peps.common.vo.CopayTierAssocVo;
import gov.va.med.pharmacy.peps.common.vo.CopayTierVo;
import gov.va.med.pharmacy.peps.common.vo.DrugClassGroupVo;
import gov.va.med.pharmacy.peps.common.vo.DrugClassVo;
import gov.va.med.pharmacy.peps.common.vo.EntityType;
import gov.va.med.pharmacy.peps.common.vo.Environment;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.common.vo.ManagedItemVo;
import gov.va.med.pharmacy.peps.common.vo.ModDifferenceVo;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.Role;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.GroupDifference;
import gov.va.med.pharmacy.peps.common.vo.diff.test.DiffCopayTierTest;
import gov.va.med.pharmacy.peps.common.vo.validator.ErrorKey;
import gov.va.med.pharmacy.peps.common.vo.validator.Errors;
import gov.va.med.pharmacy.peps.common.vo.validator.ValidationError;
import gov.va.med.pharmacy.peps.common.vo.validator.ValidationErrorLevel;
import gov.va.med.pharmacy.peps.service.common.capability.ItemModDifferences;
import gov.va.med.pharmacy.peps.service.common.utility.ManagedItemCapabilityFactory;


/**
 * Integration Test for ManagedItemCapabilityImpl.  
 * 
 * Details of ManagedItemCapabilityTest's operations, special dependencies
 * or protocols developers shall know about when using the class.
 *
 */
public class ManagedItemCapabilityIT {

    /** The Constant LOG. */
    private static final Logger LOG = LogManager.getLogger(DiffCopayTierTest.class);

    /** The managed item capability. */
    private ManagedItemCapabilityImpl managedItemCapability;

    /** The managed item capability factory. */
    private ManagedItemCapabilityFactory managedItemCapabilityFactory;

    /** The rules capability. */
    private ProductRulesCapabilityImpl rulesCapability;

    /** The environment utility. */
    private EnvironmentUtility environmentUtility;

    /** The Constant USER. */
    private static final UserVo USER = new UserVo();

    /** The copay tier list. */
    private final List<CopayTierVo> copayTierList = new CopayTierGenerator().generate();

    /** The drug class list. */
    private final List<DrugClassGroupVo> drugClassList = new DrugClassGenerator().generate();

    /**
     * Creates the mock managedItemCapability for the category cascade test.
     *
     * @throws Exception Exception
     */
    @Before
    public final void setUp() throws Exception {
        managedItemCapability = EasyMock.partialMockBuilder(ManagedItemCapabilityImpl.class)
        		  .addMockedMethods("toggleProductSupplySpecialHandling", 
        				  "saveItemAuditHistoryCascade","getRulesCapability" ).createMock();

        environmentUtility = EasyMock.createMock(EnvironmentUtility.class);
        managedItemCapabilityFactory = EasyMock.createMock(ManagedItemCapabilityFactory.class);
        managedItemCapability.setManagedItemCapabilityFactory(managedItemCapabilityFactory);
        //Needed to avoid NOT_AUTHORIZED Error
        USER.addRole(Role.PSS_PPSN_MANAGER);
    }

    /**
     * Cleans up the mock managedItemCapability.
     */
    @After
    public final void cleanUp() {
        managedItemCapability = null;
    }

    /**
     * Scenario: Template and target have no copay tiers.
     * Add copay tier to template and propagate to targets.
     * Expect that the old value = "No Selection Made" and New Value = the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TemplateEmpty_TargetEmpty() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TemplateEmpty_TargetEmpty";
        int[] TEMPLATE_START = {};
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 2 };
        int[] TARGET_START = {};
        int[] TARGET_RESULT = { 2 };
        int[] TARGET_WITH_END_DATE = {};

        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE));
    }

    /**
     * Scenario: Template and target have existing matching copay tiers.
     * Add copay tier to template and propagate to targets.
     * Expect that the old value = item's old value and New Value = the item's old value plus the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TemplateExisting_TargetSameExisting_TwoFuture() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TemplateExisting_TargetSameExisting";
        CopayTestParameters parms = new CopayTestParameters();
        CopayTestTemplateItemParameters templateParms = new CopayTestTemplateItemParameters();
        
        templateParms.startPast = arr();
        templateParms.startActive = arr(1);
        templateParms.startFuture = arr(2);
        templateParms.add = arr(3);
        templateParms.result = arr(1, 2, 3);
        
        parms.templateParameters = templateParms;
        
        CopayTestTargetItemParameters targetParms = new CopayTestTargetItemParameters();
        targetParms.startPast = arr();
        targetParms.startActive = arr(1);
        targetParms.startFuture = arr(2);
        targetParms.result = arr(1, 2, 3);
        targetParms.resultWithEndDate = arr(1, 2);
        targetParms.productName = "TARGET1";
        
        parms.targetParametersList.add(targetParms);

        testAddCopayTierCommon2(TEST_NAME, parms);
    }

    /**
     * Scenario: Template and target have existing matching copay tiers.
     * Add copay tier to template and propagate to targets.
     * Expect that the old value = item's old value and New Value = the item's old value plus the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TemplateExisting_TargetSameExisting_OneFuture() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TemplateExisting_TargetSameExisting";
        int[] TEMPLATE_START = { 1 };
        int TEMPLATE_ADD = 3;
        int[] TEMPLATE_RESULT = { 1, 3 };
        int[] TARGET_START = { 1 };
        int[] TARGET_RESULT = { 1, 3 };
        int[] TARGET_WITH_END_DATE = { 1 };

        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE));
    }
    
    
    /**
     * Scenario: Template has existing copay tiers, but target has none.
     * Add copay tier to template and propagate to targets.
     * Expect that the old value = item's old value and New Value = the item's old value plus the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TemplateExisting_TargetEmpty() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TemplateExisting_TargetEmpty";
        int[] TEMPLATE_START = { 3 };
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 3, 2 };
        int[] TARGET_START = {};
        int[] TARGET_RESULT = { 2 };
        int[] TARGET_WITH_END_DATE = {};
        Errors templateErrors = new Errors();
        Errors targetErrors = new Errors();
        targetErrors.addError(new ValidationError(FieldKey.COPAY_TIER_ASSOC, ErrorKey.COPAY_TIER_INVALID_MULTI_EDIT_SCENARIO,
                ValidationErrorLevel.ERROR));
        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, 
        		TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE, templateErrors, targetErrors));

    }

    /**
     * Scenario: Template has copay tier 1, target has copay tier 2.
     * Add copay tier 3 to template and propagate to targets.
     * Expect that the old value = "No Selection Made" and New Value = the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TemplateExisting_TargetDiffExisting() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TemplateExisting_TargetDiffExisting";
        int[] TEMPLATE_START = { 1 };
        int TEMPLATE_ADD = 3;
        int[] TEMPLATE_RESULT = { 1, 3 };
        int[] TARGET_START = { 2 };
        int[] TARGET_RESULT = { 2, 3 };
        int[] TARGET_WITH_END_DATE = { 2 };

        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE));

    }

    /**
     * Scenario: Template has copay tier 1, target has copay tier 2.
     * Add copay tier 3 to template and propagate to targets.
     * Expect that the old value = "No Selection Made" and New Value = the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TemplateMultiExist_TargetMultiDiffExist_SameActive() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TemplateMultiExist_TargetMultiDiffExist_SameActive";
        int[] TEMPLATE_START = { 1 };
        int TEMPLATE_ADD = 3;
        int[] TEMPLATE_RESULT = { 1, 3 };
        int[] TARGET_START = { 2, 1 };
        int[] TARGET_RESULT = { 2, 1, 3 };
        int[] TARGET_WITH_END_DATE = { 2, 1 };

        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE));

    }

    /**
     * Scenario: Template has copay tier 1, target has copay tier 2.
     * Add copay tier 3 to template and propagate to targets.
     * Expect that the old value = "No Selection Made" and New Value = the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TemplateMultiExist_TargetMultiSame_ListHasOldRows() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TemplateMultiExist_TargetMultiSame_ListHasOldRows";
        int[] TEMPLATE_START = { 3, 2, 3 };
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 3, 2, 3, 2 };
        int[] TARGET_START = { 3, 2, 3 };
        int[] TARGET_RESULT = { 3, 2, 3, 2 };
        int[] TARGET_WITH_END_DATE = { 3, 2, 3 };

        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE));

    }

    /**
     * Scenario: Template has copay tier 1, target has copay tier 2.
     * Add copay tier 3 to template and propagate to targets.
     * Expect that the old value = "No Selection Made" and New Value = the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TemplateMultiExist_TargetMultiSame() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TemplateMultiExist_TargetMultiSame";
        int[] TEMPLATE_START = { 2, 3 };
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 2, 3, 2 };
        int[] TARGET_START = { 2, 3 };
        int[] TARGET_RESULT = { 2, 3, 2 };
        int[] TARGET_WITH_END_DATE = { 2, 3 };

        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE));

    }

    /**
     * Scenario: Template has copay tier 1, target has copay tier 2.
     * Add copay tier 3 to template and propagate to targets.
     * Expect that the old value = "No Selection Made" and New Value = the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddCopayTier_TargetActive() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TargetActive";
        int[] TEMPLATE_START = {};
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 2 };
        int[] TARGET_START = { 2 };
        int[] TARGET_RESULT = { 2 };
        int[] TARGET_WITH_END_DATE = {};
        Errors templateErrors = new Errors();
        Errors targetErrors = new Errors();
        targetErrors.addError(new ValidationError(FieldKey.COPAY_TIER_ASSOC, ErrorKey.COPAY_TIER_INVALID_MULTI_EDIT_SCENARIO,
                ValidationErrorLevel.ERROR));        
        //TODO : Need to assert the validation error, but this test doesn't call the validation.  Otherwise, at least assert that the target's original effective date was not overwritten?
        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, 
        		TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE, templateErrors, targetErrors));
    }

    /**
     * Scenario: Template has copay tier 1, target has copay tier 2.
     * Add copay tier 3 to template and propagate to targets.
     * Expect that the old value = "No Selection Made" and New Value = the added copay tier info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testRemoveCopayTier_TargetActive() throws Exception {
        final String TEST_NAME = "testAddCopayTier_TargetActive";
        int[] TEMPLATE_START = {};
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 2 };
        int[] TARGET_START = { 2 };
        int[] TARGET_RESULT = { 2 };
        int[] TARGET_WITH_END_DATE = {};
        Errors templateErrors = new Errors();
        Errors targetErrors = new Errors();
        targetErrors.addError(new ValidationError(FieldKey.COPAY_TIER_ASSOC, ErrorKey.COPAY_TIER_INVALID_MULTI_EDIT_SCENARIO,
                ValidationErrorLevel.ERROR));
        
        testAddCopayTierCommon(new TestAddCopayTierCommonParameters(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT, 
        		TARGET_START, TARGET_RESULT, TARGET_WITH_END_DATE, templateErrors, targetErrors));
    }    
    
    /**
     * Scenario: Template and target have no existing drug class.  
     * Normally, this is no longer allowed by the application, but there are records that exist without drug class today.
     * Add drug class to template and propagate to targets.
     * Expect that the old value = "No Selection Made" and New Value = the added drug class info.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddDrugClass_TemplateAndTargetEmpty() throws Exception {
        final String TEST_NAME = "testAddDrugClass_TemplateAndTargetEmpty";
        int[] TEMPLATE_START = { };
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 2 };
        int[] TARGET_START = { };
        int[] TARGET_RESULT = { 2 };

        testAddDrugClassCommon(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT,
            TARGET_START, TARGET_RESULT);

    }

    /**
     * Scenario: Template and target have same existing drug class.
     * Add drug class to template and propagate to targets.
     * Expect that the old value = the existing drug class and New Value = The existing plus the added drug class.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddDrugClass_TemplateExisting_TargetSameExisting() throws Exception {
        final String TEST_NAME = "testAddDrugClass_TemplateExisting_TargetSameExisting";
        int[] TEMPLATE_START = { 1 };
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 1, 2 };
        int[] TARGET_START = { 1 };
        int[] TARGET_RESULT = { 1, 2 };

        testAddDrugClassCommon(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT,
            TARGET_START, TARGET_RESULT);

    }
    
    
    /**
     * Scenario: Template and target have same existing drug class.
     * Add drug class to template and propagate to targets.
     * Expect that the old value = the existing drug class and New Value = The existing plus the added drug class.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testAddDrugClass_AlreadyExistsOnTarget() throws Exception {
        final String TEST_NAME = "testAddDrugClass_AlreadyExistsOnTarget";
        int[] TEMPLATE_START = { 1 };
        int TEMPLATE_ADD = 2;
        int[] TEMPLATE_RESULT = { 1, 2 };
        int[] TARGET_START = { 1, 2 };
        int[] TARGET_RESULT = { 1, 2 };

        testAddDrugClassCommon(TEST_NAME, TEMPLATE_START, TEMPLATE_ADD, TEMPLATE_RESULT,
            TARGET_START, TARGET_RESULT);

    }    
    
    /**
     * Scenario: Template and target have existing matching drug classes.
     * Swap out drug class on template and propagate to targets.
     * Expect that the old value = Drug Class 1 and New Value = Drug Class 2 on both template and target.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testChangeDrugClass() throws Exception {
        LOG.info("Start " + "testChangeDrugClass");
        //template original
        ProductVo templateOriginal = new ProductVo();
        templateOriginal.setVaProductName("TEMPLATE");
        templateOriginal.setId("1");
        templateOriginal.setDrugClasses(createDrugClassList(templateOriginal, 1));
        Collection<ManagedItemVo> targets = new ArrayList<ManagedItemVo>();

        //template modified
        ProductVo templateModified = (ProductVo) templateOriginal.clone();
        assertThat(templateModified.getDrugClasses().size(), is(1));
        templateModified.setDrugClasses(createDrugClassList(templateOriginal, 2));
        assertThat(templateModified.getDrugClasses().size(), is(1));
        assertThat(templateOriginal.getDrugClasses().size(), is(1));

        //targets
        ProductVo target = new ProductVo();
        target.setVaProductName("TARGET1");
        target.setId("2");
        target.setDrugClasses(createDrugClassList(target, 1));
        targets.add(templateOriginal);
        targets.add(target);
        prepareMockExpectations();

        Collection<ItemModDifferences> differencesCollection = mimicPartOfSubmit(templateOriginal, targets, templateModified);

        for (ItemModDifferences itemDifferences : differencesCollection) {
            if (!((ProductVo) itemDifferences.getItem()).getVaProductName().equals(templateModified)) {

                for (ModDifferenceVo modDifferenceVo : itemDifferences.getDifferences()) {
                    LOG.debug(modDifferenceVo.getDifference().getOldValueShortString());
                    LOG.debug(modDifferenceVo.getDifference().getNewValueShortString());
                    assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedDrugClassRegEx(1)));
                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedDrugClassRegEx(2)));
                }
            }
        }

        mimicPartOfCommit(templateModified, differencesCollection);
        LOG.info("End " + "testAddDrugClassToExistingSame");
    }

    /**
     * Test add drug class common.
     *
     * @param testName the test name
     * @param TEMPLATE_START the template start
     * @param TEMPLATE_ADD the template add
     * @param TEMPLATE_RESULT the template result
     * @param TARGET_START the target start
     * @param TARGET_RESULT the target result
     * @throws Exception the exception
     */
    private void testAddDrugClassCommon(String testName, final int[] TEMPLATE_START, final int TEMPLATE_ADD,
        final int[] TEMPLATE_RESULT, final int[] TARGET_START, final int[] TARGET_RESULT) throws Exception {
        LOG.info("Start " + testName);

        //template original
        ProductVo templateOriginal = new ProductVo();
        templateOriginal.setVaProductName("TEMPLATE");
        templateOriginal.setId("1");
        templateOriginal.setDrugClasses(createDrugClassList(templateOriginal, TEMPLATE_START));

        //template modified
        ProductVo templateModified = (ProductVo) templateOriginal.clone();
        assertThat(templateModified.getDrugClasses().size(), is(TEMPLATE_START.length));
        templateModified.setDrugClasses(addDrugClass(templateModified, templateModified.getDrugClasses(),
            TEMPLATE_ADD));
        assertThat(templateModified.getDrugClasses().size(), is(TEMPLATE_RESULT.length));
        assertThat(templateOriginal.getDrugClasses().size(), is(TEMPLATE_START.length));

        //targets
        Collection<ManagedItemVo> targets = new ArrayList<ManagedItemVo>();
        ProductVo target = new ProductVo();
        target.setVaProductName("TARGET1");
        target.setId("2");
        target.setDrugClasses(createDrugClassList(target, TARGET_START));
        targets.add(templateOriginal);
        targets.add(target);

        prepareMockExpectations();

        Collection<ItemModDifferences> differencesCollection = mimicPartOfSubmit(templateOriginal, targets, templateModified);
        for (ItemModDifferences itemDifferences : differencesCollection) {
            ProductVo product = (ProductVo) itemDifferences.getItem();

            for (ModDifferenceVo modDifferenceVo : itemDifferences.getDifferences()) {
                LOG.debug("Asserting " + product.getVaProductName());
                LOG.debug(modDifferenceVo.getDifference().getOldValueShortString());
                LOG.debug(modDifferenceVo.getDifference().getNewValueShortString());
                if (product.getVaProductName().equals("TEMPLATE")) {
                    assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedDrugClassRegEx(TEMPLATE_START)));
                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedDrugClassRegEx(TEMPLATE_RESULT)));
                } else if (product.getVaProductName().equals("TARGET1")) {
                    assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedDrugClassRegEx(TARGET_START)));
                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedDrugClassRegEx(TARGET_RESULT)));
                }
            }
        }

        //CommitAll
        Collection<ProductVo> committedProducts = mimicPartOfCommit(templateModified, differencesCollection);
        assertThat(committedProducts.size(), is(2));

        for (ProductVo committedProduct : committedProducts) {
            LOG.debug("Asserting Committed " + committedProduct.getVaProductName());
            if (committedProduct.getVaProductName().equals("TEMPLATE")) {
                assertCommittedProductDrugClasses(committedProduct, TEMPLATE_RESULT);
                assertThat(committedProduct.getDrugClasses(), is(templateModified.getDrugClasses()));
            } else if (committedProduct.getVaProductName().equals("TARGET1")) {
                assertCommittedProductDrugClasses(committedProduct, TARGET_RESULT);
            } else {
                Assert.fail("Unexpected Product: " + committedProduct.getVaProductName());
            }
        }
        LOG.info("End " + testName);
    }

    /**
     * Creates the string list from int array.
     *
     * @param intArray the int array
     * @return the list
     */
    private List<String> createStringListFromIntArray(int[] intArray) {
        List<String> list = new ArrayList<String>(intArray.length);
        for (int intValue : intArray) {
            list.add(String.valueOf(intValue));
        }
        return list;
    }

    public static class TestAddCopayTierCommonParameters {
		private String testName;
		private int[] templateStart;
		private int templateAdd;
		private int[] templateResult;
		private int[] targetStart;
		private int[] targetResult;
		private int[] targetWithEndDate;
		private Errors templateErrors = new Errors();
		private Errors targetErrors = new Errors();

		public TestAddCopayTierCommonParameters(String testName, int[] templateStart, int templateAdd,
				int[] templateResult, int[] targetStart, int[] targetResult, int[] targetWithEndDate) {
			this.testName = testName;
			this.templateStart = templateStart;
			this.templateAdd = templateAdd;
			this.templateResult = templateResult;
			this.targetStart = targetStart;
			this.targetResult = targetResult;
			this.targetWithEndDate = targetWithEndDate;
		}

		public TestAddCopayTierCommonParameters(String testName, int[] templateStart, int templateAdd,
				int[] templateResult, int[] targetStart, int[] targetResult, int[] targetWithEndDate, Errors templateErrors, Errors targetErrors) {
			this.testName = testName;
			this.templateStart = templateStart;
			this.templateAdd = templateAdd;
			this.templateResult = templateResult;
			this.targetStart = targetStart;
			this.targetResult = targetResult;
			this.targetWithEndDate = targetWithEndDate;
			this.templateErrors = templateErrors;
			this.targetErrors = targetErrors;
		}
		
		
		public String getTestName() {
			return testName;
		}

		public void setTestName(String testName) {
			this.testName = testName;
		}

		public int[] gettemplateStart() {
			return templateStart;
		}

		public int gettemplateAdd() {
			return templateAdd;
		}

		public int[] gettemplateResult() {
			return templateResult;
		}

		public int[] gettargetStart() {
			return targetStart;
		}

		public int[] gettargetResult() {
			return targetResult;
		}

		public int[] gettargetWithEndDate() {
			return targetWithEndDate;
		}

		public int[] getTemplateStart() {
			return templateStart;
		}

		public void setTemplateStart(int[] templateStart) {
			this.templateStart = templateStart;
		}

		public int getTemplateAdd() {
			return templateAdd;
		}

		public void setTemplateAdd(int templateAdd) {
			this.templateAdd = templateAdd;
		}

		public int[] getTemplateResult() {
			return templateResult;
		}

		public void setTemplateResult(int[] templateResult) {
			this.templateResult = templateResult;
		}

		public int[] getTargetStart() {
			return targetStart;
		}

		public void setTargetStart(int[] targetStart) {
			this.targetStart = targetStart;
		}

		public int[] getTargetResult() {
			return targetResult;
		}

		public void setTargetResult(int[] targetResult) {
			this.targetResult = targetResult;
		}

		public int[] getTargetWithEndDate() {
			return targetWithEndDate;
		}

		public void setTargetWithEndDate(int[] targetWithEndDate) {
			this.targetWithEndDate = targetWithEndDate;
		}

		public Errors getTemplateErrors() {
			return templateErrors;
		}

		public void setTemplateErrors(Errors templateErrors) {
			this.templateErrors = templateErrors;
		}

		public Errors getTargetErrors() {
			return targetErrors;
		}

		public void setTargetErrors(Errors targetErrors) {
			this.targetErrors = targetErrors;
		}
	}


	/**
     * Test add copay tier common.
     * @param parameters The TestAddCopayTierCommonParameters specified for this test.
     *
     * @throws Exception the exception
     */
    private void testAddCopayTierCommon(TestAddCopayTierCommonParameters parameters) throws Exception {
        LOG.info("Start " + parameters.getTestName());

        boolean hasErrors = false;
        //template original
        ProductVo templateOriginal = new ProductVo();
        templateOriginal.setVaProductName("TEMPLATE");
        templateOriginal.setId("1");
        templateOriginal.setCopayTierAssocs(createCopayTierAssocVoList(templateOriginal, parameters.getTemplateStart()));

        //template modified
        ProductVo templateModified = (ProductVo) templateOriginal.clone();
        assertThat(templateModified.getCopayTierAssocs().size(), is(parameters.getTemplateStart().length));
        templateModified.setCopayTierAssocs(addCopayTier(templateModified, templateModified.getCopayTierAssocs(),
            parameters.getTemplateAdd()));
        assertThat(templateModified.getCopayTierAssocs().size(), is(parameters.getTemplateResult().length));
        assertThat(templateOriginal.getCopayTierAssocs().size(), is(parameters.getTemplateStart().length));

        //targets
        Collection<ManagedItemVo> targets = new ArrayList<ManagedItemVo>();
        ProductVo target = new ProductVo();
        target.setVaProductName("TARGET1");
        target.setId("2");
        target.setCopayTierAssocs(createCopayTierAssocVoList(target, parameters.getTargetStart()));
        targets.add(templateOriginal);
        targets.add(target);

        prepareMockExpectations();

        Collection<ItemModDifferences> differencesCollection = mimicPartOfSubmit(templateOriginal, targets, templateModified);
        for (ItemModDifferences itemDifferences : differencesCollection) {
            ProductVo product = (ProductVo) itemDifferences.getItem();
            LOG.debug("Asserting " + product.getVaProductName());

            if (product.getVaProductName().equals("TEMPLATE")) {
            	if (parameters.getTemplateErrors().hasErrors()){
            		assertThat(itemDifferences.getErrors(), is(parameters.getTemplateErrors()));
            		hasErrors = true;
            	}
            	else{            		
		            for (ModDifferenceVo modDifferenceVo : itemDifferences.getDifferences()) {
		                LOG.debug("Old Value: " + modDifferenceVo.getDifference().getOldValueShortString());
		                LOG.debug("New Value: " + modDifferenceVo.getDifference().getNewValueShortString());
		                    assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
		                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(parameters.getTemplateStart())));
		                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
		                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(parameters.getTemplateResult())));
		            }
            	}
            }
        	else if (product.getVaProductName().equals("TARGET1")) {
            	if (parameters.getTargetErrors().hasErrors()){
            		assertThat(itemDifferences.getErrors(), is(parameters.getTargetErrors()));
            		hasErrors = true;
            	}
            	else{    
            		for (ModDifferenceVo modDifferenceVo : itemDifferences.getDifferences()) {
		                LOG.debug("Old Value: " + modDifferenceVo.getDifference().getOldValueShortString());
		                LOG.debug("New Value: " + modDifferenceVo.getDifference().getNewValueShortString());
	            		assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
	                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(parameters.getTargetStart())));
	                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
	                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(parameters.getTargetResult())));
            		}
                }
            }
        }

        if (!hasErrors){
	        //CommitAll
	        Collection<ProductVo> committedProducts = mimicPartOfCommit(templateModified, differencesCollection);
	        assertThat(committedProducts.size(), is(2));
	
	        for (ProductVo committedProduct : committedProducts) {
	            if (committedProduct.getVaProductName().equals("TEMPLATE")) {
	                LOG.debug("Assert committedTemplate: ");
	                assertThat(committedProduct.getCopayTierAssocs(), is(templateModified.getCopayTierAssocs()));
	            } else if (committedProduct.getVaProductName().equals("TARGET1")) {
	                assertCommittedProductCopayTiers(committedProduct, parameters.gettargetResult(), parameters.gettargetWithEndDate());
	            } else {
	                Assert.fail("Unexpected Product: " + committedProduct.getVaProductName());
	            }
	        }
        }
        LOG.info("End " + parameters.getTestName());
    }

    /**
     * Assert committed product copay tiers.
     *
     * @param committedProduct the committed product
     * @param expectedTiersInts the expected tiers ints
     * @param expectedTiersIntsWithEndDate the expected tiers ints with end date
     */
    private void assertCommittedProductCopayTiers(ProductVo committedProduct, int[] expectedTiersInts,
        int[] expectedTiersIntsWithEndDate) {
        List<String> expectedTiers = createStringListFromIntArray(expectedTiersInts);
        List<String> expectedTiersWithEndDate = createStringListFromIntArray(expectedTiersIntsWithEndDate);
        LOG.debug("Assert committed " + committedProduct.getVaProductName() + ": ");
        assertThat("Number of Copay Tiers on result does not match expected number.",
            committedProduct.getCopayTierAssocs().size(), is(expectedTiers.size()));
        for (String expectedTier : expectedTiers) {
            boolean found = false;
            for (CopayTierAssocVo actualTier : committedProduct.getCopayTierAssocs()) {
                if (actualTier.getCopayTier().getValue().equals(expectedTier)) {
                    found = true;
                    if (expectedTiersWithEndDate.contains(expectedTier)) {
                        assertThat("Expected " + committedProduct.getVaProductName() + " to have an end date on copay tier "
                            + expectedTier + ".", actualTier.getEndDate(), notNullValue());
                    }
                    break;
                }
            }
            assertThat("Expected " + committedProduct.getVaProductName() + " to have copay tier " + expectedTier + ".",
                found, is(true));
        }
    }

    /**
     * Assert committed product drug classes.
     *
     * @param committedProduct the committed product
     * @param expectedDrugsClassIdInts the expected drugs class id ints
     */
    private void assertCommittedProductDrugClasses(ProductVo committedProduct, int[] expectedDrugsClassIdInts) {
        List<String> expectedDrugClassIds = createStringListFromIntArray(expectedDrugsClassIdInts);        
        assertThat("Number of Drug Classes on result does not match expected number.", committedProduct.getDrugClasses().size(),
            is(expectedDrugClassIds.size()));
        for (String expectedDrugClassId : expectedDrugClassIds) {
            boolean found = false;
            for (DrugClassGroupVo actualDrugClass : committedProduct.getDrugClasses()) {
                if (actualDrugClass.getDrugClass().getId().equals(expectedDrugClassId)) {
                    found = true;
                    break;
                }
            }
            assertThat("Expected " + committedProduct.getVaProductName() + " to have Drug Class " + expectedDrugClassId + ".",
                found, is(true));
        }
    }

    /**
     * Test reg ex matcher.
     */
    @Test
    public final void testRegExMatcher() {
        assertThat("Copay Tier: 2; Effective Date: 2016-06-16; End Date: 2016-07-15  \n"
            + "Copay Tier: 3; Effective Date: 2016-07-16; End Date: 2016-08-14  \n"
            + "Copay Tier: 1; Effective Date: 2016-08-15  \n",
            RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(2, 3, 1)));

        assertThat("No Selection Made", RegExMatcher.matchesRegex(createExpectedCopayTierRegEx()));
    }

    /**
     * Test reg ex matcher date.
     */
    @Test
    public final void testRegExMatcherDate() {
        String date = "((19|20)\\d\\d)-(0?[1-9]|1[012])-(0?[1-9]|[12][0-9]|3[01])";
        assertThat("2016-06-16", RegExMatcher.matchesRegex(date));
    }

    /**
     * Test create group mod difference.
     *
     * @throws Exception the exception
     */
    @Test
    public final void testCreateGroupModDifference() throws Exception {
        LOG.info("Start " + "testCreateGroupModDifference");
        //template original
        ProductVo templateOriginal = new ProductVo();
        templateOriginal.setVaProductName("TEMPLATE");
        templateOriginal.setId("1");
        templateOriginal.setCopayTierAssocs(createCopayTierAssocVoList(templateOriginal, 1, 2));
        Collection<ManagedItemVo> targets = new ArrayList<ManagedItemVo>();

        //template modified
        ProductVo templateModified = (ProductVo) templateOriginal.clone();
        assertThat(templateModified.getCopayTierAssocs().size(), is(2));
        templateModified.setCopayTierAssocs(addCopayTier(templateModified, templateModified.getCopayTierAssocs(), 3));
        assertThat(templateModified.getCopayTierAssocs().size(), is(3));
        assertThat(templateOriginal.getCopayTierAssocs().size(), is(2));

        //targets
        ProductVo target = new ProductVo();
        target.setVaProductName("TARGET1");
        target.setId("2");
        target.setCopayTierAssocs(createCopayTierAssocVoList(target, 1, 2));
        targets.add(templateOriginal);
        targets.add(target);
        prepareMockExpectations();

        Method method = ManagedItemCapabilityImpl.class.getDeclaredMethod("createGroupModDifference",
            ManagedItemVo.class, ModDifferenceVo.class, UserVo.class, Errors.class);
        method.setAccessible(true);

        Collection<ItemModDifferences> differencesCollection =
            mimicPartOfSubmit(templateOriginal, targets, templateModified);
        for (ItemModDifferences itemDifferences : differencesCollection) {
            for (ModDifferenceVo modDifferenceVo : itemDifferences.getDifferences()) {
                Errors errors = new Errors();
                ModDifferenceVo newModDifferenceVo = (ModDifferenceVo) method.invoke(managedItemCapability,
                    itemDifferences.getItem(), modDifferenceVo, USER, errors);
            }
        }
        LOG.info("End " + "testCreateGroupModDifference");
    }

    /**
     * Creates the expected copay tier reg ex.
     *
     * @param tiers the tiers
     * @return the string
     */
    private String createExpectedCopayTierRegEx(int... tiers) {
        if (tiers.length == 0) {
            return "No Selection Made";
        }

        String regex = "";
        String date = "((19|20)\\d\\d)-(0?[1-9]|1[012])-(0?[1-9]|[12][0-9]|3[01])";
        for (int i = 0; i < tiers.length; i++) {
            regex += "Copay Tier: " + tiers[i] + "; Effective Date: " + date;
            if (i != tiers.length - 1) {
                regex += "; End Date: " + date;
            }
            if (tiers.length > 1) {
                regex += "  \\R";
            }

        }

        return regex;
    }

    /**
     * Creates the expected drug class reg ex.
     *
     * @param drugClassIndex the drug class index
     * @return the string
     */
    private String createExpectedDrugClassRegEx(int... drugClassIndex) {
        if (drugClassIndex.length == 0) {
            return "No Selection Made";
        }        
        
        String regex = "";

        for (int i = 0; i < drugClassIndex.length; i++) {
            DrugClassVo drugClass = drugClassList.get(drugClassIndex[i] - 1).getDrugClass();
            regex += "<li style='list-style: none; margin-left:0px; display: block; padding: 0px;'>"
                + drugClass.getCode() + " - " + drugClass.getClassification() + "</li>";
            if (drugClassIndex.length > 1) {
                regex += "  \\R";
            }
        }
        return regex;
    }

    /**
     * Mimic part of submit.
     *
     * @param templateOriginal the template original
     * @param targets the targets
     * @param templateModified the template modified
     * @return the collection
     * @throws ItemNotFoundException the item not found exception
     */
    private Collection<ItemModDifferences> mimicPartOfSubmit(ProductVo templateOriginal, Collection<ManagedItemVo> targets,
        ProductVo templateModified) throws ItemNotFoundException {
        Collection<ModDifferenceVo> templateModDifferences = templateOriginal.compareDifferences(templateModified);
        ItemModDifferences templateItemModDifferenceVo = new ItemModDifferences(templateModified, templateModDifferences);
        Collection<ItemModDifferences> itemModDifferences =
            managedItemCapability.submitAllModifications(templateItemModDifferenceVo, targets, USER);
        return itemModDifferences;
    }

    /**
     * Prepare mock expectations.
     */
    private void prepareMockExpectations() {
        EasyMock.expect(environmentUtility.getEnvironment()).andReturn(Environment.NATIONAL).anyTimes();
        EasyMock.expect(environmentUtility.isLocal()).andReturn(false).anyTimes();
        EasyMock.expect(environmentUtility.getSiteNumber()).andReturn("999").anyTimes();
        EasyMock.expect(managedItemCapabilityFactory.getRulesCapability(EntityType.PRODUCT)).andReturn(rulesCapability)
            .anyTimes();
        //rulesCapability.enforceRules(EasyMock.anyObject(ProductVo.class), 
        //EasyMock.anyObject(UserVo.class), EasyMock.anyBoolean());
        EasyMock.replay(environmentUtility, managedItemCapabilityFactory);
    }

    /**
     * Mimic part of commit.
     *
     * @param templateModified the template modified
     * @param differencesCollection the item mod differences
     * @return the collection
     */
    private Collection<ProductVo> mimicPartOfCommit(ProductVo templateModified,
        Collection<ItemModDifferences> differencesCollection) {
        //mimic some of the logic from commitAllModifications
        Collection<ProductVo> items = new ArrayList<ProductVo>(differencesCollection.size());
        for (ItemModDifferences itemDifferences : differencesCollection) {
            ManagedItemVo itemCopy = itemDifferences.getItem().copy();

            for (ModDifferenceVo modDifference : itemDifferences.getDifferences()) {
                assertThat(modDifference.getDifference(), CoreMatchers.instanceOf(GroupDifference.class));
            }

            itemCopy.updateFromModDifferences(itemDifferences.getDifferences(), USER, environmentUtility);

            ProductVo product = (ProductVo) itemCopy;
            items.add(product);
//            for (CopayTierAssocVo copayTierAssocVo : product.getCopayTierAssocs()) {
//                LOG.debug(product.getVaProductName() + " - " + copayTierAssocVo.toShortString());
//            }            
//            LOG.debug("Asserting product: " + product.getVaProductName());
//            StringBuffer sbActual = new StringBuffer("Actual   Copay Tiers: ");
//            StringBuffer sbExpected = new StringBuffer("Expected Copay Tiers: ");
//            for (CopayTierAssocVo copayTierAssocVo : product.getCopayTierAssocs()) {
//                sbActual.append(copayTierAssocVo.getCopayTier().getDescription());
//                sbActual.append(" ");
//            }
//            for (CopayTierAssocVo copayTierAssocVo : templateModified.getCopayTierAssocs()) {
//                sbExpected.append(copayTierAssocVo.getCopayTier().getDescription());
//                sbExpected.append(" ");
//            }
//            LOG.debug(sbActual);
//            LOG.debug(sbExpected);
//            
//            assertThat(product.getCopayTierAssocs(), is(templateModified.getCopayTierAssocs()));
        }
        return items;
    }

    /**
     * Create a list containing the specified list of copay tiers.
     * All entries have an end date except the last one.
     *
     * @param product the product
     * @param tiers the tiers
     * @return the collection
     */
    private Collection<CopayTierAssocVo> createCopayTierAssocVoList(ProductVo product, int... tiers) {
        Collection<CopayTierAssocVo> list = new ArrayList<CopayTierAssocVo>(tiers.length);
        Date effectiveDate = DateUtils.addMonths(new Date(), tiers.length * -1);
        for (int tier : tiers) {
            effectiveDate = DateUtils.addMonths(effectiveDate, 1);
            addCopayTierWithSpecificEffectiveDate(product, list, tier, effectiveDate);
        }

        return list;
    }

    /**
     * Adds a copay tier to the list.
     * The last entry will not have an end date.
     *
     * @param product the product
     * @param list the list
     * @param tierToAdd the tier to add
     * @return the collection
     */
    private Collection<CopayTierAssocVo> addCopayTier(ProductVo product, Collection<CopayTierAssocVo> list, int tierToAdd) {

        Date lastEffectiveDate = new Date();
        for (CopayTierAssocVo copayTier : list) {
            if (copayTier.getEndDate() == null) {
                lastEffectiveDate = copayTier.getEffectiveDate();
                break;
            }
        }

        return addCopayTierWithSpecificEffectiveDate(product, list, tierToAdd, DateUtils.addMonths(lastEffectiveDate, 1));
    }

    /**
     * Adds a copay tier to the list.
     * The last entry will not have an end date.
     *
     * @param product the product
     * @param list the list
     * @param tierToAdd the tier to add
     * @return the collection
     */
    private Collection<CopayTierAssocVo> addCopayTierWithSpecificEffectiveDate(ProductVo product,
        Collection<CopayTierAssocVo> list, int tierToAdd, Date effectiveDate) {

        for (CopayTierAssocVo copayTier : list) {
            if (copayTier.getEndDate() == null) {
                copayTier.setEndDate(DateUtils.addDays(effectiveDate, -1));
                break;
            }
        }

        CopayTierAssocVo newCopayTier = new CopayTierAssocVo();
        newCopayTier.setId(String.valueOf(list.size()));
        newCopayTier.setProduct(product);
        newCopayTier.setCopayTier(copayTierList.get(tierToAdd));
        newCopayTier.setEffectiveDate(effectiveDate);

        list.add(newCopayTier);
        return list;
    }

    /**
     * Create a list containing the specified list of copay tiers.
     * All entries have an end date except the last one.
     *
     * @param product the product
     * @param ids the ids
     * @return the collection
     */
    private Collection<DrugClassGroupVo> createDrugClassList(ProductVo product, int... ids) {
        Collection<DrugClassGroupVo> list = new ArrayList<DrugClassGroupVo>(ids.length);
        for (int id : ids) {
            addDrugClass(product, list, id);
        }

        return list;
    }

    /**
     * Adds a copay tier to the list.
     * The last entry will not have an end date.
     *
     * @param product the product
     * @param list the list
     * @param idToAdd the id to add
     * @return the collection
     */
    private Collection<DrugClassGroupVo> addDrugClass(ProductVo product, Collection<DrugClassGroupVo> list,
        int idToAdd) {
        list.add(drugClassList.get(idToAdd - 1));
        return list;
    }

    /**
     * Test add copay tier common.
     *
     * @param testName the test name
     * @param TEMPLATE_START the template start
     * @param TEMPLATE_ADD the template add
     * @param TEMPLATE_RESULT the template result
     * @param TARGET_START the target start
     * @param TARGET_RESULT the target result
     * @param TARGET_WITH_END_DATE the target with end date
     * @throws Exception the exception
     */
    private void testAddCopayTierWithLongList(String testName, final int[] TEMPLATE_START, final int TEMPLATE_ADD,
        final int[] TEMPLATE_SUMMARY, final int[] TEMPLATE_RESULT, final int[] TARGET_START,
        final int[] TARGET_SUMMARY, final int[] TARGET_RESULT, final int[] TARGET_WITH_END_DATE) throws Exception {
        LOG.info("Start " + "testName");

        //template original
        ProductVo templateOriginal = new ProductVo();
        templateOriginal.setVaProductName("TEMPLATE");
        templateOriginal.setCopayTierAssocs(createCopayTierAssocVoList(templateOriginal, TEMPLATE_START));

        //template modified
        ProductVo templateModified = (ProductVo) templateOriginal.clone();
        assertThat(templateModified.getCopayTierAssocs().size(), is(TEMPLATE_START.length));
        templateModified.setCopayTierAssocs(addCopayTier(templateModified, templateModified.getCopayTierAssocs(),
            TEMPLATE_ADD));
        assertThat(templateModified.getCopayTierAssocs().size(), is(TEMPLATE_RESULT.length));
        assertThat(templateOriginal.getCopayTierAssocs().size(), is(TEMPLATE_START.length));

        //targets
        Collection<ManagedItemVo> targets = new ArrayList<ManagedItemVo>();
        ProductVo target = new ProductVo();
        target.setVaProductName("TARGET1");
        target.setCopayTierAssocs(createCopayTierAssocVoList(target, TARGET_START));
        targets.add(templateOriginal);
        targets.add(target);

        prepareMockExpectations();

        Collection<ItemModDifferences> differencesCollection = mimicPartOfSubmit(templateOriginal, targets, templateModified);
        for (ItemModDifferences itemDifferences : differencesCollection) {
            ProductVo product = (ProductVo) itemDifferences.getItem();

            for (ModDifferenceVo modDifferenceVo : itemDifferences.getDifferences()) {
                LOG.debug(modDifferenceVo.getDifference().getOldValueShortString());
                LOG.debug(modDifferenceVo.getDifference().getNewValueShortString());
                if (product.getVaProductName().equals("TEMPLATE")) {
                    assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(TEMPLATE_START)));
                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(TEMPLATE_SUMMARY)));
                } else if (product.getVaProductName().equals("TARGET1")) {
                    assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(TARGET_START)));
                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(TARGET_SUMMARY)));
                }
            }
        }

        //CommitAll
        Collection<ProductVo> committedProducts = mimicPartOfCommit(templateModified, differencesCollection);
        assertThat(committedProducts.size(), is(2));

        for (ProductVo committedProduct : committedProducts) {
            if (committedProduct.getVaProductName().equals("TEMPLATE")) {
                LOG.debug("Assert committedTemplate: ");
                assertThat(committedProduct.getCopayTierAssocs(), is(templateModified.getCopayTierAssocs()));
            } else if (committedProduct.getVaProductName().equals("TARGET1")) {
                assertCommittedProductCopayTiers(committedProduct, TARGET_RESULT, TARGET_WITH_END_DATE);
            } else {
                Assert.fail("Unexpected Product: " + committedProduct.getVaProductName());
            }
        }
        LOG.info("End " + "testName");
    }

    /**
     * Test add copay tier common.
     *
     * @param testName the test name
     * @param TEMPLATE_START the template start
     * @param TEMPLATE_ADD the template add
     * @param TEMPLATE_RESULT the template result
     * @param TARGET_START the target start
     * @param TARGET_RESULT the target result
     * @param TARGET_WITH_END_DATE the target with end date
     * @throws Exception the exception
     */
    private void testAddCopayTierCommon2(String testName, 
        CopayTestParameters parms) throws Exception {
        LOG.info("Start " + testName);
            
//        //template original
//        ProductVo templateOriginal = new ProductVo();
//        templateOriginal.setVaProductName("TEMPLATE");
//        templateOriginal.setId("1");
//        templateOriginal.setCopayTierAssocs(createCopayTierAssocVoList2(templateOriginal, parms.targetParametersList));
//    
//        //template modified
//        ProductVo templateModified = (ProductVo) templateOriginal.clone();
//        assertThat(templateModified.getCopayTierAssocs().size(), is(TEMPLATE_START.length));
//        templateModified.setCopayTierAssocs(addCopayTier(templateModified, templateModified.getCopayTierAssocs(),
//            TEMPLATE_ADD));
//        assertThat(templateModified.getCopayTierAssocs().size(), is(TEMPLATE_RESULT.length));
//        assertThat(templateOriginal.getCopayTierAssocs().size(), is(TEMPLATE_START.length));
//    
//        //targets
//        Collection<ManagedItemVo> targets = new ArrayList<ManagedItemVo>();
//        ProductVo target = new ProductVo();
//        target.setVaProductName("TARGET1");
//        target.setId("2");
//        target.setCopayTierAssocs(createCopayTierAssocVoList2(target, TARGET_START));
//        targets.add(templateOriginal);
//        targets.add(target);
//    
//        prepareMockExpectations();
//    
//        Collection<ItemModDifferences> itemModDifferences = mimicPartOfSubmit(templateOriginal, targets, templateModified);
//        for (ItemModDifferences itemModDifferenceVo : itemModDifferences) {
//            ProductVo product = (ProductVo) itemModDifferenceVo.getItem();
//    
//            for (ModDifferenceVo modDifferenceVo : itemModDifferenceVo.getDifferences()) {
//                LOG.debug("Asserting " + product.getVaProductName());
//                LOG.debug("Old Value: " + modDifferenceVo.getDifference().getOldValueShortString());
//                LOG.debug("New Value: " + modDifferenceVo.getDifference().getNewValueShortString());
//                if (product.getVaProductName().equals("TEMPLATE")) {
//                    assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
//                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(TEMPLATE_START)));
//                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
//                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(TEMPLATE_RESULT)));
//                } else if (product.getVaProductName().equals("TARGET1")) {
//                    assertThat(modDifferenceVo.getDifference().getOldValueShortString(),
//                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(TARGET_START)));
//                    assertThat(modDifferenceVo.getDifference().getNewValueShortString(),
//                        RegExMatcher.matchesRegex(createExpectedCopayTierRegEx(TARGET_RESULT)));
//                }
//            }
//        }
//    
//        //CommitAll
//        Collection<ProductVo> committedProducts = mimicPartOfCommit(templateModified, itemModDifferences);
//        assertThat(committedProducts.size(), is(2));
//    
//        for (ProductVo committedProduct : committedProducts) {
//            if (committedProduct.getVaProductName().equals("TEMPLATE")) {
//                LOG.debug("Assert committedTemplate: ");
//                assertThat(committedProduct.getCopayTierAssocs(), is(templateModified.getCopayTierAssocs()));
//            } else if (committedProduct.getVaProductName().equals("TARGET1")) {
//                assertCommittedProductCopayTiers(committedProduct, TARGET_RESULT, TARGET_WITH_END_DATE);
//            } else {
//                Assert.fail("Unexpected Product: " + committedProduct.getVaProductName());
//            }
//        }
//        LOG.info("End " + testName);
    }

    private int[] arr(int ...tiers)
    {
        return tiers;
    }
    
    /**
     * Create a list containing the specified list of copay tiers.
     * All entries have an end date except the last one.
     *
     * @param product the product
     * @param tiers the tiers
     * @return the collection
     */
    private Collection<CopayTierAssocVo> createCopayTierAssocVoList2(ProductVo product, int[][] tiers) {
        Collection<CopayTierAssocVo> list = new ArrayList<CopayTierAssocVo>(tiers.length);
        Date effectiveDate = DateUtils.addMonths(new Date(), tiers[0].length * -1);
        //past
        for (int tier : tiers[0]) {
            addCopayTierWithSpecificEffectiveDate(product, list, tier, effectiveDate);
            effectiveDate = DateUtils.addMonths(effectiveDate, 1);
        }
        //current
        for (int tier: tiers[1]) {
            addCopayTierWithSpecificEffectiveDate(product, list, tier, effectiveDate);
            effectiveDate = DateUtils.addMonths(effectiveDate, 1);
        }
        //future
        for (int tier: tiers[2]) {
            addCopayTierWithSpecificEffectiveDate(product, list, tier, effectiveDate);
            effectiveDate = DateUtils.addMonths(effectiveDate, 1);
        }        
        
        return list;
    }
    
    class CopayTestParameters{
        CopayTestTemplateItemParameters templateParameters;
        List<CopayTestTargetItemParameters> targetParametersList = new ArrayList<CopayTestTargetItemParameters>();
        private CopayTestTargetItemParameters findItemParameters(String productName){
            for (CopayTestTargetItemParameters itemParameters : targetParametersList) {
                if (itemParameters.productName.equals(productName)){
                    return itemParameters;
                }
            }
            return null;
        }
    }

    class CopayTestTemplateItemParameters{
        String productName;
        int[] startPast;
        int[] startActive;
        int[] startFuture;
        int[] add;
        int[] result;
        int[] resultWithEndDate;
        
        int[] getStart(){
            return mergeArrays(startPast, startActive, startFuture);
        }
        
        private int[] mergeArrays(int[] ...arrays){
            int size = 0;            
            for (int[] a : arrays) {
                size+= a.length;
            }                

            int[] result = new int[size];
            
            int destPos = 0;
            for (int i = 0; i < arrays.length; i++){
                if (i > 0 ){
                    destPos += arrays[i-1].length;
                }
                int length = arrays[i].length;
                System.arraycopy(arrays[i], 0, result, destPos, length);
            }
            return result;
        }
    }
    
    
    class CopayTestTargetItemParameters{
        String productName;
        int[] startPast;
        int[] startActive;
        int[] startFuture;
        int[] result;
        int[] resultWithEndDate;
        
        int[] getStart(){
            return mergeArrays(startPast, startActive, startFuture);
        }
        
        private int[] mergeArrays(int[] ...arrays){
            int size = 0;            
            for (int[] a : arrays) {
                size+= a.length;
            }                

            int[] result = new int[size];
            
            int destPos = 0;
            for (int i = 0; i < arrays.length; i++){
                if (i > 0 ){
                    destPos += arrays[i-1].length;
                }
                int length = arrays[i].length;
                System.arraycopy(arrays[i], 0, result, destPos, length);
            }
            return result;
        }
    }
}
