package gov.va.med.pharmacy.peps.service.common.capability.impl;

import static org.easymock.EasyMock.capture;
import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.eq;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.expectLastCall;
import static org.easymock.EasyMock.newCapture;
import static org.easymock.EasyMock.partialMockBuilder;
import static org.easymock.EasyMock.replay;
import static org.easymock.EasyMock.verify;
import static org.junit.Assert.fail;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.easymock.Capture;
import org.easymock.CaptureType;
import org.easymock.EasyMock;
import org.junit.Test;

import gov.va.med.pharmacy.peps.common.exception.AuthorizationException;
import gov.va.med.pharmacy.peps.common.exception.ItemNotFoundException;
import gov.va.med.pharmacy.peps.common.exception.ValidationException;
import gov.va.med.pharmacy.peps.common.test.CollectionMatcher;
import gov.va.med.pharmacy.peps.common.vo.Category;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.common.vo.FormularyDesignatorVo;
import gov.va.med.pharmacy.peps.common.vo.ManagedItemVo;
import gov.va.med.pharmacy.peps.common.vo.ModDifferenceVo;
import gov.va.med.pharmacy.peps.common.vo.NdcVo;
import gov.va.med.pharmacy.peps.common.vo.OrderableItemVo;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.Role;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.Difference;
import gov.va.med.pharmacy.peps.domain.common.capability.GenericNameDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.IngredientDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.NdcDomainCapability;
import gov.va.med.pharmacy.peps.domain.common.capability.ProductDomainCapability;
import gov.va.med.pharmacy.peps.service.common.capability.RematchCapability;

/**
 * JUnit Test Class for ManagedItemCapabilityImpl.  Uses a National ManagedItemCapabilityImpl to test.
 * {@link gov.va.med.pharmacy.peps.service.common.capability.impl.ManagedItemCapabilityImpl}
 * {@link gov.va.med.pharmacy.peps.service.national.capability.impl.ManagedItemCapabilityImpl}
 */
public class ManagedItemCapabilityImplTest {

  /**
   * Test checkModificationPreconditions for the role authorization logic.
   * 
   * {@link gov.va.med.pharmacy.peps.service.common.capability.impl.ManagedItemCapabilityImpl#checkModificationPreconditions(
   * Collection, gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, UserVo, java.util.function.Predicate)}
   */
  @Test
  public void testCheckModificationPreconditionsAuthorization() {
    ManagedItemCapabilityImpl impl = getImpl();
    UserVo user = new UserVo();
    //The user has Migrator role
    user.addRole(Role.PSS_PPSN_MIGRATOR);
    Collection<ModDifferenceVo> modDifferences = new ArrayList<ModDifferenceVo>();
    
    //Test that using the predicate with isEitherManager() will result in an AuthorizationException
    try {
      impl.checkModificationPreconditions(modDifferences, new OrderableItemVo(), user, u -> u.isEitherManager());
    } catch (ValidationException e) {
      if(! AuthorizationException.NOT_AUTHORIZED.equals(e.getKey())){
        fail("Expected AuthorizationException.NOT_AUTHORIZED");
      }
    }
    
   //Test that using the predicate that checks if the user has Migrator role does not result in an AuthorizationException
    try {
      impl.checkModificationPreconditions(modDifferences, new OrderableItemVo(), user, u -> u.hasRole(Role.PSS_PPSN_MIGRATOR));
    } catch (ValidationException e) {
      if(AuthorizationException.NOT_AUTHORIZED.equals(e.getKey())){
        fail("Expected AuthorizationException.NOT_AUTHORIZED");
      }
    }
    
  }

//  @Test
//  public void testToggleProductSupplySpecialHandling() {
//    fail("Not yet implemented");
//  }
//
//  @Test
//  public void testUpdateCascadedChild() {
//    fail("Not yet implemented");
//  }
//
//  @Test
//  public void testCascadeChangesToChildren() {
//    fail("Not yet implemented");
//  }

  /**
   * Test cascadeOiChangesToSingleProduct where the propagation will update categories and national formulary indicator.
   *
   *{@link gov.va.med.pharmacy.peps.service.common.capability.impl.ManagedItemCapabilityImpl
   *  #cascadeOiChangesToSingleProduct(List, ProductVo, UserVo, NdcDomainCapability)}
   *
   * @throws ItemNotFoundException the item not found exception
   * @throws ValidationException the validation exception
   */
  @SuppressWarnings("unchecked")
  @Test
  public void testCascadeOiChangesToSingleProductUpdateCategoriesAndFI() throws ItemNotFoundException, ValidationException {

    UserVo user = new UserVo();
    ModDifferenceVo categoryModDiff = new ModDifferenceVo();
    List<Category> newCategories = Arrays.asList(Category.MEDICATION, Category.COMPOUND);
    categoryModDiff.setDifference(new Difference(FieldKey.CATEGORIES, Arrays.asList(Category.MEDICATION), newCategories));
    categoryModDiff.setModificationReason("Reason Categories");
    
    ModDifferenceVo md2 = new ModDifferenceVo();
    md2.setDifference(new Difference(FieldKey.OI_TEXT, "Old OI Text", "New OI Text"));
    md2.setModificationReason("Reason OI Text");
    
    ModDifferenceVo md3 = new ModDifferenceVo();
    FormularyDesignatorVo oldFD = new FormularyDesignatorVo();
    oldFD.setName("Old FD Name");
    FormularyDesignatorVo newFD = new FormularyDesignatorVo();
    newFD.setName("New FD Name");
    md3.setDifference(new Difference(FieldKey.OI_FORMULARY_DESIGNATOR, oldFD, newFD));
    md3.setModificationReason("Reason FD");
    
    ModDifferenceVo md4 = new ModDifferenceVo();    
    md4.setDifference(new Difference(FieldKey.NATIONAL_FORMULARY_INDICATOR, Boolean.valueOf(false), Boolean.valueOf(true)));
    md4.setModificationReason("Reason FI");
    
    //This FieldKey is not processed by the method under test, just a distraction
    ModDifferenceVo md5 = new ModDifferenceVo();    
    md5.setDifference(new Difference(FieldKey.FDA_MED_GUIDE, "Old FDA Med Guide", "New FDA Med Guide"));
    md5.setModificationReason("Reason Med Guide");
    
    List<ModDifferenceVo> cascadingFieldOiDifferences = Arrays.asList(categoryModDiff, md2, md3, md4, md5);
    
    //this product will already match FD and OI Text but not match FI and Categories
    ProductVo product1 = new ProductVo();
    product1.setId("product1");
    FormularyDesignatorVo newProductFD = new FormularyDesignatorVo();
    newProductFD.setName("New FD Name");    
    product1.setProductFormularyDesignator(newProductFD);
    product1.setProductText("New OI Text");
    product1.setCategories(Arrays.asList(Category.COMPOUND));
    product1.setNationalFormularyIndicator(false);
    
    ProductVo updatedProduct1 = new ProductVo();
    updatedProduct1.setId("product1");
    updatedProduct1.setProductFormularyDesignator(newProductFD);
    updatedProduct1.setProductText("New OI Text");
    updatedProduct1.setCategories(Arrays.asList(Category.MEDICATION, Category.COMPOUND));
    updatedProduct1.setNationalFormularyIndicator(true);
    
    Map<FieldKey, Difference> expectedProduct1Differences = new HashMap<FieldKey, Difference>();
    expectedProduct1Differences.put(FieldKey.CATEGORIES, new Difference(FieldKey.CATEGORIES, Arrays.asList(Category.COMPOUND), 
        Arrays.asList(Category.MEDICATION, Category.COMPOUND)));
    expectedProduct1Differences.put(FieldKey.NATIONAL_FORMULARY_INDICATOR, new Difference(
        FieldKey.NATIONAL_FORMULARY_INDICATOR, Boolean.valueOf(false), Boolean.valueOf(true)));

    Map<FieldKey, String> expectedProduct1Reasons = new HashMap<FieldKey, String>();
    expectedProduct1Reasons.put(FieldKey.CATEGORIES, "Reason Categories");   
    expectedProduct1Reasons.put(FieldKey.NATIONAL_FORMULARY_INDICATOR, "Reason FI");
    
    ManagedItemCapabilityImpl managedItemCapability = partialMockBuilder(ManagedItemCapabilityImpl.class)
        .addMockedMethods("saveItemAuditHistoryCascade", "cascadeOiChangesToProductNdcs", "toggleProductSupplySpecialHandling")
        .addMockedMethod("updateCascadedChild", ManagedItemVo.class, Collection.class, UserVo.class, Boolean.class)
        .createMock();
    
    ProductDomainCapability productDomainCapability = createMock(ProductDomainCapability.class);
    NdcDomainCapability ndcDomainCapability = createMock(NdcDomainCapability.class);

    expect(managedItemCapability.toggleProductSupplySpecialHandling(product1)).andReturn(null);
    managedItemCapability.cascadeOiChangesToProductNdcs(updatedProduct1, user, ndcDomainCapability, categoryModDiff, 
        FieldKey.CATEGORIES, newCategories);
    expectLastCall();
    
    expect(
        managedItemCapability.updateCascadedChild(
            eq(updatedProduct1), 
            (Collection<Difference>)CollectionMatcher.collectionEq(expectedProduct1Differences.values()), 
            eq(user), 
            eq(Boolean.TRUE)))
    .andReturn(updatedProduct1);
    
    Capture<Map<FieldKey, String>> capturedReasons = newCapture(CaptureType.ALL);
    Capture<Map<FieldKey, Difference>> capturedModDiffs = newCapture(CaptureType.ALL);
    
    managedItemCapability.saveItemAuditHistoryCascade(capture(capturedReasons), 
        eq(updatedProduct1), capture(capturedModDiffs), eq(user));
    expectLastCall();
    replay(managedItemCapability, productDomainCapability, ndcDomainCapability);
    
    managedItemCapability.cascadeOiChangesToSingleProduct(cascadingFieldOiDifferences, product1, user, ndcDomainCapability, true);    
    
    verify(managedItemCapability, productDomainCapability);
  }
  
  /**
   * Test cascadeOiChangesToSingleProduct where the propagation will update Formulary Designator and Product Text.
   *
   *{@link gov.va.med.pharmacy.peps.service.common.capability.impl.ManagedItemCapabilityImpl
   *  #cascadeOiChangesToSingleProduct(List, ProductVo, UserVo, NdcDomainCapability)}
   *
   * @throws ItemNotFoundException the item not found exception
   * @throws ValidationException the validation exception
   */  
  @SuppressWarnings("unchecked")
  @Test
  public void testCascadeOiChangesToSingleProductUpdateFDAndText() throws ItemNotFoundException, ValidationException {

    UserVo user = new UserVo();
    ModDifferenceVo categoryModDiff = new ModDifferenceVo();
    List<Category> newCategories = Arrays.asList(Category.MEDICATION, Category.COMPOUND);
    categoryModDiff.setDifference(new Difference(FieldKey.CATEGORIES, Arrays.asList(Category.MEDICATION), newCategories));
    categoryModDiff.setModificationReason("Reason Categories");
    
    ModDifferenceVo md2 = new ModDifferenceVo();
    md2.setDifference(new Difference(FieldKey.OI_TEXT, "Old OI Text", "New OI Text"));
    md2.setModificationReason("Reason OI Text");
    
    ModDifferenceVo md3 = new ModDifferenceVo();
    FormularyDesignatorVo oldFD = new FormularyDesignatorVo();
    oldFD.setName("Old FD Name");
    FormularyDesignatorVo newFD = new FormularyDesignatorVo();
    newFD.setName("New FD Name");
    md3.setDifference(new Difference(FieldKey.OI_FORMULARY_DESIGNATOR, oldFD, newFD));
    md3.setModificationReason("Reason FD");
    
    ModDifferenceVo md4 = new ModDifferenceVo();    
    md4.setDifference(new Difference(FieldKey.NATIONAL_FORMULARY_INDICATOR, Boolean.valueOf(false), Boolean.valueOf(true)));
    md4.setModificationReason("Reason FI");
    
    //This FieldKey is not processed by the method under test, just a distraction
    ModDifferenceVo md5 = new ModDifferenceVo();    
    md5.setDifference(new Difference(FieldKey.FDA_MED_GUIDE, "Old FDA Med Guide", "New FDA Med Guide"));
    md5.setModificationReason("Reason Med Guide");
    
    List<ModDifferenceVo> cascadingFieldOiDifferences = Arrays.asList(categoryModDiff, md2, md3, md4, md5);
    
    //this product will already match FI and Categories but not match FD and OI Text 
    ProductVo product1 = new ProductVo();
    product1.setId("product1");
    FormularyDesignatorVo prodFD = new FormularyDesignatorVo();
    oldFD.setName("Other FD Name");    
    product1.setProductFormularyDesignator(prodFD);
    product1.setProductText("Old Product Text");
    product1.setCategories(Arrays.asList(Category.MEDICATION, Category.COMPOUND));
    product1.setNationalFormularyIndicator(true);
    
    ProductVo updatedProduct1 = new ProductVo();
    updatedProduct1.setId("product1");
    updatedProduct1.setProductFormularyDesignator(newFD);
    updatedProduct1.setProductText("New OI Text");
    updatedProduct1.setCategories(Arrays.asList(Category.MEDICATION, Category.COMPOUND));
    updatedProduct1.setNationalFormularyIndicator(true);
        
    Map<FieldKey, Difference> expectedProduct1Differences = new HashMap<FieldKey, Difference>();
    expectedProduct1Differences.put(FieldKey.PRODUCT_TEXT, new Difference(FieldKey.PRODUCT_TEXT, 
        "Old Product Text", "New OI Text"));
    expectedProduct1Differences.put(FieldKey.PRODUCT_FORMULARY_DESIGNATOR, 
        new Difference(FieldKey.PRODUCT_FORMULARY_DESIGNATOR, prodFD, newFD));

    Map<FieldKey, String> expectedProduct1Reasons = new HashMap<FieldKey, String>();
    expectedProduct1Reasons.put(FieldKey.PRODUCT_TEXT, "Reason OI Text");   
    expectedProduct1Reasons.put(FieldKey.PRODUCT_FORMULARY_DESIGNATOR, "Reason FD");
    
    ManagedItemCapabilityImpl managedItemCapability = partialMockBuilder(ManagedItemCapabilityImpl.class)
        .addMockedMethods("saveItemAuditHistoryCascade", "cascadeOiChangesToProductNdcs",
            "toggleProductSupplySpecialHandling")
        .addMockedMethod("updateCascadedChild", ManagedItemVo.class, Collection.class, UserVo.class, Boolean.class)
        .createMock();
    
    ProductDomainCapability productDomainCapability = createMock(ProductDomainCapability.class);
    NdcDomainCapability ndcDomainCapability = createMock(NdcDomainCapability.class);

    managedItemCapability.cascadeOiChangesToProductNdcs(updatedProduct1, user, ndcDomainCapability, categoryModDiff, 
        FieldKey.CATEGORIES, newCategories);
    expectLastCall();

    expect(managedItemCapability.updateCascadedChild(eq(updatedProduct1), 
        (Collection<Difference>)CollectionMatcher.collectionEq(expectedProduct1Differences.values()), eq(user), eq(Boolean.TRUE))).
        andReturn(updatedProduct1);

    Capture<Map<FieldKey, String>> capturedReasonsHistory = newCapture(CaptureType.ALL);
    Capture<Map<FieldKey, Difference>> capturedModDiffsHistory = newCapture(CaptureType.ALL);
    
    managedItemCapability.saveItemAuditHistoryCascade(capture(capturedReasonsHistory), eq(updatedProduct1), 
        capture(capturedModDiffsHistory), eq(user));
    expectLastCall();
    replay(managedItemCapability, productDomainCapability, ndcDomainCapability);
    
    managedItemCapability.cascadeOiChangesToSingleProduct(cascadingFieldOiDifferences, product1, user, ndcDomainCapability, true);    
    
   // assertThat()
    verify(managedItemCapability, productDomainCapability);
  }  
  
  /**
   * Test cascadeOiChangesToProductNdcs.
   *
   *{@link gov.va.med.pharmacy.peps.service.common.capability.impl.ManagedItemCapabilityImpl
   *  #cascadeOiChangesToProductNdcs(ProductVo, UserVo, NdcDomainCapability, ModDifferenceVo, FieldKey, List)
   * @throws ValidationException the validation exception
   */
  @SuppressWarnings("unchecked")
  @Test
  public void testCascadeOiChangesToProductNdcs() throws ValidationException {
    ProductVo product = new ProductVo();    
    NdcVo ndc1Lite = new NdcVo();
    ndc1Lite.setId("ndc1");
    NdcVo ndc2Lite = new NdcVo();
    ndc2Lite.setId("ndc2");
    product.setNdcs(Arrays.asList(ndc1Lite, ndc2Lite));
    UserVo user = new UserVo();
    
    NdcVo ndc1 = new NdcVo();
    ndc1.setId("ndc1");
    ndc1.setCategories(Arrays.asList(Category.MEDICATION));
    NdcVo ndc2 = new NdcVo();
    ndc2.setId("ndc2");
    ndc2.setCategories(Arrays.asList(Category.COMPOUND, Category.MEDICATION));
    
    ModDifferenceVo oiDiff = new ModDifferenceVo();
    oiDiff.setDifference(new Difference(FieldKey.CATEGORIES, Arrays.asList(Category.MEDICATION), 
        Arrays.asList(Category.MEDICATION, Category.COMPOUND) ));
    oiDiff.setModificationReason("Unit Test");
        
    List<Category> newCategories = Arrays.asList(Category.MEDICATION, Category.COMPOUND);
    
    Map<FieldKey, Difference> expectedNdcDifferences = new HashMap<FieldKey, Difference>();
    expectedNdcDifferences.put(FieldKey.CATEGORIES, new Difference(FieldKey.CATEGORIES, Arrays.asList(Category.MEDICATION), 
        Arrays.asList(Category.MEDICATION, Category.COMPOUND)));

    Map<FieldKey, String> expectedNdcReasons = new HashMap<FieldKey, String>();
    expectedNdcReasons.put(FieldKey.CATEGORIES, "Unit Test");

    NdcVo updatedNdc1 = new NdcVo();
    updatedNdc1.setId("ndc1");
    updatedNdc1.setCategories(Arrays.asList(Category.MEDICATION, Category.COMPOUND));
    
    
    ManagedItemCapabilityImpl managedItemCapability = partialMockBuilder(ManagedItemCapabilityImpl.class)
        .addMockedMethod("saveItemAuditHistoryCascade")
        .addMockedMethod("updateCascadedChild", ManagedItemVo.class, Collection.class, UserVo.class)
        .addMockedMethod("updateCascadedChild", ManagedItemVo.class, Collection.class, UserVo.class, Boolean.class)
        .createMock();
    
    NdcDomainCapability ndcDomainCapability = createMock(NdcDomainCapability.class);
    expect(ndcDomainCapability.retrieve("ndc1")).andReturn(ndc1);
        
    expect(managedItemCapability.updateCascadedChild(eq(updatedNdc1), 
        (Collection<Difference>)CollectionMatcher.collectionEq(expectedNdcDifferences.values()), eq(user))).
    andReturn(updatedNdc1);

    managedItemCapability.saveItemAuditHistoryCascade(expectedNdcReasons, ndc1, expectedNdcDifferences, user);
    expectLastCall();
    expect(ndcDomainCapability.retrieve("ndc2")).andReturn(ndc2);
    replay(managedItemCapability, ndcDomainCapability);
    
    managedItemCapability.cascadeOiChangesToProductNdcs(product, user, ndcDomainCapability, oiDiff, FieldKey.CATEGORIES, 
        newCategories);    
    
    verify(managedItemCapability, ndcDomainCapability);
  }

  private ManagedItemCapabilityImpl getImpl() {
    ManagedItemCapabilityImpl impl = new gov.va.med.pharmacy.peps.service.national.capability.impl.ManagedItemCapabilityImpl(
        // We don't need these capabilities. Just mock them out to avoid passing null.
        EasyMock.createMock(IngredientDomainCapability.class),
        EasyMock.createMock(GenericNameDomainCapability.class),
        EasyMock.createMock(RematchCapability.class));
    return impl;
  }  

}
