package gov.va.med.pharmacy.peps.service.common.capability.impl;

import static org.easymock.EasyMock.expect;
import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;

import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import org.easymock.EasyMockSupport;
import org.junit.Test;

import gov.va.med.pharmacy.peps.common.utility.test.generator.ProductGenerator;
import gov.va.med.pharmacy.peps.common.utility.test.generator.UserGenerator;
import gov.va.med.pharmacy.peps.common.vo.FieldKey;
import gov.va.med.pharmacy.peps.common.vo.IngredientVo;
import gov.va.med.pharmacy.peps.common.vo.ItemStatus;
import gov.va.med.pharmacy.peps.common.vo.MatchReason;
import gov.va.med.pharmacy.peps.common.vo.OrderableItemVo;
import gov.va.med.pharmacy.peps.common.vo.ProductVo;
import gov.va.med.pharmacy.peps.common.vo.RematchVo;
import gov.va.med.pharmacy.peps.common.vo.RequestVo;
import gov.va.med.pharmacy.peps.common.vo.UserVo;
import gov.va.med.pharmacy.peps.common.vo.diff.Difference;
import gov.va.med.pharmacy.peps.common.vo.diff.test.DifferenceGenerator;
import gov.va.med.pharmacy.peps.domain.common.capability.RematchMgtDomainCapability;

/**
 * Unit tests for {@link RematchCapabilityImpl}.
 */
public class RematchCapabilityImplTest extends EasyMockSupport {
  
  /**
   * Just for convenience since many tests use this.
   */
  private final ProductGenerator productGenerator = new ProductGenerator();
  
  /**
   * Just for convenience since many tests use this.
   */
  private final UserGenerator userGenerator = new UserGenerator();
  
  /**
   * Tests {@link
   * RematchCapabilityImpl#getExistingRematchForDifferences(gov.va.med.pharmacy.peps.common.vo.ManagedItemVo,
   * Collection)}.
   */
  @Test
  public void testGetExistingRematchForDifferences() {

    /* Setup */

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);
    
    // Create the existing rematch.
    final RematchVo existingRematch = createExistingRematch(MatchReason.FUTURE_INACTIVATION);
    final ProductVo testProduct = existingRematch.getFromProduct();

    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);
    
    final Collection<Difference> inactivationDiffs = Arrays.asList(
        new Difference(FieldKey.ITEM_STATUS, ItemStatus.ACTIVE, ItemStatus.INACTIVE));
    final Collection<Difference> futureInactivationDiffs = Arrays.asList(
        DifferenceGenerator.makeDifferenceForDataField(FieldKey.PROPOSED_INACTIVATION_DATE, null, new Date()));

    // Expected method calls
    expect(mockRematchDomainCapability.getRematchForProduct(Long.parseLong(testProduct.getId())))
      .andReturn(Optional.of(existingRematch)).times(2); // Combining two tests in one here.
    
    replayAll();
    
    /* Behavior & verification */
    
    assertThat(
        objUnderTest.getExistingRematchForDifferences(testProduct, inactivationDiffs),
        equalTo(Optional.empty()));
    
    assertThat(
        objUnderTest.getExistingRematchForDifferences(testProduct, futureInactivationDiffs),
        equalTo(Optional.of(existingRematch)));
    
    verifyAll();

  }
  
  /**
   * Tests calling {@link RematchCapabilityImpl#updateUponProposedInactivation(
   * gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, UserVo)} for a non-product.
   */
  @Test
  public void testUpdateUponProposedInactivation_NonProduct() {
    /* Setup */

    final UserVo updatingUser = userGenerator.getNationalManagerOne();

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);

    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);

    // No expected method calls
    
    replayAll();
    
    /* Behavior & verification */
    
    objUnderTest.updateUponProposedInactivation(new OrderableItemVo(), updatingUser);
    objUnderTest.updateUponProposedInactivation(new IngredientVo(), updatingUser);
    
    verifyAll();
  }

  @Test
  public void testUpdateUponProposedInactivation_NoExisting() {
    /* Setup */

    final UserVo updatingUser = userGenerator.getNationalManagerTwo();
    final ProductVo testProduct = productGenerator.getFirst();

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);

    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);

    // Expected method calls
    expect(mockRematchDomainCapability.getRematchForProduct(Long.parseLong(testProduct.getId())))
      .andReturn(Optional.empty());
    
    replayAll();
    
    /* Behavior & verification */
    
    objUnderTest.updateUponProposedInactivation(testProduct, updatingUser);
    
    verifyAll();
  }

  @Test
  public void testUpdateUponProposedInactivation_ExistingFutureInact() {
    /* Setup */

    final UserVo updatingUser = userGenerator.getNationalManagerOne();
    final RematchVo existingRematch = createExistingRematch(MatchReason.FUTURE_INACTIVATION);
    final ProductVo testProduct = existingRematch.getFromProduct();

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);

    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);

    // Expected method calls
    expect(mockRematchDomainCapability.getRematchForProduct(Long.parseLong(testProduct.getId())))
      .andReturn(Optional.of(existingRematch));
    expect(mockRematchDomainCapability.saveOrUpdate(existingRematch, updatingUser)).andReturn(existingRematch);
    
    replayAll();
    
    /* Behavior & verification */
    
    objUnderTest.updateUponProposedInactivation(testProduct, updatingUser);
    
    verifyAll();

    assertThat("should have changed match reason", existingRematch.getMatchReason(), equalTo(MatchReason.INACTIVATION));
  }

  @Test
  public void testUpdateUponProposedInactivation_ExistingOther() {
    /* Setup */

    final MatchReason initialMatchReason = MatchReason.OTHER_CHANGE;
    final RematchVo existingRematch = createExistingRematch(initialMatchReason);
    final ProductVo testProduct = existingRematch.getFromProduct();

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);

    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);

    // Expected method calls
    expect(mockRematchDomainCapability.getRematchForProduct(Long.parseLong(testProduct.getId())))
      .andReturn(Optional.of(existingRematch));
    
    replayAll();
    
    /* Behavior & verification */
    
    objUnderTest.updateUponProposedInactivation(testProduct, userGenerator.getNationalManagerOne());
    
    verifyAll();
    
    assertThat("should not have changed match reason", existingRematch.getMatchReason(), equalTo(initialMatchReason));
  }

  /**
   * Tests that {@link RematchCapabilityImpl#saveOrUpdateUncommittedRematch(Collection, RematchVo, UserVo)} deletes the
   * rematch if it is obsolete.
   */
  @Test
  public void testSaveOrUpdateUncommittedRematch_DeleteObsolete() {
    
    /* Setup */
    
    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);
    
    final List<ProductVo> testProducts = productGenerator.getList();
    final ProductVo fromProduct = testProducts.get(0);
    final ProductVo toProduct = testProducts.get(1);
    final UserVo testUser = userGenerator.getNationalManagerTwo();

    final Collection<Difference> differences = Arrays.asList(
        // Red herring difference.
        new Difference(FieldKey.HERBAL, Boolean.FALSE, Boolean.TRUE));
    final RematchVo testRematch = RematchVo.automaticRematch(fromProduct, toProduct, MatchReason.INACTIVATION);
    testRematch.setEplRematchId(null);  // Unsaved.
    
    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);
    
    // No expected method calls, should just return an empty optional.
    
    replayAll();
    
    /* Behavior & verification */
    
    assertThat(
        objUnderTest.saveOrUpdateUncommittedRematch(differences, testRematch, testUser),
        equalTo(Optional.empty()));
    
    verifyAll();
  }

  /**
   * Tests that {@link
   * RematchCapabilityImpl#saveOrUpdateCommittedRematch(gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, Collection,
   * Optional, UserVo)} does not delete an existing rematch if it shouldn't.
   */
  @Test
  public void testSaveOrUpdateCommittedRematch_PreserveExisting() {

    /* Setup */

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);
    
    final ProductVo testProduct = productGenerator.getFirst();
    final UserVo testUser = userGenerator.getNationalManagerTwo();
    
    final Collection<Difference> differences = Arrays.asList(
        // The item has a rematch and we are inactivating it. This shouldn't actually happen in product, but it's a good
        // edge case for this test.
        new Difference(FieldKey.ITEM_STATUS, ItemStatus.ACTIVE, ItemStatus.INACTIVE));
    
    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);
    
    // No expected method calls: should _not_ call delete.
    
    replayAll();
    
    /* Behavior & verification */
    
    assertThat(
        objUnderTest.saveOrUpdateCommittedRematch(testProduct, differences, Optional.empty(), testUser),
        equalTo(Optional.empty()));
    
    verifyAll();
  }

  
  /**
   * Tests that {@link
   * RematchCapabilityImpl#saveOrUpdateCommittedRematch(gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, Collection,
   * Optional, UserVo)} deletes the existing rematch if we are reversing the trigger.
   */
  @Test
  public void testSaveOrUpdateCommittedRematch_DeleteReversed() {

    /* Setup */

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);
    
    // Create the existing rematch.
    final RematchVo existingRematch = createExistingRematch(MatchReason.INACTIVATION);
    final UserVo testUser = userGenerator.getNationalManagerTwo();
    final ProductVo testProduct = existingRematch.getFromProduct();

    final Collection<Difference> differences = Arrays.asList(
        new Difference(FieldKey.ITEM_STATUS, ItemStatus.INACTIVE, ItemStatus.ACTIVE));
    
    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);
    
    // Expected method calls
    expect(mockRematchDomainCapability.getRematchForProduct(Long.parseLong(testProduct.getId())))
      .andReturn(Optional.of(existingRematch));
    mockRematchDomainCapability.delete(existingRematch);
    
    replayAll();
    
    /* Behavior & verification */
    
    assertThat(
        objUnderTest.saveOrUpdateCommittedRematch(testProduct, differences, Optional.empty(), testUser),
        equalTo(Optional.empty()));
    
    verifyAll();
  }
  
  /**
   * Tests that {@link
   * RematchCapabilityImpl#saveOrUpdateCommittedRematch(gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, Collection,
   * Optional, UserVo)} works properly if we are reversing the trigger and there is no existing rematch.
   */
  @Test
  public void testSaveOrUpdateCommittedRematch_DeleteReversedNonExisting() {

    /* Setup */
    
    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);
    
    final ProductVo testProduct = productGenerator.getFirst();
    final UserVo testUser = userGenerator.getNationalManagerOne();

    final Collection<Difference> differences = Arrays.asList(
        new Difference(FieldKey.ITEM_STATUS, ItemStatus.INACTIVE, ItemStatus.ACTIVE));
    
    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);

    // Expected method calls
    expect(mockRematchDomainCapability.getRematchForProduct(Long.parseLong(testProduct.getId())))
      .andReturn(Optional.empty());
    
    replayAll();
    
    /* Behavior & verification */
    
    assertThat(
        objUnderTest.saveOrUpdateCommittedRematch(testProduct, differences, Optional.empty(), testUser),
        equalTo(Optional.empty()));
    
    verifyAll();
  }

  /**
   * Tests that {@link
   * RematchCapabilityImpl#saveOrUpdateCommittedRematch(gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, Collection,
   * Optional, UserVo)} deletes the rematch if it is obsolete.
   */
  @Test
  public void testSaveOrUpdateCommittedRematch_DeleteObsolete() {
    
    /* Setup */
    
    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);
    
    final List<ProductVo> testProducts = productGenerator.getList();
    final ProductVo fromProduct = testProducts.get(0);
    final ProductVo toProduct = testProducts.get(1);
    final UserVo testUser = userGenerator.getNationalManagerTwo();

    final Collection<Difference> differences = Arrays.asList(
        // Red herring difference.
        new Difference(FieldKey.HERBAL, Boolean.FALSE, Boolean.TRUE));
    final RematchVo testRematch = RematchVo.automaticRematch(fromProduct, toProduct, MatchReason.INACTIVATION);
    testRematch.setEplRematchId("1999");
    
    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);
    
    // Expected method calls
    mockRematchDomainCapability.delete(testRematch);
    
    replayAll();
    
    /* Behavior & verification */
    
    assertThat(
        objUnderTest.saveOrUpdateCommittedRematch(fromProduct, differences, Optional.of(testRematch), testUser),
        equalTo(Optional.empty()));
    
    verifyAll();
  }
  
  /**
   * Tests that {@link
   * RematchCapabilityImpl#saveOrUpdateCommittedRematch(gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, Collection,
   * Optional, UserVo)} does not delete the rematch if it is not obsolete.
   */
  @Test
  public void testSaveOrUpdateCommittedRematch_KeepCurrent() {
    
    /* Setup */

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);
    
    final List<ProductVo> testProducts = productGenerator.getList();
    final ProductVo fromProduct = testProducts.get(0);
    final ProductVo toProduct = testProducts.get(1);
    final UserVo testUser = userGenerator.getNationalManagerTwo();
    final Collection<Difference> differences = Arrays.asList(
        new Difference(FieldKey.ITEM_STATUS, ItemStatus.ACTIVE, ItemStatus.INACTIVE));
    final RematchVo testRematch = RematchVo.automaticRematch(fromProduct, toProduct, MatchReason.INACTIVATION);
    
    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);
    
    // Expected method calls.
    expect(mockRematchDomainCapability.saveOrUpdate(testRematch, testUser)).andReturn(testRematch);
    
    replayAll();
    
    /* Behavior & verification */
    
    assertThat(
        objUnderTest.saveOrUpdateCommittedRematch(fromProduct, differences, Optional.of(testRematch), testUser),
        equalTo(Optional.of(testRematch)));
    
    verifyAll();
  }
  
  /**
   * Tests that {@link RematchCapabilityImpl#updateUponCommittingItemChanges(RequestVo,
   * gov.va.med.pharmacy.peps.common.vo.ManagedItemVo, java.util.Collection)} updates the rematch if the trigger reason
   * changed.
   */
  @Test
  public void testSaveOrUpdateCommittedRematch_UpdateCurrent() {
    
    /* Setup */

    // Mock the RematchMgtDomainCapability.
    final RematchMgtDomainCapability mockRematchDomainCapability = strictMock(RematchMgtDomainCapability.class);
    
    final List<ProductVo> testProducts = productGenerator.getList();
    final ProductVo fromProduct = testProducts.get(0);
    final ProductVo toProduct = testProducts.get(1);
    final UserVo testUser = userGenerator.getNationalManagerOne();
    final Collection<Difference> differences = Arrays.asList(
        DifferenceGenerator.makeDifferenceForDataField(FieldKey.PROPOSED_INACTIVATION_DATE, null, new Date()),
        new Difference(FieldKey.ITEM_STATUS, ItemStatus.ACTIVE, ItemStatus.INACTIVE));
    final RematchVo testRematch = RematchVo.automaticRematch(fromProduct, toProduct, MatchReason.FUTURE_INACTIVATION);
    final RematchVo expectedRematch = RematchVo.automaticRematch(fromProduct, toProduct, MatchReason.INACTIVATION);
    
    final RematchCapabilityImpl objUnderTest = new RematchCapabilityImpl(mockRematchDomainCapability);
    
    // Expected method calls.
    expect(mockRematchDomainCapability.saveOrUpdate(expectedRematch, testUser)).andReturn(expectedRematch);
    
    replayAll();
    
    /* Behavior & verification */
    
    assertThat(
        objUnderTest.saveOrUpdateCommittedRematch(fromProduct, differences, Optional.of(testRematch), testUser),
        equalTo(Optional.of(expectedRematch)));
    
    verifyAll();
  }
  
  private RematchVo createExistingRematch(final MatchReason matchReason) {
    
    // Create the existing rematch.
    final List<ProductVo> testProducts = productGenerator.getList();
    final ProductVo fromProduct = testProducts.get(0);
    final ProductVo toProduct = testProducts.get(1);
    final RematchVo existingRematch = RematchVo.automaticRematch(fromProduct, toProduct, MatchReason.INACTIVATION);
    existingRematch.setMatchReason(matchReason);
    existingRematch.setEplRematchId("1244");
    
    return existingRematch;
  }

}
